package eve.database;
import java.util.Vector;

import eve.data.PlainDate;
import eve.data.PlainTime;
import eve.math.Decimal;
import eve.sys.DayOfYear;
import eve.sys.Reflection;
import eve.sys.Time;
import eve.sys.TimeOfDay;
import eve.util.ByteArray;

//##################################################################
public class DatabaseUtils implements DatabaseTypes{
//##################################################################

//-------------------------------------------------------------------
DatabaseUtils(){}
//-------------------------------------------------------------------

static Time time = new Time();

//===================================================================
public synchronized static long getNewOID()
//===================================================================
{
	Time t = time.setToCurrentTime();
	int val = (t.month*31+t.day)*24*60*60;
	val += t.hour*60*60;
	val += t.minute*60;
	val += t.second;
	val <<= 5;
	val |= ((int)(java.lang.Math.random() * 32))%32;
	int val2 = (t.millis << 22) | (int)(4194304 * java.lang.Math.random());
	return (long)val << 32 | ((long)val2 & 0xffffffffL);	
	/*
	specs = getSpecs();
	if (specs == null) return 0;
	specs.lastAssigned++;
	specsEntry.saveObject("Specs",specs);
	return (long)specs.myOID << 32 | (long)specs.lastAssigned;
	*/
}
//===================================================================
public static int toCriteria(int field, int type, int options)
//===================================================================
{
	return (options & 0xff) << 24 | (type & 0xff) << 16 | (field & 0xffff);
}
//===================================================================
public static int criteriaToField(int criteria)
//===================================================================
{
	return criteria & 0xffff;
}
//===================================================================
public static int criteriaToType(int criteria)
//===================================================================
{
	return (criteria >> 16) & 0xff;
}
//===================================================================
public static int criteriaToOptions(int criteria)
//===================================================================
{
	return (criteria >> 24) & 0xff;
}

/**
 * Returns if a quick search using the specified criteria can be done on
 * a set of entries which are sorted by the sortedBy criteria.
 * @param criteria the search criteria.
 * @param sortCriteria the sorted criteria.
 * @return if a quick search using the specified criteria can be done on
 * a set of entries which are sorted by the sortedBy criteria.
 */
//===================================================================
public static boolean searchIsCompatibleWithSort(int[] criteria, int[] sortCriteria)
//===================================================================
{
	if (criteria.length > sortCriteria.length) return false;
	for (int i = 0; i<criteria.length; i++){
		if (criteriaToField(criteria[i]) != criteriaToField(sortCriteria[i])) return false;
		if (criteriaToType(criteria[i]) != criteriaToType(sortCriteria[i])) return false;
		//if ((criteriaToOptions(criteria[i]) & ~SORT_DESCENDING) != (criteriaToOptions(sortCriteria[i]) & ~SORT_DESCENDING)) return false;
	}
	return true;
}
//===================================================================
public static int[] getCriteriaSubset(int[] sortCriteria, int numberOfCriteria)
//===================================================================
{
	if (numberOfCriteria > sortCriteria.length || numberOfCriteria < 0)
		throw new IllegalArgumentException();
	int [] criteria = new int[numberOfCriteria];
	System.arraycopy(sortCriteria,0,criteria,0,numberOfCriteria);
	return criteria;
}
//===================================================================
public static int[] copyCriteria(int[] criteria)
//===================================================================
{
	if (criteria == null) return criteria;
	return getCriteriaSubset(criteria,criteria.length);
}
//===================================================================
public static int[] getCriteriaSubset(Database db,int[] criteria,Object searchData)
//===================================================================
{
	int num = 1;
	if (searchData instanceof Vector) num = ((Vector)searchData).size();
	else if (searchData instanceof Object[]) num = ((Object[])searchData).length;
	else if (searchData instanceof DatabaseEntry) num = criteria.length;
	else if (db.isInstanceOfDataObject(searchData)) num = criteria.length;
	if (num > criteria.length) num = criteria.length;
	return getCriteriaSubset(criteria,num);
}
//===================================================================
public static boolean isCollection(Object searchData)
//===================================================================
{
	return ((searchData instanceof Object[]) || (searchData instanceof Vector));
}
//===================================================================
public static int lengthOfCollection(Object searchData)
//===================================================================
{
	if (searchData instanceof Vector) return ((Vector)searchData).size();
	if (searchData instanceof Object[]) return ((Object[])searchData).length;
	throw new IllegalArgumentException();
}
//===================================================================
public static Object getInCollection(Object searchData, int index)
//===================================================================
{
	if (searchData instanceof Object[])
		return ((Object[])searchData)[index];
	if (searchData instanceof Vector)
		return ((Vector)searchData).get(index);
	throw new IllegalArgumentException();
}
private static Class[] classes;
private static int[] types;

/**
 * Convert a Java Class into a database type.
 * @param javaType the type of the Object.
 * @return one of the DatabaseTypes type values.
 */
public synchronized static int javaTypeToDatabaseType(Class javaType)
{
	if (javaType == null) return 0;
	if (classes == null){
		classes = new Class[]{
			null,String.class,TimeOfDay.class,DayOfYear.class,
			Decimal.class,TimeStamp.class,Time.class,ByteArray.class,
			Integer.TYPE,Double.TYPE,Boolean.TYPE,Long.TYPE,
			PlainTime.class, PlainDate.class
		};
		classes[0] = Reflection.forEncodedName("[B");
		types = new int[]{
				BYTE_ARRAY,STRING,TIME,DATE,
				DECIMAL,TIMESTAMP,DATE_TIME,BYTE_ARRAY,
				INTEGER,DOUBLE,BOOLEAN,LONG,
				PLAIN_TIME, PLAIN_DATE
		};
	}
	for (int i = 0; i<classes.length; i++){
		if (classes[i] == javaType || classes[i].isAssignableFrom(javaType))
			return types[i];
	}
	return JAVA_OBJECT;
}
/*
 * 				if (Reflection.isTypeOf(r,"Leve/sys/DayOfYear;")) return DATE;
				if (Reflection.isTypeOf(r,"Leve/math/Decimal;")) return DECIMAL;
				if (Reflection.isTypeOf(r,"Leve/database/TimeStamp;")) return TIMESTAMP;
				if (Reflection.isTypeOf(r,"Leve/sys/Time;")) return DATE_TIME;
				if (Reflection.isTypeOf(r,"Leve/util/ByteArray;")) return BYTE_ARRAY;

 */
//===================================================================
public static String getTypeSpecifier(int type,boolean ignoreString)
//===================================================================
{
	switch(type){
		case STRING: return ignoreString ? "" : "Ljava/lang/String;";
		case INTEGER: return "I";
		case BOOLEAN: return "Z";
		case DOUBLE: return "D";
		case LONG: return "J";
		case DATE_TIME: return "Leve/sys/Time;";
		case TIME: return "Leve/sys/TimeOfDay;";
		case DATE: return "Leve/sys/DayOfYear;";
		case TIMESTAMP: return "Leve/database/TimeStamp;";
		case DECIMAL: return "Leve/math/Decimal;";
		case BYTE_ARRAY: return "Leve/util/ByteArray;";
		case PLAIN_DATE: return "Leve/data/PlainDate;";
		case PLAIN_TIME: return "Leve/data/PlainTime;";
		case JAVA_OBJECT: return "Ljava/lang/Object;";
		default:
			return "";
	}
}
/**
 * This returns a field spec string for a particular fieldID in the form:
 * "Header|FieldName$Type"
 * @param fieldID the fieldID to look for.
 * @return the field spec string or null if the fieldID is invalid.
 */
//===================================================================
public static String getFieldSpec(Database db,int fieldID)
//===================================================================
{
	String name = db.getFieldName(fieldID);
	if (name == null) return null;
	String header = db.getFieldHeader(fieldID);
	if (header == null) return null;
	String ret = header+"|"+name+"$";
	ret += getTypeSpecifier(db.getFieldType(fieldID),true);
	return ret;
}
//##################################################################
}
//##################################################################

