package eve.fx.gui;
/**
This interface defines the system flags used with the <b>modifier</b> field in the
Event class. Start all custom flags at 0x00010000
*/
//##################################################################
public interface IEventModifiers{
//##################################################################

/** modifier for alt key */
public static final int ALT      = 0x1;
/** modifier for control key */
public static final int CONTROL  = 0x2;
/** modifier for shift key */
public static final int SHIFT    = 0x4;
/** modifier for non-ascii keys */
public static final int SPECIAL =  0x8;
/** modifier used when specifying hot keys - indicates that the key alone with no modifier must be pressed. */
public static final int ONLY =  0x10;
/** modifier indicating that the key is a special GUI navigation key (e.g. MENU_KEY) */
//public static final int NAVIGATION = 0x200;
/** modifier used when specifying hot keys - indicates that the key should not be underlined. */
public static final int INVISIBLE = (1 << 5);
/** For PRESSED KeyEvents, indicates that the key press is generated
because the key is being held down.*/
public static final int REPEATED = 0x100;
/** For PRESSED KeyEvents, indicates that the text element has a multi-character
string that should be considered as entered by the user.
**/
public static final int MULTICHARACTER = 0x200;
/** For PRESSED KeyEvents, indicates that the key that was typed should be considered
"conditional" - that is it may be superceded by another key.
**/
public static final int CONDITIONAL = 0x400;
/** For PRESSED KeyEvents, indicates that the key that was typed should be considered
to replace the previously sent key press. This <b>may</b> be used in combination with
CONDITIONAL to indicate that this replacement may also be conditional.
**/
public static final int REPLACEMENT = 0x800;
/** This indicates that the KeyEvent was generated by an InputMethod.
**/
public static final int FROM_INPUT_METHOD = 0x1000;

/**
 * If you call WindowSurface.postRawEvent() with this modifier, it will
 * definitely put it at the back of the queue and so will be guaranteed
 * to be delivered before another event posted afterwards. Without this
 * it cannot be determined whether it is placed at the front or back.
 */
public static final int POST_TO_BACK = 0x2000;

/** For PenEvents, indicates that the right button is being pressed. */
public static final int RIGHT_BUTTON = 0x8;
/** For PenEvents, indicates that the middle button is being pressed. */
public static final int MIDDLE_BUTTON = 0x10;
/** For PenEvents, indicates that the event is coming from another window (for drag and drop) */
public static final int FROM_OTHER_WINDOW = 0x20;
/** For PenEvents, indicates that the event is coming from another control*/
public static final int TRANSFERRED_PRESS = 0x40;


//##################################################################
}
//##################################################################

