/*
 * Created on Feb 11, 2007
 *
 * Michael L Brereton - www.ewesoft.com
 * 
 * 
 */
package eve.ui;

import eve.sys.Vm;
import eve.ui.event.ControlEvent;

/**
 * @author Michael L Brereton
 *
 */
//####################################################
public class IconMenuForm extends Form{

	IconMenu fmenu;
	protected SoftKeyBar softKeys;
	protected int moreKey = 0;
public String cancelButtonName = "Cancel";
public String menuButtonName = "Menu";
public String moreButtonName = "More";
public String backButtonName = "Back";
public static final int OPTION_NO_CONTEXT_MENU = 0x1;
public static final int OPTION_NO_CANCEL_BUTTON = 0x2;
int options;
	/**
	 * Create a FullScreenMenuForm using the provided IFullScreenMenu object.
	 * The getBestScroller() method will be called on the IFullScreenMenu
	 * to get the Control to add to the Form.
	 * @param m the IFullScreenMenu to use.
	 * @param options options for the FullScreenMenuForm 
	 */
	public IconMenuForm(IconMenu m,int options)
	{
		fmenu = m;
		addLast(m.getBestScroller());
		this.options = options;
	}
	public void make(boolean reMake)
	{
		SoftKeyBar skb = softKeys = new SoftKeyBar();
		int nextKey = 1;
		boolean hasContext = (options & OPTION_NO_CONTEXT_MENU) == 0;
		boolean backInContext = fmenu.needsExternalBack();
		if ((Vm.getParameter(Vm.VM_FLAGS) & Vm.VM_FLAG_HAS_BACK_BUTTON) != 0)
			backInContext = false;
		boolean cancelInContext = (options & OPTION_NO_CANCEL_BUTTON) == 0;
		if (fmenu.needsExternalMore()){
			moreKey = 3-nextKey++;
			skb.setKey(moreKey,moreButtonName,null,null);
		}
		if (backInContext && nextKey == 1){
			skb.setKey(3-nextKey++,backButtonName,null,null);
			backInContext = false;
		}
		if (cancelInContext && nextKey == 1){
			skb.setKey(3-nextKey++,cancelButtonName,null,null);
			cancelInContext = false;
		}
		final boolean bic = backInContext;
		final boolean cic = cancelInContext;
		if (backInContext || cancelInContext || hasContext){
			skb.setKey(3-nextKey++,"Menu",null,new Menu(){
				protected boolean prepareForDisplay(){
					removeAll();
					if (bic) addItem(backButtonName);
					if (cic) addItem(cancelButtonName);
					if (bic || cic) addItem("-");
					fmenu.addToContextMenu(this,fmenu.getSelectedItem());
					SoftKeyBar.fixMenu(this);
					return true;
				}
			});
		}
		setSoftKeyBarFor(null,skb);
		super.make(reMake);
	}
	/**
	 * Override this to handle the "Cancel" event. By default
	 * it exits with return value IDCANCEL.
	 */
	protected void handleCancel()
	{
		exit(IDCANCEL);
	}
	public void onControlEvent(ControlEvent ev)
	{
		if (ev.target == fmenu){
			if (ev.type == ControlEvent.CANCELLED)
				handleCancel();
			else if (ev.type == ControlEvent.STRUCTURE_CHANGED){
				if (moreKey != 0){
					Control c = softKeys.getKey(moreKey);
					if (c != null) {
						c.set(Disabled,!fmenu.hasMorePages());
						c.repaintNow();
					}
				}
			}
			super.onControlEvent(ev);
		}else
			super.onControlEvent(ev);
	}
	void handleContextMenuItem(MenuItem selected, int selectedItem)
	{
		fmenu.handleContextMenuItem(selected,selectedItem);
	}
	public void onSoftKey(int key, String action, MenuItem selected)
	{
		if (action.equals(cancelButtonName)){
			fmenu.doCancel();
		}else if (action.equals(moreButtonName)){
			fmenu.nextPage(true,true);
		}else if (action.equals(backButtonName)){
			fmenu.nextPage(false,true);
		}else if (selected != null){
			handleContextMenuItem(selected,fmenu.getSelectedItem());
		}
	}
	public IconMenu getIconMenu()
	{
		return fmenu;
	}
	/*
	public static void main(String args[])
	{
		Application.startApplication(args);
		PhoneMenu ph = new PhoneMenu();//IconMenu();
		IconMenu im = new IconMenu();
		//im.setDisplayMode(im.DISPLAY_LINES,80,28,24,24);
		//ph.dontShowMore = true;
		//IFullScreenMenu pm = im;//ph;
		IconMenu pm = ph;
		IImage eve = loadImage("eve/eve48.png");
		for (int i = 0; i<19; i++){
			pm.addItem(pm.makeItem(i%3 == 0 ? null : eve,"Item: "+(i+1)+"\nSecond Line"));
		}
		pm.setSelectedItem(0,false);
		Form f = new IconMenuForm(pm,0);
		f.windowFlagsToSet |= WindowConstants.FLAG_MAXIMIZE_ON_PDA;
		f.title = "Test Phone Menu";
		f.execute();
		System.exit(0);
	}
	*/
}
//####################################################
