package eve.ui;
import eve.fx.Color;
import eve.fx.Graphics;
import eve.fx.IImage;
import eve.fx.ImageBuffer;
import eve.fx.ImageRefresher;
import eve.fx.OnScreenImage;
import eve.fx.Rect;
import eve.util.ByteArray;
import eve.util.IntArray;

//##################################################################
public class ImageControl extends Canvas implements ImageRefresher{
//##################################################################

protected IImage image;
{
	isFullScrollClient = false;
	//modify(PreferredSizeOnly,0);
}

/**
* If an image is smaller than the control, this option will expand it.
**/
public static final int STRETCH = 0x1;
/**
* If an image is larger than the control, this option will shrink it.
**/
public static final int SHRINK = 0x2;
/**
* If an image is being stretched or shrunk, this option tells it not
* to change the aspect ration.
**/
public static final int FIT = 0x4;
/**
* When the setData() method is called with the encoded image bytes, this option tells
* the ImageControl to limit the decoded size if possible.
**/
public static final int LIMIT_DECODED_SIZE = 0x8;
/**
* This options tells the ImageControl to scale the image using a rough scaling instead
* of smooth scaling. Rough scaling is faster than smooth scaling.
**/
public static final int ROUGH_SCALING = 0x10;

/**
By default this is STRETCH|SHRINK|FIT - telling the control to stretch or shrink the image
into the display, but preserve the aspect ration.
**/
public int options = STRETCH|SHRINK|FIT;

/**
* If this is true then the control will have a background equal to the
* image background IF the image has a background color associated with 
* it (e.g. PNG images allow a background color to be specified).
**/
public boolean showImageBackground = true;
//===================================================================
public ImageControl(IImage what){setImage(what);}
//===================================================================
public IImage getImage() {return image;}
//===================================================================
public void setImage(IImage im) 
//===================================================================
{
	if ((image instanceof OnScreenImage) && (im != image))
		((OnScreenImage)image).changeRefresher(null,this);
	image = im;
	if (im instanceof OnScreenImage){
		((OnScreenImage)im).setRefresher(this);
	}
	repaintNow();
}

/*
//===================================================================
public void shown()
//===================================================================
{
	super.shown();
	if (image instanceof AniImage)
		((AniImage)image).shown();
}
		*/
/*
//-------------------------------------------------------------------
protected void formClosing()
//-------------------------------------------------------------------
{
	super.formClosing();
	if (image instanceof AniImage)
		((AniImage)image).closing();
}
		*/
//FIXME ImageControl.setData()
//===================================================================
public void setData(Object data)
//===================================================================
{
	if (data instanceof IImage) {
		setImage((IImage)data);
	}else if (data instanceof byte [] || data instanceof ByteArray){
		try{
			ByteArray ba = (data instanceof byte []) ? new ByteArray((byte [])data) : (ByteArray)data;
			if (((options & LIMIT_DECODED_SIZE) != 0) && width > 0 && height > 0){
				try{
					//setImage(new Image(ba,0,width,height));
				}catch(IllegalArgumentException e){
					//setImage(new Image(ba,0));
				}
			}else;
				//setImage(new Image(ba,0));
		}catch(Exception e){
			setImage(null);
		}
	}
}
ImageBuffer buffer = new ImageBuffer();

//===================================================================
Rect getScale(boolean fit,int width,int height)
//===================================================================
{
	Rect r = new Rect();
	int mw = this.width-borderWidth*2;
	int mh = this.height-borderWidth*2;
	if (!fit) {
		r.width = mw;
		r.height = mh;
	}else{
		double xscale = width == 0 ? 1 : (double)mw/width;
		double yscale = height == 0 ? 1: (double)mh/height;
		double scale = Math.min(xscale,yscale);
		r.width = (int)(scale*width);
		r.height = (int)(scale*height);
	}
	return r;
}
static IntArray pixels = new IntArray();
//static Image scaled;
/*
//-------------------------------------------------------------------
private Image fromIImage(IImage what)
//-------------------------------------------------------------------
{
	if (what == null || what.usesAlpha() || !(what instanceof Drawing)) return null;
	return ((Drawing)what).image;
}
*/
//===================================================================
public void doPaint(Graphics g,Rect area)
//===================================================================
{
/* - Rotates the image.
	Image image = new Image(this.image.getWidth(),this.image.getHeight(),Image.TRUE_COLOR);
	Graphics gr = new Graphics(image);
	gr.setColor(getBackground());
	gr.fillRect(0,0,image.getWidth(),image.getHeight());
	this.image.draw(gr,0,0,0);
	image = image.rotate(null,90);
*/
	int flags = getModifiers(true);
	if (!((flags & Invisible) == 0)) return;
	Color oldBack = backGround;
	Color iback = image == null ? null : image.getBackground();
	if (iback == null) iback = getBackground();
	backGround = iback;
	super.doPaint(g,area);
	int width = this.width-borderWidth*2;
	int height = this.height-borderWidth*2;
	if (image != null){
		//ewe.sys.Vm.debug(width+", "+height);
		int w = image.getWidth();
		int h = image.getHeight();
		//ewe.sys.Vm.debug(" and "+w+", "+h);
	  /*
		Graphics g2 = buffer.get(w,h,true);
		g2.setColor(backGround);
		g2.fillRect(0,0,width,height);
		image.draw(g2,0,0,0);
		*/
		Rect rect = new Rect(0,0,w,h);
		if (w > width || h > height)
			if ((options & SHRINK) != 0) 
				rect = getScale(((options & FIT) != 0),w,h);
		if (w < width || h < height)
			if ((options & STRETCH) != 0) 
				rect = getScale(((options & FIT) != 0),w,h);
		rect.x = borderWidth+(width-rect.width)/2;
		rect.y = borderWidth+(height-rect.height)/2;
		boolean full = rect.width == w && rect.height == h;
		if (area != null){
			if (area.width < width || area.height < height)
				full = false;
		}
		if (full){// && false){
			image.draw(g,rect.x,rect.y,0);//g.drawImage(image,null,null,new Rect(0,0,w,h),rect,0);
		}else
			g.drawImageData(image,rect.x,rect.y,rect.width,rect.height,area,0);
	}
	backGround = oldBack;
	//if (image != this.image) image.free();
}

//-------------------------------------------------------------------
protected void calculateSizes()
//-------------------------------------------------------------------
{
	preferredWidth = preferredHeight = 10;
	if (image != null) {
		preferredWidth = image.getWidth()+borderWidth*2;
		preferredHeight = image.getHeight()+borderWidth*2;
	}

	//System.out.println("Image ps: "+Geometry.toString(preferredSize));
}

//##################################################################
}
//##################################################################

