package eve.ui;
import java.util.Vector;

import eve.fx.Dimension;
import eve.fx.Graphics;
import eve.fx.Insets;
import eve.sys.Vm;
import eve.ui.Button;
import eve.ui.CellPanel;
import eve.ui.CheckBox;
import eve.ui.CheckBoxGroup;
import eve.ui.Choice;
import eve.ui.ComboBox;
import eve.ui.Control;
import eve.ui.Input;
import eve.ui.InputForm;
import eve.ui.Label;

//##################################################################
public class InputStack extends CellPanel {
//##################################################################
public boolean doPopupInput = false;
public boolean doubleLined = false;
public int inputLength = 10;
{
	//borderWidth = 2;
	//borderStyle = EDGE_ETCHED;
	defaultTags.set(TAG_INSETS,new Insets(1,1,1,1));
	columns = 1;
}
public Vector controls = new Vector(), prompts = new Vector();

private Vector extras;

public void addExtraControlToLastAdded(Control c)
{
	if (c == null) return;
	int idx = controls.size()-1;
	if (idx < 0) return;
	if (extras == null) extras = new Vector();
	extras.setSize(idx);
	extras.add(c);
}
//===================================================================
public void make(boolean remake)
//===================================================================
{
	int col = columns;
	for (int i = 0; i<controls.size(); i++){
		col--;
		boolean last = false;
		if (col == 0) {
			last = true;
			col = columns;
		}
		Control c = (Control)controls.get(i);
		String p  = (String)prompts.get(i);
		Control ca = c;
		if (extras != null && extras.size() >= i+1){
			Control x = (Control)extras.get(i);
			if (x != null){
				CellPanel cc = new CellPanel();
				cc.addNext(c).setCell(STRETCH).setControl(HFILL|WEST);
				cc.addLast(x).setCell(DONTSTRETCH);
				ca = cc;
			}
		}
		Control l = null;
		if (p != null){
			int hAlign = EAST;
			if (doubleLined) hAlign = WEST;
			l = getLabel(p);
			addNext(l,doubleLined).setCell(DONTSTRETCH).setControl(DONTFILL|hAlign);
		}
		addNext(ca,last).setCell(HSTRETCH).setControl(HFILL|WEST);
		if (p == null && !doubleLined) ca.setTag(TAG_SPAN,new Dimension(2,1));
		c.prompt = p;
		c.setPromptControl(l);
		if (l != null) 
			if (l instanceof Label) ((Label)l).control = c;
	}
	super.make(remake);
}

//==================================================================
public void add(Control c,String prompt)
//==================================================================
{
	controls.add(c);
	prompts.add(prompt);
	if (!(c instanceof InputStack)) c.columns = inputLength;
}

//===================================================================
public static String appendToPrompt(String promptWithHotkey,String moreText)
//===================================================================
{
	int idx = promptWithHotkey.indexOf('$');
	if (idx == -1 || idx != promptWithHotkey.length()-2) return promptWithHotkey+moreText;
	else return promptWithHotkey.substring(0,idx)+moreText+promptWithHotkey.substring(idx);
}
//------------------------------------------------------------------
protected Control getLabel(String text)
//------------------------------------------------------------------
{
	return new Label(text);
}
//------------------------------------------------------------------
protected void popup(Input what) {if (doPopupInput) InputForm.attach(what);}
//------------------------------------------------------------------
protected ComboBox addComboBox(ComboBox cb,String p)
//------------------------------------------------------------------
{
	add(cb,p);
	popup(cb.input);
	return cb;
}
//==================================================================
public Label addLabel(String prompt) {Control l = getLabel(prompt); add(l,null); return (Label)l;}
public Input addInput(String prompt,String initial) {Input i = new Input(initial); add(i,prompt); popup(i);return i;}
public CheckBox addCheckBox(String prompt) {return addCheckBox(prompt,null,false);}
public CheckBox addCheckBox(String prompt,CheckBoxGroup group,boolean state) 
{CheckBox mc = new CheckBox(); add(mc,prompt); if (group != null) mc.setGroup(group); mc.setState(state); return mc;}
public Choice addChoice(String prompt) {Choice mc = new Choice(); add(mc,prompt); return mc;}
public Choice addChoice(String prompt,String [] choices,int initialChoice) {Choice mc = (Choice)addChoice(prompt); mc.set(choices,initialChoice); return mc;}
public Button addButton(String prompt) {Button mb = new Button(prompt); add(mb,null); return mb;}
public InputStack addInputStack(String prompt,int columns) {InputStack is = new InputStack(); add(is,prompt); is.columns = columns; is.borderStyle = Graphics.EDGE_SUNKEN; return is;}
public CellPanel addPanel(String prompt,boolean spanBoth) {CellPanel mp = new CellPanel(); if (spanBoth) {if (prompt != null) add(getLabel(prompt),null); add(mp,null);} else add(mp,prompt); return mp;}
//public TextArea addTextArea(String prompt,int rows,int columns,boolean scrollable) {if (prompt != null) add(getLabel(prompt),null); TextArea mt = new TextArea(rows,columns); if (scrollable) add(mt.getScrollablePanel(),null); else add(mt,null); return mt;}
public ComboBox addComboBox(String prompt) {return addComboBox(new ComboBox(),prompt);}
public ComboBox addComboBox(String prompt,String [] choices,int initialChoice) {return addComboBox(new ComboBox(choices,initialChoice),prompt);}
public ComboBox addComboBox(String prompt,String [] choices,String initialChoice) {return addComboBox(new ComboBox(choices,initialChoice),prompt);}
public CheckBoxGroup addChecks(String prompt,String choices [],int columns)
{
	CellPanel p2 = new CellPanel();
	CheckBoxGroup cg = new CheckBoxGroup();
	cg.makePanel(choices,p2,columns);
	add(p2,prompt);
	return cg;
}
/**
 * This converts a name with underscores to a prompt with capital letters and spaces where
 * the underscores were.
 * @param name 
 * @return The converted name.
 */
//===================================================================
public static String nameToPrompt(String name)
//===================================================================
{
	if (name == null) return null;
	char [] all = Vm.getStringChars(name);
	char [] another = new char[all.length*2];
	boolean cap = true;
	boolean hasUnder = false;
	int j = 0;
	for (int i = 0; i<all.length; i++){
		if (all[i] == '$') return new String(another,0,j);
		if (all[i] == '_') {
			hasUnder = true;
			another[j++] = ' ';
			cap = true;
		}else{
			if (cap) another[j++] = Character.toUpperCase(all[i]);
			else if (i != 0 && !hasUnder && Character.isUpperCase(all[i])){
				if (Character.isUpperCase(all[i-1])) another[j++] = all[i];
				else {
					another[j++] = ' ';
					another[j++] = all[i];
				}
			}else another[j++] = all[i];
			cap = false;
		}
	}
	return new String(another,0,j);
}
//##################################################################
}
//##################################################################

