package eve.ui;

import eve.fx.Point;
import eve.ui.event.PenEvent;

/**
* This class provides an mInput along with a drop-down tool menu, that is meant
* to aid the input in some way. For example it may contain special hard to remember
* codes or it may provide some kind of auxillary input.
**/
//##################################################################
public class InputWithTools extends Holder{
//##################################################################

public Input input;
public ArrowButton tool;
//-------------------------------------------------------------------
protected void setup()
//-------------------------------------------------------------------
{
	tool = new ArrowButton(Right){
		public void doPenPress(Point p){startDropMenu(p);}
		public boolean willShowFrame(PenEvent ev)
		{
			if ((ev.type == ev.PEN_DOWN) && !menuIsActive()) return true;
			return super.willShowFrame(ev);
		}
		public void popupMenuEvent(Object selected)
		{
			if (selected instanceof MenuItem) handleMenu((MenuItem)selected);
		}
	};
	addNext(input);
	addLast(tool).setCell(VSTRETCH);
	tool.modify(NoFocus,TakesKeyFocus);
}
//===================================================================
public void setToolsMenu(Menu m)
//===================================================================
{
	tool.setMenu(m);
}
//===================================================================
public void make(boolean remake)
//===================================================================
{
	if (made && !remake) return;
	int h = input.getPreferredSize(null).height;	
	tool.setPreferredSize(15,h);
	tool.setControl(VCONTRACT|HCONTRACT|NORTHWEST);
	super.make(remake);
}

/**
 * Create an mInputWithTools but with no tool menu selected. Call setToolsMenu() to set the tools.
 */
//===================================================================
public InputWithTools()
//===================================================================
{
	input = new Input();
	setup();
}
/**
 * Create an mInputWithTools with a specific tools menu.
 */
//===================================================================
public InputWithTools(Menu m)
//===================================================================
{
	this();
	setToolsMenu(m);
}
/**
 * Create an mInputWithTools with a specific tools menu derived from the toolOptions.
 */
//===================================================================
public InputWithTools(String [] toolOptions)
//===================================================================
{
	this();
	setToolsMenu(new Menu(toolOptions,""));
}
/**
 * This requests the text to insert into the mInput for the specified MenuItem.
 * It will be called when a tool item is selected via the handleMenu() method. 
 * @param forMenuItem The selected item.
 * @return Text to insert into the mInput, or null for no insertion.
 */
//-------------------------------------------------------------------
protected String getInsertionText(MenuItem forMenuItem)
//-------------------------------------------------------------------
{
	return forMenuItem.toString();
}
/**
 * Use this to insert text into the mInput at the current cursor position, overwriting
 * any text that might be selected.
 * @param toInsert The text to insert.
 */
//-------------------------------------------------------------------
protected void insertIntoInput(String toInsert)
//-------------------------------------------------------------------
{
	if (toInsert != null) {
		input.replaceSelection(toInsert);
		input.update();
		input.updateText(true);
	}
}
/**
 * This gets called when a tool item is selected. By default it
 * will call getInsertionText() and then call insertIntoInput(). You can override it however
 * to do anything else.
 * @param forMenuItem The item selected.
 */
//-------------------------------------------------------------------
protected void handleMenu(MenuItem forMenuItem)
//-------------------------------------------------------------------
{
	insertIntoInput(getInsertionText(forMenuItem));
}
//===================================================================
public String getText() {return input.getText();}
//===================================================================
public void setText(String text) {input.setText(text);}
//===================================================================

/*
public static void main(String args[])
{
	ewe.sys.Vm.startEwe(args);
	Form f = new Form();
	mInputWithTools mi;
	f.addNext(mi = new mInputWithTools(new String[]{"One","Two","-","Three"}));
	f.addLast(new mInputWithTools());
	f.execute();
	ewe.sys.Vm.exit(0);
}
*/
//##################################################################
}
//##################################################################

