/*
 * Created on Aug 13, 2005
 *
 * Michael L Brereton - www.ewesoft.com
 * 
 * 
 */
package eve.ui;

import eve.fx.Color;
import eve.fx.Dimension;
import eve.fx.FontMetrics;
import eve.fx.FormattedTextSpecs;
import eve.fx.Graphics;
import eve.fx.Metrics;
import eve.fx.Point;
import eve.fx.Rect;
import eve.sys.Event;
import eve.ui.event.PenEvent;
import eve.util.CharArray;
import eve.util.StringList;
import eve.util.mString;

/**
 * @author Michael L Brereton
 *
 */
//####################################################
public class Label extends Control {

	public FormattedTextSpecs format = null;
	public boolean isActive = false;
	public int alignment = Left;
	public int anchor = NORTHWEST;
	public int spacing = 2;
	/**
	* This is used to associate another Control with this label. This is usually used when
	* a label is acting as a prompt for another control (e.g. an mInput). If this label has
	* a hot key associated with it, then pressing that hot key will give focus to the
	* associated control and, if that control is a button, will cause it to fire.
	**/
	public Control control;
	protected boolean isData = false;
	{
		modify(NoFocus,0);
		rows = columns = 0;
		//defaultAddMeCellConstraints = mPanel.DontStretch;
		//defaultAddMeControlConstraints = mPanel.Fill;
		//defaultAddMeAnchor = mPanel.NorthWest;
	}
//	===================================================================
	public Label(int rows,int columns) {this.rows = rows; this.columns = columns;}
	public Label(String what) {this(what,true);}
//	===================================================================
	public Label(String what,boolean containsHotKey)
//	===================================================================
	{
		if (containsHotKey){
			text = Gui.getTextFrom(what); 
			if (text == null) text = "";
			char hot = Gui.getHotKeyFrom(what);
			if (hot != 0) setHotKey(0,hot);
		}else{
			text = what;
			if (text == null) text = "";
		}
	}
	protected int lineHeight = 0;
//	-------------------------------------------------------------------
	protected int getScreenRows()
//	-------------------------------------------------------------------
	{
		if (lineHeight == 0){
			FontMetrics fm = getFontMetrics();
			lineHeight = fm.getHeight()+fm.getLeading();
		}
		return (height-spacing*2)/lineHeight;
	}
//	-------------------------------------------------------------------
	protected void calculateSizes()
//	-------------------------------------------------------------------
	{
		Dimension r = Dimension.getCached(0,0);
		try{
			if (text == null || text.length() == 0) text = " ";
			//if (text.length() == 0) text = " "; <= Add this in.
		 	if (rows != 0 && columns != 0) {
				// v--- Add this in.
				//r = Gui.getAverageSize(getFontMetrics(),rows,columns,spacing+borderWidth,spacing+borderWidth);
		 		Metrics.getAverageSize(getFontMetrics(),rows,columns,spacing,spacing,r);
			}
			else if (text.trim().length() == 0){
				preferredWidth = spacing+borderWidth*2;
				preferredHeight = getFontMetrics().getHeight()+spacing+borderWidth*2;
				return;
			}else {
				String[] got = mString.split(text,'\n');
				Metrics.
				getSize(getFontMetrics(),got,spacing+borderWidth,spacing+borderWidth,r);
			}
			preferredWidth = r.width; preferredHeight = r.height;
		}finally{
			
			r.cache();
		}
	}

	//protected static String [] _lines = new String[1];
//	===================================================================
	//public String [] getLines() {_lines[0] = text; return _lines;}
	
	public StringList getLines(StringList destination)
	{
		if (destination == null) destination = new StringList();
		CharArray ca = makeHot(text,(CharArray)getCached(CharArray.class));
		destination.add(ca);
		cache(ca);
		return destination;
	}
//	===================================================================
	protected int startLine = 0, xPos = 0;
//	==================================================================
	public void setText(String text) {super.setText(text); repaintNow();}
//	==================================================================
	protected void doPaintData(Graphics g) { if (isData) doPaint(g,null);}
//	===================================================================
	public void doPaint(Graphics g,Rect area)
//	===================================================================
	{
		int flags = getModifiers(true);
		if (!((flags & Invisible) == 0)) return;
		doBackground(g);
		doBorder(g);
		StringList sl = getLines(((StringList)getCached(StringList.class)).zero());
		Rect d = getDim((Rect)getCached(Rect.class));
		try{
			int gap = (spacing*2)+(borderWidth*2);
			d.width -= gap; d.height -= gap;
			d.x = d.y = gap/2;
			g.setColor((flags & Disabled) == 0 ? getForeground() : Color.DarkGray);
			g.setFont(getFont());
			FontMetrics fm = getFontMetrics();
			//d.y -= (fm.getHeight()+fm.getLeading())*startLine;
			d.x -= xPos;
			Object [] ar = sl.toTempSubStringArray(true);
			//System.out.println(mString.toString(ar));
			Metrics.drawText(g,fm,ar,d,alignment,anchor,startLine,startLine+getScreenRows()+1,format);
		}finally{
			cache(d);
			cache(sl);
		}
		//lines.length);//
	}

	public void onEvent(Event ev)
	{
		if (ev instanceof PenEvent || control == null) super.onEvent(ev);
		else control.onEvent(ev);
	}
	public void onPenEvent(PenEvent ev)
	{
		if (checkPenTransparent(ev)) return;
		if (control == null) {
			super.onPenEvent(ev);
			return;
		}
		Point p = new Point(ev.x,ev.y);
		ev.x = ev.y = 0;
		if (!isOnMe(p)) ev.x = ev.y = -1;
		control.onLabelPenEvent(ev);
	}
	public void penClicked(Point where)
	{
		if (isActive) notifyAction();
	}
	public void doAction(int how)
	{
		if (control != null) {
			control.takeFocus(how);
			if (control instanceof ButtonControl){
				control.doAction(ByKeyboard);
				control.notifyAction();
			}
		}
	}

}

//####################################################
