package eve.ui;
import eve.fx.Color;
import eve.fx.FontMetrics;
import eve.fx.Graphics;
import eve.fx.ImageBuffer;
import eve.fx.Rect;
import eve.sys.Handle;
import eve.sys.HandleStoppedException;
import eve.sys.Time;
//##################################################################
public class ProgressBar extends Control{
//##################################################################

int drawWidth = 0;
float progress = 0;
int percent = 0;
/**
 * Set this to true so that the progress is always shown to be -1, which
 * makes the bar bounce back and forth instead of progressing.
 */
public boolean alwaysBounce = false;
public boolean showPercent = true;
public boolean showTimeLeft = false;
/**
* If you set this to a non null value, then a message will be displayed along
* with the percentage value.
**/
public String doing = null;

public long timeStarted;

public Color incompleteColor = Color.White;
public Color barColor = new Color(0,0,0x80);

{
	borderStyle = GuiStyle.inputEdge|BF_RECT;
}

//-------------------------------------------------------------------
protected void calculateSizes()
//-------------------------------------------------------------------
{
	preferredWidth = 100; preferredHeight = getFontMetrics().getHeight()+6;
}
//===================================================================
public void reset()
//===================================================================
{
	drawWidth = 0;
	doing = null;
	repaintNow();
}

private long lastSet = System.currentTimeMillis();

public int bounceSteps = 10;
public int bounceTimeMillis = 2000;

public void set(float progress,String doing)
{
	this.doing = doing;
	set(progress,0);
	repaintNow();
}
/**
 * Set the bar progress to the specified value.
 * @param progress A value between 0.0 and 1.0 or -1.
 */
//===================================================================
public void set(float progress)
//===================================================================
{
	set(progress,0);
}
/**
 * Set the bar progress to the specified value.
 * @param progress A value between 0.0 and 1.0 or -1.
* @param startTime A value indicating when the operation started. This
* may be zero if it is unknown.
*/
//===================================================================
public void set(float progress,long startTime)
//===================================================================
{
	if (alwaysBounce) progress = -1;
	timeStarted = startTime;
	long now = System.currentTimeMillis();
	float lastProgress = this.progress;
	this.progress = progress;
	float value = progress;
	if (value < 0){
		if (now-lastSet >= bounceTimeMillis/bounceSteps){
			lastSet = now;
			if (percent >= 0) percent = -1;
			else percent--;
			if (percent <= (-bounceSteps*2)) percent = -2;
			repaint();
		}else if (percent >= 0) percent = -1;
		return;
	}
	percent = (int)(value*100);
	int newWidth = (int)((float)(width-4)*value);
	if (newWidth != drawWidth || lastProgress < 0){
		drawWidth = newWidth;
		repaint();
		lastSet = now;
	}else if ((now-lastSet >= 1000) && showTimeLeft){
		repaint();
		lastSet = now;
	}
}
static ImageBuffer buffer;

Time now;
StringBuffer sb;
//-------------------------------------------------------------------
String timeLeft()
//-------------------------------------------------------------------
{
	if (timeStarted == 0 || progress <= 0) return null;
	if (now == null) now = new Time();
	now.setToCurrentTime();
	long taken = now.getTime()-timeStarted;
	int need = (int)(((taken/progress)-taken)/1000);
	if (need < 0) return null;
	int hours = need/(60*60);
	need = need%(60*60);
	int minutes = need/60;
	need = need%60;
	int seconds = need;
	if (sb == null) sb = new StringBuffer();
	sb.setLength(0);
	if (hours != 0){
		sb.append(hours); sb.append(" hr. ");
	}
	if (minutes != 0 || hours != 0){
		sb.append(minutes); sb.append(" min. ");
	}
	if (hours == 0){
		sb.append(seconds); sb.append(" sec. ");
	}
	sb.append("left.");
	return sb.toString();
}
//===================================================================
public void doPaint(Graphics g,Rect area)
//===================================================================
{
	//System.out.println(System.identityHashCode(Thread.currentThread()));
	Rect r2 = Rect.getCached();
	//synchronized(Gui.staticPaintLock)
	try{
		Rect r = getDim(r2);
		int flags = getModifiers(true);
		int blockWidth = 0, blockPos = 0;
		if ((borderStyle & BF_SOFT) != 0) doBackground(g);
		g.draw3DRect(
				getDim(r2),
				borderStyle, //Graphics.EDGE_SUNKEN|Graphics.BDR_OUTLINE,
				(flags & DrawFlat) != 0,
				incompleteColor,
				borderColor);
				
		if (percent < 0){
			double w = (r.width-4)/(double)(bounceSteps);
			int p = (-percent)-1;
			blockWidth = (int)(r.width-4-((bounceSteps-1)*w)-1);
			if (p >= bounceSteps) p = bounceSteps*2-p-2;
			if (p < 0) p = 0;
			blockPos = (int)(p*w+2);
			g.setColor(barColor);
			g.fillRect(blockPos,2,blockWidth,r.height-4);
		}else{
			if (drawWidth > r.width-4) drawWidth = r.width-4;
			else if (drawWidth < 0) drawWidth = 0;
			if (false && ((borderStyle & BF_SOFT) != 0) && !showPercent && doing == null){
		/*
				r = new Rect().set(Rect.buff);
				if (buffer == null) buffer = new ImageBuffer();
				Graphics gi = buffer.get(drawWidth,r.height-4,true);
				gi.setColor(incompleteColor);
				gi.fillRect(0,0,drawWidth,r.height-4);
				int rw = 12;
				int num = drawWidth/rw;
				int h = r.height-4, w = rw-1;
				for (int i = 0; i<num; i++)
						gi.draw3DRect(Rect.buff.set(rw*i,0,w,h),borderStyle,false,barColor,null);
				g.drawImage(buffer.image,2,2);
		*/
			}else{
				g.setColor(barColor);
				g.fillRect(2,2,drawWidth,r.height-4);
			}
		}
		String toPrint = new String();
		if (showPercent && percent >= 0) toPrint += percent+"%"+ (doing == null ? "" : " - ");
		if (doing != null) toPrint += doing;
		if (showTimeLeft){
			String left = timeLeft();
			if (left != null){
				if (toPrint.length() != 0) toPrint += " - ";
				toPrint += left;
			}
		}
		if (toPrint.length() != 0) {
			toPrint = toPrint.trim();
			FontMetrics fm = getFontMetrics();
			String s = toPrint;
			int w = fm.getTextWidth(s);
			int h = fm.getAscent()+fm.getDescent();
			int xpos = (doing == null ? (r.width-w)/2 : 4);
			g.setFont(fm.getFont());
			Rect old = new Rect();
			if (percent >= 0){
				g.setColor(incompleteColor);
				Rect oc = g.reduceClip(0,0,drawWidth+2,r.height,old);
				g.drawText(s,xpos,((r.height-h)/2)/*+fm.getAscent()*/);//,(d.height-h)/2);
				g.restoreClip(oc);
				oc = g.reduceClip(drawWidth+2,0,r.width-drawWidth-4,r.height,old);
				g.setColor(barColor);
				g.drawText(s,xpos,((r.height-h)/2)/*+fm.getAscent()*/);//,(d.height-h)/2);
				g.restoreClip(oc);
			}else{
				g.setColor(barColor);
				g.drawText(s,xpos,((r.height-h)/2));
				g.setColor(incompleteColor);
				Rect oc = g.reduceClip(blockPos,0,blockWidth,r.height,old);
				g.drawText(s,xpos,((r.height-h)/2));
				g.restoreClip(oc);
			}
		}
	}finally{
		r2.cache();
	}
}
/**
 * This is called by the monitor() method when the monitored Handle
 * has changed. By default it updates the progress and "doing" values.
 * @param h 
 */
protected void monitoredHandleChanged(Handle h)
{
	set(h.progress,h.doing);
	h.changeAccepted();
}
/**
 * This will start a Thread that will monitor the Handle provided and
 * update the bar accordingly. When the Handle is marked as Stopped
 * the thread will exit.
 * @param h the Handle to monitor
 */
public void monitor(final Handle h)
{
	new Thread(){
		public void run(){
			set(0,0);
			while(true) try{
				monitoredHandleChanged(h);
				h.waitOn(h.Changed);
			}catch(HandleStoppedException e){
				break;
			}catch(InterruptedException e){
				break;
			}
		}
	}.start();
}
//##################################################################
}
//##################################################################

