package eve.ui;
import eve.fx.Color;
import eve.fx.Graphics;
import eve.fx.Point;
import eve.fx.Rect;
import eve.ui.event.ScrollEvent;

//##################################################################
public class ScrollTrack extends TrackControl{
//##################################################################
public int minThumb = 10;
//==================================================================
public void doPaint(Graphics g,Rect area)
//==================================================================
{
	int flags = getModifiers(true);
	doBackground(g);
	boolean flat = ((flags & DrawFlat) != 0);
	boolean soft = ((GuiStyle.buttonEdge & BF_SOFT) == BF_SOFT);
	boolean etched = ((GuiStyle.standardEdge & EDGE_ETCHED) == EDGE_ETCHED);
	if (GuiStyle.globalPalmStyle){
		g.setColor(Color.Black);
		if (type == Horizontal) {
			g.drawLine(0,height/2,width-1,height/2);
			g.drawLine(0,height/2-2,0,height/2+2);
			g.drawLine(width-1,height/2-2,width-1,height/2+2);
		}else{
			g.drawLine(width/2,0,width/2,height-1);
			g.drawLine(width/2-2,0,width/2+2,0);
			g.drawLine(width/2-2,height-1,width/2+2,height-1);
		}
	}else {
		synchronized(Rect.gBuffer1){
			g.draw3DRect(
				getDim(Rect.gBuffer1),
				soft ? GuiStyle.buttonEdge & ~BF_BUTTON : (etched ? GuiStyle.standardEdge :
				Graphics.EDGE_SUNKEN|((GuiStyle.standardEdge & BDR_OUTLINE) != 0 ? BDR_OUTLINE : 0)) ,
				flat,
				soft ? getBackground() : Color.LighterGray,
				Color.Black);
			g.draw3DRect(
				getThumbArea(Rect.gBuffer1),
				GuiStyle.buttonEdge,
				flat,
				flat ? getForeground() : getBackground(),
				Color.Black);
		}
	}
}
/**
Create a new ScrollTrack.
@param type Either IScroll.Horizontal or IScroll.Vertical
*/
//===================================================================
public ScrollTrack(int type) 
//===================================================================
{
	this(type, 0);
}

/** The options specified in the constructor.**/
protected int options;

/**
Create a new ScrollTrack.
@param type Either IScroll.Horizontal or IScroll.Vertical
@param trackOptions Any of the IScroll.OPTION_XXX values OR'ed together
*/
//===================================================================
public ScrollTrack(int type, int trackOptions)
//===================================================================
{
	this.type = type;
	this.options = trackOptions;
	//if ((options & OPTION_INDICATOR_ONLY) != 0)
		//modify(SmallControl|DrawFlat,0);
}

//==================================================================
protected void calculateSizes()
//==================================================================
{
	boolean etched = ((GuiStyle.standardEdge & EDGE_ETCHED) == EDGE_ETCHED);
	int emt = 14, et = 10;
	if (isDoubleSized()) {
		emt *= 2;
		et *= 2;
	}
	if (etched && minThumb < emt) minThumb = emt;
	if (minThumb < et) minThumb = et;
	int w = 30, h = 15;
	if (hasModifier(SmallControl,true)) {
		w = 15;
		h = 10;
	}else if (isDoubleSized()){
		w *= 2;
		h *= 2;
	}
	if (type == Horizontal) {
		preferredWidth = w;
		preferredHeight = h;
	}else if (type == Vertical) {
		preferredWidth = h;
		preferredHeight = w;
	}
}

Point thumbPos = new Point();
int thumbThickness = 10;

public Rect getThumbArea(Rect r)
{
	r = Rect.unNull(r);
	if (thumbThickness == 0) return r.set(-100,-100,1,1);
	int w = thumbThickness, h = thumbThickness;
	/*
	if (isDoubleSized()) {
		w *= 2;
		h *= 2;
	}*/
	if (type == Horizontal) h = getDim(r).height;
	else w = getDim(r).width;
	r.set(thumbPos.x,thumbPos.y,w,h);
	boolean etched = ((GuiStyle.standardEdge & EDGE_ETCHED) == EDGE_ETCHED);
	if (etched) {
		if (type == Horizontal || true){
			r.y+=2;
			r.height-=4;
		}
		if (type == Vertical || true){
			r.x+=2;
			r.width-=4;
		}
	}
	return r;
}
public void setThumbPos(int x,int y)
{
	Rect r = getDim(getCachedRect());
	if (type == Horizontal) {
		y = 0;
		int max = r.width-thumbThickness;
		if (x > max) x = max;
		if (x < 0) x = 0;
	}else{
 		x = 0;
		int max = r.height-thumbThickness;
		if (y > max) y = max;
		if (y < 0) y = 0;
	}
	thumbPos.move(x,y);
	cache(r);
}
public void generate(int what,int value)
{
	postEvent(new ScrollEvent(this,what,value));
	//System.out.println(what+","+value);
}

protected int visible, actual, current, scrollable;

//==================================================================
public void recalculate()
//==================================================================
{
	Rect r = getRect(getCachedRect());
	//if (visible >= actual) thumbThickness = 0;
	
	if (actual <= visible && current > 0){
		actual += current;
		//System.out.println("Compensating: "+visible+", "+actual);
	}
	scrollable = actual-visible;
	int max = r.width;
	if (type == Vertical) max = r.height;
	//System.out.println(visible+" "+actual+" "+current+" "+max);
	int et = 10;
	//if (isDoubleSized()) et *= 2;
	if (scrollable == 0) scrollable = et;
	if (actual == 0) actual = et;
	thumbThickness = (visible*max)/actual;
	//System.out.println(visible+" "+actual+" "+current);
	if (thumbThickness < minThumb) thumbThickness = minThumb;
	int pos = (int)(((double)current/scrollable)*(max-thumbThickness));
	if (!amTracking)
		if (type == Vertical) setThumbPos(0,pos);
		else setThumbPos(pos,0);
	cache(r);
}
//==================================================================
public void resizeTo(int width,int height)
//==================================================================
{
	super.resizeTo(width,height);
	recalculate();
}
/**
 * Set the postion and size of the "thumb" control in the track, using units of the users choice.
 * For example, if the ScrollTrack is being used to scroll a number of lines of text
 * in a Text editor - then each unit may represent one line of text. If some kind of Image
 * is being scrolled, then each unit may represent one pixel.
 * @param vs (visible) this represents the number of units that are "visible" on screen.
 * For instance the number of lines that are displayed on screen of a text editor,
 * or pixel lines in an image. 
 * @param ac (actual) this represents the total number of units contained in the data being displayed.
 * For instance the total number of lines in a text editor, or total number of pixel
 * lines in an image.
 * @param cur (current) this represents the current unit position being viewed. 
 * For instance the line number of the line at the top of the display in a text editor,
 * or the pixel line number (Y-co-ordinate) of the line at the top of the display in an
 * image.  
 * @param repaint true to have the control repaint itself, false if not.
 */
//==================================================================
public void setPositions(int vs,int ac,int cur,boolean repaint)
//==================================================================
{
	if (visible == vs && actual == ac && current == cur) return;
	if (vs > ac) vs = ac;
	visible = vs;
	actual = ac;
	current = cur;
	recalculate();
	if (!amTracking && repaint) repaintNow();
}
//==================================================================
public int getTrackTo()
//==================================================================
{
	int max = width, cp = thumbPos.x;
	if (type == Vertical) {
		max = height;
		cp = thumbPos.y;
	}
	int avail = max-thumbThickness;
	if (avail == 0 || cp == 0) return 0;
	if (cp == avail) return scrollable;
	return (int)(((double)cp/avail)*scrollable);
}
//##################################################################
}
//##################################################################

