package eve.ui;

import eve.fx.Dimension;
/**
* A SingleContainer is used to hold a single component which can be
* changed dynamically. It is difficult to change the contents of a 
* CellPanel dynamically, but any elements which do need to be changed
* can have a SingleContainer placed in it, and the contents of the
* SingleContainer can change. Call setControl() to change the contents.
**/
//##################################################################
public class SingleContainer extends Container{
//##################################################################

public Control client;
//===================================================================
public void resizeTo(int width,int height) 
//===================================================================
{
	super.resizeTo(width,height);
	if (client != null) client.setRect(borderWidth,borderWidth,width-borderWidth*2,height-borderWidth*2);
}
/**
* This changes the control. It does a make() on the control after it has
* been added. It will <b>not</b> repaint the control after - call repaintNow()
* for that.
* @param who The new control to display.
**/
//===================================================================
public void setControl(Control who)
//===================================================================
{
	if (client != null) remove(client);
	add(who);
	who.make(false);
	client = who;
	client.setRect(borderWidth,borderWidth,width-borderWidth*2,height-borderWidth*2);
}
/**
* This changes the control and optionally repaints itself. It does a make() on the control after it has
* been added.
* @param who The new control to display.
* @param repaint if true a repaint will be done.
*/
//===================================================================
public void setControl(Control who,boolean repaint)
//===================================================================
{
	setControl(who);
	if (repaint) repaintNow();
}
/**
* This changes the control and relays out a parent control
* to account for the possibly changed size of the SingleContainer. 
* It does a make() on the control after it has
* been added.
* @param who The new control to display.
* @param parentToRelayout if not null then relayout(true) will be called on
* that Parent.
*/
public void setControl(Control who, Container parentToRelayout)
{
	setControl(who);
	if (parentToRelayout != null)
		parentToRelayout.relayout(true);
}
//===================================================================
public Dimension getPreferredSize(Dimension ps)
//===================================================================
{
	if (client != null) {
		ps = client.getPreferredSize(ps);
		ps.width += borderWidth*2;
		ps.height += borderWidth*2;
		return ps;
	}
	if (ps == null) ps = new Dimension(0,0);
	return ps.set(borderWidth*2,borderWidth*2);
}
//===================================================================
public void make(boolean reMake)
//===================================================================
{
	if (client != null) client.make(reMake);
}

//===================================================================
public void dismantle(Control downTo)
//===================================================================
{
	if (this == downTo) return;
	super.dismantle(downTo);
	client = null;
}
//===================================================================
public void formClosing()
//===================================================================
{
	super.formClosing();
}
public ControlIterator getSubControls(ControlIterator it)
{
	return ControlIterator.unNull(it).setForSingleControl(client);
}


//##################################################################
}
//##################################################################

