package eve.ui;
import java.util.Vector;

import eve.data.DateTimeUtils;
import eve.data.ITime;
import eve.data.PlainTime;
import eve.data.Value;
import eve.fx.Color;
import eve.fx.Point;
import eve.sys.Time;
import eve.ui.event.ControlEvent;
import eve.util.Range;
import eve.util.Timeable;
import eve.util.mVector;
/**
This inputs a time of day value using a set of UpDownInputs.
**/
//##################################################################
public class TimeUpDownInput extends Holder implements Timeable, ITime{
//##################################################################

UpDownInput hour, minute, second, ampm;
private static Range hourRange = new Range(1,12);
private static Range hour24Range = new Range(0,23);
private static Range minuteRange = new Range(0,59);
private static Vector ampmRange = mVector.newVector(new String[]{"am","pm"});
protected boolean is24Hours = false;

//-------------------------------------------------------------------
Label getLabel(String what)
//-------------------------------------------------------------------
{
	Label l = new Label(what);
	l.spacing = 1;
	setControl(DONTFILL|CENTER);
	return l;
}
//===================================================================
public TimeUpDownInput(boolean showSeconds, boolean hours24)
//===================================================================
{
	is24Hours = hours24;
	addNext(hour = new UpDownInput(2));
	addNext(getLabel(":"));
	addNext(minute = new UpDownInput(2));
	if (showSeconds){
		addNext(getLabel(":"));
 		addNext(second = new UpDownInput(2));
	}	
	if (!hours24){
		addNext(getLabel(" "));
 		addNext(ampm = new UpDownInput(2));
	}
	hour.integerValues = hours24 ? hour24Range : hourRange;
	hour.integerDigits = 2;
	hour.zeroFillInteger = hours24;
	minute.integerValues = minuteRange;
	minute.integerDigits = 2;
	minute.zeroFillInteger = true;
	if (ampm != null) {
		ampm.textValues = ampmRange;
		ampm.hasSplit = false;
	}
	if (second != null){
 		second.integerValues = minuteRange;
		second.integerDigits = 2;
		second.zeroFillInteger = true;
	}
	setTime(new Time());
}


/**
Set the Time the input displays.
@param t The Time to display.
*/
//===================================================================
public void setTime(Time t)
//===================================================================
{
	setTime(t.hour,t.minute,t.second,0);
	/*
	if (is24Hours){
		hour.setInt(t.hour);
	}else{
		if (t.hour < 12) ampm.setText("am");
		else ampm.setText("pm");
		if (t.hour == 0) hour.setInt(12);
		else if (t.hour > 12) hour.setInt(t.hour-12);
		else hour.setInt(t.hour);
	}
	minute.setInt(t.minute);
	if (second != null) second.setInt(t.second);
	*/
}
/**
Get the time displayed/entered.
@param dest an optional destination Time which may be null.
@return the destination Time or a new Time if dest is  null.
*/
//===================================================================
public Time getTime(Time dest)
//===================================================================
{
	if (dest == null) dest = new Time();
	dest.hour = hour.getInt();
	if (!is24Hours){
		if (ampm.getText().equals("pm")) dest.hour += 12;
		if (dest.hour == 24) dest.hour = 12;
		else if (dest.hour == 12) dest.hour = 0;
	}
	dest.minute = minute.getInt();
	dest.second = second == null ? 0 : second.getInt();
	dest.update();
	return dest;
}
public PlainTime getTime(PlainTime dest) {
	if (dest == null) dest = new PlainTime();
	dest.hour = hour.getInt();
	if (!is24Hours){
		if (ampm.getText().equals("pm")) dest.hour += 12;
		if (dest.hour == 24) dest.hour = 12;
		else if (dest.hour == 12) dest.hour = 0;
	}
	dest.minute = minute.getInt();
	dest.second = second == null ? 0 : second.getInt();
	return dest;
}
public void setTime(int hour, int minute, int second, int millis) {
	if (is24Hours){
		this.hour.setInt(hour);
	}else{
		if (hour < 12) ampm.setText("am");
		else ampm.setText("pm");
		if (hour == 0) this.hour.setInt(12);
		else if (hour > 12) this.hour.setInt(hour-12);
		else this.hour.setInt(hour);
	}
	this.minute.setInt(minute);
	if (this.second != null) this.second.setInt(second);
}

public void setValue(Value data)
{
	if (data instanceof Time) setTime((Time)data);
	else if (data instanceof ITime) DateTimeUtils.transferTime((ITime)data,this);
}
public void getValue(Value data)
{
	if (data instanceof Time) getTime((Time)data);
	else if (data instanceof ITime) DateTimeUtils.transferTime(this,(ITime)data);
}

private static ControlPopupForm pop, pop24, popS, pop24S;

//===================================================================
public static ControlPopupForm getPopup(boolean showSeconds, boolean hours24)
//===================================================================
{
	if (showSeconds)
		if (hours24)
			return pop24S == null ? (pop24S = new TimeUpDownInputPopup(showSeconds,hours24)) : pop24S;
		else
			return popS == null ? (popS = new TimeUpDownInputPopup(showSeconds,hours24)) : popS;
	else
		if (hours24)
			return pop24 == null ? (pop24 = new TimeUpDownInputPopup(showSeconds,hours24)) : pop24;
		else
			return pop == null ? (pop = new TimeUpDownInputPopup(showSeconds,hours24)) : pop;
}
/**
 * Use this to popup a Time Up/Down Input in a Frame.
 * @param parent the parent Frame - must be non-null. Use getFrame() on any Control to
 * determine its Frame.
 * @param initial an optional initial Time. If this is non-null the new Date entered
 * will be placed in here too.
 * @param location an optional point to display the popup in the Frame. If it is null
 * the popup will be centered.
 * @param showSeconds true to show seconds.
 * @param hours24 use 24 hour format.
 * @return a Time value containing the new date or null if the user cancelled.
 */
public static Time popupTimeInput(Frame parent, Time initial, Point location, boolean showSeconds, boolean hours24)
{
	if (initial == null) initial = new Time();
	DateTimeInput dti = new DateTimeInput();
	dti.setTime(initial);
	ControlPopupForm cp = getPopup(showSeconds,hours24);
	if (location == null)
		cp.putByClient = false;
	else{
		dti.setRect(location.x,location.y,1,1);		
	}
	dti.modify(Control.Invisible,0);
	parent.addDirectly(dti);
	cp.setFor(dti);
	int got = cp.waitUntilClosed();
	parent.remove(dti);
	if (got == Form.IDCANCEL) return null;
	return dti.getTime(initial);
}

//##################################################################
public static class TimeUpDownInputPopup extends ControlPopupForm{
//##################################################################
TimeUpDownInput input;
//===================================================================
public TimeUpDownInputPopup(boolean showSeconds, boolean hours24)
//===================================================================
{
	//putByClient = false;
	setBorder(BDR_OUTLINE|BF_RECT,1);
	backGround = Color.White;
	addMainControls(input = new TimeUpDownInput(showSeconds,hours24));
	addCloseControls();
}
//===================================================================
public void onControlEvent(ControlEvent ev)
//===================================================================
{
	if (ev.target == input){
		if (ev.type == ev.PRESSED)
			exit(IDOK);
		else if (ev.type == ev.CANCELLED)
			exit(IDCANCEL);
	}
	else super.onControlEvent(ev);
}
//-------------------------------------------------------------------
protected void transferToClient(Control client)
//-------------------------------------------------------------------
{
	if (client instanceof DateTimeInput)
		((DateTimeInput)client).setTime(input.getTime((Time)null));
}
/**
* This is called by setFor(Control who) and gives you an opportunity to
* modify the Form based on the client control.
* @param who The new client control.
*/
//------------------------------------------------------------------
protected void startingInput(Control who)
//-------------------------------------------------------------------
{
	if (who instanceof DateTimeInput)
		input.setTime(((DateTimeInput)client).getTime((Time)null));
}
//##################################################################
}
//##################################################################
/*
//=================================================================
public static void main(String[] args)
//=================================================================
{
	ewe.sys.Vm.startEwe(args);
	Form f = new Form();
	f.addLast(new TimeUpDownInput(false,false));
	f.execute();
	ewe.sys.Vm.exit(0);
}
*/
//##################################################################

public void makeTimeInvalid() {
	hour.setInt(-1);
}
public boolean timeIsValid() {
	return DateTimeUtils.timeIsValid(this);
}

}
//##################################################################

