package eve.ui;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import eve.fx.Dimension;
import eve.fx.Font;
import eve.fx.Fx;
import eve.fx.IFxHandler;
import eve.fx.ISurface;
import eve.io.File;
import eve.io.FileChooserParameters;
import eve.io.Io;
import eve.io.StreamUtils;
import eve.sys.Convert;
import eve.sys.Device;
import eve.sys.Handle;
import eve.sys.IConsole;
import eve.sys.IRegistryKey;
import eve.sys.Registry;
import eve.sys.Time;
import eve.sys.Type;
import eve.sys.Vm;
import eve.ui.launcher.VMApp;

//##################################################################
public class VMOptions extends eve.sys.options.VMOptions implements IFxHandler{
//##################################################################

public boolean showIcon = false;
public boolean launcherTree = false;
public boolean useJavaVM = false;
static final int EVESYNC_ENABLED = 0;
static final int EVESYNC_TEMPORARY_DISABLED = 1;
static final int EVESYNC_DISABLED = 2;
public int eveSyncState = 0;
static final int CONNECTION_GUEST_ONLY = 0;
static final int CONNECTION_NORMAL = 1;
public int guestConnections = 0;
public String javaVM = "";
public String launcherStart = "\\";
public boolean exitAfterLaunch = false;//Device.isMobile();
public double javaFontScale = 0.8;
public boolean overrideSystemFonts = false;
public String systemFontName = "Helvetica";
public int systemFontSize = !Device.isMobile() ? 16 : 12;
public int guiFontStyle = 0;
public int style = GuiStyle.STYLE_ETCHED;
public boolean useDefaultColors = false;
//public FontChooser fontChooser = new FontChooser(false);
public String fixedFontName = "Courier";
public String _fields = "javaFontScale,overrideSystemFonts,launcherTree,systemFontName,systemFontSize,guiFontStyle,showIcon,style,exitAfterLaunch,useJavaVM,javaVM,eveSyncState,guestConnections,useDefaultColors";
public static String currentDir = "C:\\";
//private FontInput fontInput;

public String eveProgramDirectory;
public File eveDirectory;

public VMOptions()
{
	Fx.setFxHandler(this);
	if (Vm.getApplet() == null && java.io.File.pathSeparatorChar == ':')
		javaFontScale = 1.4;
	else
		javaFontScale = 0.8;
}
public boolean yieldToEvents(int max)
{
	Gui.yieldToEvents(max);
	return true;
}
protected void initialize()
{
	try{
	Font sf = Font.getSystemFont(Font.SYSTEM_FONT_DEFAULT);
	if (sf != null) {
		systemFontName = sf.getName();
		systemFontSize = sf.getSize();
	}
	sf = Font.getSystemFont(Font.SYSTEM_FONT_MONOSPACED);
	//
	// Temp. bug fix. Windows returns "Courier" as the ANSI_FIXED_FONT
	// but "Courier" is not fixed on many systems. "Courier New" is however.
	// Right now getSystemFont() only works on Win32 systems.
	/*
	if (sf != null) {
		fixedFontName = sf.getName();
		if (fixedFontName.equals("Courier"))
			fixedFontName = "Courier New";
	}
	*/
	eveProgramDirectory = File.getProgramDirectory();
	eveDirectory = File.getNewFile(eveProgramDirectory);
	}catch(Throwable t){
		t.printStackTrace();
	}
	try{
		boolean found = false;
		if (shouldUsePathToEve()){
			String p = pathToEve;
			File pvm = File.getNewFile(p);
			if (pvm.exists() && !pvm.isDirectory()){
				p = pvm.getParent();
				if (p != null){
					found = true;
					eveProgramDirectory = p;
					eveDirectory = File.getNewFile(p);
				}
			}
		}
		if (!found){
			IRegistryKey k = Vm.getRegistryKey(true,"HKEY_CLASSES_ROOT\\EveFile10\\DLL",false,false);
			File f = File.getNewFile((String)k.getValue(null));
			eveProgramDirectory = f.getParent();
			eveDirectory = File.getNewFile(eveProgramDirectory);
		}
	}catch(Exception e){
	}
}
/**
 * Determine if System fonts only should be used.
 * @return true then the overrideSystemFonts is false or the VM Flag
 * VM_FLAG_SYSTEM_FONTS_ONLY is set.
 */
public static boolean useSystemFonts()
{
	return ((!((VMOptions)vmOptions).overrideSystemFonts) ||
			((Vm.getParameter(Vm.VM_FLAGS) & Vm.VM_FLAG_SYSTEM_FONTS_ONLY) != 0));
}
//===================================================================
public static VMOptions getVMOptions()
//===================================================================
{
	return (VMOptions)vmOptions;
}
private static Type launcher;
/**
 * Get the default launcher Class. If a non-null class is returned
 * then eveMain() and main() are called on that class at startup.
 * @return a default Class for launching.
 */
public Type getLaunchType()
{
	if (launcher == null) {
		if (Device.hasGUI())
			launcher = new Type("eve.ui.launcher.VMLauncher");
		else
			launcher = new Type((String)null);
	}
	return launcher;
}
/**
 * Get the default Font for the Application depending on the settings
 * for overrideSystemFonts.
 */
public static Font getApplicationDefaultFont()
{
	Font f = null;
	VMOptions v = (VMOptions)vmOptions;
	if ((!v.overrideSystemFonts) ||
		((Vm.getParameter(Vm.VM_FLAGS) & Vm.VM_FLAG_SYSTEM_FONTS_ONLY) != 0)){
		f = Font.getSystemFont(Font.SYSTEM_FONT_DEFAULT);
	}
	if (f == null)
		f = new Font(v.systemFontName,Font.PLAIN,v.systemFontSize);
	return f;
}
/*
//===================================================================
public void readAndApply()
//===================================================================
{
	read();
	apply();
}
*/
private Font toFont(boolean useSys, int which, int newStyle, int sizeChange,String appName)
{
	Font f = (useSys && which != -1) ? Font.getSystemFont(which) : null;
	if (f == null) f = new Font(systemFontName,newStyle,systemFontSize+sizeChange);
	try{
	if (appName != null) Application.addFont(f,appName);
	}catch(Throwable t){
		t.printStackTrace();
	}
	return f;
}
private static boolean applied = false;

public static void doApply()
{
	boolean doApply = false;
	//
	synchronized(VMOptions.class){
		doApply = !applied;
		applied = true;
	}
	//
	if (doApply) try {
		getVMOptions().apply();
	}catch(Throwable t){
	}
}
public void apply()
{
	super.apply();
	Vm.setProperty("java.font.scale", Convert.toString(javaFontScale));
	GuiStyle.setStyle(style);
	Gui.resetFlags();
	boolean use = useSystemFonts();
	Font f = toFont(use,Font.SYSTEM_FONT_DEFAULT,Font.PLAIN,0,"system");
	int size = f.getSize();
	Application.addFont(f,"text");
	toFont(use,Font.SYSTEM_FONT_GUI,guiFontStyle,0,"gui");
	f = toFont(use,Font.SYSTEM_FONT_MONOSPACED,Font.PLAIN,0,"fixed");
	if (!Fx.isMonospaced(f)) {
		f = Fx.toMonospacedFont(null,f.getSize());
		if (f != null) Application.addFont(f,"fixed"); 
	}
	toFont(use,Font.SYSTEM_FONT_SERIF,Font.PLAIN,0,"serif");
	f = toFont(use,Font.SYSTEM_FONT_SANS_SERIF,Font.PLAIN,0,"sans serif");
	Application.addFont(f,"sans");
	Application.addFont(f,"sansserif");
	toFont(use,Font.SYSTEM_FONT_DEFAULT,Font.PLAIN,-2,"small");
	toFont(use,Font.SYSTEM_FONT_DEFAULT,Font.PLAIN,+2,"plain");
/*	
	Font f = use ? Font.getSystemFont(Font.SYSTEM_FONT_DEFAULT) : null;
	if (f == null) f = sys;
	Application.addFont(f,"system");
	Application.addFont(f,"text");
	Application.addFont(sys.changeStyle(guiFontStyle),"gui");
	Application.addFont(sys.changeNameAndSize(fixedFontName,sys.getSize()+2),"fixed");
	Application.addFont(sys.changeNameAndSize(null,sys.getSize()-2),"small");
	Application.addFont(sys.changeNameAndSize(null,sys.getSize()+2),"big");
*/
}
//public String aTest = "This is a test";

//===================================================================
public static String [] getExtraActions()
//===================================================================
{
	return new String[]{};//"Unload Vm from Memory$u"};
}
public static void unloadVM(Frame parent)
{
	try{
		Device.preloadVm(false);
	}catch(Throwable e){}
	new MessageBox("VM Unloaded","The VM has been unloaded from memory.",MessageBox.MBOK).execute();
}
//===================================================================
public static boolean doExtraAction(String actionName,Frame parent)
//===================================================================
{
	if (actionName.equals("Unload Vm from Memory")){
		unloadVM(parent);
		return true;
	}
	return false;
}
//===================================================================
public static String [] getExtraContextActions()
//===================================================================
{
	return new String [0];
}

//===================================================================
public static boolean doExtraContextAction(VMApp app,String action,Frame parent)
//===================================================================
{
	return false;
}
//===================================================================
public Font toFont()
//===================================================================
{
	Font f = null;
	VMOptions v = this;
	if ((!v.overrideSystemFonts) ||
		((Vm.getParameter(Vm.VM_FLAGS) & Vm.VM_FLAG_SYSTEM_FONTS_ONLY) != 0)){
		f = Font.getSystemFont(Font.SYSTEM_FONT_DEFAULT);
	}
	if (f == null)
		f = new Font(v.systemFontName,Font.PLAIN,v.systemFontSize);
	return f;
}

private Control addDouble(CellPanel dest,FieldHandler fh,String field, String label, Control c)
{
	boolean doubled = (Gui.isSmartPhone || !Gui.screenIs(Gui.WIDE_SCREEN)); 
	c.promptControl = dest.addNext(new Label(label));
	if (!doubled) c.promptControl.setCell(c.DONTSTRETCH).setControl(c.DONTFILL|c.EAST);
	if  (doubled) dest.endRow();
	return dest.addLast(fh.add(field,c));
}
boolean copyOver(File src, File dest)
{
	if (dest.isDirectory()){
		dest.mkdirs();
		dest = dest.getChild(src.getFileExt());
	}else{
		File f = dest.getParentFile();
		if (f != null) f.mkdirs();
	}
	try{
		InputStream in = src.toReadableStream();
		OutputStream out = dest.toWritableStream(false);
		StreamUtils.transfer(null,in,out);
		in.close();
		out.close();
		dest.setModified(src.getModified(new Time()));
		return true;
	}catch(Exception e){
		return false;
	}
}
void applyEveSync(boolean onStartup)
{
	/*
	#!/bin/bash
	if [ "connect" = $1 ]
	then
	     eve eve.net.RemoteServer -activesyncdesktop &
	fi
	disown
	*/
	if (onStartup){
		if (eveSyncState == EVESYNC_TEMPORARY_DISABLED)
			eveSyncState = EVESYNC_ENABLED;
	}
	if (true || (File.separatorChar == '/' && new File("/usr/bin").exists())){//FIXME
		String h = Registry.getUserHomeDirectory();
		if (h != null){
			File f = new File(h).getChild(".synce/scripts");
			if (eveSyncState == EVESYNC_ENABLED) f.mkdirs();
			f = f.getChild("evesync.sh");
			if (eveSyncState == EVESYNC_ENABLED){
				try{
					String pte = shouldUsePathToEve() ? pathToEve : File.toSystemDependantPath(Device.getPathToVm());
					PrintWriter pw = Io.getPrintWriter(f);
					pw.println(
"#!/bin/bash\nif [ \"connect\" = $1 ]\nthen\n\t"+pte+ 
" eve.net.RemoteServer -activesyncdesktop &\nfi\ndisown"
							);
					pw.close();
					f.setPermissionsAndFlags(File.OWNER_EXECUTE, 0);
				}catch(IOException e){}
			}else{
				if (f.exists()) f.delete();
			}
		}
	}
	IRegistryKey rk = Vm.getRegistryKey(true,"HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\Windows CE Services\\AutoStartOnConnect",true,true);
	if (rk != null){
		if (eveSyncState == EVESYNC_ENABLED){
			rk.setValue("EveSync",
					File.toSystemDependantPath(eveDirectory.getChild("EveSync.exe").getAbsolutePath())
					);
		}else
			rk.deleteValue("EveSync");
	}
	rk = Vm.getRegistryKey(true,"HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\Windows CE Services",true,false);
	if (rk != null){
		rk.setValue("GuestOnly",guestConnections == CONNECTION_GUEST_ONLY ? 1 : 0);
	}
}
/**
 * This tells the VM to setup the Shell\Command\Open entry to
 * use either the native VM (eve.exe) or the Java VM to run
 * ".eve" files.
 * @return true on success, false on failure.
 */
public boolean applySystemOptions()
{
	super.applySystemOptions();
	applyEveSync(true);
	if (useJavaVM) return setupJavaVMDir(File.getNewFile(javaVM));
	else return setupEveVMDir();
}
boolean copyFromProgram(String file, File dest)
{
	return copyOver(File.getNewFile(eveProgramDirectory).getChild(file),dest);
}
boolean setupEveVMDir()
{
	try{
		IRegistryKey k = Vm.getRegistryKey(true,"HKEY_CLASSES_ROOT\\EveFile10\\DLL",false,false);
		if (k == null) return false;
		String me = (String)k.getValue(null);
		if (me != null){
			me = me.substring(0,me.length()-3)+"exe";
			k = Vm.getRegistryKey(true,"HKEY_CLASSES_ROOT\\EveFile10\\Shell\\Open\\Command",true,true);
			if (k != null){
				String cmd = "\""+me+"\" \"%1\"";
				k.setValue(null,cmd);
			}
			return true;
		}
		return false;
	}catch(Exception e){
		return false;
	}
}
boolean setupJavaVMDir(File theVM)
{
	File pd = File.getNewFile(eveProgramDirectory);
	while(true){
		File eveJar = pd.getChild("eve.jar"); 
		if (eveJar.isFile()) break;
		if (new MessageBox("Missing Java Library",
				"The eve.jar library was not found in the Eve program directory.\n"+
				"This file is necessary for running Eve applications with a Java VM\n"+
				"Select Yes to try to locate and copy it, select No to cancel."
				,MessageBox.MBYESNO).execute() == MessageBox.IDNO)
			return false;
		File ej = Gui.chooseFile(false,"Locate eve.jar","eve.jar - Eve Java Library",null,FileChooserParameters.OPTION_QUICK_SELECT,null);
		if (ej == null) return false;
		if (ej.isFile()){
			if (!copyOver(ej,pd)){
				new MessageBox("Error Copying File",
						"There was an error copying the eve.jar file."
						,MessageBox.MBYESNO).execute();
				return false;
			}
		}
	}
	/* Not necessary. java_eve.dll embedded in eve.jar
	 * 
	File jd = theVM.getParentFile();
	if (!jd.isDirectory()) return true;
	String[] dlls = jd.list("java_eve.*",0);
	boolean ok = dlls != null && dlls.length != 0;
	if (ok){
		File javaEve = pd.getChild(dlls[0]); 
		if (javaEve.isFile()){
			Time t = new Time();
			try{
				javaEve.getModified(t);
				long tt = t.getTime();
				File d = jd.getChild(dlls[0]);
				if (d.getModified(t).getTime() < tt){
					ok = false;
				}
			}catch(IOException e){
				ok = false;
			}
		}
	}
	if (!ok){
		boolean did = false;
		if (copyFromProgram("java_eve.dll",jd)) did = true;
		if (copyFromProgram("java_eve.so",jd)) did = true;
		if (!did){
			new MessageBox("Missing DLL",
					"The java_eve DLL could not be copied to the JVM directory.\n"+
					"Java will stil run your Eve applications but this\n"+
					"may affect applications that use certain native APIs."
					,MessageBox.MBOK).execute();
		}
	}
	*/
	String cmd = "\""+File.toSystemDependantPath(theVM.getAbsolutePath())+"\" -cp \""+
	File.toSystemDependantPath(eveDirectory.getChild("eve.jar").getAbsolutePath())+"\" Eve \"%1\"";
	IRegistryKey k = Vm.getRegistryKey(true,"HKEY_CLASSES_ROOT\\EveFile10\\Shell\\Open\\Command",true,true);
	if (k != null) 
		k.setValue(null,cmd);
	return true;
}
boolean setupJavaVM()
{
	String start = null;
	File f = File.getNewFile(javaVM);
	if (f.exists()){
		f = f.getParentFile();
		if (f != null && f.isDirectory())
			start = f.toString();
	}
	File jvm = Gui.chooseFile(false,"Select Java Virtual Machine","*.* - Executable File",start,FileChooserParameters.OPTION_QUICK_SELECT,null);
	if (jvm == null) return false;
	if (!setupJavaVMDir(jvm)) return false;
	javaVM = jvm.toString();
	return true;
}
public int whichColor = 2;

public static int LauncherEditor = 0;
public static int StandAloneEditor = 1;

public Form getEditor(int whichEditor)
{
	Form f = new Form();
	CellPanel f2 = new CellPanel();
	f.title = Gui.isSmartPhone ? "Eve Options" : "Eve VM/Launcher Options";
	f.windowFlagsToClear |= Window.FLAG_HAS_CLOSE_BUTTON;
	f.windowFlagsToSet |= Window.FLAG_MAXIMIZE_ON_PDA;
	f.modifyForSmartPhone();
	FieldHandler fh = new FieldHandler(this){
		protected void fieldAction(Control c, Form f)
		{
			if ("selectVM".equals(c.fieldTransfer.fieldName)){
				File ej = Gui.chooseFile(false,"Locate Eve Virtual Machine","*.* - All Files",null,FileChooserParameters.OPTION_QUICK_SELECT,null);
				if (ej != null) {
					pathToEve = ej.getAbsolutePath();
					toControls("pathToEve");
					getControlFor("pathToEve").notifyDataChange();
					
				}
			}else if ("save".equals(c.fieldTransfer.fieldName)){
				save(f);
			}else if ("setupVM".equals(c.fieldTransfer.fieldName)){
				if (setupJavaVM())
					toControls("javaVM");
			}
		}
		protected void fieldChanged(Control c, Form f)
		{
			String fn = c.fieldTransfer.fieldName;
			if ("pathToEve".equals(fn)){
				File fl = File.getNewFile(pathToEve);
				if (fl.isDirectory()){
					File ch = fl.getChild("eve");
					if (!ch.exists()) ch = fl.getChild("Eve");
					if (ch.exists()) {
						pathToEve = ch.getAbsolutePath();
						c.fromField();
					}else {
						Gui.flashMessage("Please specify the VM Executable, not a directory.",f);
						Gui.takeFocus(c,c.ByRequest);
					}
				}else if (!fl.exists()){
					Gui.flashMessage("That file does not exist.",f);
					Gui.takeFocus(c,c.ByRequest);
				}
			}else if ("javaFontScale".equals(fn)){
				if (javaFontScale > 3) javaFontScale = 3;
				else if (javaFontScale < 0.5) javaFontScale = 0.5;
				toControls("javaFontScale");
			}else if ("eveSyncState".equals(fn) || "guestConnections".equals(fn)){
				applyEveSync(false);
			}else if ("whichColor".equals(fn)){
				IRegistryKey rk = Vm.getRegistryKey(true,"HKEY_LOCAL_MACHINE\\System\\GWE",false,false);
				if (rk != null){
					Object got = rk.getValue("SysColor");
					if (got instanceof byte[])
						f.backGround = Gui.getColorFrom((byte[])got,whichColor); 
					f.repaintNow();
				}
			}else if ("font".equals(fn)){
				FontSelector fi = (FontSelector)c;
				Font font = fi.toFont();
				systemFontName = font.getName();
				systemFontSize = font.getSize();
			}else if ("useJavaVM".equals(fn)){
				if (!useJavaVM){
					setupEveVMDir();
				}else if (useJavaVM){
					if (javaVM.trim().length() == 0){
						try{
							File jvm = Registry.getJavaVM();
							if (jvm != null){
								javaVM = jvm.getAbsolutePath();
								toControls("javaVM");
								if (setupJavaVMDir(File.getNewFile(javaVM)))
									return;
							}
						}catch(Exception e){}
					}else{
						if (setupJavaVMDir(File.getNewFile(javaVM)))
							return;
					}
					if (!setupJavaVM()){
						useJavaVM = false;
						c.fromField();
					}else
						toControls("javaVM");
				}
			}
		}
	};
	f.addFieldHandler(fh);
	CellPanel cp = new CellPanel();
	Choice lf = new Choice(new String[]{"3-D","Etched","Soft","Palm"},0);
	Choice gs = new Choice(new String[]{"Plain","Bold","Italic","Bold & Italic"},0);
	//int uc = sp ? 0 : FontInput.OPTION_USE_CHOICE;
	FontSelector fi = new FontSelector(FontInput.OPTION_NO_STYLE);
	fi.fromFont(new Font(systemFontName,0,systemFontSize));
	boolean mb = Device.isMobile(); 
	if (!Gui.isSmartPhone){
		if (mb && (whichEditor != StandAloneEditor) && !Gui.hasSoftKeys) f2.addLast(fh.add("save",new Button("Save Options","eve/savesmall.png"))).setCell(f2.HSTRETCH);
		if (shouldUsePathToEve()){
			CellPanel cp3 = new CellPanel();
			cp3.addNext(fh.add("selectVM",new Button("Locate VM Path:"))).setCell(cp3.DONTSTRETCH);
			cp3.addLast(fh.add("pathToEve",new Input()));
			f2.addLast(cp3).setCell(f2.HSTRETCH);
		}
		if (!mb)
			cp.setText("Options");
		else
			cp.setBorder(cp.EDGE_ETCHED,2);
		cp.addNext(fh.add("singleWindowed",new CheckBox("Single Window")));
		//cp.addLast(fh.add("keepVmResident",new CheckBox(mb ? "Keep Resident":"Keep VM Resident")));
		cp.endRow();
		cp.addNext(fh.add("useSIP",(new CheckBox("Use SIP"))));
		cp.addLast(fh.add("fixedSIPButton",new CheckBox(mb ? "Fixed SIP" : "Fixed SIP Button")));
		cp.addNext(fh.add("exitAfterLaunch",new CheckBox("Exit After Launch")));
		cp.addLast(fh.add("useDefaultColors",new CheckBox(mb  ? "Default Colors":"Use Default Colors")));
		//if (!Gui.screenIs(Gui.WIDE_SCREEN)) cp.endRow();
		cp.addNext(fh.add("showEveSyncConsole",new CheckBox("Show EveSync Console")));
		cp.endRow();
		//Input in = new Input(); in.wantReturn = true; cp.addLast(fh.add("whichColor",in));
		f2.addLast(cp).setCell(f.HSTRETCH);//.setControl(f.HCONTRACT|f.CENTER);
		if (!Device.isMobile()){
			cp = new CellPanel();
			CellPanel cp3 = new CellPanel();
			//cp.setBorder(cp.EDGE_ETCHED,2);
			cp.setText("Use Java VM");
			cp.addLast(fh.add("useJavaVM",new CheckBox("Use Java VM to Run Apps"))).setTag(cp.TAG_SPAN,new Dimension(-1,1));
			cp3.addNext(fh.add("setupVM",new Button("Select JVM","eve/ChooseFile.png"))).setCell(cp.DONTSTRETCH);
			cp3.addNext(fh.add("javaVM",new Input()));
			cp.addLast(cp3).setCell(cp.HSTRETCH);
			f2.addLast(cp).setCell(f.HSTRETCH);//.setControl(f.DONTFILL|f.CENTER);
			
			cp = new CellPanel();
			cp.setText("EveSync/ActiveSync");
			//cp.setBorder(cp.EDGE_ETCHED,2);
			addDouble(cp,fh,"eveSyncState","EveSync",new Choice("Enabled|Temporarily Disabled|Disabled",0)).setControl(cp.HCONTRACT|cp.LEFT);
			addDouble(cp,fh,"guestConnections","Unknown Connections",new Choice("Guest Only|Normal",0)).setControl(cp.HCONTRACT|cp.LEFT);
			f2.addLast(cp).setCell(f.HSTRETCH);//.setControl(f.DONTFILL|f.CENTER);
		}		
		cp = new CellPanel();
		if (!mb)
			cp.setText("Appearance");
		else
			cp.setBorder(cp.EDGE_ETCHED,2);
		addDouble(cp,fh,"javaFontScale","Java Font Scale",new Input()).setControl(cp.HCONTRACT|cp.LEFT);
		addDouble(cp,fh,"overrideSystemFonts","Override System Font",new CheckBox()).setControl(cp.HCONTRACT|cp.LEFT);
		addDouble(cp,fh,"font","System Font",fi).setControl(cp.HCONTRACT|cp.LEFT);
		addDouble(cp,fh,"guiFontStyle","GUI Font Style",gs).setControl(cp.HCONTRACT|cp.LEFT);
		addDouble(cp,fh,"style","Look and Feel",lf).setControl(cp.HCONTRACT|cp.LEFT);
		f2.addLast(cp).setCell(f.HSTRETCH);//.setControl(f.DONTFILL|f.CENTER);
		cp = new CellPanel();
		if (!mb && (whichEditor != StandAloneEditor)) cp.addLast(fh.add("save",new Button("Save","eve/savesmall.png"))).setControl(cp.DONTFILL);
		f2.addLast(cp).setCell(f.HSTRETCH);//.setControl(f.DONTFILL|f.CENTER);
		if (mb && Gui.hasSoftKeys) Gui.setOKCancel(f);
	}else{
		//addDouble(cp,fh,"keepVmResident","Keep VM Resident",new CheckBox(""));
		addDouble(cp,fh,"exitAfterLaunch","Exit After Launch",new CheckBox(""));
		addDouble(cp,fh,"useDefaultColors","Use Default Colors",new CheckBox(""));
		addDouble(cp,fh,"showEveSyncConsole","Show EveSync Console",new CheckBox(""));
		addDouble(cp,fh,"overrideSystemFonts","Override System Font",new CheckBox(""));
		addDouble(cp,fh,"font","System Font",fi);
		addDouble(cp,fh,"guiFontStyle","GUI Font Style",gs).setControl(cp.HCONTRACT|cp.LEFT);
		addDouble(cp,fh,"style","Look and Feel",lf).setControl(cp.HCONTRACT|cp.LEFT);
		f2.addLast(cp).setCell(f.HSTRETCH);//.setControl(f.DONTFILL|f.CENTER);
		Gui.setOKCancel(f);
	}
	VerticalScrollPanel vp = new VerticalScrollPanel(f2);
	f.addLast(vp);//.setControl(f.HCONTRACT|f.CENTER);
	return f;
}
//===================================================================
public boolean save(Control ed)
//===================================================================
{
	if (!save()) return false;
	applySystemOptions();
	if (ed != null) Gui.flashMessage("Saved. Restart to take effect...",1000,ed,Gui.FLASH_BEEP);
	return true;
}
private static String guiName(String name)
{
	return Device.hasGUI() ? name : null;
}
public void runMain(String[] args, String startClassName)
{
	Type t = new Type(startClassName);
	if (!t.exists()) return;
	Class c = t.getReflectedClass();
	if (Vm.isJavaVM())
		t.invoke(null,"eveMain([Ljava/lang/String;)V",new Object[]{args});
	Type ft = new Type(guiName("eve.ui.Form"),c);
	Type lt = new Type(guiName("eve.ui.data.LiveData"),c);
	Type lo = new Type(guiName("eve.ui.data.LiveObject"),c);
	Class formClass = ft.exists() ? ft.getReflectedClass() : null;
	Class liveClass = lt.exists() ? lt.getReflectedClass() : null;
	boolean isForm = formClass == null ? false : formClass.isAssignableFrom(c); 
	boolean isLive = liveClass == null ? false : liveClass.isAssignableFrom(c);
	try{
		Method m =
			isForm||isLive ? 
			c.getDeclaredMethod("main",new Class[]{args.getClass()}) :
			c.getMethod("main",new Class[]{args.getClass()});
		m.invoke(null,new Object[]{args});
		//Reflection.invokeMethod(Class.forName(className),"main([Ljava/lang/String;)V",new Object[]{args});
	}catch(Throwable e){
		if (e instanceof NoSuchMethodException && isForm){
			ft.invoke(null,"formMain(Ljava/lang/String;[Ljava/lang/String;)V",
					new Object[]{startClassName,args});
			Vm.exit(0);
		}else if (e instanceof NoSuchMethodException && isLive && lo.exists()){
			lo.invoke(null,"liveMain(Ljava/lang/String;[Ljava/lang/String;)V",
					new Object[]{startClassName,args});
			Vm.exit(0);
		}else{
			if (e instanceof InvocationTargetException){
				((InvocationTargetException)e).getCause().printStackTrace();
			}else
				e.printStackTrace();
			if (Vm.getApplet() == null){
				Device.tryMessageBox("Application Error","An error occured in main().\n"+Vm.getStackTrace(e)+"\nThe application will now exit.",true);
			}
			Vm.exit(-1);
		}
	}
}
/* (non-Javadoc)
 * @see eve.fx.IFxHandler#getDefaultFont()
 */
public Font getDefaultFont() {
	Application a = Application.mainApp;
	return a == null ? null : a.getFont();
}

/* (non-Javadoc)
 * @see eve.fx.IFxHandler#getDefaultSurface()
 */
public ISurface getDefaultSurface() {
	Application a = Application.mainApp;
	return a == null ? null : a.getSurface();
}
/* (non-Javadoc)
 * @see eve.fx.IFxHandler#getDefaultWindowTitle()
 */
public String getDefaultWindowTitle() {
	return Application.getDefaultWindowTitle();
}
/* (non-Javadoc)
 * @see eve.fx.IFxHandler#getConsole()
 */
public IConsole getConsole(String title, int options, int maxLines) 
{
	if (!Device.hasGUI()) return null;
	Type cs = new Type("eve.ui.Console");
	if (!cs.exists()) return null;
	return (IConsole)cs.invoke(null,"getConsole(Ljava/lang/String;II)Leve/sys/IConsole;",
			new Object[]{
			title, new Integer(options), new Integer(maxLines)
			});
}
/*
public static void main(String[] args) throws IOException
{
	if (true){
		Vm.startEve(args);
		VMOptions vo = (VMOptions)getNewVMOptions();
		vo.pathToEve = "C:/program files/eve/evew.exe";
		vo.singleWindowed = false;
		vo.eveSyncState = EVESYNC_ENABLED;
		vo.setupInstall();
		Vm.exit(0);
	}
	Application.startApplication(args);
	VMOptions vmo = getVMOptions();
	Form f = vmo.getEditor(StandAloneEditor);
	Gui.setOKCancel(f);
	if (f.execute() == f.IDOK){
		vmo.save(null);
	}
	Application.exit(0);
}
*/
//##################################################################
public Handle printerDialog(Object printerProperties) {
	Type t = new Type("eve.ui.print.PrinterDialog");
	if (!t.exists()) return new Handle(Handle.Failed,null);
	Form f = (Form)t.newInstance("Leve/fx/print/PrinterProperties;", new Object[]{printerProperties});
	return f.exec(true);
}
}
//##################################################################

