/*
 * Created on Jul 31, 2005
 *
 * Michael L Brereton - www.ewesoft.com
 * 
 * 
 */
package eve.ui;

import java.util.Stack;
import java.util.Vector;

import eve.fx.BufferedGraphics;
import eve.fx.Color;
import eve.fx.Font;
import eve.fx.FontMetrics;
import eve.fx.Graphics;
import eve.fx.ISurface;
import eve.fx.Image;
import eve.fx.ImageBuffer;
import eve.fx.Pen;
import eve.fx.PixelBuffer;
import eve.fx.Point;
import eve.fx.Rect;
import eve.fx.Sound;
import eve.fx.gui.Cursor;
import eve.fx.gui.IKeys;
import eve.fx.gui.RawEvent;
import eve.fx.gui.SurfaceKeyEvent;
import eve.fx.gui.SurfacePaintEvent;
import eve.fx.gui.SurfacePointerEvent;
import eve.fx.gui.SurfaceResizeEvent;
import eve.fx.gui.SurfaceSIPEvent;
import eve.fx.gui.SurfaceTextEvent;
import eve.fx.gui.SurfaceWindowEvent;
import eve.fx.gui.WindowConstants;
import eve.fx.gui.WindowCreationData;
import eve.fx.gui.WindowSurface;
import eve.sys.Cache;
import eve.sys.Countdown;
import eve.sys.DeferredEvent;
import eve.sys.Event;
import eve.sys.EventListener;
import eve.sys.Handle;
import eve.sys.Task;
import eve.sys.TimeOut;
import eve.sys.TimerProc;
import eve.sys.Vm;
import eve.sys.WeakRef;
import eve.ui.event.KeyEvent;
import eve.ui.event.PenEvent;
import eve.ui.event.WindowEvent;
import eve.util.Tag;
import eve.util.mString;

/**
 * @author Michael L Brereton
 *
 * TODO Implement taskbar Icon in Window
 */
//####################################################
public class Window extends Container implements WindowConstants{

private final static Object openCloseLock = new Object();//WindowExecManager.openCloseLock;	
WindowRawEventThread rawEventThread;	
WindowEventThread activeThread;
static Object windowReleaseLock = new Object();
public static boolean bufferFullWindow = false;
/*
EventBuffer eventBuffer = new EventBuffer(null){
	{
		//setPointerMoveDelay(1000);
		setRepaintDelay(100);
		//writeThrough = true;
	}
	boolean lastWasDown = false;
	protected void deliver(Event ev, boolean isDirect){
		//super.deliver(ev,isDirect);
		if (isDirect) {
			Device.setThreadIsEventHandler();
			handleSurfaceEvent(ev);
			if (!ev.isReused()) ev.cache();
		}else{
			//int num = Vm.countObjects(false);
			//Vm.debug(Convert.toString(Vm.countObjects(false)-num));
			Event ev2 = (Event)Reflection.getCachedCopy(ev);//(Event)ev.getCopy();
			queueEvent(ev2);
		}
	}
};
*/
/**
 * Set this to be a defaultWindowIcon if you wish. It is most efficient to
 * set this to be a DeviceIcon using Device.createIcon() or Device.toIcon(),
 * but you can also set it to be any of the convertible values. 
 */
public static Object defaultWindowIcon = null;

protected WindowSurface surface = WindowSurface.getNew();
protected static Vector openWindows = new Vector();
protected Window containing;
Control focus;
//static boolean inEventThread;
Vector frames = new Vector();
public boolean canDisplay, firstDisplay;
public WindowCreationData creationData;
public Frame contents = new Frame();
private boolean hasClosed;

public boolean hasPendingEvents()
{
	WindowSurface s = surface;
	return s != null && s.hasPendingEvents();
}
static int visibleWidth, visibleHeight;

private static Window activeWindow, lastWindow;
/*
public void requestPaint(int x, int y, int width, int height)
{
	WindowSurface s = surface;
	if (s != null) s.requestPaint(x,y,width,height);
}
*/

boolean holdEvents;

Window myMainApp = null;

public Window()
{
	this(true);
}
public Window(boolean holdEvents)
{
	font = Application.guiFont;
	this.holdEvents = holdEvents;
	myMainApp = Application.mainApp;
}
/*
private Graphics myGraphics;
private boolean graphicsAvailable = true; 
	
class savedGraphics extends Graphics {

	public savedGraphics(){
		super(Window.this.surface);
	}
	
	public void free(){
		synchronized(Window.this){
			if (myGraphics == this) {
				graphicsAvailable = true;
				//Vm.debug("Replaced!");
				return;
			}
		}
		super.free();
	}
	
}
*/

private Countdown cursorBlink;

/**
 * Only one control per Window will get this timer. The ticked() method
 * will be called in the Window Event Thread.
 * @return an Object used as an id for the TimerProck.ticked() method.
 */
public static Object requestCursorBlink(Control c,long howLong)
{
	final Window w = c.getWindow();
	if (w == null) return null;
	synchronized(w){
		if (w.cursorBlink == null){ 
			w.cursorBlink = new Countdown(){
				protected void ticked(){
					final TimerProc tc = tickProc;
					tickProc = null;
					final Countdown cd = this;
					try{
						if (tc != null) new WindowThreadTask(){
							protected void doRun(){
								//System.out.println("WTT!");
								if (w.hasBeenPainted())
									tc.ticked(cd,elapsed());
							}
						}.run((Control)tc,true);
					}catch(Throwable t){}
				}
			};
			w.cursorBlink.idleQuitTime = -1;//2000;
		}
		if (c != null) w.cursorBlink.requestTick(c,howLong);
		return w.cursorBlink;
	}
}
public static void cancelCursorBlink(Control c)
{
	Window w = c.getWindow();
	if (w == null) return;
	synchronized(w){
		if (w.cursorBlink != null)
			w.cursorBlink.cancelTick(c);
	}
}
/*
public void relayout(boolean repaint)
{
	contents.relayoutAndResize(false);
}
*/
public boolean isCurrentEventThread(Thread t)
{
	return t == activeThread;
}
public boolean inCurrentEventThread()
{
	return Thread.currentThread() == activeThread;
}
public static boolean inEventThread()
{
	boolean it = Thread.currentThread() instanceof WindowEventThread;
	//System.out.println(it);
	return it;
	//return inEventThread;
}
public Object getInfo(int infoCode,Object parameters,Object dest,int options)
{
	return surface.getInfo(infoCode,parameters,dest,options);
}
public Object getInfo(int infoCode,Object parameters)
{
	return surface.getInfo(infoCode,parameters,null,0);
}
//FIXME implement Window.supportsMultiple()
public static boolean supportsMultiple()
{
	return (Vm.getParameter(Vm.VM_FLAGS) & Vm.VM_FLAG_NO_WINDOWS) == 0;
}
/*
protected void finalize()
{
	Vm.debug("Window finalized!");
}
*/
/**
* This closes the window and destroys the native Window component. Do not re-use the window after closing it.
**/
//===================================================================
public boolean close()
//===================================================================
{
//	System.out.println("Must close: "+this);
	synchronized(openCloseLock){
		focus = null;
		if (lastWindow == this) lastWindow = null;
		if (inFront == this) inFront = null;
		if (activeWindow == this) activeWindow = null;
		if (Gui.sipWindow == this) Gui.sipWindow = null;
		/*
		_sipEvent.window = null;
		_keyEvent.window = null;
		_controlEvent.window = null;
		_penEvent.window = null;
		*/
		openWindows.remove(this);
		//windowThread.closing(Window.this, true);
	}
	surface.close();
	canDisplay = false;
	//
	// Stop all event threads.
	//
	if (rawEventThread != null)
		synchronized(rawEventThread){
			//System.out.println("Must stop: "+this);
			rawEventThread.closed = true;
			try{
				rawEventThread.interrupt();
			}catch(SecurityException e){}
			rawEventThread.notifyAll();
		}
	//eventBuffer.close();
	//System.out.println("Must shutdown: "+this);
	shutdown();
	//System.out.println("Must return: "+this);
	return true;
}
Object taskbarIcon;
/**
* This is called either before the window is closed or before the application exits.
**/
//-------------------------------------------------------------------
protected synchronized void shutdown()
//-------------------------------------------------------------------
{
	if (surface != null){
		if (taskbarIcon != null){
			surface.setTaskbarIcon(null);
		}
		//discardGraphics();
	}
	synchronized(this){
		hasClosed = true;
		notifyAll();
	}
}
//===================================================================
protected void shutdownAll()
//===================================================================
{
	for (int i = 0; i<openWindows.size(); i++)
		try{
			((Window)openWindows.get(i)).shutdown();
		}catch(Throwable t){}
}

/**
This returns true if the Window has been created (its native component has been created)
and has not been closed.
**/
//===================================================================
public boolean isCreated()
//===================================================================
{
	for (int i = 0; i<openWindows.size(); i++)
		if (openWindows.get(i) == this) return true;
	return false;
}

/**
* If this Window is contained in another Window, this will return the containing Window. Calling getParent() on
* a Window should always return null.
**/
//===================================================================
public Window getContainingWindow()
//===================================================================
{
	return containing;	
}
/**
 * Return the Window flags for this window.
 */
public int getWindowFlags()
{
	return surface.getFlags();
}

/**
 * If this is true then if toFront() is called on the window, and it
 * happens to be in an invisible state, then the window will be
 * faced to be visible. 
 */
public boolean forceVisibleOnToFront = false;

/**
 * Bring the window to the front.
 */
public boolean toFront()
{
	if (forceVisibleOnToFront) setVisible(true);
	surface.setState(STATE_TO_FRONT);
	return true;
}
/**
 * Minimize all windows in the application.
 */
public static void minimizeApplication()
{
	Vector v = new Vector();
	synchronized(openCloseLock){
		for (int i = openWindows.size()-1; i>= 0; i--){
			Window w = (Window)openWindows.get(i);
			if ((w.getWindowFlags() & FLAG_IS_VISIBLE) == 0) continue;
			v.add(w);
		}
	}
	for (int i = v.size()-1; i>= 0; i--){
		((Window)v.get(i)).setState(STATE_MINIMIZED);
	}
}

/**
 * This will attempt to bring the topmost window of the application to the front.
 * @return true if successful.
 */
//===================================================================
public static boolean applicationToFront()
//===================================================================
{
	WindowSurface ws = WindowSurface.getModal();
	Window found = null;
	synchronized(openCloseLock){
		for (int i = openWindows.size()-1; found == null && i>= 0; i--){
			Window w = (Window)openWindows.get(i);
			if (ws != null){
				if (w.surface == ws)
					found = w;
			}else{
				// No modal window.
				if ((w.getWindowFlags() & FLAG_IS_VISIBLE) != 0) continue;
				found = w;
			}
		}
		if (found == null)
			if (Application.mainApp != null){
				if (Application.mainApp.forceVisibleOnToFront || (Application.mainApp.getWindowFlags() & FLAG_IS_VISIBLE) != 0)
					found = Application.mainApp;
			}
	}
	if (found != null) {
		return found.toFront();
	}
	return false;
}
/**
* This returns the currently active window.
**/
//===================================================================
public static Window getActiveWindow()
//===================================================================
{
	WindowSurface ws = WindowSurface.getModal();
	if (ws != null) {
		return (Window)ws.applicationWindow;
	}
	//synchronized(openCloseLock){
		for (int i = openWindows.size()-1; i>= 0; i--){
			Window w = (Window)openWindows.get(i);
			if ((w.getWindowFlags() & FLAG_IS_VISIBLE) == 0) continue;
			return w;
		}
		if (Application.mainApp != null){
			if ((Application.mainApp.getWindowFlags() & FLAG_IS_VISIBLE) != 0)
				return Application.mainApp;
		}
		return null;
	//}
	
//	return activeWindow;
}
/**
 * Get the window's full or client Rect.
 * @param dest a destination Rect
 * @param clientArea if this is true then the client Rect will be returned. The X and Y
 * coordinates in the Client rect represent the offset into the whole Rect that the
 * client area is in.
 * @return null if the Rect could not be determined, otherwise it returns the destination Rect. 
 */
//===================================================================
public Rect getWindowRect(Rect dest,boolean clientArea)
//===================================================================
{
	if (dest == null) dest = new Rect();
	boolean got = !clientArea ? surface.getBounds(dest) : surface.getClientRect(dest);
	return got ? dest : null;
}
/**
* Set the window full or client rectangle. Do not pass a null destination rectangle.
 * @param where the new Rect for the Window or client area.
 * @param clientArea if this is true then the client Area will be placed in the specified
 * position in the parent and the Window will be placed and sized so that its client
 * area is as specified.
 * **/
//===================================================================
public boolean setWindowRect(Rect where,boolean clientArea)
//===================================================================
{
	if (clientArea) return surface.setClientRect(where.x,where.y,where.width,where.height);
	else return surface.setBounds(where.x,where.y,where.width,where.height);
}
/**
* Set specified window flags. This is only meaningful after a create() call.
**/
//===================================================================
public int setWindowFlags(int flags)
//===================================================================
{
	return surface.setFlags(flags,0);
}
/**
* Clear specified window flags. This is only meaningful after a create() call.
**/
//===================================================================
public int clearWindowFlags(int flags)
//===================================================================
{
	return surface.setFlags(0,flags);
}

public void setTitle(String title)
{
	surface.setTitle(title);
}
public int getFlagsForSize(int requestedWidth, int requestedHeight, int flagsToSet, int flagsToClear)
{
	return surface.getFlagsForSize(requestedWidth,requestedHeight,flagsToSet,flagsToClear);
}
/**
 * Set window state bits.
 * @param state one of the STATE_XXX values.
 */
public void setState(int state)
{
	surface.setState(state);
}
public WindowSurface getWindowSurface()
{
	return surface;
}
/**
 * Set the visibility state of the window.
 * @param visible true to make the window visible, false to make it invisible.
 */
public void setVisible(boolean visible)
{
	surface.setVisible(visible);
}
/**
 * Return if the Window is visible.
 */
public boolean isVisible()
{
	return surface.isVisible();
}
//
private int waiting = 0;
private int curCursor = Cursor.DEFAULT_CURSOR;
/**
 * Show or hide the wait cursor for this window.  
 * This overrides all
 * other cursors as long as it is active. showWait() is considered active
 * as long as there are calls made with showOrHide true without 
 * corresponding calls with showOrHide false.
 * @param showOrHide true to show the wait, false to hide it. For 
 * every call with showOrHide true there must be a corresponding call
 * with showOrHide false in order to restore the cursor to normal.
 */
public synchronized void showWait(boolean showOrHide)
{
	if (showOrHide){
		waiting++;
		getWindowSurface().setCursor(Cursor.WAIT_CURSOR);
	}else{
		if (waiting > 0) {
			waiting--;
			if (waiting == 0) getWindowSurface().setCursor(curCursor);
		}
	}
}
/**
 * Set the cursor for this window. This will not override the WAIT_CURSOR
 * if the wait cursor was shown using showWait(). As long as showWait()
 * is active the wait cursor is shown until it is no longer active, at
 * which point the most recent cursor type specified to setCursor() is shown.
 * @param type the cursor type.
 */
public synchronized void setCursor(int type)
{
	curCursor = type;
	if (waiting <= 0) getWindowSurface().setCursor(type);
}
/*
public synchronized Graphics getGraphics()
{
	if (true) return createGraphics();
	if (surface == null || hasClosed) return null;
	try{
		if (myGraphics == null){
			myGraphics = new savedGraphics();
			graphicsAvailable = true;
		}
		if (graphicsAvailable){
			graphicsAvailable = false;
			myGraphics.reset();
			//Vm.debug("Getting: "+myGraphics.hashCode()+", "+myGraphics.isValid());
			//Vm.debug("REuse!");
			return myGraphics;
		}
		return createGraphics();
	}catch(Exception e){
		return null;
	}
}
public Graphics createGraphics()
{
	if (surface == null || hasClosed) return null;
	try{
		return new Graphics(surface);
	}catch(Exception t){
		return null;
	}
}
*/
public ISurface getSurface()
{
	return surface;
}
//===================================================================
public void transferPenPress(PenEvent ev,Control toWho,int dx,int dy)
//===================================================================
{
	focus = toWho;
	Control.firstPress = true;
	//ev.target = toWho;
	ev.x += dx;
	ev.y += dy;
	//FIXME ev.TRANSFERRED_PRESS
	//ev.modifiers |= ev.TRANSFERRED_PRESS; 
	toWho.postUIEvent(ev);
}
/*
//-------------------------------------------------------------------
void makeModal()
//-------------------------------------------------------------------
{
	windowThread wt = new windowThread(this,true);
	wt.runHandle.waitUntilCompletion();
}
//-------------------------------------------------------------------
void closeModal(boolean closingActualWindow)
//-------------------------------------------------------------------
{
	windowThread.closing(this,closingActualWindow);
}
*/
protected static Window inFront;
Point mouseOffset = null;
static Point dragxy = new Point();
int lastMoveX = -1, lastMoveY = -1;
private boolean inKeyPress;
/**
* This is the point at which the mouse was first pressed in a current press and hold/drag operation.
**/
Point downPoint = new Point();
private static Rect rbuff = new Rect(); 
private boolean gotFirstResize = false; 
private Object firstResizeLock = new Object();

static Control lastDraggedOver;
public static Object dragAndDropData;
public static Object dragAndDropSource;
public static int dragAndDropCursor;
/**
* This is called by a control to start a drag and drop operation. You should not call this directly,
* rather you should call Control.startDragAndDrop() from Control
**/
//===================================================================
public boolean beginDragAndDrop(Control source,Object data,int dragCursor)
//===================================================================
{
	dragAndDropData = data;	
	dragAndDropCursor = dragCursor;
	dragAndDropSource = source;
	//releaseMouseCapture();
	return true;
}
/**
* This is called by a control to start a drag and drop operation. You should not call this directly,
* rather you should call Control.startDragAndDrop() from Control
**/
//===================================================================
public boolean beginDragAndDrop(Control source,Object data,boolean isMultiple,boolean isCopy)
//===================================================================
{

	return beginDragAndDrop(source,data,
		isMultiple ?
			isCopy ?
				Cursor.COPY_MULTIPLE_CURSOR : Cursor.DRAG_MULTIPLE_CURSOR
		:
			isCopy ?
				Cursor.COPY_SINGLE_CURSOR : Cursor.DRAG_SINGLE_CURSOR);
}
/**
 * This is called from within handleSIPEvent() and it calls
 * onEvent(ev) on the topmost Frame in this Window. Then it calls
 * Gui.ensureVisible() on the Control returned by Gui.focusedControl().
 * @param ev the SurfaceSIPEvent received.
 */
protected void sendSIPToChildren(SurfaceSIPEvent ev)
{
	Frame f = Gui.topFrame(this);
	if (f != null) f.onEvent(ev);
	Control c = Gui.focusedControl();
	if (c != null) Gui.ensureVisible(c);
}
/**
 * This is the default SIP handler method, but you should not override
 * to customize SIP handling. Instead override sendSIPToChildren() - that
 * will allow the window to continue simulating the SIP on a desktop and
 * to correctly keep track of its visible width and height.
 * @param ev the SurfaceSIPEvent received.
 */
protected void handleSIPEvent(SurfaceSIPEvent ev)
{
	visibleWidth = ev.type == ev.SIP_SHOWN ? ev.visibleWidth : 0;//width;
	visibleHeight = ev.type == ev.SIP_SHOWN ? ev.visibleHeight : 0;//height;
	//
	sendSIPToChildren(ev);
	//
	if (visibleHeight != 0 && (Vm.getParameter(Vm.SIMULATE_SIP) != 0)){
		if (sipThread == null){
			sipThread = new Task(){
				protected void doRun(){
					while(!shouldStop && !hasClosed){
						try{
							showSipCoverage();
							sleep(250);
						}catch(Exception e){}
					}
				}
			}.start();
		}
	}else if (sipThread != null){
		sipThread.stop(0);
		sipThread = null;
		repaintNow();
	}
}
//-------------------------------------------------------------------
static void dragOverNew(Control newControl)
//-------------------------------------------------------------------
{
	if (lastDraggedOver != newControl)
		if (lastDraggedOver != null) 
			lastDraggedOver.dataDraggedOff(dragAndDropData);
	lastDraggedOver = newControl;
}
//-------------------------------------------------------------------
protected Window tryWindow(Window win,int x,int y,Point where)
//-------------------------------------------------------------------
{
	if (win == null || mouseOffset == null) return null;
	int absX = x+mouseOffset.x;
	int absY = y+mouseOffset.y;
	Rect wr = win.getWindowRect(new Rect(),false);
	//Vm.debug("Check: "+wr+", "+absX+","+absY);
	if (wr.isIn(absX,absY)){
		//Vm.debug("OK, now in.");
		int cx = absX-wr.x, cy = absY-wr.y;
		Rect in = win.getWindowRect(new Rect(),true);
//		Vm.debug("client: "+in);
		cx -= in.x; cy -= in.y;
		Control target = win.findChild(cx,cy);
		if (target != win && target != null){
			if (win != lastWindow) win.toFront();
			lastWindow = win;
			where.x = cx; where.y = cy;
			return win;
		}
	}
	return null;	
}
private void sendDragAndDropPenEvent(int type, int key, int x, int y, int modifiers, long timeStamp, Object data)
{
	doPostEvent(type, key, x, y, modifiers, timeStamp, data);
}

/*
private void _postEvent(int type, int key, int x, int y, int modifiers, long timeStamp, Object data)
{
	//if ((modifiers & PenEvent.FROM_OTHER_WINDOW) != 0) Vm.debug("FROM OTHER: "+timeStamp);
	//Vm.debug("_postEvent to "+hashCode()+": "+type+", "+x+", "+y+" "+((modifiers & PenEvent.FROM_OTHER_WINDOW) != 0));
	//if (nativeDialogCount != 0) return;
	if (type == WindowEvent.ACTIVATE){
		activeWindow = this;
		checkAppKeys();
		Frame.checkCurrentSip();
		Control fc = focus;//Gui.focusedControl();
		if (fc != null)
			if (fc.hasModifier(ShowSIP,false) && fc.canEdit() && ((getWindowFlags() & FLAG_SHOW_SIP_BUTTON) == 0)){
				surface.setSIP(2);//,this);
				//new Exception().printStackTrace();
			}
	}
	if (type >= WindowEvent.CLOSE && type <= WindowEvent.HIBERNATE)
		if (Device.isMobile())
			mobileWindowEvent(type,key,x,y,modifiers,timeStamp);

	if (type == WindowEvent.APP_TO_FRONT) applicationToFront();
	else
		new windowThread.queuedEvent(this,type,key,x,y,modifiers,timeStamp,null);
}
*/

protected void mobileWindowEvent(int type, int key, int x, int y, int modifiers, long timeStamp)
{
	
}
private static Point hoverSpot = new Point();
private static int hoverMods;
private static WeakRef shouldSendHoverTo = new WeakRef(), sentHoverTo = new WeakRef();
private static RawEvent hoverRawEvent;
private static Countdown controlHoverCountdown = new Countdown(){
	{
		idleQuitTime = -1;
	}
	protected void ticked(){
		Window ht = (Window)shouldSendHoverTo.get();
		if (ht != null){
			sentHoverTo.set(ht);
			if (hoverRawEvent == null) hoverRawEvent = new RawEvent();
			hoverRawEvent.rawEventType = hoverRawEvent.POINTER_EVENT;
			hoverRawEvent.surfaceEventType = PenEvent.MOUSE_HOVER_OVER;
			hoverRawEvent.x = hoverSpot.x;
			hoverRawEvent.y = hoverSpot.y;
			hoverRawEvent.modifiers = hoverMods;
			ht.surface.postRawEvent(hoverRawEvent);
			// This was posting an event out of the window event queue.
			//ht.doPostEvent(PenEvent.MOUSE_HOVER_OVER,0,hoverSpot.x,hoverSpot.y,hoverMods,System.currentTimeMillis(),null);
		}
	}
};

private void handleHover(int eventType, int x, int y, int modifiers)
{
	if (!PenEvent.isSurfacePointerEvent(eventType)) return;
	if (eventType == PenEvent.MOUSE_HOVER_OVER || eventType == PenEvent.MOUSE_STOP_HOVER) return;
	synchronized(controlHoverCountdown){
		Window ht = (Window)sentHoverTo.get();
		if (ht != null) ht.doPostEvent(PenEvent.MOUSE_STOP_HOVER,0,hoverSpot.x,hoverSpot.y,hoverMods,System.currentTimeMillis(),null);
		sentHoverTo.clear();
		shouldSendHoverTo.clear();
		/*
		WeakSet me = null;
		if (shouldSendHoverTo.get() == this)
			me = shouldSendHoverTo;
		if (me == null && sentHoverTo.get() == this)
			me = shouldSendHoverTo;
		shouldSendHoverTo.clear(); 
		sentHoverTo.clear();
		*/
		controlHoverCountdown.cancelTick();
		if (eventType == PenEvent.MOUSE_MOVE){
			hoverSpot.set(x,y);
			hoverMods = modifiers;
			shouldSendHoverTo.set(this);
			controlHoverCountdown.requestTick(500);
		}
	}
}
/**
 * Wait for the the window to get the resize event for the specified number
 * of milliseconds.
 * @return true if it got the resize, false if not.
 */
public boolean waitForResize(int howLong)
{
	try{
		synchronized(firstResizeLock){
			if (!gotFirstResize){
				try{
					if (howLong < 0) firstResizeLock.wait();
					else firstResizeLock.wait(howLong);
				}catch(InterruptedException e){}
				gotFirstResize = true;
			}
		}
		return gotFirstResize;
	}catch(Throwable t){
		return false;
	}
}
/**
 * Wait for the the window to get the resize event for 1 second.
 * @return true if it got the resize, false if not.
 */
public boolean waitForResize()
{
	return waitForResize(1000);
}
static WindowEventThread findOrMakeWindowEventThread()
{
	Thread th = Thread.currentThread();
	return th instanceof WindowEventThread ? (WindowEventThread)th : new WindowEventThread(true);
}

public static void penDragToMove()
{
	WindowEventThread._inPenDrag = false;
	WindowEventThread._freeDragging = true;
}

static boolean actuallyInPress = false;
PenHandler penHandler = new PenHandler();

public static void resetCursor()
{
	synchronized(openCloseLock){
		for (int i = 0; i<openWindows.size(); i++)
			((Window)openWindows.get(i)).penHandler.resetCursor();
	}
}
static void topFrameChanging()
{
	synchronized(openCloseLock){
		for (int i = 0; i<openWindows.size(); i++)
			((Window)openWindows.get(i)).penHandler.topFrameChanging();
	}
}
public void removeTip()
{
	penHandler.removeTip();
}
public boolean tipIsDisplayed()
{
	return penHandler.tipIsDisplayed();
}
public void setCursorFor(Control c, int cursor)
{
	penHandler.setCursorFor(this,c,cursor);
}
//===================================================================
void doPostEvent(int type, int key, int x, int y, int modifiers, long timeStamp, Object data)
//===================================================================
{
	WindowEventThread wt = findOrMakeWindowEventThread();
	handleHover(type,x,y,modifiers);
	Event event = null;
	Control target = focus;
	try{
	if (target == null) target = this;
	if (type != PenEvent.PEN_MOVE && type != PenEvent.MOUSE_HOVER_OVER && type != PenEvent.MOUSE_STOP_HOVER) 
		removeTip();

	if (type >= WindowEvent.CLOSE && type <= WindowEvent.DATA_DROPPED){
		target = this;
		//if (type == WindowEvent.CLOSE) Vm.messageBox("Close!","Closing: "+getClass().getName(),0);
		//if (type == WindowEvent.HIBERNATE) Vm.messageBox("Hibernate!","Hibernating: "+getClass().getName(),0);
		WindowEvent we = new WindowEvent();
		we.type = type;
		we.x = x;
		we.y = y;
		we.data = data;
		event = we;
		event.window = this;
	}else if (type >= SurfaceTextEvent.TEXT_EVENT_FIRST && type <= SurfaceTextEvent.TEXT_EVENT_LAST){
		SurfaceTextEvent te = new SurfaceTextEvent();
		target = (Control)te.inputParameters.control;//inputingText;
		if (target == null) return;
		/*
		if (type == SurfaceTextEvent.TEXT_ENTERED){
			surface.exitNativeDialog();
			inputingText = null;
		}
		*/
		te.type = type;
		te.entered = TextInputParameters.fixEditedText((String)data);
		te.flags = modifiers;
		event = te;
	}else if (type == KeyEvent.KEY_PRESS || type == KeyEvent.KEY_RELEASE)
		{
		if (Gui.hasSoftKeys){
			if (key == IKeys.F1) key = IKeys.SOFTKEY1;
			if (key == IKeys.F2) key = IKeys.SOFTKEY2;
		}
		Application.lastKeyPress = System.currentTimeMillis();
		KeyEvent _keyEvent = wt._keyEvent;
		_keyEvent.type = type;
		_keyEvent.key = key;
		_keyEvent.modifiers = modifiers;
		if (type == KeyEvent.KEY_PRESS)
			if (inKeyPress)
				_keyEvent.modifiers |= IKeys.REPEATED;
			else
				inKeyPress = true;
		else
			inKeyPress = false;
		event = _keyEvent;
		event.window = this;
		if (focus == null)
			focus = this;
		if (_keyEvent.key == IKeys.SOFTKEY1 || _keyEvent.key == IKeys.SOFTKEY2){
			if (SoftKeyBar.onKeyEvent(_keyEvent)) return;				
		}
		//System.out.println(_keyEvent);
		
		}
	//MLB
	else if (type == SurfaceSIPEvent.SIP_SHOWN || type == SurfaceSIPEvent.SIP_HIDDEN){
		SurfaceSIPEvent _sipEvent = wt._sipEvent;
		_sipEvent.type = type;
		_sipEvent.visibleWidth = type == SurfaceSIPEvent.SIP_SHOWN ? x : width;
		_sipEvent.visibleHeight = type == SurfaceSIPEvent.SIP_SHOWN ? y : height;
		//_sipEvent.desktopHeight = type  == _sipEvent.SIP_SHOWN ? key : height;
		event = _sipEvent;
		event.window = this;
		final Event sipEv = (Event)_sipEvent.getCopy(); 
		if (!gotFirstResize) {
			new Thread(){
				public void run(){
					waitForResize(1000);
					gotFirstResize = true;
					queueEvent(new DeferredEvent(sipEv,Window.this));
				}
			}.start();
			return;
		}
		if (type == SurfaceSIPEvent.SIP_SHOWN) {
			_sipEvent.handlingSipOn = true;
			/* Can't do this - resize in same thread now.
			synchronized(firstResizeLock){
				if (!gotFirstResize){
					try{
						Vm.debug("I wait!");
						firstResizeLock.wait(1000);
					}catch(InterruptedException e){}
					gotFirstResize = true;
				}
			}
			*/
		}
		handleSIPEvent(_sipEvent);
		return;
	}
	//END MLB
	else
		{
		PenEvent _penEvent = wt._penEvent;
		_penEvent.window = this;
		_penEvent.modifiers = modifiers;
		HoldDownDisplay.hd.cancel(type,x,y);
		// set focus to new control
		if (type == PenEvent.SCROLL_DOWN || type == PenEvent.SCROLL_UP){
			Application.lastKeyPress = System.currentTimeMillis();
			Control c = findChild(x, y);
			if (c != null)
				if (c.amOnTopFrame()){
					target = c;
					if (c != focus && !c.hasModifier(Disabled,true))
						focus = c;
				}else
					return;
		}else if (type == PenEvent.PEN_DOWN){
			WindowEventThread._freeDragging = false;
			actuallyInPress = true;
			downPoint.set(x,y);
			lastWindow = this;
			inFront = this;
			Rect r = getWindowRect(rbuff,false);
			if (mouseOffset == null) mouseOffset = new Point(r.x,r.y);
			else mouseOffset.set(r.x,r.y);
			r = getWindowRect(rbuff,true);
			mouseOffset.y += r.x; mouseOffset.y += r.y;
			//toFront();
			Application.lastKeyPress = System.currentTimeMillis();
			Control c = findChild(x, y);
			target = c;
			if (c != focus && !c.hasModifier(Disabled,true)){
				focus = c;
			}
			wt._inPenDrag = true;
			}
		else if (type == PenEvent.PEN_MOVED_OFF){
			//Vm.debug("Off!");
			_penEvent.window = this;
			_penEvent.type = PenEvent.PEN_MOVED_OFF;
			if (wt._inMouseMove || wt._inMouseSize) return;
			penHandler.handlePenMove(this,_penEvent,x,y);
			target = this;
		}else if (type == PenEvent.PEN_MOVED_ON){
			return; // Defer to PEN_MOVE
		}else if (type == PenEvent.PEN_MOVE){
			if (!Gui.hasMousePointer && !actuallyInPress) return;
			if (wt._inMouseMove || wt._inMouseSize) return;
			if (lastMoveX == x && lastMoveY == y && x != -1 && y != -1) return;
			lastMoveX = x; lastMoveY = y;
			if (wt._inPenDrag){
				type = PenEvent.PEN_DRAG;
				/*
				if (inFront != this) {
					inFront = this;
					toFront();
				}*/
				if (dragAndDropData != null){
					if ((modifiers & PenEvent.FROM_OTHER_WINDOW) != 0){
						target = findChild(x,y);
					}else{
						Rect r = getWindowRect(rbuff,false);
						mouseOffset = new Point(r.x,r.y);
						r = getWindowRect(rbuff,true);
						mouseOffset.y += r.x; mouseOffset.y += r.y;
						Window shouldGet = tryWindow(lastWindow,x,y,dragxy);
						if (shouldGet == this){
							target = findChild(x,y);
						}else if (shouldGet != null){
							shouldGet.sendDragAndDropPenEvent(PenEvent.PEN_MOVE,key,dragxy.x,dragxy.y,modifiers|PenEvent.FROM_OTHER_WINDOW,timeStamp,null);
							return;
						}else{
							for (int i = 0; i<openWindows.size(); i++){
								Window win = (Window)openWindows.get(i);
								if (win == lastWindow) continue;
								if ((win.getWindowFlags() & FLAG_IS_VISIBLE) == 0) continue;
								shouldGet = tryWindow(win,x,y,dragxy);
								if (shouldGet == this){
									target = findChild(x,y);
									break;
								}else if (shouldGet != null){
									shouldGet.sendDragAndDropPenEvent(PenEvent.PEN_MOVE,key,dragxy.x,dragxy.y,modifiers|PenEvent.FROM_OTHER_WINDOW,timeStamp,null);
									return;
								}
							}
							target = this;
						}
					}
				}
			}else{
				_penEvent.timeStamp = timeStamp;
				_penEvent.modifiers = modifiers;
				_penEvent.type = type;
				penHandler.handlePenMove(this,_penEvent,x,y);
				return;
			}
		}else if (type == PenEvent.PEN_UP){
			actuallyInPress = false;
			Control.firstPress = true;
			if (wt._inMouseMove||wt._inMouseSize){
				Window.resetCursor();
				int dx = x-downPoint.x;
				int dy = y-downPoint.y;
				if (wt._inMouseMove){
					Rect r = getWindowRect(new Rect(),false);
					r.x += dx;
					r.y += dy;
					setWindowRect(r,false);
				}else{
					if (x >= 4 && y >= 4){
						Rect r = getWindowRect(new Rect(),true);
						r.width = x;
						r.height = y;
						setWindowRect(r,true);

					}
				}
				wt._inMouseMove = wt._inMouseSize = false;
				return;
			}
			if (wt._inPenDrag){
				if (dragAndDropData != null){
					if ((modifiers & PenEvent.FROM_OTHER_WINDOW) != 0){
						target = findChild(x,y);
					}else{
						Window shouldGet = tryWindow(lastWindow,x,y,dragxy);
						if (shouldGet == this){
							target = findChild(x,y);
						}else if (shouldGet != null){
							shouldGet.sendDragAndDropPenEvent(PenEvent.PEN_UP,key,dragxy.x,dragxy.y,modifiers|PenEvent.FROM_OTHER_WINDOW,timeStamp,null);
							return;
						}else{
							for (int i = 0; i<openWindows.size(); i++){

								Window win = (Window)openWindows.get(i);
								if (win == lastWindow) continue;
								if ((win.getWindowFlags() & FLAG_IS_VISIBLE) == 0) continue;
								shouldGet = tryWindow(win,x,y,dragxy);
								if (shouldGet == this){
									target = findChild(x,y);
									break;
								}else if (shouldGet != null){
									shouldGet.sendDragAndDropPenEvent(PenEvent.PEN_UP,key,dragxy.x,dragxy.y,modifiers|PenEvent.FROM_OTHER_WINDOW,timeStamp,null);
									return;
								}
							}
							target = this;
						}
					}
				}
			}
			wt._inPenDrag = false;
			if (target == null || target == this)
				if (dragAndDropData != null){
					dragOverNew(null);
					((Control)dragAndDropSource).dataTransferCancelled(dragAndDropData);
					dragAndDropData = null;
					Window.resetCursor();
				}
		}
		_penEvent.type = type;
		_penEvent.x = x;
		_penEvent.y = y;

		// translate x, y to coordinate system of target
		Control c = target;//focus;
		while (c != null)
			{
			_penEvent.x -= c.x;
			_penEvent.y -= c.y;
			c = c.parent;
			}

		_penEvent.modifiers = modifiers;
		event = _penEvent;

		}
	event.target = target;//focus;
	event.timeStamp = timeStamp;
	if (event.target != null){
		//if (event.type == PenEvent.PEN_UP) Vm.debug(event.target.getClass().toString());
		KeyEvent _keyEvent = wt._keyEvent;
		PenEvent _penEvent = wt._penEvent;
		if (event == wt._keyEvent){
			if (_keyEvent.key == IKeys.PDA_CANCEL && _keyEvent.type == _keyEvent.KEY_PRESS && event.target != null){
				_keyEvent.key = IKeys.ESCAPE;
				if (!((Control)event.target).isSomeonesHotKey(_keyEvent))
				doPostEvent(WindowEvent.CLOSE,0,0,0,0,timeStamp,null);
				return;
			}
			if (_keyEvent.type == _keyEvent.KEY_PRESS && ((_keyEvent.modifiers & IKeys.CONTROL) != 0)){
				if (((Control)event.target).isSomeonesHotKey(_keyEvent))
					return;
			}
		}
		if (event == _keyEvent || event == _penEvent){
	
			if (WindowEventThread.canDeliverTo((Control)event.target))
				((Control)event.target).postUIEvent(event);
			else{
				if ((event == _keyEvent && event.type == KeyEvent.KEY_PRESS)
					|| (event == _penEvent && event.type == PenEvent.PEN_DOWN))
				Sound.beep(Sound.BEEP_GUI);//System.out.println("Rejecting: "+event);
			}
		}else
			((Control)event.target).postUIEvent(event);
	}
	if (type == PenEvent.PEN_UP){
		dragAndDropData = null;
		focus = Gui.focusedControl();
	}
	}catch(Throwable t){
		t.printStackTrace();
//		new ReportException(t,null,null,false).exec();
	}finally{
		if (event != null) {
			event.window = null;
			event.target = null;
		}
		/*
		if (Gui.setSipTo != -1) {
			//Vm.debug("Window Setting to: "+Gui.setSipTo);
			surface.setSIP(Gui.setSipTo);//,activeWindow);
		}else{
			//Vm.debug("Is -1");
		}
		*/
	}
	}
/**
 * Send a pen event. The target of the event must already be set.
 * @param ev
 */
//-------------------------------------------------------------------
public void dispatchPenEvent(PenEvent ev)
//-------------------------------------------------------------------
{
	Control c = (Control)ev.target;
	if (!WindowEventThread.canDeliverTo(c)){
		//System.out.println("Rejecting: "+ev);
		return;
	}
	while (c != null){
		ev.x -= c.x;
		ev.y -= c.y;
		c = c.parent;
	}
	if (ev.target != null)((Control)ev.target).postUIEvent(ev);
}
/*
public void queuePenEvent(Control forTarget, int xInControl, int yInControl, int modifiers, int type)
{
	Rect r = Gui.getAppRect(forTarget);
	_postEvent(type,0,r.x+xInControl,r.y+yInControl,modifiers,System.currentTimeMillis(),null);
}
*/
/*
public void queuePenEvent(int x, int y, int modifiers, int type)
{
	_postEvent(type,0,x,y,modifiers,System.currentTimeMillis(),null);
}
*/
/*
protected void doPostEvent(int type, int key, int x, int y, int modifiers, int timestamp, Object data)
{
	
}
*/
/** Returns the font metrics for a given font. */
public FontMetrics getFontMetrics(Font font)
{
	return surface.getFontMetrics(font);
}
/**
* This gets set to be true when the VM calls _doPaint() on the Window.
**/
Handle paintedHandle = new Handle();

/**
 * Wait until the Window receives its first Paint message.
 * @param timeOut The length of time in milliseconds to wait.
 * @return true if the paint message was received within the timeOut period, false if not.
 */
//===================================================================
public boolean waitUntilPainted(int timeOut)
//===================================================================
{
	if (!getWindowSurface().isVisible()) return false;
	try{
		while(true){
			try{
				return paintedHandle.waitUntilStopped(new TimeOut(timeOut));
				//return true;
			}catch(InterruptedException e){}
		}
	}finally{
	}
}
/**
 * Returns true if the Window has received its paint event yet.
 */
public boolean hasBeenPainted()
{
	return ((paintedHandle.check() & Handle.Stopped) != 0);
}
//===================================================================
public void resizeTo(int newWidth,int newHeight)
//===================================================================
{
	super.resizeTo(newWidth,newHeight);
	if (contents != null) {
		contents.setRect(0,0,this.width,this.height);
	}
}
/**
 * Queue an Event to the Window at the front of the Queue. This Event will be delivered
 * to the Window via its event thread and so will not conflict with other event
 * processing. If the target of the Event is not the Window, it will be delivered 
 * directly to the target.
 * @param ev the Event to deliver.
 */
public void queueEvent(Event ev)
{
	rawEventThread.queueEvent(ev);
}
private static Object windowRepaintLock = new Object();
private static boolean dpe = false;
/*
private synchronized void discardGraphics()
{
	//Vm.debug("Clearing: "+hashCode());
	if (myGraphics != null && graphicsAvailable){
		Graphics old = myGraphics;
		myGraphics = null;
		old.free();
	}else{ //Someone else may have the graphics.
		myGraphics = null;
	}
}
*/
void wasResizedTo(int newWidth, int newHeight)
{
	if (this.width == newWidth && this.height == newHeight) return;
	//discardGraphics();
	if (newWidth > 0 && newHeight > 0){
		resizeTo(newWidth,newHeight);
		synchronized(firstResizeLock){
			if (!gotFirstResize){
				gotFirstResize = true;
				firstResizeLock.notifyAll();
			}
		}
		if ((surface.getGuiFlags() & surface.GUI_FLAG_DONT_REPAINT_ON_RESIZE) == 0){
			Rect r = getCachedRect(0,0,width,height);
			repaintNow(null,r);
			cache(r);
		}
	}
}

int checkPlease = 0;

static int eventNumber;
/**
 * This calls handleWindowEvents(int minTime, int maxTime)
 * with a min time of 100ms an a maxTime of 1 second.
 */
public static void handleWindowEvents()
{
	handleWindowEvents(100,1000);
}
/**
 * Put the current Thread to sleep to allow window events to be handled.
 * @param minTime the minimum time to wait.
 * @param maxTime the maximum time to wait.
 */
public static void handleWindowEvents(int minTime, int maxTime)
{
	long now = System.currentTimeMillis();
	int ev = eventNumber;
	while(true){
		try{
			Thread.sleep(minTime);
		}catch(InterruptedException e){
			return;
		}
		if (ev == eventNumber) return;
		if (System.currentTimeMillis()-now >= maxTime) return;
	}
}
public synchronized void contentsAdded()
{
	holdEvents = false;
	notifyAll();
}

protected void handleSurfaceEvent(Event ev)
{
	synchronized(this){
		while(holdEvents && !hasClosed){
			try{
				wait();
			}catch(Exception e){}
		}
		if (hasClosed) return;
	}
	if (ev instanceof WindowEvent){
		System.out.println("What!");
		Object dt = ((WindowEvent)ev).data;
		if (dt != null) System.out.println("Dropped: "+mString.toString(dt));
	}
	eventNumber++;
	//Gui.setSipTo = -1;
	//int num = Vm.countObjects(false);
	//if (ev instanceof SurfaceResizeEvent || ev instanceof SurfacePaintEvent)
		//System.out.println(hashCode()+": "+ev.getClass().getName());
	try{
		//System.out.println(Thread.currentThread()+" In: "+ev.getClass());
	if (!(Thread.currentThread() instanceof WindowEventThread))
		System.out.println("Not a window thread: "+ev.getClass().getName());
	if (ev.target instanceof EventListener && ev.target != this && ev.target != surface){
		((EventListener)ev.target).onEvent(ev);
		return;
	}
	/*
	if (checkPlease != 0){
		System.out.println("Track!-------------------------------");
		System.out.println(Vm.trackObject(checkPlease,0));
		checkPlease = 0;
	}
	*/
	if (ev instanceof SurfacePaintEvent){
		//if (!gotFirstResize) System.out.println("Did not resize yet!");
		SurfacePaintEvent pe = (SurfacePaintEvent)ev;
		Rect area = getCachedRect(pe.x,pe.y,pe.width,pe.height);
		try{
			repaintNow(null,area);
		}finally{
			cache(area);
		}
		paintedHandle.set(Handle.Succeeded);
	}else if (ev instanceof SurfaceResizeEvent){
		SurfaceResizeEvent re = (SurfaceResizeEvent)ev;
		Gui.closePopups(this);
		wasResizedTo(re.newWidth,re.newHeight);
	}else if (ev instanceof SurfacePointerEvent){
		if (!dpe) {
			//Vm.debug(null,Vm.DEBUG_START_OBJECT_CREATION);
			//Vm.debug(Graphics.class,Vm.DEBUG_START_STACK_TRACE_FOR_CLASS);
			dpe = true;
		}
		SurfacePointerEvent sp = (SurfacePointerEvent)ev;
		//if (sp.type == SurfacePointerEvent.PEN_UP) Vm.debug("Release");
		/*
		if (ev.type == sp.PEN_DOWN){
			System.out.println(++checkPlease);
			new Exception().printStackTrace();
		}
		*/
		doPostEvent(ev.type,0,sp.x,sp.y,sp.modifiers,sp.timeStamp,null);
	}else if (ev instanceof SurfaceKeyEvent){
		SurfaceKeyEvent sk = (SurfaceKeyEvent)ev;
		doPostEvent(ev.type,sk.key,0,0,sk.modifiers,sk.timeStamp,null);
	}else if (ev instanceof SurfaceWindowEvent){
		SurfaceWindowEvent sw = (SurfaceWindowEvent)ev;
		if (sw.type == sw.CAN_DISPLAY){
			canDisplay = true;
		}
		if (sw.type == sw.CLOSE && Application.appHandler != null)
			if (Application.appHandler.handleMobileAppClosing(this,sw.flags))
				return;
			//Gui.flashMessage("Can't exit with that button!",this);
		doPostEvent(ev.type,0,sw.x,sw.y,0,sw.timeStamp,sw.data);
	}else if (ev instanceof SurfaceSIPEvent){
		SurfaceSIPEvent se = (SurfaceSIPEvent)ev;
		doPostEvent(se.type,se.visibleHeight,se.visibleWidth,se.visibleHeight,0,se.timeStamp,null);
	}else if (ev instanceof SurfaceTextEvent){
		SurfaceTextEvent se = (SurfaceTextEvent)ev;
		((Control)se.inputParameters.control).onEvent(se);
	}else
		onEvent(ev);
	}catch(Throwable t){
		if (Application.mainApp == myMainApp){
			t.printStackTrace();
		}
	}finally{
		if (ev instanceof SurfaceSIPEvent){
			SurfaceSIPEvent ss = (SurfaceSIPEvent)ev;
			if (!ss.forwarded) synchronized(openWindows){
				int sz = openWindows.size();
				for (int i = 0; i<sz; i++){
					Window w = (Window)openWindows.get(i); 
					if (w == this) continue;
					SurfaceSIPEvent sc = (SurfaceSIPEvent)ss.getCopy();
					sc.forwarded = true;
					w.queueEvent(sc);
				}
			}
		}
		//System.out.println(Thread.currentThread()+" Out: "+ev.getClass());
	}
		//num = Vm.countObjects(false)-num;
		//System.out.println(num);
}

/**
 * This gets called when WindowSurface events are generated by the underlying window surface.
 * @param ev the generated event.
 */
/*
protected void windowSurfaceEvent(Event ev)
{
	if (ev instanceof SurfacePaintEvent){
		SurfacePaintEvent pe = (SurfacePaintEvent)ev;
		synchronized(windowRepaintLock){
			repaintNow(null,new Rect(pe.x,pe.y,pe.width,pe.height));
		}
		paintedHandle.set(Handle.Succeeded);
	}else if (ev instanceof SurfaceResizeEvent){
		SurfaceResizeEvent re = (SurfaceResizeEvent)ev;
		if (re.newWidth > 0 && re.newHeight > 0){
			resizeTo(re.newWidth,re.newHeight);
			synchronized(firstResizeLock){
				if (gotFirstResize = false){
					gotFirstResize = true;
					firstResizeLock.notifyAll();
				}
			}
			repaintNow(null,new Rect(0,0,width,height));
		}
	}else if (ev instanceof SurfaceWindowEvent){
		SurfaceWindowEvent wse = (SurfaceWindowEvent)ev;
		if (wse.type == wse.CAN_DISPLAY){
			canDisplay = true;
		}
		_postEvent(wse.type,0,0,0,0,wse.timeStamp,null);
	}else if (ev instanceof SurfacePointerEvent){
		SurfacePointerEvent pe = (SurfacePointerEvent)ev;
		_postEvent(pe.type,0,pe.x,pe.y,pe.modifiers,pe.timeStamp,null);
	}else if (ev instanceof SurfaceKeyEvent){
		SurfaceKeyEvent se = (SurfaceKeyEvent)ev;
		_postEvent(se.type,se.key,0,0,se.modifiers,se.timeStamp,null);
	}else if (ev instanceof SurfaceSIPEvent){
		SurfaceSIPEvent se = (SurfaceSIPEvent)ev;
		_postEvent(se.type,se.desktopHeight,se.visibleWidth,se.visibleHeight,0,se.timeStamp,null);
	}
}
*/
//==================================================================
public void onEvent(Event ev)
//==================================================================
{
	if (ev instanceof DeferredEvent)
		ev = ev.cause;
	if (ev instanceof SurfaceSIPEvent){
		handleSIPEvent((SurfaceSIPEvent)ev);
		return;
	}
	try{
	ev.consumed = false;
	if (ev instanceof WindowEvent){
		Frame f = Gui.topFrame(this);
		if (f instanceof FormFrame){
			Form form = ((FormFrame)f).myForm;
			form.onEvent(ev);
			if (ev.consumed) return; 
		}
	}
	if (ev.type == WindowEvent.CLOSE){
		Frame f = Gui.topFrame(this);
		if (f instanceof FormFrame && f.isModal){
			((FormFrame)f).myForm.exit(-1);
			return;
		}
		ControlIterator it = contents.getCachedChildren(false);
		while (it.hasNext()){
			Object c = it.next();
			if (c instanceof FormFrame){
				((FormFrame)c).myForm.exit(-1);
			}
		}
		Cache.put(it);
	}else if (ev.type == WindowEvent.APP_MINIMIZE){
		minimizeApplication();
	}else if (ev.type == WindowEvent.APP_TO_FRONT){
		applicationToFront();
	}else if (ev.type == WindowEvent.ACTIVATE){
		//System.out.println("-Active: "+this+" f: "+focus+" -");
		if (focus != null) Gui.takeFocus(focus,ByRequest);
		Gui.notifyTop(this,true);
	}else if (ev.type == WindowEvent.DEACTIVATE){
		//System.out.println("-Deactivate: "+this+"-");
		Gui.notifyTop(this,false);
		/*
	}else if (ev.type == ControlEvent.FOCUS_IN){
		ewe.sys.Vm.debug("Focus in: "+ev.target.getClass());		
	}else if (ev.type == ControlEvent.FOCUS_OUT){
		ewe.sys.Vm.debug("Focus out: "+ev.target.getClass());		
		*/
	}else{
		super.onEvent(ev);
	}
	}catch(Throwable t){
		t.printStackTrace();
	}
}

public boolean create(WindowSurface alreadyCreated)
{
	return create(alreadyCreated,null);
}
public boolean create(WindowCreationData wcd)
{
	return create(null,wcd);
}
private boolean create(WindowSurface surfaceToUse, WindowCreationData wcd)
{
	/*
	surface.addListener(new EventListener(){
		public void onEvent(Event ev){
			windowSurfaceEvent(ev);
		}
	});
	*/
	Window ma = Application.mainApp;
	if (ma != null && ma.holdEvents && this != ma){
		synchronized(ma){
			ma.holdEvents = false;
			ma.notifyAll();
		}
	}
	if (wcd != null) text = wcd.title;
	if (text == null) text = "";
	//if (!(this instanceof MainWindow)) 
	add(contents);
	contents.make(false);
	if ((modifiers & DontBuffer) != 0)
		contents.modifyAll(DontBuffer,0);
		if (surfaceToUse != null) {
			surface = surfaceToUse;
			surface.applicationWindow = this;
		}else{
			surface.applicationWindow = this;
			wcd.flagsToSet |= FLAG_CAN_RESIZE;
			wcd.flagsToClear &= ~ FLAG_CAN_RESIZE;
			surface.create(wcd);
		}
		bufferFullWindow = (surface.getGuiFlags() & WindowConstants.GUI_FLAG_REPAINT_FULL_CONTROL) != 0;
		synchronized(openCloseLock){
			rawEventThread = new WindowRawEventThread(this);
			new WindowEventThread(this);
			firstDisplay = canDisplay = true;
			openWindows.add(Window.this);
		}	
		if (wcd != null)
			if ((wcd.flagsToClear & FLAG_IS_VISIBLE) == 0){
				activeWindow = Window.this;
				checkAppKeys();
			}
	Rect r = new Rect();
	surface.getClientRect(r);
	wasResizedTo(r.width,r.height);
	return true;
}
/**
* This creates the native window associated with this Window object.
**/
//===================================================================
public boolean create(Rect location,final String title,final int flagsToSet,final int flagsToClear,final Object extra)
//===================================================================
{
	//windowThread.making(Window.this);
	WindowCreationData wcd = (extra instanceof WindowCreationData) ?
		(WindowCreationData)extra : new WindowCreationData();
	containing = (Window)wcd.parentWindow;
	wcd.bounds = new Rect().set(location);
	wcd.title = title;
	wcd.flagsToSet = flagsToSet;
	wcd.flagsToClear = flagsToClear;
	return create(wcd);
}
//
private static boolean capturedKeys = false;
//
private static void captureAppKeys(int options,Window win)
{
	if (win == null) win = Application.mainApp;
	win.surface.captureAppKeys(options);
}
/**
* This will capture the WindowCE special App keys (the non-cursor buttons generally
* found on the bottom). Calling this with options set to 1 will cause WinCE to
* not capture the App keys as ordinary keypresses, instead of special application
* launch buttons. These keypresses are then passed via a normal key event to your
* application with the APP0 to APP15 key codes. Calling this with options set to 0 will
* cause WinCE to no longer capture these keys and they will be re-associated with
* their application launch functionality.
* @param options A value of 1 causes WinCE to capture the keys. A value of 0 cause WinCE to
* release the keys.
* @return The same value as options.
*/
//===================================================================
public static int captureAppKeys(int options)
//===================================================================
{
	if (options != 0) capturedKeys = true;
	else capturedKeys = false;
	captureAppKeys(options, capturedKeys ? activeWindow : null);
	return options;
}

//-------------------------------------------------------------------
private static void checkAppKeys()
//-------------------------------------------------------------------
{
	if (capturedKeys) captureAppKeys(1,activeWindow);
}

public static void clearQueuedMoves()
{
	//FIXME Window.clearQueuedMoves()
}
public static boolean nextMouseIsMove()
{
	//FIXME Window.nextMouseIsMove()
	return false;
}
//private native boolean _nativeTextInput(TextInputParameters tip);

//private static Control inputingText;
/**
This returns whether the system is within a native input or native dialog.
If this is the case, then no Gui input will be sent to the Ewe Gui system.
**/
//===================================================================
public static boolean inNativeInput()
//===================================================================
{
	return WindowSurface.inNativeTextInput() || WindowSurface.inNativeDialog();
	//return inputingText != null || WindowSurface.inNativeDialog();
}
/**
 * This starts a native text input for a control.
 * @param tip the TextInputParameters to use.
 * @return true if the input started successfully, false if it could not start because
 * another native input is already active.
 * @throws NoSuchMethodError if native text input is not possible at all.
 */
//===================================================================
public boolean textInput(final TextInputParameters tip)
throws NoSuchMethodError
//===================================================================
{
	return surface.startNativeInput(tip);
}

public static void enterNativeDialog(){WindowSurface.enterNativeDialog();}
public static void exitNativeDialog(){WindowSurface.exitNativeDialog();}
private Stack sipMods = new Stack();
protected boolean undoSipChanges(boolean repaint)
{
	boolean resized = false;
	synchronized(sipMods){
		//Vm.debugStackTrace();
		if (sipMods.empty()) return false;
		while(!sipMods.empty()){
			Tag t = (Tag)sipMods.pop();
			if (t.value == null){
				contents.y += t.tag;
				contents.height -= t.tag;
			}else{
				Control c = (Control)t.value;
				if (c instanceof SIPFiller){
					while(c != null && !(c instanceof Form))
						c = c.parent;
					if (c instanceof Form)
						((Form)c).unfillSip();
				}
				c.y -= t.tag;
				c.height += t.tag;
				resized = true;
			}
		}
	}
	if (repaint) contents.repaintNow();
	return resized;
}

//private WeakRef bottom = new WeakRef();
/**
 * You can only have this work for a single Control per window.
 * @param c the Control to keep below the SIP.
 */
/*
public void setBelowSip(Control c)
{
	bottom.set(c);
}
private void contentsShifted()
{
	if (true) return;
	Control c = (Control)bottom.get();
	//System.out.println("Bottom: "+c);
	if (c != null){
		c.y = contents.height-c.height;
	}
}
*/
/**
* This returns null if the control is not covered by the SIP. If it is
* covered it returns a Rect representing the uncovered area of the screen.
**/
//===================================================================
public Rect checkSipCoverage(Control who)
//===================================================================
{
	if (true) return null;
	Vm.debugStackTrace();
	int cy = contents.y;
	if (visibleHeight == 0) return null;
	Window w = null;
	Frame f = null;
	Rect r = getCachedRect(0,0,who.width,who.height);
	try{
	for (Control c = who; c != null; c = c.parent){
		if (c instanceof Window) {
			w = (Window)c;
			break;
		}else if (f == null && c instanceof Frame)
			f = (Frame)c;
		r.x += c.x;
		r.y += c.y;
	}
	/*
	Rect r = Gui.getRectInWindow(who,new Rect(),false);//true);
	if (r == null) return null;
	*/
	if (w == null || f == null) return null;
	if (f.isPopup()) return null;
	boolean rp = undoSipChanges(false);
	if (rp){
		r.width = who.width;
		r.height = who.height;
		for (Control c = who; c != null; c = c.parent){
			r.x += c.x;
			r.y += c.y;
		}
	}
	try{
	ControlGroup cg = ControlGroup.getExposeGroupFor(who);
	if (cg != null) cg.getRectInParent(w,r);
	Rect r2 = new Rect();
	Rect r3 = new Rect();
	w.getWindowRect(r3,false);
	w.getWindowRect(r2,true);
	r2.x += r3.x; r2.y += r3.y;
	if (r2.y+r.y+r.height <= visibleHeight) return null;
	//
	// Try to move the contents Frame so that the control is 
	// no longer covered.
	//
	
	if (r3.y > 0) {
		// If the window is not all the way up, move it all the way up
		// and try again.
		r3.x = r3.y = 0;
		w.setWindowRect(r3,false);
		w.getWindowRect(r3,false);
		w.getWindowRect(r2,true);
		r2.x += r3.x; r2.y += r3.y;
		if (r2.y+r.y+r.height <= visibleHeight) return null;
	}
	
	int moveUp = r2.y-(visibleHeight-r.y-r.height);
	if (moveUp <= 0) return null;
	contents.y -= moveUp;
	contents.height += moveUp;
	sipMods.push(new Tag(moveUp,null));
	if (r.y < -contents.y){
		int adjust = (-contents.y-r.y);
		who.y += adjust;
		who.height -= adjust;
		sipMods.push(new Tag(adjust,who));
	}
	return null;
	}finally{
		if (rp ||(contents.y != cy)){
			//Vm.debug("Repaint: "+rp +", "+cy+","+contents.y);
			repaintNow();
		}
	}
	}finally{
		r.cache();
	}
	//return r.set(0,0,visibleWidth,visibleHeight);

/*
	Rect r = Gui.getRectInWindow(who,new Rect(),true);
	if (r == null) return null;
	Window w = who.getWindow();
	if (w == null) return null;
	//Rect r2 = w.getWindowRect(new Rect(),true);
	if (r.y+r.height < visibleHeight || visibleHeight == 0) return null;
	return r.set(0,0,visibleWidth,visibleHeight);
*/
}
private Handle sipThread;
private Image sipOverLay;
private ImageBuffer sipBuffer;

//-------------------------------------------------------------------
private void showSipCoverage()
//-------------------------------------------------------------------
{
	if (visibleHeight != 0 && Vm.getParameter(Vm.SIMULATE_SIP) != 0){
		//System.out.println("VH: "+visibleHeight);
		Rect r2 = getWindowRect(new Rect(),true);
		Rect r3 = getWindowRect(new Rect(),false);
		r2.x += r3.x; r2.y += r3.y;
		int yy = visibleHeight-r2.y;
		if (yy > height) yy = height;
		if (yy >= height) return;
		//System.out.println(visibleHeight);
		int h = height-yy;
		if (h <= 26) return;
		/*
		BufferedGraphics bgr = getGraphics();
		if (bgr == null) return;
		Graphics gr = bgr.getGraphics();
		*/
		if (sipBuffer == null) sipBuffer = new ImageBuffer();
		Graphics gg = sipBuffer.get(width,h,true);
		if (sipOverLay == null || sipOverLay.getHeight() != h-26){
			if (sipOverLay != null) sipOverLay.free();
			int hh = h-26;
			PixelBuffer pb = new PixelBuffer(width,hh);
			Graphics g = pb.getDrawingBuffer(null,null,0.5);
			g.setColor(Color.LightBlue);
			g.fillRect(0,0,width,hh);
			g.changePen(Color.Black,Pen.SOLID,3);
			g.drawLine(0,1,width,1);
			g.drawLine(0,hh-2,width,hh-2);
			g.changePen(Color.Black,Pen.DASH,3);
			g.drawLine(0,hh/2,width,hh/2);
			pb.putDrawingBuffer(pb.PUT_SET);
			sipOverLay = pb.toImage();
			pb.free();
		}
		
		gg.translate(0,-yy);
		repaintNow(gg,new Rect(0,yy,width,h));
		gg.translate(0,yy);
		sipOverLay.draw(gg,0,0,0);
		BufferedGraphics bg = getGraphics(0,yy,sipBuffer.image.getWidth(),sipBuffer.image.getHeight());
		if (bg == null) return;
		bg.getGraphics().drawImage(sipBuffer.image,0,yy);
		bg.release();
		
	}
}
/**
 * Use this to enable or disable the accepting of dropped files.
 * @param accept true to enable, false to disable.
 */
public void acceptDroppedFiles(boolean accept)
{
	if (surface != null) surface.acceptDroppedFiles(accept);
}
/*
//===================================================================
public void handleSipEvent(SurfaceSIPEvent ev)
//===================================================================
{
	visibleWidth = ev.type == ev.SIP_SHOWN ? ev.visibleWidth : 0;//width;
	visibleHeight = ev.type == ev.SIP_SHOWN ? ev.visibleHeight : 0;//height;
	Control c = Gui.focusedControl();
	if (c != null){
		if (contains(c)){
			Frame f = c.getFrame();
			if (f != null)
				if (!f.isPopup())
					f.onEvent(ev);
			if (ev.type == ev.SIP_SHOWN)
				if (c instanceof EditControl)
					((EditControl)c).checkSipCoverage();
		}
	}
	Frame.checkSip(ev);
	if (visibleHeight != 0 && (Vm.getParameter(Vm.SIMULATE_SIP) != 0)){
		if (sipThread == null){
			sipThread = new Task(){
				protected void doRun(){
					while(!shouldStop){
						showSipCoverage();
						sleep(250);
					}
				}
			}.start();
		}
	}else if (sipThread != null){
		sipThread.stop(0);
		sipThread = null;
		repaintNow();
	}
}
*/
}
//####################################################
