/*
 * Created on Feb 7, 2006
 *
 * Michael L Brereton - www.ewesoft.com
 * 
 * 
 */
package eve.ui.advanced.database;

import eve.database.Database;
import eve.database.DatabaseTypes;
import eve.database.DatabaseUtils;
import eve.ui.CellPanel;
import eve.ui.CheckBox;
import eve.ui.Control;
import eve.ui.DateTimeInput;
import eve.ui.Input;
import eve.ui.InputStack;
import eve.ui.data.Editor;
import eve.ui.event.ControlEvent;

/**
 * @author Michael L Brereton
 *
 */
//####################################################
public class DatabaseUIUtils implements DatabaseTypes{
	
//	##################################################################
	static class IncludedField extends CellPanel{
//	##################################################################

	CheckBox enabler;
	Control input;
//	===================================================================
	IncludedField(Database db, int field, Editor ed)
//	===================================================================
	{
		int type = db.getFieldType(field);
		addLast(enabler = new CheckBox(db.getFieldHeader(field)));
		String fieldName = db.getFieldName(field);
		if (ed != null)
			ed.addField(enabler,"_include_"+fieldName+"$Z");
		Control c = getFieldInput(type);
		if (ed != null){
			fieldName += '$'+DatabaseUtils.getTypeSpecifier(type,true);
			ed.addField(c,fieldName);
		}
		c.modify(c.Disabled,0);
		addLast(c);
		input = c;
	}
//	===================================================================
	public void onControlEvent(ControlEvent ev)
//	===================================================================
	{
		if (ev.type == ev.PRESSED && ev.target == enabler){
			if (enabler.getState()) input.modify(0,input.Disabled);
			else input.modify(input.Disabled,0);
			input.repaintNow();
		}
		super.onControlEvent(ev);
	}
//	##################################################################
	}
//	##################################################################
//	===================================================================
	public static Control getIncludedField(Database db,int field,Editor ed)
//	===================================================================
	{
		return new IncludedField(db,field,ed);	
	}
//	===================================================================
	public static Control getFieldInput(int fieldType)
//	===================================================================
	{
		switch(fieldType){
			case DATE: case TIME: case DATE_TIME: case TIMESTAMP:
				//return new ewe.ui.DateDisplayInput();
				return new DateTimeInput();
			case BOOLEAN:
				return new CheckBox();
			default:
				return new Input();
		}
	}
//	===================================================================
	public static Control addToInputStack(Database db,int field,InputStack is,Editor ed)
	throws IllegalArgumentException
//	===================================================================
	{
		String fieldHeader = db.getFieldHeader(field);
		int type = db.getFieldType(field);
		Control c = getFieldInput(type);
		is.add(c,fieldHeader);
		if (ed != null){
			String fieldName = db.getFieldName(field);
			fieldName += '$'+DatabaseUtils.getTypeSpecifier(type,true);
			ed.addField(c,fieldName);
		}
		return c;
	}

}

//####################################################
