package eve.ui.advanced.database;
import java.util.Vector;

import eve.data.PropertyList;
import eve.database.Database;
import eve.database.DatabaseTypes;
import eve.database.DatabaseUtils;
import eve.database.EntrySelector;
import eve.database.FoundEntries;
import eve.math.Decimal;
import eve.sys.DayOfYear;
import eve.sys.Time;
import eve.sys.TimeOfDay;
import eve.ui.Frame;
import eve.ui.data.Editor;
import eve.ui.data.EditorStack;
import eve.ui.data.InputObject;
import eve.util.IntArray;

//##################################################################
public class GetSearchCriteria extends InputObject implements DatabaseTypes{
//##################################################################

Database db;
int[] criteria;
public boolean _filter = false;

//===================================================================
public GetSearchCriteria(Database db)
//===================================================================
{
	this(db,null);
}
//===================================================================
public GetSearchCriteria(Database db, int sortID)
//===================================================================
{
	this(db,db.toCriteria(sortID));
}
//===================================================================
public GetSearchCriteria(Database db, int[] criteria)
//===================================================================
{
	this.criteria = DatabaseUtils.copyCriteria(criteria);
	this.db = db;
}
//===================================================================
public GetSearchCriteria(FoundEntries fe)
//===================================================================
{
	this(fe.getDatabase(),fe.getSortCriteria());
}
protected void addingField(String name,int type)
{
	PropertyList pl = getProperties();
	switch(type){
		case TIMESTAMP: case DATE_TIME:
			pl.defaultTo(name,new Time()); break;
		case DATE:
			pl.defaultTo(name,new DayOfYear()); break;
		case TIME:
			pl.defaultTo(name,new TimeOfDay()); break;
		case DECIMAL:
			pl.defaultTo(name,new Decimal(0)); break;
	}
}
//-------------------------------------------------------------------
protected void setupInputStack(EditorStack is,Editor ed)
//-------------------------------------------------------------------
{
	is.setPreferredSize(200,-1);
	if (criteria == null){	
		ed.enableScrolling(true,true);
		int[] all = db.getFields();
		for (int i = 0; i<all.length; i++){
			if (db.getFieldName(all[i]).startsWith("_")) continue;
			is.add(DatabaseUIUtils.getIncludedField(db,all[i],ed),null);
			addingField(db.getFieldName(all[i]),db.getFieldType(all[i]));
		}
	}else{
		for (int i = 0; i<criteria.length; i++){
			int field = DatabaseUtils.criteriaToField(criteria[i]);
			DatabaseUIUtils.addToInputStack(db,field,is,ed);
			addingField(db.getFieldName(field),db.getFieldType(field));
		}
		ed.addField(is.addCheckBox("Filter Entries"),"_filter$Z");
	}
	super.setupInputStack(is,ed);
}
//===================================================================
public EntrySelector input(Frame parent,String title)
//===================================================================
{
	PropertyList pl = new PropertyList();
	pl = input(parent,title,pl);
	if (pl == null) return null;
	Vector v = new Vector();
	IntArray fields = new IntArray();
	//ewe.sys.Vm.debug(pl.toString());
	if (criteria == null){
		int[] all = db.getFields();
		for (int i = 0; i<all.length; i++){
			if (db.getFieldName(all[i]).startsWith("_")) continue;
			String f = db.getFieldName(all[i]);
			if (!pl.getBoolean("_include_"+f,false)) continue;
			Object value = pl.getValue(f,null);
			if (value != null) {
				v.add(value);
				fields.add(all[i]);
			}
		}
		int[] c = new int[fields.length];
		for (int i = 0; i<fields.length; i++){
			c[i] = DatabaseUtils.toCriteria(fields.data[i],db.getFieldType(fields.data[i]),db.SORT_IGNORE_CASE);
		}
		EntrySelector es = new EntrySelector(db,v,c,true);
		es.forceAFilterSearch = _filter;//pl.getBoolean("_filter",false);
		return es;
	}else{
		for (int i = 0; i<criteria.length; i++){
			int field = DatabaseUtils.criteriaToField(criteria[i]);
			String fieldName = db.getFieldName(field);
			Object value = pl.getValue(fieldName,null);
			if (value != null) v.add(value);
		}
		//ewe.sys.Vm.debug(v.toString());
		//System.out.println(v+", "+getProperties());
		EntrySelector es = new EntrySelector(db,v,criteria,true);
		es.forceAFilterSearch = _filter;//pl.getBoolean("_filter",false);
		return es;
	}
	
}
//===================================================================
public EntrySelector input(String title)
//===================================================================
{
	return input(null,title);
}
//##################################################################
}
//##################################################################

