package eve.ui.data;
import java.lang.reflect.Method;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import eve.sys.DataConverter;
import eve.sys.Event;
import eve.sys.FieldTransfer;
import eve.sys.Reflection;
import eve.sys.Vm;
import eve.sys.Wrapper;
import eve.ui.Application;
import eve.ui.Button;
import eve.ui.CellPanel;
import eve.ui.Control;
import eve.ui.Form;
import eve.ui.Gui;
import eve.ui.IFileSaver;
import eve.ui.Menu;
import eve.ui.MenuItem;
import eve.ui.SoftKeyBar;
import eve.ui.WindowThreadTask;
import eve.ui.event.ControlEvent;
import eve.ui.event.DataChangeEvent;
import eve.ui.event.ListEvent;
import eve.ui.event.MenuEvent;
import eve.ui.event.SoftKeyEvent;
import eve.util.ObjectIterator;
import eve.util.VectorIterator;
import eve.util.mString;
/**
* An Editor is a special type of Form that can automatically transfer data between
* public fields in an Object and Gui Controls within the Editor.<p>
* You should consult the Eve Programming Guide for instructions on how
* to use this correctly.<p>
**/
//##################################################################
public class Editor extends Form implements FieldListener{
//##################################################################
/**
* This holds a list of all the FieldTransfers used in the Editor.
**/
public Vector transfers = new Vector();
/**
* This holds a list of all the fieldListeners listening to the Editor.
**/
public Vector fieldListeners = new Vector();
/**
* @deprecated.
**/
public int coroutineStackSize = 0;//100;
/**
* This must be set before adding fields.
**/
public Class objectClass;
/**
* This is only necessary if you are using the objects _getField() and _setField()
* methods for field transfer. If not, you can leave this as null.
**/
public Object sampleObject;
/**
* This is the object currently being edited. Use setObject() to set it.
**/
public Object myObject;

FieldTransfer myTransfer = new FieldTransfer("_editor_");
boolean dontSendEachFieldTransferEvent = false;
boolean fieldsSet = false;
{
	maximizeOnPDA();
}
private boolean canUseField(String name)
{
	if (name == null || name.length() == 0) return false;
	FieldTransfer ft = findFieldTransfer(name);
	return ft == null; 
}
private void addSoftKeyMenu(Menu m)
{
	int num = m.itemsSize();
	for (int i = 0; i<num; i++){
		MenuItem mi = m.getItemAt(i);
		int opt = mi.modifiers;
		if ((opt & mi.Separator) != 0) continue;
		if (mi.subMenu != null){
			addSoftKeyMenu(mi.subMenu);
			continue;
		}
		if (mi.action == null) mi.action = mi.label;
		String nm = mi.action;
		if (!canUseField(nm)) continue;
		/*
		if (mi.data != null) continue;
		mi.data = new Button(nm);
		addExternalField((Control)mi.data,nm);
		*/
		if (mi.data == null)
			mi.data = new Button(nm);
		if (mi.data instanceof Button)
			addExternalField((Control)mi.data,nm);
	}
}
/**
 * Call this after setting up a SoftKeyBar and after setting all fields. If there are any buttons
 * in the SoftKeyBar that were added as fields and also added to the SoftKeyBar
 * then this will ensure that their generated proxy events will be sent to
 * the Editor. Otherwise, each item in the SoftKeyBar is added as an Action
 * item.
 * @param b the SoftKeyBar created for the Editor.
 */
public void addSoftKeysAsActions(SoftKeyBar b)
{
	if (b == null) return;
	int n = b.numberOfKeys();
	for (int i = 0; i<n; i++){
		Control c = b.getKey(i+1);
		if (c == null) continue;
		if (c instanceof Button){
			Button bc = (Button)c;
			if (bc.promptControl != null && findFieldTransfer(bc.promptControl) != null){
				bc.promptControl.addListener(this);
			}else{
				String name = bc.action;
				if (name == null) name = bc.text;
				if (!canUseField(name)) continue;
				addExternalField(bc,name);
			}
		}else if (c instanceof Menu){
			addSoftKeyMenu((Menu)c);
		}
	}
}
/**
 * Switch the focus to a particular field.
 * @param fieldName the name of the field.
 * @param inCurrentThread if this is true then Gui.takeFocus() is called
 * directly. If it is false it is called as a separate event in the Window thread.
 */
public void focusOn(String fieldName,boolean inCurrentThread)
{
	final Control c = findControlFor(fieldName);
	if (c == null) return;
	if (inCurrentThread) Gui.takeFocus(c, ByRequest);
	else new WindowThreadTask(){
		protected void doRun() {
			Gui.takeFocus(c, ByRequest);
		}
	}.run(c,true);
}
//===================================================================
public void dismantle(Control stopAt)
//===================================================================
{
	if (stopAt == this) return;
	transfers.clear();
	fieldListeners.clear();
	super.dismantle(stopAt);
}
/**
If a ScrollablePanel has been setup in the Editor, and the property "MainScrollBarPanel" has
been added to the PropertyList of this Editor, then this will either enable or disable the
scrolling of the scrollbar.
@param enable true to enable the scrolling, false to disable it.
@return true if a "MainScrollBarPanel" was defined, false if it was not.
@deprecated use Form.enableScrolling(boolean horizontally, boolean vertically) instead.
**/
//===================================================================
public boolean enableScrolling(boolean enable)
//===================================================================
{
	return enableScrolling(enable,enable);
}
/**
 * Create a new Editor that will initially be set to edit itself.
 */
//===================================================================
public Editor() {this(null);}
//===================================================================
/**
 * Create a new Editor to edit the type of Object specified.
 * @param objectOrClassOrReflectToEdit Can be either an instance of the
	Object to edit, or a Class or Class Object that represents the Object class to edit.
 */
//===================================================================
public Editor(Object objectOrClassOrReflectToEdit)
//===================================================================
{
	if (objectOrClassOrReflectToEdit == null) objectOrClassOrReflectToEdit = this;
	objectClass = Reflection.toClass(objectOrClassOrReflectToEdit);
	setObject(Reflection.toObject(objectOrClassOrReflectToEdit));
}
//-------------------------------------------------------------------
protected boolean canExit(int exitCode)
//-------------------------------------------------------------------
{
	IFileSaver fs = getFileSaver();
	if (fs == null) return true;
	if (!getProperties().getBoolean("CheckFileSaverOnExit",false)) return true;
	return fs.checkExit(this,getProperties().getValue("ObjectToSave",myObject),exitCode);
}

//===================================================================
public void setFields(Object objectOrClassOrReflectToEdit)
//===================================================================
{
	setFields(null,objectOrClassOrReflectToEdit);
}
public void setFields(CellPanel destination,Object objectOrClassOrReflectToEdit)
{
	if (destination == null) destination = this;
	objectClass = Reflection.toClass(objectOrClassOrReflectToEdit);
	LiveObject.addObjectToPanel(destination,this,objectOrClassOrReflectToEdit,false).close(true);
	setObject(Reflection.toObject(objectOrClassOrReflectToEdit));
}
/**
* This is used to set the Object to be edited by the Editor. This will automatically
cause a toControls() to be called, which causes the on-screen Controls to reflect the data
in the Object.
* @param obj The Object to be edited.
*/
//===================================================================
public void setObject(Object obj)
//===================================================================
{

	if (obj != null){
		if (objectClass == null){
			System.err.println("eve.ui.data.Editor.setObject() called but no objectClass defined.");
			return;
		}else if (!objectClass.isInstance(obj)){
			System.err.println("eve.ui.data.Editor.setObject() called but the object: "+obj.getClass().getName()+" is not compatible with objectClass: "+objectClass.getName());	
			return;
		}
	}
	//if (obj instanceof Vector) ewe.sys.Vm.debug(objectClass.getReflectedClass().getName());
	if (myObject != null && myObject != this)
		fieldListeners.remove(myObject);
	myObject = obj;
	if (obj != null) 
		if (obj instanceof FieldListener && obj != this) 
			fieldListeners.add(obj);
	if (fieldsSet) postEditorEvent(EditorEvent.OBJECT_SET,obj);
	if (made) 
		toControls();
	int on = 0, off = 0;
	if (obj == null && !hasModifier(Disabled,false)) on = Disabled;
	else if (obj != null && hasModifier(Disabled,false)) off = Disabled;
	if ((on|off) != 0){
		modify(on,off);
		repaintNow();
	}
}
//===================================================================

static class fieldHandlerMethod{
Class target;
String fieldName;
String extra;
Method m;
Wrapper[] pars;
Wrapper ret;
boolean pauseEvents;

fieldHandlerMethod set(Class t, String f, String e)
{
	target = t; fieldName = f; extra = e;
	return this;
}
public boolean equals(Object other)
{
	fieldHandlerMethod fm = (fieldHandlerMethod)other;
	return (fm.target == target) && fm.fieldName.equals(fieldName) && fm.extra.equals(extra);
}
public int hashCode()
{
	return target.hashCode()+fieldName.hashCode()+extra.hashCode();
}
boolean findMethod(String nm)
{
	m = Reflection.getMethod(target,nm,"(Leve/ui/data/Editor;)V",false);
	if (m == null) {
		m = Reflection.getMethod(target,nm,"()V",false);
		pars = Wrapper.noParameter;
	}else{
		pars = new Wrapper[1];
		pars[0] = new Wrapper();
	}
	if (m != null) ret = new Wrapper();
	return m != null;
}

}

static fieldHandlerMethod searchFieldHandler = new fieldHandlerMethod();

static Hashtable fieldHandlers;

public static boolean invokeFieldHandlerMethod(Object target, String fieldName, Editor ed, String extra)
{
	Class me = target.getClass();
	fieldHandlerMethod found = null;
	Object paused = null;
	try{
		synchronized(searchFieldHandler){
			if (fieldHandlers == null) fieldHandlers = new Hashtable();
			found = (fieldHandlerMethod)fieldHandlers.get(searchFieldHandler.set(me,fieldName,extra));
		}
		if (found == null){
			found = new fieldHandlerMethod().set(me,fieldName,extra);
			String nm = fieldName.replace(' ','_')+"_"+extra;
			if (!found.findMethod(nm))
				if (found.findMethod("execute_"+nm))
					found.pauseEvents = true;
			fieldHandlers.put(found,found);
		}
		if (found.m == null) return false;
		if (found.pauseEvents) paused = ed.pauseCurrentEventThread();
		try{
			Wrapper [] p = found.pars;
			if (p.length == 1){
				p = new Wrapper[1];
				p[0] = new Wrapper().setObject(ed);
			}
			Reflection.invoke(target,found.m,p,found.ret);
			return true;
		}catch(Exception e){
			return false;
		}
	}finally{
		if (paused != null) ed.resumeCurrentEventThread(paused);
	}
}
/**
* This gets called by action(FieldTransfer ft,Editor ed) and by fieldChanged(FieldTransfer ft,Editor ed)
* if the field changed is a menu item.
* By default it will look for
* a method called void <fieldName>_action(Editor ed) and if it finds it, it will be executed.
**/
//===================================================================
public void action(String fieldName,Editor ed)
//===================================================================
{
	invokeFieldHandlerMethod(this,fieldName,ed,"action");
}
/**
* This gets called by fieldChanged(FieldTransfer ft,Editor ed). 
* By default it will look for
* a method called void <fieldName>_changed(Editor ed) and if it finds it, it will be executed.
**/
//===================================================================
public void fieldChanged(String fieldName,Editor ed)
//===================================================================
{
	invokeFieldHandlerMethod(this,fieldName,ed,"changed");
}

/**
* This implements the action() method in FieldListener. By default it
* calls action(String name,Editor ed)
**/
//===================================================================
public void action(FieldTransfer ft,Editor ed) {action(ft.fieldName,ed);}
//===================================================================
/**
* This implements the fieldChanged() method in FieldListener. If the field is of
type MenuItem then action(String fieldName,Editor ed) method is called - where
fieldName is set to be the "action" value for the MenuItem.
**/
//===================================================================
public void fieldChanged(FieldTransfer ft,Editor ed)
//===================================================================
{
	if (MenuItem.class.equals(ft.fieldType)){
		Object got = ft.getFieldValue(ed.myObject);
		if (got instanceof MenuItem){
			if (((MenuItem)got).action != null)
				action(((MenuItem)got).action,ed);
		}
	}//else
	fieldChanged(ft.fieldName,ed);
}
//===================================================================
/**
* This implements fieldEvent() in FieldListener. By default it does nothing.
**/
//===================================================================
public void fieldEvent(FieldTransfer ft,Editor ed,Object event) {}
//===================================================================
/**
* This calls setObject(obj).
**/
//===================================================================
public void setData(Object obj)
//===================================================================
{
	setObject(obj);
}
/**
* This calls fromControls(obj).
**/
//===================================================================
public void getData(Object obj)
//===================================================================
{
	fromControls(obj);
}
//===================================================================
public void make(boolean reMake)
//===================================================================
{
	super.make(reMake);
	toControls();
}

/**
* Set this true if you are adding fields with Controls which themselves will not
* be added as a child of the Editor. The Editor will still listen to events from the controls.
**/
public boolean fieldsWillNotBeAdded = false;
//===================================================================
public void addField(FieldTransfer ft){fieldsSet = true; transfers.add(ft); ft.setMethodsFor(MenuItem.class,"getLastSelected()Leve/ui/MenuItem;",null);}
//===================================================================

public Control addField(Control control,String fieldName,DataConverter converter, boolean isExternalControl)
{
	if (isExternalControl) return addExternalField(control,fieldName,converter);
	else return addField(control,fieldName,converter);
}
/**
 * This is the main way of adding a field to the Editor - although it does not add the control as a
	child of the Editor. This only associates the Control with the Field of the Object to be edited with the
	specified field name. The Control must be added to the Editor either directly or somewhere within the child
	tree.
 * @param control The control to associate with the field.
 * @param fieldName The field to be associated with the Control.
 * @return The control itself.
 */
//===================================================================
public Control addField(Control control,String fieldName)
//===================================================================
{return addField(control,fieldName,(DataConverter)null);}
/**
 * This is the main way of adding a field to the Editor - although it does not add the control as a
	child of the Editor. This only associates the Control with the Field of the Object to be edited with the
	specified field name. The Control must be added to the Editor either directly or somewhere within the child
	tree.
 * @param control The control to associate with the field.
 * @param fieldName The field to be associated with the Control.
	@param converter An optional DataConverter to convert the data during a data transfer.
 * @return The control itself.
 */
//===================================================================
public Control addField(Control control,String fieldName,DataConverter converter)
//===================================================================
{
	if (converter == null) converter = Gui.getDefaultConverter();
	addField(new FieldTransfer(objectClass,sampleObject,fieldName,control,converter));
	if (control != null && fieldsWillNotBeAdded) control.addListener(this);
	return control;
}
/**
 * Use this to add a Control for a field when the Control will not be added as a child of the
 * Editor but will be added to some external Container. This will tell the Editor to listen
 * to events from the Control.
 * @param control The Control for the field.
 * @param fieldName The name of the field.
 * @param convert An optional DataConverter.
 * @return The Control
 */
//===================================================================
public Control addExternalField(Control control,String fieldName,DataConverter convert)
//===================================================================
{
	boolean fwa = fieldsWillNotBeAdded;
	fieldsWillNotBeAdded = true;
	addField(control,fieldName,convert);
	fieldsWillNotBeAdded = fwa;
	return control;
}
/**
 * Use this to add a Control for a field when the Control will not be added as a child of the
 * Editor but will be added to some external Container. This will tell the Editor to listen
 * to events from the Control.
 * @param control The Control for the field.
 * @param fieldName The name of the field.
 * @return The Control
 */
//===================================================================
public Control addExternalField(Control control,String fieldName)
//===================================================================
{
	return addExternalField(control,fieldName,(DataConverter)null);
}

/**
 * This method is called on a SoftKeyEvent. If the event has a Control proxy
 * associated with it which happens to be one of the fields of the Editor,
 * then it will call fireProxyAction() only and no further action will be taken.
 * Otherwise it will call onSoftKeyEvent(int, String, MenuItem);
 * @param se the SoftKeyEvent.
 */
//===================================================================
public void onSoftKey(SoftKeyEvent se)
//===================================================================
{
	fromControls();
	se.fireProxyAction();
	if (se.proxy == null || findFieldTransfer(se.proxy) == null) onSoftKey(se.whichKey, se.action, se.selectedItem);
}

//static WeakSet ws = new WeakSet();
//-------------------------------------------------------------------
protected void postEditorEvent(int type,Object parameter)
//-------------------------------------------------------------------
{
	EditorEvent ev = new EditorEvent(type,this,parameter);
	ObjectIterator it = ObjectIterator.getCached(myTransfer);
	fireFieldEvent(it,ev,2);
	it.cache();
	postEvent(ev);
}
//===================================================================
public void shown()
//===================================================================
{
	super.shown();
	postEditorEvent(EditorEvent.SHOWN,null);
}
//-------------------------------------------------------------------
protected void formClosing()
//-------------------------------------------------------------------
{
	postEditorEvent(EditorEvent.CLOSED,null);
	super.formClosing();
}
//===================================================================
public void transfer(Object obj,Iterator it,int direction)
//===================================================================
{
	while(it.hasNext()){
		FieldTransfer ft = (FieldTransfer)it.next();
/*
		if (ft.fieldName.equals("programName")) {
			ewe.sys.Vm.debug("+++"+ft.fieldName+" "+ft.dataInterface.hashCode());
			((Control)ft.dataInterface)._debug = true;
		}
*/
//
// Bug fix - Jan 2004
		// ft.transfer(myObject,direction);
		try{
			ft.transfer(obj,direction);
		}catch(RuntimeException e){
			System.out.println(e.toString()+" - "+ft.fieldName+" = ");//+obj.getClass().getName());
			//e.printStackTrace();
			throw e;
		}
		
/*
		if (ft.fieldName.equals("programName")) {
			ewe.sys.Vm.debug("---"+ft.fieldName+" "+ft.dataInterface.hashCode());
			((Control)ft.dataInterface)._debug = false;

		}
*/
		if (!dontSendEachFieldTransferEvent) postEditorEvent(direction == FieldTransfer.FROM_OBJECT ? EditorEvent.TO_CONTROLS : EditorEvent.FROM_CONTROLS,ft);
		if (direction == FieldTransfer.FROM_OBJECT && ft.dataInterface instanceof Control){
			((Control)ft.dataInterface).updateData();
		}
	}
	if (dontSendEachFieldTransferEvent) postEditorEvent(direction == FieldTransfer.FROM_OBJECT ? EditorEvent.TO_CONTROLS : EditorEvent.FROM_CONTROLS,null);
}
//===================================================================
/**
* Transfer all edited fields from the Object currently being edited to the on-screen controls.
**/
public void toControls() {toControls(myObject);}
/**
* Transfer all edited fields to the Object currently being edited from the on-screen controls.
**/
public void fromControls() {fromControls(myObject);}
/**
* Transfer all edited fields from the specified Object to the on-screen controls.
**/
public void toControls(Object obj) {dontSendEachFieldTransferEvent = true; transfer(obj,transfers.iterator(),FieldTransfer.FROM_OBJECT);}
/**
* Transfer all edited fields to the specified Object from the on-screen controls.
**/
public void fromControls(Object obj) {dontSendEachFieldTransferEvent = true; transfer(obj,transfers.iterator(),FieldTransfer.TO_OBJECT);}
/**
* Transfer specified fields from the Object currently being edited to the on-screen controls. The "fields"
argument should be pure field names separated by commas (,).
**/
public void toControls(String fields) {transfer(myObject,fields,null,FieldTransfer.FROM_OBJECT);}
/**
* Transfer specified fields to the Object currently being edited from the on-screen controls. The "fields"
argument should be pure field names separated by commas (,).
**/
public void fromControls(String fields) {transfer(myObject,fields,null,FieldTransfer.TO_OBJECT);}
/**
* Transfer specified fields from the specified Object to the on-screen controls. The "fields"
argument should be pure field names separated by commas (,).
**/
public void toControls(Object obj,String fields){transfer(obj,fields,null,FieldTransfer.FROM_OBJECT);}
/**
* Transfer specified fields to the specified Object from the on-screen controls. The "fields"
argument should be pure field names separated by commas (,).
**/
public void fromControls(Object obj,String fields){transfer(obj,fields,null,FieldTransfer.TO_OBJECT);}
/**
 * This calls toControls(fields) and then notifyDataChange().
 * @param fields a comma separated field list.
 */
//===================================================================
public void dataChanged(String fields)
//===================================================================
{
	toControls(fields);
	notifyDataChange();
}
/**
 * This calls toControls() and then notifyDataChange().
 */
//===================================================================
public void dataChanged()
//===================================================================
{
	toControls();
	notifyDataChange();
}
//===================================================================
//private Vector tx = new Vector(), xfer = new Vector();

private Hashtable ht;
private VectorIterator vi;
private Vector xfer;
//===================================================================
protected Vector findFields(String fieldNames,String exclude,Vector dest)
//===================================================================
{
	if (ht == null) ht = new Hashtable();
	String[] all = (String[])ht.get(fieldNames);
	if (all == null){
		all = mString.split(fieldNames,',');
		ht.put(fieldNames,all);
	}
	if (dest == null) dest = new Vector();
	else dest.clear();
	for (int i = 0; i<all.length; i++){
		String f = all[i];
		for (int j = 0; j<transfers.size(); j++){
			FieldTransfer ft = (FieldTransfer)transfers.get(j);
			if (!ft.isField(f)) continue;
			if (ft.isField(exclude)) continue;
			dest.add(ft);
		}
	}
	return dest;
}
//===================================================================
protected void transfer(Object obj,String fields,String exclude,int direction)
//===================================================================
{
	dontSendEachFieldTransferEvent = false;
	if (xfer == null){
		xfer = new Vector();
		vi = new VectorIterator();
	}
	findFields(fields,exclude,xfer);
	transfer(obj,vi.set(xfer),direction);
}
/**
 * Find the FieldTransfer associated with the specified field.
 * @param name The field to look for.
 * @return The FieldTransfers associated with the field or null if it was not found.
 */
//===================================================================
public FieldTransfer findFieldTransfer(String name)
//===================================================================
{
	for (int i = 0; i<transfers.size(); i++){
		FieldTransfer ft = (FieldTransfer)transfers.get(i);
		if (ft.fieldName.equals(name)) return ft;
	}
	return null;
}
/**
 * Find the FieldTransfer associated with the specified control.
 * @param c The Control associated with the field.
 * @param menuItem true if you are looking for the FieldTransfer associated with
	a MenuItem field for the Control.
 * @return 
 */
//===================================================================
public FieldTransfer findFieldTransfer(Control c,boolean menuItem)
//===================================================================
{
	for (int i = 0; i<transfers.size(); i++){
		FieldTransfer ft = (FieldTransfer)transfers.get(i);
		if (ft.dataInterface == c) {
			boolean isMenu = MenuItem.class.equals(ft.fieldType);
			if ((isMenu && !menuItem) || (!isMenu && menuItem)) continue;
			return ft;
		}
	}
	return null;
}
/**
 * Find the FieldTransfer associated with the specified control.
 * @param c The Control associated with the field.
 * @return 
 */
//===================================================================
public FieldTransfer findFieldTransfer(Control c) {return findFieldTransfer(c,false);}
//===================================================================
/**
 * Find the Control associated with the field.
 * @param fieldName The name of the field.
 * @return The Control associated with the field, or null if one was not found.
 */
//===================================================================
public Control findControlFor(String fieldName)
//===================================================================
{
	FieldTransfer ft = findFieldTransfer(fieldName);
	if (ft == null) return null;
	return (Control)ft.dataInterface;
}
//===================================================================
public void onDataChangeEvent(DataChangeEvent ev)
//===================================================================
{
	if (ev.target == this) return;
	//System.out.println("DC: "+ev.target);
	currentEvent = ev;
	ObjectIterator it = ObjectIterator.getCached(findFieldTransfer((Control)ev.target));
	boolean doFire = it.hasNext() && !hasModifier(TakeControlEvents,false) && !((Control)ev.target).hasModifier(NotAnEditor,false);
	fireFieldEvent(it,null,DoDataTransfer|0);
	it.cache();
	if (doFire){
		DataChangeEvent dce = new DataChangeEvent(DataChangeEvent.DATA_CHANGED,this);
		dce.cause = ev;
		postEvent(dce);
	}
	currentEvent = null;
}
public static final int DoDataTransfer = 0x80000000;

/**
* This is the event that caused the current fieldChanged() or action() event.
**/
public Event currentEvent;


//===================================================================
public void onControlEvent(ControlEvent ev)
//===================================================================
{
	currentEvent = ev;
	ObjectIterator it = ObjectIterator.getCached(null);
	if ((ev.type == ev.PRESSED) && (ev.target instanceof Control)){
		fireFieldEvent(it.setFor(findFieldTransfer((Control)ev.target)),ev,1);
	}
	if (ev.type == MenuEvent.SELECTED && !(ev instanceof ListEvent)){
		fireFieldEvent(it.setFor(findFieldTransfer((Control)ev.target,true)),ev,DoDataTransfer|0);
	}
	if ((ev.type > ev.CANCELLED) && (ev.target instanceof Control)){
		fireFieldEvent(it.setFor(findFieldTransfer((Control)ev.target)),ev,2);
	}
	it.cache();
	currentEvent = null;
	super.onControlEvent(ev);
}
//===================================================================
public void fireFieldChanged(String fieldName)
//===================================================================
{
	fireFieldEvent(findFields(fieldName,null,null).iterator(),null,0);
}

//-------------------------------------------------------------------
protected  void fireFieldEvent(FieldListener fl,FieldTransfer ft,Object event,int type)
//-------------------------------------------------------------------
{
	type &= ~DoDataTransfer;
	ft.fieldEvent = event;
	if (!Vm.inSystemThread()) {
		switch(type){
			case 0: fl.fieldChanged(ft,this); break;
			case 1: fl.action(ft,this); break;
			case 2: fl.fieldEvent(ft,this,event); break;
		}
		ft.fieldEvent = null;
	}else {
		new editorFirerer(ft,this,fl,type,event);
	}
}
/**
 * This is called if an Exception is thrown while transferring data from the
 * UI control to the object. If this method returns false then the Editor
 * will transfer from the object back to the control.
 * @param e the Exception that was thrown.
 * @param ft the FieldTransfer for the field.
 * @param direction the direction - either FieldTransfer.TO_OBJECT or FieldTransfer.FROM_OBJECT.
 * @return return true to indicate that the error has been handled, return false
 * to indicate that the Editor should refresh the control with the Object
 * data.
 */
protected boolean handleTransferError(Exception e,FieldTransfer ft,int direction)
//-------------------------------------------------------------------
{
	return false;
}
private Form topmost;

/**
 * Normally during Event processing, a call to pauseEventThread() is
 * made to the topmost Form. This disables all events to that Form except
 * for Paint and Resize events.
 * If this fields is set true then Events for the Editor (and it's containing
 * Forms) are not paused during Event processing.  
 */
//public boolean dontPauseEvents;

public Object pauseCurrentEventThread()
{
	return getTopmostForm().pauseEventThread();
}
public void resumeCurrentEventThread(Object fromPause)
{
	getTopmostForm().resumeEventThread(fromPause);
}
private Object paused;
private Control top;
/**
 * You can call this method from a fieldChanged(), action() or fieldEvent()
 * handling Thread. It allows another Thread to take over Paint/Resize
 * events while some background processing is taking place.<p>
 * You can call resumeThisEventThread() to resume the current Thread as
 * the event handling Thread, otherwise it is called automatically when
 * the fieldChanged(), action() or fieldEvent() method returns.
 * @deprecated this is not recursive  - use pauseCurrentEventThread() and resumeCurrentEventThread() instead.
 */
public void pauseThisEventThread()
{
	if (top == null){
		top = getTopmostForm();
		paused = top.pauseEventThread();
	}
}
/**
 * You can call this method from a fieldChanged(), action() or fieldEvent()
 * handling Thread after a call to pauseThisEventThread(). It 
 * allows the current Thread to resume as event handling Thread. Calling
 * this method is optional, if it was not called during the field handling
 * method it is automatically called when return.
 * @deprecated this is not recursive - use pauseCurrentEventThread() and resumeCurrentEventThread() instead.
 */
public void resumeThisEventThread()
{
	if (top != null) top.resumeEventThread(paused);
	top = null;
	paused = null;
}
//===================================================================
public void fireFieldEvent(Iterator it,Object event,int type)
//===================================================================
{
	/*
	Object p = null;
	Control top = null;
	if (!dontPauseEvents && ((type & 3) != 2)){
		if (topmost == null) topmost = getTopmostForm();
		top = topmost == null ? this : topmost;
		p = top.pauseEventThread();
	}
	*/
	if (top != null){
		new Exception().printStackTrace();
	}
	try{
		if (event instanceof ControlEvent){
			Control c = Gui.focusedControl();
			if (c != ((ControlEvent)event).target){
				FieldTransfer ft = findFieldTransfer(c);
				if (ft != null) ft.transfer(myObject,ft.TO_OBJECT);
			}
		}
		while(it.hasNext()){
			FieldTransfer ft = (FieldTransfer)it.next();
			if ((type & DoDataTransfer) != 0) {
				try{
					ft.transfer(myObject,ft.TO_OBJECT);
					postEditorEvent(EditorEvent.FROM_CONTROLS,ft);
				}catch(Exception e){
					if (!handleTransferError(e,ft,ft.TO_OBJECT))
						ft.transfer(myObject,ft.FROM_OBJECT);
				}
			}
			for (int i = 0; i<fieldListeners.size(); i++)
				fireFieldEvent((FieldListener)fieldListeners.get(i),ft,event,type);
			fireFieldEvent(this,ft,event,type);
		}
	}finally{
		if (top != null) {
			System.out.println("Resuming!");
			top.resumeEventThread(paused);
		}
		top = null;
	}
}
/**
* This will modify field controls in the editor conditionally. If "condition" is true, then the
* turnOn flags will be set and the turnOff flags will be cleared. If it is false then the reverse will
* be done. If the flags have changed then the control will be repainted if repaint is true.
**/
//===================================================================
public void modifyFields(String fields,boolean condition,int turnOn,int turnOff,boolean repaint)
//===================================================================
{
	String [] all = mString.split(fields,',');
	for (int i = 0; i<all.length; i++){
		FieldTransfer ft = findFieldTransfer(all[i]);
		if (ft == null) continue;
		Control c = (Control)ft.dataInterface;
		if (c == null) continue;
		int flags = c.getModifiers(false);
		if (condition) c.modify(turnOn,turnOff);
		else c.modify(turnOff,turnOn);
		if (!repaint) continue;
		if (c.getModifiers(false) == flags) continue;
		c.repaintNow();
	}
}

//===================================================================
public static void main(String args[]) throws Exception
//===================================================================
{
	Application.startApplication(args);
	Editor ed = new Editor();
	//ed.addLast(new CellPanel()).setPreferredSize(100,200).setCell(HSTRETCH);
	Class c = Class.forName(args[0]);
	ed.setFields(c.newInstance());
	ed.title = c.getName();
	ed.modifyAll(KeepSIP,0);
	Form f2 = new Form();
	CellPanel[] cp = f2.addSIPAdjustment();
	CellPanel tools = null, main = null;
	if (cp == null){
		main = f2;
	}else{
		main = cp[1];
	}
	main.addLast(ed.getScrollablePanel());
	//cp[0].addLast(new Button("Help!"));
	f2.execute();
	//ed.execute();
	
	System.exit(0);
}
public boolean convertData(Wrapper source, Class sourceType, Wrapper dest,
		Class destType) {
	return Wrapper.doConvertData(source, sourceType, dest, destType);
}

//##################################################################
}
//##################################################################

//##################################################################
class editorFirerer implements Runnable{
//##################################################################

int which;
FieldTransfer ft;
Editor ed;
FieldListener fl;
Object event;
public editorFirerer(FieldTransfer ft,Editor ed,FieldListener fl,int which,Object event)
{
	this.which = which;
	this.ft = ft;
	this.ed = ed;
	this.fl = fl;
	this.event = event;
	new Thread(this).start();
}

public void run()
{
	switch(which){
		case 0: fl.fieldChanged(ft,ed); return;
		case 1: fl.action(ft,ed); return;
		case 2: fl.fieldEvent(ft,ed,event); return;
	}
	ft.fieldEvent = null;
}
//##################################################################
}
//##################################################################


