package eve.ui.event;

import eve.sys.Event;
import eve.ui.Actionable;
import eve.ui.Control;

/**
 * ControlEvent is an event posted by a control.
 */

public class ControlEvent extends Event
{
/** The event type for a pressed event. */
public static final int PRESSED = 300;
/** The event type for a focus in event. */
public static final int FOCUS_IN = 301;
/** The event type for a focus out event. */
public static final int FOCUS_OUT = 302;
/** The event type for a timer event. */
public static final int TIMER = 303;
/** The event type for a Cancel event (e.g. ESC pressed). */
public static final int CANCELLED = 304;
/** The event type for a Exit event (i.e. one of the exit keys was pressed). */
public static final int EXITED = 305;
/** The event type for when a control shows its associated menu. */
public static final int MENU_SHOWN = 306;
/** Used to indicate that the Control has changed its structure rather
 * than the user make a change.
 */
public static final int STRUCTURE_CHANGED = 307;

public ControlEvent()
	{
	}

/**
 * Constructs a control event of the given type.
 * @param type the type of event
 * @param c the target control
 */
public ControlEvent(int type, Control c)
{
	this.type = type;
	target = c;
	timeStamp = System.currentTimeMillis();
	if (type == PRESSED){
		if (c instanceof Actionable)
			action = ((Actionable)c).getAction();
		if (action == null && c != null) 
			action = c.getText();
	}
}

//MLB
public static final int /*TEXT_CHANGED = 40000,*/ OPERATION_CANCELLED = 40001, POPUP_CLOSED = 40002;

/**
* This is set if one of the exit keys was pressed.
**/
public int exitKey;
/**
* This is set if one of the exit keys was pressed.
**/
public int exitKeyModifiers;
/**
* If the event is a FOCUS_IN then this is the control that is losing focus. If it is a FOCUS_OUT then this
* is the control that is taking away the focus.
**/
public Control oldOrNewFocus; 
/**
* If the event is a PRESSED event, and an mButton generated the event, then 
* this value will be equal to the "action" value of the button, or the button's text
* if its action is null.
**/
public String action;
}

