/*
 * Created on Jul 31, 2005
 *
 * Michael L Brereton - www.ewesoft.com
 * 
 * 
 */
package eve.ui.event;

import java.util.Vector;

import eve.fx.gui.SurfacePointerEvent;
import eve.sys.Vm;
import eve.sys.WeakSet;
import eve.sys.Wrapper;
import eve.ui.Control;
import eve.ui.Window;
import eve.util.Tag;
import eve.util.TagList;

/**
 * @author Michael L Brereton
 *
 */
//####################################################
public class PenEvent extends SurfacePointerEvent {//implements ObjectFinder{

	public static final int WANT_PEN_MOVE_TAG = 0x80000000;
	public static final int WANT_PEN_MOVED_ONOFF = 0x1;
	public static final int WANT_PEN_MOVED_INSIDE = 0x2;
	public static final int WANT_PEN_MOVED_OUTSIDE = 0x4;
	public static final int WHEN_NOT_ON_TOP_FRAME = 0x8;
	public static final int OVERRIDE_NO_MOUSE = 0x10;
	

//public static Control tipDisplayed = null;
/**
* This is the tool tip delay time in milliseconds. By default it is 500.
**/
public static int TipDelay = 500;

//-------------------------------------------------------------------
static WeakSet moved = new WeakSet();
//-------------------------------------------------------------------

//-------------------------------------------------------------------
public static synchronized void setCursor(Control who,int cursor)
//-------------------------------------------------------------------
{
	wantPenMoved(who,WANT_PEN_MOVED_ONOFF,true);
	setOptions(who,cursor << 16,0xffff0000);
	Window w = who.getWindow();
	if (w != null) w.setCursorFor(who,cursor);
	/*
	if (lastMovedIn == who){
		Gui.setCursor(who,cursor);
	}
	*/
}

//-------------------------------------------------------------------
protected static void setOptions(Control who,int toSet,int toClear)
//-------------------------------------------------------------------
{
	int opts = getOptionsFor(who);
	opts &= ~toClear;
	opts |= toSet;
	Wrapper w = (Wrapper)who.tags.getValue(WANT_PEN_MOVE_TAG,null);
	if (w == null) w = new Wrapper();
	w.setInt(opts);
	who.tags.set(WANT_PEN_MOVE_TAG,w);
}
/*
//-------------------------------------------------------------------
static Tag getListener(Control who,PenEvent ev)
//-------------------------------------------------------------------
{
	if (ev == null) ev = new PenEvent();
	//look = who;
	Object got = moved.find(ev);
	if (got == null) return null;
	return ((Control)got).tags.get(WANT_PEN_MOVE_TAG,null);
}
*/
/**
* It is necessary to call this method if a Control wishes to receive Pen/Mouse movement
* notifications.
* @param who The control requesting notifications.
* @param options any of the WANT_PEN_MOVED_XXX options ORed together.
* @param enable true to enable notification, false to disable them.
*/
//===================================================================
public static synchronized void wantPenMoved(Control who,int options,boolean enable)
//===================================================================
{
	if (
		((Vm.getParameter(Vm.VM_FLAGS) & Vm.VM_FLAG_NO_MOUSE_POINTER) != 0)
		&&((options & OVERRIDE_NO_MOUSE) == 0)
		) enable = false;
	if (moved == null) moved = new WeakSet();
	if (!enable) moved.remove(who);
	else moved.add(who);
	if (who.tags == null) who.tags = new TagList();
	setOptions(who,options,0);
}

private static Tag tag = new Tag();
//-------------------------------------------------------------------
public static int getOptionsFor(Control who)
//-------------------------------------------------------------------
{
	if (who == null) return 0;
	if (who.tags == null) return 0;
	synchronized(tag){
		Tag tg = who.tags.get(WANT_PEN_MOVE_TAG,tag);
		if (tg == null) return 0;
		return (int)((Number)tg.value).intValue();
	}
}
/**
 * Get hard references to all Controls that want some kind of PEN_MOVE
 * event.
 * @param dest a destination Vector or null to create and return a new one.
 * @return the destination Vector or a new one if it was null.
 */
public static Vector getWantPenMoves(Vector dest)
{
	if (dest == null) dest = new Vector();
	moved.getRefs(dest);
	return dest;
}
}

//####################################################
