package eve.ui.filechooser;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Hashtable;
import java.util.Vector;

import eve.data.DataUtils;
import eve.data.Property;
import eve.data.PropertyList;
import eve.data.TreeNode;
import eve.fx.Color;
import eve.fx.Dimension;
import eve.fx.FontMetrics;
import eve.fx.Graphics;
import eve.fx.IImage;
import eve.fx.IconAndText;
import eve.fx.ImageDecoder;
import eve.fx.Insets;
import eve.fx.Point;
import eve.fx.Rect;
import eve.fx.gui.IKeys;
import eve.fx.gui.WindowConstants;
import eve.io.File;
import eve.io.FileChooserParameters;
import eve.io.FileClipboard;
import eve.io.FileComparer;
import eve.io.Io;
import eve.io.filestore.EveFile;
import eve.sys.Device;
import eve.sys.Event;
import eve.sys.FieldTransfer;
import eve.sys.Handle;
import eve.sys.Registry;
import eve.sys.SlowTaskException;
import eve.sys.Task;
import eve.sys.Time;
import eve.sys.Type;
import eve.sys.Vm;
import eve.ui.Application;
import eve.ui.Button;
import eve.ui.ButtonBar;
import eve.ui.ButtonCheckBox;
import eve.ui.ButtonPullDownMenu;
import eve.ui.CellPanel;
import eve.ui.CheckBox;
import eve.ui.CheckBoxGroup;
import eve.ui.Choice;
import eve.ui.ComboBox;
import eve.ui.Control;
import eve.ui.FileChooserLink;
import eve.ui.FileClipboardProgress;
import eve.ui.Form;
import eve.ui.Frame;
import eve.ui.Gui;
import eve.ui.GuiStyle;
import eve.ui.IWebBrowser;
import eve.ui.ImageControl;
import eve.ui.Input;
import eve.ui.InputStack;
import eve.ui.Label;
import eve.ui.List;
import eve.ui.Menu;
import eve.ui.MenuBar;
import eve.ui.MenuItem;
import eve.ui.MessageBox;
import eve.ui.ReportException;
import eve.ui.ScrollBarPanel;
import eve.ui.SoftKeyBar;
import eve.ui.SplittablePanel;
import eve.ui.Window;
import eve.ui.data.Editor;
import eve.ui.data.InputBox;
import eve.ui.data.InputObject;
import eve.ui.data.Notepad;
import eve.ui.event.KeyEvent;
import eve.ui.table.HexTableModel;
import eve.ui.table.TableControl;
import eve.ui.table.TableEvent;
import eve.ui.table.TreeControl;
import eve.ui.table.TreeEvent;
import eve.ui.table.TreeTableModel;
import eve.util.FormattedDataSource;
import eve.util.mString;
import eve.util.mVector;
import eve.zipfile.ZipFile;
/**
* The FileChooser class is used to choose files or a directory on a file system. It provides the
* following features:
* <ul>
* <li>You can assign custom icons to file types.
* <li>You can set multiple file types using standard file masks (e.g. *.txt).
* <li>You can set a default file type.
* <li>The FileChooser will browse any file system that provides a File object. This includes Zip files, Ewe files
* and the FakeFileSystem (which maintains a file system in memory).
* <li>You can extend FileChooser to add custom controls (e.g. the ImageFileChooser).
* </ul>
* The simplest way to use a FileChooser is to construct it with the desired type and directory, and then
* call one of the execute() methods. If execute() returns IDCANCEL, then the user canceled the operation. If
* the user did not cancel, then you can use getChosen() or getChosenFile() to get the file chosen.
* <pre>
* FileChooser fc = new FileChooser(FileChooser.SAVE,null);
* fc.title = "Save Your Data Please...";
* if (fc.execute() == fc.IDCANCEL) return null;
* File toSaveTo = fc.getChosenFile();
* return toSaveTo;
* </pre>
* You can also allow MULTI_SELECT with the OPEN type to allow multiple file selection. In this case, you
* should use the chosenFiles Vector variable along with the chosenDirectory variable after execute() returns
* to determine the files chosen.
* <p>
**/

//##################################################################
public class FileChooser extends FileChooserBase{
//##################################################################
/**
* Use this to get the selected file - do NOT use the "file" variable. If multiple files are selected this will return the first one. Otherwise
* it will return the full path of the selected file. 
**/
//===================================================================
public String getChosen()
//===================================================================
{
	File got = getChosenFile();
	if (got == null) return null;
	return got.getFullPath();
}
/**
* Get all the chosen file names - even if only a single file was selected. Each item
* in the array is the file name only, without the directory path the selected item is in.
* Use getChosenDirectory() to get the directory the selected items are in.
**/
//===================================================================
public String [] getAllChosen()
//===================================================================
{
	if (chosenDirectory == null || chosenFiles == null) return new String[0];
	String [] all = new String[chosenFiles.size()];
	chosenFiles.copyInto(all);
	return all;
}
/**
 * Get the directory the chosen file(s) is in.
 */
//===================================================================
public File getChosenDirectory()
//===================================================================
{
	if (chosenDirectory == null) return null;
	File got = afile.getNew(chosenDirectory);
	if (got == null) return null;
	if (!got.isDirectory()) return null;
	return got;
}
/**
* Get all chosen files, even if only one file was selected.
**/
//===================================================================
public File [] getAllChosenFiles()
//===================================================================
{
	String [] ch = getAllChosen();
	File d = getChosenDirectory();
	if (d == null) return new File[0];
	File [] ret = new File[ch.length];
	for (int i = 0; i<ret.length; i++)
		ret[i] = d.getChild(ch[i]);
	return ret;
}
/**
* Use this to get the selected file - do NOT use the "file" variable. If multiple files are selected this will return the first one. Otherwise
* it will return the full path of the selected file. 
**/
//===================================================================
public File getChosenFile()
//===================================================================
{
	if (chosenDirectory == null || chosenFiles == null) return null;
	if (chosenFiles.size() < 1) return null;
	String cf = (String)chosenFiles.get(0);
	if (chosenDirectory.equals("/") && (cf.indexOf('\\') != -1 || cf.indexOf('/') != -1))
		return afile.getNew(cf);
	return afile.getNew(chosenDirectory).getChild(cf);
}
/**
* Do NOT use this to variable. Use getChosen() or getChosenFile() to get the file that the user selected.
**/
public String file = "\\";
/**
* Do NOT use this variable. Use defaultExtension to set a default extention for a file.
**/
public String fileType = "";
/**
* If no extension is given then this will be appended on to the selected file. By default this is null. When you
* set it, do not put the '.'
**/
public String defaultExtension = null;

List list = new List(10,10,false);
/**
* This is the table (list) of files/directory being displayed.
**/
public TableControl table;
/**
* This is the tree of directory being displayed.
**/
public TreeControl tree;
/**
* This is the FileListTableModel used by the table.
**/
public FileListTableModel files;
/**
* This is used internally.
**/
public Time time = new Time();

protected CellPanel dirNavigation;
/**
* Set this to an application specific name (e.g. "Ewesoft-Jewel") to keep and
* retrieve a persistent history for the FileChooser.
**/
public String persistentHistoryKey;

ButtonPullDownMenu roots = null;//new ButtonPullDownMenu("Drives",new Menu()), 
ButtonPullDownMenu toolMenu;
String [] sroots;
int type = 2;

/**
* Put any extra controls in this panel.
**/
public CellPanel extraPanel;

/**
* These are links for the FileChooser - each of which must be of type FileChooserLink. It will initially be null, unless there are
* links inside of globalLinks. In that case a new Vector is created and all the links inside of globalLinks will
* be added to it. You can call addLink() to add links to it safely.
**/
public Vector links;

{
	dismantleOnClose = true;
}

//===================================================================
public void addLink(FileChooserLink link)
//===================================================================
{
	if (link != null)
		links = mVector.add(links,link);
}
private static String newItems = "Folder|File";
static {
	try{
		if ((new Type("eve.sys.registry.Registry")).exists())
			newItems += "|Shortcut";
	}catch(Exception e){
		
	}
}

/**
* This associates an icon with the specified file extension. Do not place the leading '.' in the
* extension. e.g. associateIcon("exe",FileChooser.loadImage("myicon/programicon.png"));
* @param extension The extension WITHOUT the leading '.'
* @param icon The icon to be displayed. Should be a 16x16 icon.
*/
//===================================================================
public static void associateIcon(String extension,IImage icon)
//===================================================================
{
	FileChooserParameters.associateFileIcon(extension,icon,null);
}

protected static String baseDir = File.getProgramDirectory();

/**
 * Get the associated icon for a file name.
 * @param fileName The file.
 * @return The icon associated with the file.
 */
//===================================================================
public static IImage getIconForFile(String fileName)
//===================================================================
{
	if (fileName == readingDirectory[0]) return loadImage("eve/WaitSmall.png");
	int idx = fileName.lastIndexOf('.');
	if (idx != -1){
		String extension = fileName.substring(idx+1);
		IImage got = (IImage)FileChooserParameters.getFileIcon(extension,null,TreeTableModel.page);
		if (got == null) return got;
		IImage ret = (IImage)doubleSizeIt(got);
		if (ret != got) FileChooserParameters.doubleSizeAll(null);
		return ret;
	}else
		return TreeTableModel.page;
	/*
	if (extensions.size() == 0) return TreeTableModel.page;
	int idx = fileName.lastIndexOf('.');
	if (idx != -1){
		Locale l = Vm.getLocale();
		char [] look = Vm.getStringChars(fileName);
		int offset = idx+1;
		int length = fileName.length()-offset;
		for (int i = 0; i<extensions.size(); i++){
			char [] got = (char [])extensions.get(i);
			int c = l.compare(look,offset,length,got,0,got.length,Locale.IGNORE_CASE);
			if (c == 0) {
				return (IImage)icons.get(i);
			}
		}
	}
	return TreeTableModel.page;
	*/
}

/**
* This is a copy of the File model being used by the FileChooser. You should not set this, but you can
* reference it if you need to.

**/
public File afile;
/**
* This indicates that access to the file system will be slow.
**/
public boolean slowAccess = false;
/**
* This indicates that the wait cursor should be shown when listing is being done.
**/
public boolean showWait = false;


boolean hasFileList = true;
boolean masksSet = false;
/**
Set this true to initially show the FileChooser in list mode.
**/
public static boolean listModeDefault = false;

/**
* Do not set this directly.
**/
public boolean iconMode = true;
/**
* Do not set this directly.
**/
public boolean listMode = listModeDefault;
/**
* Do not set this directly.
**/
public boolean detailMode = false;
/**
* Do not set this directly.
**/
public boolean noKeyboard = false;
/**
* Do not set this directly.
**/
public boolean noMouse = false;
//-------------------------------------------------------------------
protected CellPanel getViewToolBar(String iconsAndFieldsAndTips)
//-------------------------------------------------------------------
{
	CellPanel cp = new CellPanel();
	String all[] = mString.split(iconsAndFieldsAndTips);
	CheckBoxGroup group = new CheckBoxGroup();
	Color red = new Color(255,0,0);
	for (int i = 0; i<all.length-2; i += 3){
		CheckBox cb = new ButtonCheckBox(loadImage("eve/"+all[i]+".png"));
		if (all.length > 3) cb.setGroup(group);
		cp.addNext(addField(cb,all[i+1])).setToolTip(all[i+2]);
	}
	return cp;
}

public synchronized FileClipboard getFileClipboard(boolean forSetting)
{
	if (forSetting || clipboard == null) clipboard = new FileClipboard();
	return clipboard;
}
FileClipboard clipboard;

static String [] 
needOne = mString.split("Rename|Properties"),
viewFiles = mString.split("View File|View As|Run"),
needOneOrMore = mString.split("Copy|Cut|Delete"),
pasteOp = mString.split("Paste");
//-------------------------------------------------------------------
void fixMenu(Menu m,boolean isPopup)
//-------------------------------------------------------------------
{
	int [] all = table.getSelectedIndexes();
	boolean none = all.length == 0;
	boolean one = all.length == 1;
	boolean multi = all.length > 1;
	boolean oneOrMore = one | multi;
	m.modifyItems(needOne,one ? 0 : MenuItem.Disabled,one ? MenuItem.Disabled : 0,true);
	//if (!allowFileViewing) m.modifyItems(viewFiles,MenuItem.Disabled,0,true);
	m.modifyItems(needOneOrMore,oneOrMore ? 0 : MenuItem.Disabled,oneOrMore ? MenuItem.Disabled : 0,true);
	boolean paste = true;
	if (multi) paste = false;
	boolean cpi = getFileClipboard(false).canPasteInto(files.parent);
	if (!cpi) paste = false;
	m.modifyItems(pasteOp,paste ? 0 : MenuItem.Disabled,paste ? MenuItem.Disabled : 0,true);
}

boolean noWrite;

//-------------------------------------------------------------------
Control makeToolMenu()
//-------------------------------------------------------------------
{
	Menu mm = new Menu(), sm;
	
	String toFile =  !noWrite ? 
		"New|-|Rename|-|" : "";
	toFile += "Run";
	//if (allowFileViewing) 
	toFile += "|View File";
	mm.addItem(sm = new Menu(mString.split(toFile),"File"));
	if (!noWrite)
		sm.getItemAt(0).subMenu = new Menu(newChoices = mString.split(newItems),"New");
	//if (allowFileViewing)
	sm.addItem(new Menu(mString.split(viewable),"View as"));
	sm.addItem("Properties");
	
	toFile =  !noWrite ? 
		"Copy|Cut|Paste|Delete|-|" : "Copy|";
	toFile += "Select All|Invert Selection";
	if (noKeyboard) toFile += "|Multiple Select";
	mm.addItem(sm = new Menu(mString.split(toFile),"Edit"));
	mm.addItem(sm = new Menu(viewChoices = mString.split("Details|Names Only"),"View"));
	sm.addItem("-");
	sm.addItem("Show Hidden Files");
	sm.addItem("Pop-up File Info");
	sm.addItem("Refresh");
	mm.addItem(sm = new Menu(sortChoices = mString.split("By Name|By Date|By Type|By Length"),"Sort"));
	
	if ((type & LAUNCHER_TYPE) == LAUNCHER_TYPE){
		if ((new Type("eve.ui.launcher.LaunchPanel")).exists())
			//;if (LaunchPanel.launcher != null)
			mm.addItem("Install");
	}
	if (((type & EXPLORER_TYPE) != 0) && !((type & LAUNCHER_TYPE) == LAUNCHER_TYPE)){
		mm.addItem("-");
		mm.addItem("Close");
	}
	sm.addItems(mString.split("-|Descending"));
	addField(toolMenu = new ButtonPullDownMenu("",mm){
		public boolean checkMenu(Menu m){
			fixMenu(m,false);
			return true;
		}
	}
	,"toolField");
	toolMenu.image = tools;
	toolMenu.getMenu().checkOnlyOne(viewChoices,viewChoices[0],true);
	toolMenu.getMenu().checkOnlyOne(sortChoices,sortChoices[0],true);
	toolMenu.setToolTip("Additional Tools");
	if (true) return toolMenu;
	else{
		MenuBar mb = new MenuBar();
		mb.addMenu(toolMenu);
		return mb;
	}
	//return toolMenu;
}
/**
 * Create a FileChooser in OPEN mode.
 */
//===================================================================
public FileChooser()
//===================================================================
{
	this(OPEN,baseDir);
}
/**
 * Create a FileChooser with a certain type, in the specified initial directory.
 */
//===================================================================
public FileChooser(int type,String initial)
//===================================================================
{
	this(type,initial,Vm.newFileObject());
}
public MenuItem goBackMenu, goForwardMenu;

static String[] listRoots(File model, boolean returnSlash)
{
	File[] sr = model.roots();//getInfo(File.INFO_ROOT_LIST,null,null,0);
	if (sr == null) return returnSlash ? new String[]{"/"} : null;
	String[] sroots = new String[sr.length];
	for (int i = 0; i < sr.length; i++)
		sroots[i] = sr[i].toString();
	return sroots;
}
//-------------------------------------------------------------------
void addDirTools(CellPanel cp)
//-------------------------------------------------------------------
{
	Button dirUp = new Button();
	dirUp.image = folderUp;
	if (back != null){
		ButtonBar bb = new ButtonBar();
		bb.addNext(addField(back.setToolTip("Go back"),"goBack"));
		bb.addNext(addField(forward.setToolTip("Go forward"),"goForward"));
		addField(back,"goBackMenu");
		addField(forward,"goForwardMenu");
		cp.addNext(bb).setCell(DONTSTRETCH);
	}
	sroots = listRoots(afile,false);
		/*
	else if (got != null)
		Vm.debug("Got: "+got+"\n"+got.getClass());
		*/
	if (sroots != null && roots != null){
		Gui.iconize(roots,drive,false,getFontMetrics());
		cp.addNext(addField(roots,"drives")).setCell(DONTSTRETCH).setToolTip("Select a disk drive");
	}
	cp.addNext(addField(directories,"directories")).setCell(HSTRETCH);
	/*if ((type & 3) != DIRECTORY_SELECT && !bigScreen){
		cp.addNext(accept).setCell(DONTSTRETCH);
		cp.addNext(cancel).setCell(DONTSTRETCH);
	}*/
	cp.addNext(addField(dirUp,"dirUp")).setCell(DONTSTRETCH).setToolTip("Go to higher directory");
}
//VMOptions vmoptions = new VMOptions();

Button back, forward;
String viewable = "Text File|Zip File|Eve File|Ewe File|Image File|Binary File";

//-------------------------------------------------------------------
private static int toType(FileChooserParameters fcp)
//-------------------------------------------------------------------
{
	int ret = OPEN;
	String r = fcp.getString(fcp.TYPE,fcp.TYPE_OPEN);
	if (r.equalsIgnoreCase(fcp.TYPE_OPEN) || r.equalsIgnoreCase("load")) ret = OPEN;
	else if (r.equalsIgnoreCase(fcp.TYPE_SAVE)) ret = SAVE;
	else if (r.toLowerCase().startsWith(fcp.TYPE_DIRECTORY_SELECT)) ret = DIRECTORY_SELECT;
	else if (r.equalsIgnoreCase(fcp.TYPE_BROWSE)) ret = BROWSE;
	ret |= fcp.getInt(fcp.OPTIONS,0);
	return ret;
}

private FileChooserParameters fcp;
//===================================================================
public FileChooser(FileChooserParameters fcp)
//===================================================================
{
	this(fcp,false);
}
//===================================================================
public FileChooser(FileChooserParameters fcp, boolean extraControl)
//===================================================================
{
	this(toType(fcp)|(extraControl ? EXTRA_CONTROL : 0),fcp.getString(fcp.START_LOCATION,null),(File)fcp.getValue(fcp.FILE_MODEL,Vm.newFileObject()));
	title = fcp.getString(fcp.TITLE,title);
	icons = (Hashtable)fcp.getValue(fcp.ICONS,null);
	defaultExtension = fcp.getString(fcp.DEFAULT_EXTENSION,defaultExtension);
	for (Property p = fcp.get(fcp.FILE_MASK); p != null; p = fcp.get(fcp.FILE_MASK,p))
		addMask((String)p.value);
	persistentHistoryKey = fcp.getString(fcp.PERSISTENT_HISTORY,persistentHistoryKey);
	this.fcp = fcp;
}
/**
 * Create a new FileChooser(). After creating it, use execute() to display it. If execute() returns Form.IDCANCEL then
 * the user canceled the choose operation. Otherwise you can use getChosen() or getChosenFile() to get the chosen file.
 * <p>
 * You can also change the title of the FileChooser by setting "title" after construction.
 * @param type either OPEN, SAVE, DIRECTORY_SELECT or BROWSE OR'ed with some of the other options.
 * @param initial The initial directory.
 * @param aFileModel The file mode to use.
 */
//===================================================================
public FileChooser(int type,String initial,File aFileModel)
//===================================================================
{
	SoftKeyBar sb = null;
	if (web.exists()) viewable += "|HTML File";
	if (globalLinks != null && globalLinks.size() != 0){
		links = new Vector();
		links.addAll(globalLinks);
	}
	wideScreen = Gui.screenIs(Gui.WIDE_SCREEN);
	desktopWide = Gui.screenIs(Gui.DESKTOP_WIDTH);
	if (wideScreen){
		back = new Button(null,"eve/leftarrowsmall.png");
		forward = new Button(null,"eve/rightarrowsmall.png");
	}
/*
	if ((type & LAUNCHER_TYPE) == LAUNCHER_TYPE){
		if (vmoptions.read()){
			initial = vmoptions.launcherStart;
			if (vmoptions.launcherTree) type |= DIRECTORY_TREE;
			else type &= ~DIRECTORY_TREE;
			if (vmoptions.showIcon)
				new TaskbarWindow("Eve VM","eve/evesmall.png","Eve VM");
		}
	}
*/
	title = "File Chooser";
	afile = aFileModel;
	//Device.messageBox("Debug", initial+" - "+baseDir, Device.MB_TYPE_OK);
	if (initial == null) initial = baseDir;
	if (initial == null) initial = "";
	if ((initial.equals("\\") || initial.equals("/")) && (afile.getClass() == File.class)){
		File[] f = afile.listRoots();
		if (f != null && f.length > 1){
			File got = afile.getNew(File.getProgramDirectory());
			while(true){
				File p = got.getParentFile();
				if (p == null) break;
				got = p;
			}
			initial = got.getAbsolutePath();
		}
	}
	if (type != DIRECTORY_SELECT){
		File sf = afile.getNew(initial);
		File pf = sf.getParentFile();
		if ((pf == null || !pf.isDirectory()) && !sf.isDirectory()){
			sf = afile.getNew(baseDir).getChild(initial);
			initial = sf.getAbsolutePath();
		}
	}
	//Device.messageBox("Debug", initial+" - "+baseDir, Device.MB_TYPE_OK);
	//if (initial.trim().length() == 0) initial = baseDir;
	noKeyboard = (Vm.getParameter(Vm.VM_FLAGS) & Vm.VM_FLAG_NO_KEYBOARD) != 0;
	noMouse = (Vm.getParameter(Vm.VM_FLAGS) & Vm.VM_FLAG_NO_MOUSE_POINTER) != 0;
	noKeyboard = noMouse; //Using a pen on a hand-held device is like that of a palm.
	noWrite = (type & NO_WRITING) != 0;
	noWrite |= (afile.getFlags() & afile.FLAG_FILE_SYSTEM_IS_READ_ONLY) != 0;
	if ((type & 3) == OPEN) type |= FILE_MUST_EXIST;
	IImage newFolderImage = loadImage("eve/NewFolder.png");
	if (desktopWide && ((type & 3) != DIRECTORY_SELECT)) type |= DESKTOP_VERSION;

	this.type = type;
	if ((type & EXPLORER_TYPE) == 0)
		windowFlagsToClear = Window.FLAG_HAS_CLOSE_BUTTON;
	if (((type & LAUNCHER_TYPE) == LAUNCHER_TYPE) && Device.isMobile()){
		//title = "File C;
		windowFlagsToClear = Window.FLAG_HAS_CLOSE_BUTTON|Window.FLAG_HAS_TITLE;
	}
	slowAccess = ((afile.getFlags() & afile.FLAG_SLOW_CHILD_COUNT) != 0);
	file = initial;
	CellPanel cp;
	//boolean bigScreen = Gui.screenIs(Gui.BIG_SCREEN);
	boolean haveAccept = (type & 3) != BROWSE;
	boolean haveCancel = (type & EMBEDDED) == 0;
	FontMetrics fm = getFontMetrics();
	accept = desktopWide || Gui.hasSoftKeys ? new Button(new IconAndText(Form.tick,"Accept",fm)) : new Button(tick);
	cancel = desktopWide || Gui.hasSoftKeys ? new Button(new IconAndText(Form.cross,haveAccept ? "Cancel" : "Close",fm)) : new Button(cross);
	CellPanel treePanel = null;
	CellPanel filePanel = new CellPanel();
	CellPanel main = new CellPanel();
	exitButtonDefined = true;
	if ((type & 3) == DIRECTORY_SELECT){
		if (!Gui.hasSoftKeys || !noWrite){
			ButtonBar bb = new ButtonBar(); bb.equalWidths = false;
			if (!noWrite) bb.addNext(addField(new Button(new IconAndText(newFolderImage,"New",getFontMetrics())),"newFolder"));
			if (!Gui.hasSoftKeys){
				bb.addNext(addField(accept = new Button(new IconAndText(Form.tick,"Select Folder",getFontMetrics())),"acceptSelection"));
				bb.addNext(addField(cancel = new Button(new IconAndText(Form.cross,"Cancel",getFontMetrics())),"cancelSelection"));
			}
			addLast(bb).setCell(HSTRETCH);
		}
	} 
	if (Gui.hasSoftKeys){
		sb = new SoftKeyBar();
		if (haveAccept) sb.setKey(1,accept); 
			//sb.setKey(1,"Accept|"+Form.EXIT_IDOK,tick,null);
		if (haveCancel) sb.setKey(haveAccept ? 2 : 1,"Cancel|"+Form.EXIT_IDCANCEL,cross,null);
		if (haveAccept || haveCancel)
			exitButtonDefined = true;
	}
	
	Dimension d = Gui.getPreferredDialogSize();
	setPreferredSize(d.width,d.height);
	Rect got = null;//(Rect)Application.mainApp.getInfo(Window.INFO_PARENT_RECT,null,new Rect(),0);
	if (got != null) d.set(got.width,got.height);
	
	if ((type & DIRECTORY_TREE) != 0){
		int how = wideScreen ? SplittablePanel.HORIZONTAL : SplittablePanel.VERTICAL;
		SplittablePanel sp = null;
		if ((type & 3) != DIRECTORY_SELECT){
			sp = new SplittablePanel(how);
			addLast(sp);
			treePanel = sp.getNextPanel(0);
		}else{
			treePanel = new CellPanel();
			addLast(treePanel);
		}
		if (how == sp.HORIZONTAL) treePanel.setPreferredSize(d.width/3,50);
		else treePanel.setPreferredSize(50,d.height/2);
		treePanel.setBorder(GuiStyle.inputEdge|BF_RECT,2);
		tree = new FileTreeControl();
		tree.backGround = Color.White;
		tree.setTableModel(new FileTree(null,afile,this));
		//if (slowAccess) tree.getTreeTableModel().showWaitCursor = true;
		ScrollBarPanel sbp;
		treePanel.addLast(sbp = new ScrollBarPanel(tree));
		sbp.hbar.setFollowTracking(true);
		sbp.vbar.setFollowTracking(true);
		if ((type & 3) != DIRECTORY_SELECT){
			filePanel = sp.getNextPanel();
			sp.theSplitter.doCloseFirst = true;
			if (how == sp.HORIZONTAL) {
				filePanel.setPreferredSize(d.width*2/3,50);
				filePanel.setMinimumSize(0,0);
				if (desktopWide) filePanel.mySplitter.arrowPosition = Right;
			}else{
				filePanel.setPreferredSize(50,200);//d.height*2/3);
				filePanel.setMinimumSize(0,0);
				//filePanel.mySplitter.arrowPosition = Down;
			}
		}else{
			hasFileList = false;
		}
		addField(tree,"dirTree");
	}else{
		addLast(filePanel);
	}
	main.setPreferredSize(200,200);
	if ((type & EXTRA_CONTROL) == 0){
		filePanel.addNext(main);
	}else{
		int how = wideScreen ? SplittablePanel.HORIZONTAL : SplittablePanel.VERTICAL;
		SplittablePanel sp = new SplittablePanel(how);
		filePanel.addNext(sp);
		sp.getNextPanel().addLast(main);
		sp.getNextPanel().addLast(extraPanel = new CellPanel());
	}
	ScrollBarPanel sbp;

	ComboBox fileInput;
 	addField(fileInput = new ComboBox(),"file");
	fileInput.input.wantReturn = true;
	fileInput.input.prompt = "File/Mask";

	directories = new Choice();
	directories.useMenuItems = true;
	directories.alwaysDrop = true;
	directories.indentDropItems = true;
	directories.prompt = "Choose Folder";
	directories.menuOptions |= directories.MENU_FULL_WIDTH;
	directories.menuOptions |= directories.MENU_SHOW_TITLE_IF_EXPANDED;
	//
	// Remove this.
	//
	if ((type & DESKTOP_VERSION) == 0){
		if (desktopWide && ((type & 3) != DIRECTORY_SELECT)){
			if (!Gui.hasSoftKeys){
				ButtonBar bb = new ButtonBar();
				if (haveAccept) bb.addNext(accept);
				bb.addNext(haveCancel? cancel : new Control());
				main.addLast(bb).setCell(HSTRETCH);
			}
		}
		if ((type & (NO_DIRECTORY_CHANGE|DIRECTORY_TREE|EXPLORER_TYPE)) == 0) {
			cp = dirNavigation = new CellPanel();
			cp.defaultTags.set(TAG_INSETS,new Insets(0,1,0,0));
			if ((type & 3) == DIRECTORY_SELECT)
				cp.addNext(makeToolMenu()).setCell(DONTSTRETCH);
			addDirTools(cp);
			main.addLast(cp).setCell(HSTRETCH);
		}else if ((type & DIRECTORY_TREE) != 0)	{
		}
		
		if (((type & 3) != DIRECTORY_SELECT)){
			if (((type & EXPLORER_TYPE) == 0) && ((type & DIRECTORY_TREE) != 0)){
				cp = dirNavigation = new CellPanel();
				cp.defaultTags.set(TAG_INSETS,new Insets(0,1,0,0));
				addDirTools(cp);
				main.addLast(cp).setCell(HSTRETCH);
				cp.setPreferredSize(0,0);
			}
			cp = new CellPanel();
			cp.defaultTags.set(TAG_INSETS,new Insets(0,1,0,0));
			cp.addNext(makeToolMenu()).setCell(DONTSTRETCH);
			//toolButton.arrowDirection = Down;
			if ((type & EXPLORER_TYPE) == 0){
				cp.addNext(new Label("File:")).setCell(DONTSTRETCH);
				cp.addNext(fileInput).setCell(HSTRETCH);
			}
			maskChoices = fileInput.choice;
			if (!desktopWide){
				if ((type & EXPLORER_TYPE) == 0){
					if (!Gui.hasSoftKeys){
						if (haveAccept) cp.addNext(accept).setCell(DONTSTRETCH);
						if (haveCancel) cp.addLast(cancel).setCell(DONTSTRETCH);
					}
				}else{
					addDirTools(cp);
				}
			}
			main.addLast(cp).setCell(HSTRETCH);
		}
		main.addLast(sbp = new ScrollBarPanel(table = new FileTableControl()));
	/**
	* This is the desktop layout.
	**/
	}else{
		resizable = true;
		cp = new CellPanel();
		cp.defaultTags.set(TAG_INSETS,new Insets(2,2,2,2));
		cp.addNext(makeToolMenu()).setCell(DONTSTRETCH);			
		addDirTools(cp);
		
		/*
		sroots = (String [])afile.getNew("").getInfo(File.INFO_ROOT_LIST,null,null,0);		
		if (sroots != null){
			Gui.iconize(roots,drive,false,getFontMetrics());
			cp.addNext(addField(roots,"drives")).setCell(DONTSTRETCH).setToolTip("Select a disk drive");
		}
		cp.addNext(addField(directories,"directories")).setCell(HSTRETCH);
		Button dirUp = new Button();
		dirUp.image = folderUp;
		cp.addNext(addField(dirUp,"dirUp")).setCell(DONTSTRETCH).setToolTip("Go to higher directory");
		*/
		if (!noWrite)cp.addNext(addField(new Button(newFolderImage),"newFolder")).setCell(DONTSTRETCH).setToolTip("Create a new directory");
		cp.addNext(getViewToolBar("wideview|iconMode|View multiple columns|listview|listMode|View single column")).setCell(DONTSTRETCH);
		cp.addNext(getViewToolBar("detailview|detailMode|Show file details")).setCell(DONTSTRETCH);
		/*if ((type & 3) != DIRECTORY_SELECT && !bigScreen){
			cp.addNext(accept).setCell(DONTSTRETCH);
			cp.addNext(cancel).setCell(DONTSTRETCH);
		}*/
		main.addLast(cp).setCell(HSTRETCH);
		main.addLast(sbp = new ScrollBarPanel(table = new FileTableControl())).setTag(TAG_INSETS,new Insets(2,4,2,4));
		CellPanel selectors = new CellPanel();
		selectors.defaultTags.set(TAG_INSETS,new Insets(2,2,2,2));
		if (haveAccept){
			selectors.addNext(new Label("File name:"),DONTSTRETCH,DONTFILL|EAST);
			selectors.addNext(fileInput).setCell(HSTRETCH);
		}
		historyList = fileInput.choice;
		if (haveAccept && !Gui.hasSoftKeys)
			selectors.addLast(accept).setCell(DONTSTRETCH);
		selectors.endRow();
		selectors.addNext(new Label("Files of type:"),DONTSTRETCH,DONTFILL|EAST);
		selectors.addNext(addField(maskChoices = new Choice(),"fileType")).setCell(HSTRETCH);
		maskChoices.alwaysDrop = true;
		if (!Gui.hasSoftKeys) 
			selectors.addLast(haveCancel ? cancel : new Control()).setCell(DONTSTRETCH);
		if ((type & EXPLORER_TYPE) == 0) 
			main.addLast(selectors).setCell(HSTRETCH).setControl(HFILL|NORTHWEST);
	}
	table.listMode = 1;
	table.multiSelect = !noKeyboard;//((type & MULTI_SELECT) != 0);
	table.modify(WantHoldDown,0);
	addMask(allFilesMask);
	masksSet = false;
	sbp.hbar.setFollowTracking(true);
	sbp.vbar.setFollowTracking(true);
	addField(table,"fileTable");
	table.setClickMode(true);
	table.clickClearsItself = false;
	table.setToolTip(null);
	sbp.setBorder(GuiStyle.inputEdge|BF_RECT,2);
	table.backGround = Color.White;
	table.setTableModel(files = new FileListTableModel(this));
	files.fileCheck = afile;
	//resizable = true;
	time.format = "dd-MM-yyyy, HH:mm";
	if ((type & 3) == DIRECTORY_SELECT) {
		lastOptions.viewOptions |= lastOptions.VIEW_NAME_ONLY;
	}
	setOptions(lastOptions);
	addField(accept,"acceptSelection",null,sb != null);
	if (cancel != null) cancel.setHotKey(0,IKeys.ESCAPE);
	setHistory(history);
	amExiting = false;
	if (roots != null && sroots != null){
		roots.getMenu().addItems(sroots);
		//sroots = null;
	}
	if (sb != null) {
		setSoftKeyBarFor(null,sb);
	}
}

	protected void showError(String message,boolean execute)
	{
		showError("Error",message,execute);
	}
	//-------------------------------------------------------------------
	protected void showError(String title,String message,boolean execute)
	//-------------------------------------------------------------------
	{
		MessageBox mb = new MessageBox(title,message,MBOK);
		mb.exec(getFrame(),Gui.CENTER_FRAME);
		if (execute) mb.waitUntilClosed();
	}
//-------------------------------------------------------------------
protected boolean doHotKey(Control from,KeyEvent ev)
//-------------------------------------------------------------------
{
	if ((ev.key == IKeys.F5) && ((ev.modifiers & (IKeys.CONTROL|IKeys.SHIFT)) == 0) && (parent != null)){
		refresh();
		return true;
	}else
		return super.doHotKey(from,ev);
}
/**
* Refresh the file list.
**/
//===================================================================
public void refresh()
//===================================================================
{
	if (parent != null){
		parent.refresh();
		setFile(parent);
	}
}
//===================================================================
public static String lengthToDisplay(long len)
//===================================================================
{
	return Io.fileLengthDisplay(len);
}
	//##################################################################
	class FileTableControl extends TableControl{
	//##################################################################
	{
		autoScrollToVisible = false;

	}
	//===================================================================
	public void doPaint(Graphics g,Rect area)
	//===================================================================
	{
		boolean sw = (afile.getFlags() & File.FLAG_SLOW_LIST) != 0;
		//if (sw) Vm.showWait(true);

		super.doPaint(g,area);
		//if (sw)Vm.showWait(false);
	}
	
	//===================================================================
	public boolean checkMenu(Menu m)
	//===================================================================
	{
		fixMenu(m,true);
		return true;
	}
	//===================================================================
	public void popupMenuEvent(final Object selected)
	//===================================================================
	{
		new Task(){
		public void doRun(){
			FileChooser.this.checkMenu(mString.toString(selected),FileChooser.this);
		}
		}.start();
	}	
	//===================================================================
	public void onKeyEvent(KeyEvent ev)
	//===================================================================
	{
		if (ev.type == ev.KEY_PRESS){
			if (ev.key == IKeys.BACKSPACE) {
				if (lastInHistory > 0)
					setFile(afile.getNew(backForward.get(lastInHistory-1).toString()),true,-1);
			}else if (ev.key == IKeys.DELETE){
				int opts = 0;
				if ((ev.modifiers & IKeys.CONTROL) != 0) opts |= NOCONFIRM;
				deletePressed(opts);
			}if (ev.key == IKeys.LEFT && noKeyboard){
				dirUp();
			}
		}
		super.onKeyEvent(ev);
	}
	//##################################################################
	}
	//##################################################################


	//##################################################################
	class FileTreeControl extends TreeControl{
	//##################################################################

	public void resizeTo(int w,int h)
	{
		super.resizeTo(w,h);
		int m = FileChooser.this.modify(Invisible,0);
		if (dirNavigation != null){
			if (w < 10 || h < 10)
				dirNavigation.clearTag(dirNavigation.TAG_PREFERREDSIZE);
			else
				dirNavigation.setPreferredSize(0,0);
			dirNavigation.getParent().redisplay();
		}
		//FileChooser.this.setFile(FileChooser.this.parent,false);
		FileChooser.this.restore(m,Invisible);
	}

	//##################################################################
	}
	//##################################################################

//===================================================================
static String toFullPath(Object what)
//===================================================================
{
	if (what instanceof File) return File.fixupPath(((File)what).getFullPath());
	else return what.toString();
}
/**
* This does not affect the "history" variable. This just assigns the Strings or Files
* in the Vector to the history drop down box (desktop layout only). 
**/
//===================================================================
public void setHistory(Vector v)
//===================================================================
{
	if (historyList == null) return;
	historyList.shortenItems = true;
	historyList.items.clear();
	if (v == null) return;
	for (int i = 0; i<v.size(); i++)
		historyList.addItem(toFullPath(v.get(i)));
}

String [] foundFiles;
File parent;
Vector levels = new Vector();

//===================================================================
public void setOptions(FileChooserOptions options)
//===================================================================
{
	if (options == null) options = new FileChooserOptions();
	if (lastOptions == null) lastOptions = new FileChooserOptions();
	lastOptions.copyFrom(options);
	showHiddenFiles = (lastOptions.viewOptions & FileChooserOptions.VIEW_HIDDEN_FILES) != 0;
	MenuItem mi = toolMenu == null ? null : toolMenu.getMenu().findItem("Show Hidden Files",true);
	if (mi != null)
		if (showHiddenFiles)
			mi.modifiers |= mi.Checked;
		else
			mi.modifiers &= ~mi.Checked;
	sortBy = lastOptions.sortOptions & ~File.LIST_DESCENDING;
	descending = lastOptions.sortOptions & File.LIST_DESCENDING;
	tips = lastOptions.viewOptions & FileChooserOptions.VIEW_FILE_TIPS;
	int which = ((lastOptions.viewOptions & FileChooserOptions.VIEW_NAME_ONLY) != 0) ? 1 : 0;
	files.setShowDetails(detailMode = (which == 0));
	//iconMode = !(listMode = detailMode);
	files.fileTips = tips != 0;

	if (viewChoices != null && toolMenu != null){
		toolMenu.getMenu().checkOnlyOne(viewChoices,viewChoices[which],true);
		which = 0;
		if ((lastOptions.sortOptions & File.LIST_BY_DATE) == File.LIST_BY_DATE)
			which = 1;
		if ((lastOptions.sortOptions & File.LIST_BY_TYPE) == File.LIST_BY_TYPE)
			which = 2;
		if ((lastOptions.sortOptions & File.LIST_BY_SIZE) == File.LIST_BY_SIZE)
			which = 3;
		toolMenu.getMenu().checkOnlyOne(sortChoices,sortChoices[which],true);
		mi = toolMenu.getMenu().findItem("Descending",true);
		mi.modifiers &= ~mi.Checked;
		if ((descending & File.LIST_DESCENDING) != 0) mi.modifiers |= mi.Checked;
		mi = toolMenu.getMenu().findItem("Pop-up File Info",true);
		if (mi != null){
			mi.modifiers &= ~mi.Checked;
			if ((tips & FileChooserOptions.VIEW_FILE_TIPS) != 0) mi.modifiers |= mi.Checked;
		}
	}
	/*
	if (maskChoices != null)
	for (int i = 0; i<maskChoices.items.size(); i++)
		if (((String)maskChoices.items.get(0)).equals(lastOptions.mask)){
			mask = lastOptions.mask;
			break;
		}
		*/
}


/**
 * Use this to add a file mask to the chooser's mask list. The mask should
	be a string with a set of masks (e.g. *.png) separated by commas (no spaces in between)
	followed by a space and then any descriptive name.
	<p>For example: "*.jpg,*.png - Image files."
 * @param mask The mask to add.
 */
//===================================================================
public void addMask(String mask)
//===================================================================
{
	if (maskChoices != null) {
		if (!masksSet) {
			maskChoices.items.clear();
			this.mask = fixMask(mask);
			fileType = mask;
		}
		if (maskChoices.items.indexOf(mask) == -1)
			maskChoices.addItem(mask);
	}
	masksSet = true;
}
/**
* This returns the list of masks currently set for the file chooser. You can add, remove masks
* as necessary, but this must be done before the chooser is shown.
**/
//===================================================================
public Vector getMaskList()
//===================================================================
{
	if (maskChoices != null) return maskChoices.items;
	return null;
}
/**
* Use this if you have updated the Vector returned by getMaskList().
**/
//===================================================================
public void masksChanged()
//===================================================================
{
	if (maskChoices.items.size() != 0) {
		fileType = (String)maskChoices.items.get(0);
		mask = fixMask(fileType);
		masksSet = true;
	}else
		masksSet = false;
}
private static String getConfigInfo(String name) throws IOException
{
	return Io.getConfigInfo(name);
}
private static void saveConfigInfo(String name, String data, int options) throws IOException
{
	Io.saveConfigInfo(name,data,options); 
}
//===================================================================
public void shown()
//===================================================================
{
	Window w = getWindow();
	if (w != null) w.waitUntilPainted(3000);
	if (historyList != null && persistentHistoryKey != null)
		try{
			history = new mVector(mString.split(getConfigInfo("FileHistory\\"+persistentHistoryKey),'\n'));
			setHistory(history);
		}catch(Exception e){
		}
	File nf = afile.getNew(file); 
	setFile(nf);
	super.shown();	
}
/**
 * Setup the static history list for FileChoosers in this application using the specified
 * persistentHistoryKey. Call this method at the start of your application.
 * @param persistentHistoryKey a unique key name for your application.
 * @exception ewe.io.IOException if there was an error.
 */
//===================================================================
public static void retrieveApplicationHistory(String persistentHistoryKey)
throws IOException
//===================================================================
{
	history = new mVector(mString.split(getConfigInfo("FileHistory\\"+persistentHistoryKey),'\n'));	
}

/**
 * Store the specified list of Strings as the application history.
 * @param persistentHistoryKey a unique key name for your application.
 * @param history a Vector of Strings.
 * @exception ewe.io.IOException 
 */
//===================================================================
public static void storeApplicationHistory(String persistentHistoryKey,Vector history)
throws IOException
//===================================================================
{
	if (persistentHistoryKey != null){
		String s = "";
		for (int i = 0; i<history.size(); i++){
			if (i != 0) s += "\n";
			s += File.fixupPath(history.get(i).toString());
		}
		saveConfigInfo(s,"FileHistory\\"+persistentHistoryKey,0);//SAVE_IN_REGISTRY);
	}
}

/**
* Do not use this.
**/
public String mask = "*.*";

int sortBy = 0, descending = 0, tips;

/**
* Do not use this.
**/
//===================================================================
public String fixMask(String mask)
//===================================================================
{
	if (mask == null) return "";
	int idx = mask.indexOf(' ');
	if (idx == -1) return mask;
	return mask.substring(0,idx);
}
static final String [] readingDirectory = new String [] {"Reading directory..."};

boolean showDirsInTable()
{
	if ((type & NO_DIRECTORY_CHANGE) != 0) return false;
	if ((type & DIRECTORY_TREE) == 0) return true;
	Dimension d = tree.getSize(null);
	if (d.width < 10 || d.height < 10) return true;
	return false;
}
/**
* Do not use this.
**/
public boolean setFile(File file) {return setFile(file,true);}
boolean showHiddenFiles;
String lastDir;
int lastInHistory = -1;
Vector backForward;
/**
* Do not use this.
**/
//===================================================================
public boolean setFile(File file,boolean expandTree)
//===================================================================
{
	return setFile(file,expandTree,0);
}

boolean inList = false;

private String[] list(File who, String mask, int options)
{
	if (!showHiddenFiles) options |= File.LIST_DONT_LIST_HIDDEN_FILES;
	try{
		return (String[])SlowTaskException.checkSlowTask(
				who,File.class,"list(Ljava/lang/String;I)[Ljava/lang/String;",new Object[]{mask,new Integer(options)},250
				);
	}catch(SlowTaskException e){
		Object p = pauseEventThread();
		if (hasFileList) files.setFiles(who,readingDirectory);
		try{
			return (String[])e.waitForReturnValue();
		}finally{
			resumeEventThread(p);
		}
	}
}
private boolean setFile(final File tfile,final boolean expandTree,final int historyMoveDirection)
{
	if (tfile == null) return false;
	if (inList) return false;
	inList = true;
	File theDir = tfile.getCopy();
	String theFile = mask;
	if ((tfile.getFullPath().indexOf('*') != -1) || !tfile.isDirectory()) {
		theDir = tfile.getParentFile();
		theFile = tfile.getFileExt();
	}
	if (theDir == null) theDir = parent;//afile.getNew(baseDir);//return false;
	if (theDir == null) theDir = afile.getNew(baseDir);
	if (theFile == null) theFile = mask;
	if (theFile.indexOf('*') != -1) {
		mask = fixMask(theFile);
	}
	lastOptions.mask = mask;
	final File theParent = theDir;
	table.clearSelectedCells(null); table.clearCursor();
	//
	/*
	if (hasFileList)
		if (true || (afile.getFlags() & File.FLAG_SLOW_LIST) != 0 || Device.isMobile()){
			files.setFiles(theParent,readingDirectory);
		}
	*/
	try{
		int old = 0;
		boolean sw = (afile.getFlags() & File.FLAG_SLOW_LIST) != 0;
		if (sw) {
			//Vm.showWait(true);
			//files.setFiles(parent,readingDirectory);
		}
		String[] found = null;
		String[] dirs = null;
		if ((mask.indexOf(',') == -1 && mask.indexOf(';') == -1) || ((type & 3) == DIRECTORY_SELECT)){
			int opts = File.LIST_ALWAYS_INCLUDE_DIRECTORIES;
			if ((type & 3) == DIRECTORY_SELECT && ((type & SHOW_FILES_IN_DIRECTORY_SELECT) == 0)) opts |= File.LIST_DIRECTORIES_ONLY;
			if (!showDirsInTable()) opts = File.LIST_FILES_ONLY;
			opts |= sortBy|descending;
			found = list(theParent,mask,opts);
		}else{
			char c = mask.indexOf(',') == -1 ? ';' : ',';
			String masks [] = mString.split(mask,c);
			dirs = new String[0];
			if (showDirsInTable())
				dirs = list(theParent,"*.*",File.LIST_DIRECTORIES_ONLY);
			found = list(theParent,"*.*",File.LIST_FILES_ONLY|sortBy|descending);
	
			FileComparer [] fcs = new FileComparer[masks.length];
	
			for (int i = 0; i<masks.length; i++)
				fcs[i] = new FileComparer(theParent,Vm.getLocale(),sortBy|descending,masks[i]);
			int left = found.length;
			for (int i = 0; i<found.length; i++){
				boolean matched = false;
				for (int f = 0; f<fcs.length; f++){
					if (fcs[f].matches(found[i])){
						matched = true;
						break;
					}
				}
				if (!matched) {
					found[i] = null;
					left--;
				}
			}
			String [] isMatching = new String[dirs.length+left];
			System.arraycopy(dirs,0,isMatching,0,dirs.length);
			for (int i = 0, d = dirs.length; i<found.length; i++)
				if (found[i] != null)
					isMatching[d++] = found[i];
			found = isMatching;
		}
		doSetFile(tfile,theParent,expandTree,historyMoveDirection,dirs,found);
	}finally{
		synchronized(FileChooser.this){
			inList = false;
			FileChooser.this.notifyAll();
		}
	}
	return true;
}

//-------------------------------------------------------------------
private void doSetFile(File tfile,File parent,boolean expandTree,final int historyMoveDirection,String[] dirs, String[] found)
//-------------------------------------------------------------------
{
	File file = tfile.getCopy();
	foundFiles = found;
	this.parent = parent;
	try{
	if (back != null)
		if (backForward == null) backForward = new Vector();
	if (file == null) return;
	File theDir = parent;
	try{
		if (backForward != null){
			lastDir = theDir.getFullPath();
			int n = lastInHistory+1;
			if (historyMoveDirection > 0){
				for (n = lastInHistory+1; n < backForward.size() ; n++)
					if (backForward.get(n).equals(lastDir)) break;
			}else if (historyMoveDirection < 0){
				for (n = lastInHistory-1; n >= 0 ; n--)
					if (backForward.get(n).equals(lastDir)) break;
			}else{
				while(n < backForward.size()) backForward.removeElementAt(n);
			}
			if (n == -1){
				n = lastInHistory+1;
				while(n < backForward.size()) backForward.removeElementAt(n);
			}
			/*
			if (n > backForward.size()) n = backForward.size();
			else if (n < backForward.size()){
				FileChooserLink fcl = (FileChooserLink)backForward.get(n);
				if (!fcl.equals(lastDir))
					while(n < backForward.size()) backForward.del(n);
			}
			*/
			if (n == backForward.size()){
				if (n > 0){
					if (!backForward.get(n-1).equals(lastDir))
						backForward.add(lastDir);	
					else
						n--;
				}else	
					backForward.add(lastDir);
			}
			while (backForward.size() > 10) {
				backForward.removeElementAt(0);
				n--;
			}
			lastInHistory = n;
			FontMetrics fm = getFontMetrics();
			if (n == 0) back.modify(Disabled,0);
			else {
				back.modify(0,Disabled);
				Menu m = new Menu();
				for (int i = n-1; i >= 0; i--)
					new FileChooserLink(backForward.get(i).toString(),afile).addToMenu(m,fm);
				back.setMenu(m);
			}
			if (n == backForward.size()-1) forward.modify(Disabled,0);
			else {
				forward.modify(0,Disabled);
				Menu m = new Menu();
				for (int i = n+1; i < backForward.size(); i++)
					new FileChooserLink(backForward.get(i).toString(),afile).addToMenu(m,fm);
				forward.setMenu(m);
			}
			back.repaintNow(); forward.repaintNow();
			//Vm.debug(backForward.toString()+", "+lastInHistory);
			
		}
	}catch(Exception e){e.printStackTrace();} //Just for safety.	
	
	//new MessageBox("File",file.getFullPath()+"\n"+parent,0).exec();
	/*
	if ((type & LAUNCHER_TYPE) == LAUNCHER_TYPE)
		VMOptions.currentDir = parent.getFullPath();
		*/
	if (tree != null && expandTree){
		TreeTableModel tm = tree.getTreeTableModel();
		TreeNode root = tm.getRootObject();
		String fp = parent.getFullPath();
		//Vm.debug("FP: "+fp);
		fp = mString.removeTrailingSlash(fp);
		int [] addr = DataUtils.addressOfChild(root,fp);
		if (addr != null) tm.expandTo(addr,true);
	}
	if (hasFileList){
		files.setFiles(parent,found);
		//if (sw) Vm.showWait(false);
	}
	
	file = theDir;
	
	Vector v = new Vector();
	for(File p = file; p != null; p = p.getParentFile()){
		String s = p.getFileExt();
		if (!s.equals("."))
			v.insertElementAt(s,0);
	}
	if (directories != null){
		directories.removeAll();
		levels.clear();
		File cur = null;
		FontMetrics fm = directories.getMenuFontMetrics();
		int start = 0;
		if (links != null){
			start = links.size();
			for (int i = 0; i<start; i++){
				FileChooserLink l = (FileChooserLink)links.get(i);
				MenuItem mi = l.addToMenu(directories,fm);
				mi.indentLevel = 0;
				levels.add(afile.getNew(l.path));
			}
		}
		String [] all = sroots;
		String cd = (String)v.get(0);
		if (all == null) all = new String[]{cd};
		if (cd.indexOf(':') != -1) cd = mString.leftOf(cd,':');
		boolean did = false;
		int sel = 0;
		IImage drive = loadImage("eve/drive.png");
		IImage openFolder = loadImage("eve/openfolder.png");
		for (int r = 0; r<all.length; r++){
			if (!did && mString.leftOf(all[r],':').equalsIgnoreCase(cd)){
				did = true;
				int sz = v.size();
				for (int i = 0; i<sz; i++) {
					String dir = (String)v.get(i);
					levels.add(cur = afile.getNew(cur,dir));
					boolean isDrive = i == 0 ? dir.indexOf(':') != -1 : false;
					//if (isDrive) dir = dir.toUpperCase();
					MenuItem mi = directories.addItem(dir);
					mi.image = new IconAndText((isDrive ? drive : openFolder),dir,fm).setColor(null,null);
					mi.indentLevel = i;
				}
				sel = sz+r-1+start;
			}else{
				levels.add(afile.getNew(all[r]));
				//if (all[r].indexOf(':') != -1) all[r] = all[r].toUpperCase();
				MenuItem mi = directories.addItem(all[r]);
				mi.image = new IconAndText(drive,all[r],fm).setColor(null,null);
				mi.indentLevel = 0;
			}
		}
		//Vm.debug("Sel: "+sel);
		directories.select(sel);
	}
	String next = mask == null ? "" : mask;
	if (this.file != next){
		this.file = next;
		toControls("file");
	}
	focusToFile();
	return;
	}finally{
		//Vm.debug("CD: "+files.parent);//+", "+chosenFiles.get(0));
	}
}
public MenuItem drives,toolField;
//===================================================================
public boolean checkLength(File f,int warningLength)
//===================================================================
{
	long len = f.getLength();
	if (len > warningLength)
		if (new MessageBox("Large File","This file is: "+
		Vm.getLocale().format(len,",",0)
		+" bytes long.\nThis may take a while to open.\nDo you want to continue?",MBYESNO).execute(getFrame(),Gui.CENTER_FRAME)
		 != IDOK) return false;
	return true;
}

public boolean continueAfterError(File f, Throwable t, int errorType,boolean hasMore) {
	String err = "Could not read source file.";
	if (errorType == FileClipboard.FileClipboardSource.ERROR_COULD_NOT_CREATE_DESTINATION)
		err = "Could not write to destination file.";
	else
		err = "Could not write to destination file.";
	String full = "An error occured with file:\n"+f.getFileExt()+"\n"+err;
	ReportException re = new ReportException(t,full,hasMore ? "Continue with other files?":null,hasMore);
	re.windowFlagsToClear |= WindowConstants.FLAG_HAS_CLOSE_BUTTON;
	re.title = "File Error";
	int got = re.execute();
	return got == IDOK;
}

//===================================================================
public static boolean doPaste(File targetDirectory,FileChooser targetChooser)
//===================================================================
{
	FileClipboard fc = targetChooser.getFileClipboard(false);
	if (targetDirectory.equals(fc.sourceDir)){
		Gui.flashMessage("Cannot paste to this directory.",targetChooser);
		return false;
	}
	final Handle [] h = fc.copyTo(targetDirectory,true);
	Form f = new FileClipboardProgress(h,"Paste Files...");
	f.execute();
	fc.taken();
	fc.clear();
	targetDirectory.refresh();
	if (targetChooser != null) {
		targetChooser.setFile(targetChooser.files.parent);
		if (targetChooser.tree != null)
			targetChooser.tree.getTreeTableModel().reExpandNode(targetChooser.tree.getTreeTableModel().selectedLine);
	}
	
	//Gui.flashMessage("Cannot paste yet!",targetChooser);
	/*
	FileClipboard c = FileClipboard.clipboard;
	if (!c.hasFiles()) return false;
	if (c.sourceFiles.length == 0) return false;
	File dest = targetDirectory;
	if (!dest.isDirectory()) return false;
	ewe.io.CopyOver co = new ewe.io.CopyOver();
	co.sourceDirectory = c.sourceDir;
	co.destDirectory = dest;
	co.sourceFiles = c.sourceFiles;
	co.doMove = c.isCut;
	co.makeCopies = !c.isCut;
	co.displayTitle = c.isCut ? "Move Files" : "Copy Files";
	co.execute(targetChooser == null ? null : targetChooser.getFrame(),null);
	*/
	return true;
}
//===================================================================
void detailModeChanged()
//===================================================================
{
	files.setShowDetails(detailMode);
	lastOptions.viewOptions = (lastOptions.viewOptions & ~FileChooserOptions.VIEW_NAME_ONLY) | ((detailMode) ? 0 : FileChooserOptions.VIEW_NAME_ONLY);
	if (detailMode) {
		//iconMode = false;
		//listMode = true;
		//toControls("iconMode,listMode");
	}
	toControls("detailMode");
	table.clearSelectedCells(null); 
	table.clearCursor();
	table.update(true);
}

//===================================================================
public static String askClassName(String pathToClass,StringBuffer programDir,Frame parent)
//===================================================================
{
	String target = pathToClass;
	final String cls = target.replace('\\','.').replace('/','.').substring(0,target.length()-6);
	String className = null;
	int idx = cls.toLowerCase().indexOf("classes.");
	if (idx != -1)
		className = cls.substring(idx+8);
	else{
		PropertyList pl = new InputObject(){
			protected void setupInputStack(InputStack is,Editor ed){
				ed.windowFlagsToClear = Window.FLAG_HAS_TITLE;
				Vector v = new Vector();
				String got = cls;
				while(true){
					v.insertElementAt(got,0);
					int i = got.indexOf('.');
					if (i == -1) break;
					got = got.substring(i+1);
				}
				String [] all = new String[v.size()];
				v.copyInto(all);
				ed.addField(is.addChoice("Class:",all,0),"className$");
			}
		}.input(parent,"Select Class Name",null);
		if (pl == null) return null;
		className = pl.getString("className","<none>");
	}
	if (className == null) return null;
	String path = target.substring(0,target.length()-className.length()-1-6);
	if (path.toLowerCase().endsWith("classes")) path = path.substring(0,path.length()-8);
	if (programDir != null) programDir.append(path);
	return className;
}

public void doNewShortcut(Editor f)
{
	if (!afile.getClass().getName().equals("eve.io.File")){
		showError("Cannot create shortcut on this file system.",true);
		return;
	}
	NewShortcut ns = new NewShortcut();
	if (ns.getEditor(0).execute(f.getFrame(),Gui.CENTER_FRAME) == IDCANCEL) return;
	try{
		String sn = ns.shortcutName;
		String args = ns.arguments.trim();
		if (args.length() == 0) args = null;
		if (!sn.toLowerCase().endsWith(".lnk"))
			sn += ".lnk";
		File sc = afile.getNew(parent,sn);
		if (sc.exists()){
			showError("Shortcut already exists!",true);
			return;
		}
		Device.createShortcut(ns.target,args,sc.getFullPath());
	}catch(Exception e){
		showError("Could not create shortcut!",true);
		return;
	}
	setFile(parent);
}
//===================================================================
public void runFile(File file)
//===================================================================
{
	String f = file.toString();
	try{
		Registry.executeShellCommand(f);
	}catch(Exception e){
		new ReportException(e).execute();
	}
	//Gui.flashMessage("Cannot execute yet.",this);
	/*
	String where = file.getFullPath();
	String sl = where.toLowerCase();
	if (sl.endsWith(".ewe"))
		Vm.executeEwe(file.getFullPath(),null);
	else if (sl.endsWith(".exe"))
		Vm.execute(file.getFullPath(),null);
	else if (sl.endsWith(".class")){
		StringBuffer programDir = new StringBuffer();
		String clName = askClassName(where,programDir,null);
		if (clName == null) return;
		String all = "/d \""+programDir+"\" "+clName;
		Vm.execute(Vm.getPathToEweVM(),all);
	}else {
		showError("Cannot run that file.",true);
	}
	*/
}
//===================================================================
private void cantView()
//===================================================================
{
	showError("Cannot view that file type\non a mobile device.",true);
}
static final int NOCONFIRM = 1;
//===================================================================
public void deletePressed(int how)
//===================================================================
{
	String [] all = files.getSelectedFiles();
	if (all.length == 0) return;
	if ((how & NOCONFIRM) == 0)
	if (new MessageBox("Delete?","Delete the selected file(s)?",MBYESNO).execute(getFrame(),Gui.CENTER_FRAME) != IDOK)
		return;
	if (all.length == 1 && !files.parent.getChild(all[0]).isDirectory()){
		if (!files.parent.getChild(all[0]).delete())
			new MessageBox("Delete Error","That file could not be deleted",MBOK).execute();
	}else{
		FileClipboard fc = new FileClipboard();
		fc.set(files.parent,all,false,files);
		Handle h = fc.delete(true);
		FileClipboardProgress fcp = new FileClipboardProgress(h,"Deleting files...");
		fcp.execute();
	}	
	files.parent.refresh();
	setFile(files.parent);
	switchMultiSelect(false);
}
//===================================================================
public void checkMenu(String menuLabel,Editor f)
//===================================================================
{
		String which = menuLabel;
		if (toolMenu.getMenu().checkOnlyOne(viewChoices,which,true)){
			detailMode = which.equals(viewChoices[0]);
			detailModeChanged();
		}else if (toolMenu.getMenu().checkOnlyOne(sortChoices,which,true)){
			if (which.equals(sortChoices[0])) sortBy = 0;
			else if (which.equals(sortChoices[1])) sortBy = File.LIST_BY_DATE;
			else if (which.equals(sortChoices[2])) sortBy = File.LIST_BY_TYPE;
			else if (which.equals(sortChoices[3])) sortBy = File.LIST_BY_SIZE;
			lastOptions.sortOptions = sortBy | descending;
			setFile(parent);
		}else if (which.equals("Multiple Select")){
			switchMultiSelect(!table.multiSelect);
		}else if (which.equals("Descending")){
			MenuItem mi = toolMenu.getMenu().findItem(which,true);
			mi.modifiers ^= mi.Checked;
			if ((mi.modifiers & mi.Checked) != 0) descending = File.LIST_DESCENDING;
			else descending = 0;
			lastOptions.sortOptions = sortBy | descending;
			setFile(parent);
		}else if (which.equals("Folder")){
			doNewFolder(f);
		}else if (which.equals("File")){
			doNewFile(f);
		}else if (which.equals("Shortcut")){
			doNewShortcut(f);
		}else if (which.equals("Show Hidden Files")){
			MenuItem mi = toolMenu.getMenu().findItem(which,true);
			if (mi != null){
				mi.modifiers ^= mi.Checked;
				showHiddenFiles = ((mi.modifiers & mi.Checked) != 0);
				if (!showHiddenFiles)
					lastOptions.viewOptions &= ~FileChooserOptions.VIEW_HIDDEN_FILES;
				else
					lastOptions.viewOptions |= FileChooserOptions.VIEW_HIDDEN_FILES;
			}
			setFile(parent);
		}else if (which.equals("Refresh")){
			parent.refresh();
			setFile(parent);
		}else if (which.equals("Pop-up File Info")){
			MenuItem mi = toolMenu.getMenu().findItem(which,true);
			if (mi != null){
				mi.modifiers ^= mi.Checked;
				if ((mi.modifiers & mi.Checked) != 0) tips = FileChooserOptions.VIEW_FILE_TIPS;
				else tips = 0;
				lastOptions.viewOptions &= ~FileChooserOptions.VIEW_FILE_TIPS;
				lastOptions.viewOptions |= tips;
				files.fileTips = tips != 0;
			}
		}else if (which.equals("Properties")){
			try{
				String [] all = files.getSelectedFiles();
				if (all.length != 1) return;
				File file = files.parent.getChild(all[0]);
				String prop = file.getPropertiesString();
				/*
				Form ff = new Form();
				ff.title = "File Properties";
				mTabbedPanel mt = new mTabbedPanel();
				mTextPad tp = new mTextPad(prop);
				mt.addItem(new ScrollBarPanel(tp),"General",null);
				ff.addLast(mt);
				*/
				MessageBox ff = new MessageBox("File Properties",prop,MessageBox.MBOK);
				ff.setPreferredSize(220, 300);
				ff.doBeep = false;
				ff.useTextMessage = ff.useScrollBars = true;
				ff.textAlignment = LEFT;
				ff.resizable = true;
				ff.execute();//getFrame(),Gui.CENTER_FRAME);
			}catch(Exception e){
				new ReportException(e,null,null,false).execute();
				return;
			}
		}else if (which.equals("Rename")){
			String [] all = files.getSelectedFiles();
			if (all.length != 1) return;
			File file = files.parent.getChild(all[0]);
			int [] got = table.getSelectedIndexes();
			String newName = null;
			Rect r = null;
			if (!alwaysUsePopupInputs && !Gui.isPDA) r = files.getTextRect(got[0], null);
			if (r != null){
				newName = Input.execute(all[0],table, r.x-1, r.y-1, r.width+2, r.height+2, 0);
			}else{
				newName = new InputBox("Rename File").input(getFrame(),all[0],30);
			}
			if (newName == null) return;
			if (newName.indexOf('/') != -1 || newName.indexOf('\\') != -1 || newName.indexOf(':') != -1){
				showError("Bad file name.",false);
				return;
			}
			if (!file.rename(newName)){
				showError("The file could not be renamed.",false);
				return;
			}
			parent.refresh();
			int idx = files.indexOf(all[0]);
			if (idx == -1) return;
			files.files[idx] = newName;
			Point p = files.cellOf(idx);
			if (p == null) return;
			table.repaintCell(p.y,p.x);
		}else if (which.equals("Select All")){
			table.selectAllIndexes(true);
			table.repaintNow();
		}else if (which.equals("Invert Selection")){
			table.invertSelectedIndexes();
			table.repaintNow();
		}else if (which.equals("Cut")||which.equals("Copy")){
			boolean cut = which.equals("Cut");
			files.setFilesToCopy(cut);
			if (noKeyboard && table.multiSelect)
				switchMultiSelect(false);
			if (cut) table.repaintNow();
		}else if (which.equals("Delete")){
			deletePressed(0);
		}else if (which.equals("Touch")){
			Time t = new Time();
			try{
				String [] all = files.getSelectedFiles();
				for (int i = 0; i<all.length; i++){
					File dest = files.parent.getChild(all[i]);
					dest.setModified(t);
				}
			}catch(Exception e){
				new ReportException(e).execute();
			}
		}else if (which.equals("Paste") || which.equals("Paste In")){
			File dest = files.parent;
			if (which.equals("Paste In")){
				String [] all = files.getSelectedFiles();
				if (all.length != 1) return;
				dest = dest.getChild(all[0]);
			}
			doPaste(dest,this);
		}else if (which.equals("Close")){
			exit(0);
		}else if (which.equals("Install")){
			//FIXME
			/*
			String [] all = files.getSelectedFiles();
			if (all.length != 1) return;
			String sl = all[0].toLowerCase();
			if (sl.endsWith(".class") || sl.endsWith(".ewe")){
				if (LaunchPanel.launcher != null)
					LaunchPanel.launcher.add(files.parent.getChild(all[0]).getFullPath(),getFrame());
			}
			*/
		}else if (which.equals("Run")){
			String [] all = files.getSelectedFiles();
			if (all.length != 1) return;
			runFile(files.parent.getChild(all[0]));
			return;
		}else if (which.endsWith(" File")){
			String [] all = files.getSelectedFiles();
			if (all.length != 1) return;
			String sl = all[0].toLowerCase();
			if (which.equals("View File")){
				if (sl.endsWith(".txt")) checkMenu("Text File",f);
				if (sl.endsWith(".zip")) checkMenu("Zip File",f);
				if (sl.endsWith(".ewe")) checkMenu("Ewe File",f);
				if (sl.endsWith(".eve")) checkMenu("Eve File",f);
				if (sl.endsWith(".htm")) checkMenu("HTML File",f);
				if (sl.endsWith(".html")) checkMenu("HTML File",f);
				if (sl.endsWith(".bmp") || sl.endsWith(".png") || 
				sl.endsWith(".jpg")||
				sl.endsWith(".jpeg")||
				sl.endsWith(".gif")
				
				) checkMenu("Image File",f);
				else return;
			}
			final File file = files.parent.getChild(all[0]);
			if (which.equals("Text File")){
				if (!allowFileViewing){
					cantView();
					return;
				}
				if (!checkLength(file,60000)) return;
				InputStream is = null;
				try{
					is = file.toReadableStream();
				}catch(IOException e){}
				if (is == null) {
					showError(file.getFileExt()+"\nThat file cannot be viewed.",true);
					return;
				}
				MessageBox mb = new MessageBox("Opening...","Opening the text file.\nPlease wait",0);
				try {is.close();}catch(IOException e){}
				Notepad np = new Notepad();
				np.saver.fileModel = afile;
				mb.doBeep = false;
				mb.exec();
				
				boolean op = false;
				try{
					op = np.open(file.getFullPath());
				}catch(IOException e){
					op = false;
				}finally{
					mb.close(0);
				}
				if (op) np.exec(isModal());
				/*
				ewe.io.StreamReader br = new ewe.io.StreamReader(is);
				br.encoding = null;
				String txt = "";
				while(true){
					String line = br.readLine();
					if (line == null) break;
					txt += line+"\n";
				}
				br.close();
				final String nt = txt;
				mb.close(0);
				new Form(){
					{
						title = file.getFileExt();
						resizable = true;
						mTextPad mt = new mTextPad(10,40);
						addLast(new ScrollBarPanel(mt));
						mt.setText(nt);
						mt.modify(DisplayOnly,0);
					}
				}.exec(isModal());
				*/
			}else if (which.equals("Image File")){
				IImage im = null; 
				try{
					im = ImageDecoder.decodeScaledPicture(new FormattedDataSource().set(file.toReadableStream()), 200, 200, Device.isMobile(), true,true, null);
				}catch(Exception e){
					new ReportException(e).execute();
				}
				Form ff = new Form();
				ff.maximizeOnPDA();
				ff.title = "Image";
				ImageControl ic = new ImageControl(im);
				ff.addLast(ic).setPreferredSize(200,200);
				ff.exec(isModal());
			}else if (which.equals("Binary File")){
				if (!allowFileViewing){
					cantView();
					return;
				}
				try{
					Form ff = HexTableModel.getViewOf(file,100000,null);
					ff.execute();
					return;
				}catch(IOException e){
					showError(file.getFileExt()+"\nThat file cannot be viewed.",true);
				}
			}else if (which.equals("HTML File")){
				if (!allowFileViewing){
					cantView();
					return;
				}
				if (web.exists()){
					IWebBrowser w = (IWebBrowser)web.newInstance();
					w.showFor(file.getFullPath(),true);
				}
			}else if (which.equals("Zip File")){
				if (!allowFileViewing){
					cantView();
					return;
				}
				try{
					//new ZipFileBrowser(file,isModal(),10000000);
					new FileStoreBrowser(file,ZipFile.class,isModal(),10000000);
				}catch(Exception e){
					e.printStackTrace();
				}
				//Object got = zipView.newInstance("(Leve/io/File;ZJ)V",new Object[]{file,new Boolean(isModal()),new java.lang.Long(10000000)});
				//if (got == null) showError(file.getFileExt()+"\nThat file cannot be viewed.",true);
			}else if (which.equals("Eve File") || which.equals("Ewe File")){
				try{
					//new ZipFileBrowser(file,isModal(),10000000);
					new FileStoreBrowser(file,EveFile.class,isModal(),10000000);
				}catch(Exception e){
					e.printStackTrace();
				}
				/*
				if (!allowFileViewing){
					cantView();
					return;
				}
				Object got = eweView.newInstance("(Lewe/io/File;ZJ)V",new Object[]{file,new Boolean(isModal()),new java.lang.Long(1000000)});
				if (got == null) showError(file.getFileExt()+"\nThat file cannot be viewed.",true);
				*/
			}
		}
}
//===================================================================
public void fieldChanged(FieldTransfer ft,Editor f)
//===================================================================
{
	if (amExiting||inList) return;
	if (ft.fieldName.equals("directories")){
		int i = directories.selectedIndex;
		if (i < 0 || i >= levels.size()) return;
		setFile((File)levels.get(i));
	}
	else if (ft.fieldName.equals("file")){
		Event ev = f.currentEvent;
		if (ev != null)
			if ((ev.flags & Input.DATA_CHANGED_BY_ENTER) != 0) return;
		newFileInfo(false);
	}else if(ft.fieldName.equals("fileType")){
		file = fileType;
		f.toControls("file");
		newFileInfo(false);
	}else if (ft.fieldName.equals("drives")){
		setFile(afile.getNew(drives.label));
	}else if (ft.fieldName.equals("toolField")){
		checkMenu(toolField.label,f);
	}else if (ft.fieldName.equals("iconMode")){// || ft.fieldName.equals("listMode") || ft.fieldName.equals("detailMode")){
		files.verticalOnly = !iconMode;
		files.clearCellAdjustments();
		files.updateDisplay();
	}else if (ft.fieldName.equals("detailMode")){
		detailModeChanged();
	}else if (ft.fieldName.equals("goBackMenu")){
		setFile(afile.getNew(goBackMenu.data.toString()),true,-1);
	}else if (ft.fieldName.equals("goForwardMenu")){
		setFile(afile.getNew(goForwardMenu.data.toString()),true,+1);
	}
}

//-------------------------------------------------------------------
void focusToFile()
//-------------------------------------------------------------------
{
	try{
		if (noKeyboard) return;
		Control ip = (Control)findFieldTransfer("file").dataInterface;
		Input mi = null;
		if (ip instanceof ComboBox)
			mi = ((ComboBox)ip).input;
		else if (ip instanceof Input)
			mi = ((Input)ip);
		if (ip instanceof ComboBox)
			if (((ComboBox)ip).choice.menuIsActive())
				mi = null;
		if (mi != null){
			if (Gui.focusedControl() != mi) return;
			mi.selectAll();
			//if (Gui.focusedControl() != mi) Gui.takeFocus(mi,ByRequest);
		}
	}catch(Exception e){}
}
/**
* This gets called when a new file has been clicked or entered. This file
* has not necessarily been accepted yet.
**/
//-------------------------------------------------------------------
protected void newFileSelected(File f)
//-------------------------------------------------------------------
{
}
//===================================================================
public boolean newFileInfo(boolean isAction)
//===================================================================
{
	boolean ret = doNewFileInfo(isAction);
	if (ret && !isAction)
		newFileSelected(afile.getNew(parent,file));
	return ret;
}
//-------------------------------------------------------------------
File existsAs(File f)
//-------------------------------------------------------------------

{
	if (f.exists()) return f;
	if (defaultExtension == null) return null;
	File f2 = afile.getNew(f.getFullPath()+"."+defaultExtension);
	if (f2.exists()) return f2;
	return null;
}
//-------------------------------------------------------------------
boolean doNewFileInfo(boolean isAction)
//-------------------------------------------------------------------
{
		if ((type & 3) == DIRECTORY_SELECT) {
			file = parent.getFullPath();
			if (isAction) exit(IDOK);
			return true;
		}else{
			if (file.indexOf('*') != -1){
				mask = fixMask(file);
				setFile(parent);
			}else{
				File nf = afile.getNew(parent,file);
				if (file.indexOf(':') != -1 || file.startsWith("\\") || file.startsWith("/")){
					nf = afile.getNew(null,file);
				}
				if (nf.isDirectory()) 
					if ((type & NO_DIRECTORY_CHANGE) == 0) setFile(nf);
					else return false;
				else{
					if (!table.multipleSelected()){
						if (((type & FILE_MUST_EXIST) != 0) && isAction)
							if (existsAs(nf) == null){
								showError("Bad File","That file does not exist."
										//+"\n"+nf
										,true);
								return false;
							}
						//setFile(nf);
						if (isAction && ((type & 3) != BROWSE)) {
							if (((type & 3) == DIRECTORY_SELECT) && !nf.isDirectory())
								return false;
							file = nf.getFullPath();
							exit(IDOK);
						}
					}else{
						if ((type & MULTI_SELECT) == 0){
							showError("Too Many Files","Please select one file only.",true);
							return false;
						}
						if (isAction && ((type & 3) != BROWSE)){
							if (((type & 3) == DIRECTORY_SELECT) && !nf.isDirectory())
								return false;
							file = nf.getFullPath();
							exit(IDOK);
						}
					}
					return true;
				}
			}
		}
		focusToFile();
		return false;
}

boolean amExiting = false;

//===================================================================
public boolean exit(int value)
//===================================================================
{
	//Vm.debugStackTrace();
	if ((type & EMBEDDED) != 0) {
		close(value);
		return false;
	}else 
		return super.exit(value);
}
//===================================================================
public void close(int value)
//===================================================================
{
	if (value == IDOK){
		File ret = afile.getNew(file);
		int [] all = table.getSelectedIndexes();
		if (all.length > 1){
			chosenDirectory = files.parent.getFullPath();
			chosenFiles = new Vector();
			for (int i = 0; i<all.length; i++)
				chosenFiles.add(files.files[all[i]]);
			File f = afile.getNew(files.parent,files.files[all[0]]);
			if (f != null) {
				file = f.getFullPath();
				ret = f;
			}
		}else{
			if (!ret.exists()){
				int id = file.indexOf('.');
				if (id == -1 && defaultExtension != null){
					file = file+"."+defaultExtension;
				}else if (id != -1 && id == file.length()-1){
					file = file.substring(0,id);
				}
				ret = afile.getNew(file);
			}
			File p = ret.getParentFile();
			if (p == null) chosenDirectory = "/";
			else chosenDirectory = p.getFullPath();
			chosenFiles = new Vector();
			chosenFiles.add(ret.getFileExt());
		}
		putInHistory(history,ret,historyDirectoriesOnly);
		setHistory(history);
		storeHistory(history,persistentHistoryKey);
		//Vm.debug("Choosing: "+chosenDirectory);
		if (fcp != null){
			fcp.set(fcp.CHOSEN_FILE,getChosenFile());
			fcp.set(fcp.CHOSEN_FILES,getAllChosen());
		}
	}
	if ((type & EMBEDDED) != 0){
		if (value == IDOK) notifyAction();
	}else
		super.close(value);
}

//-------------------------------------------------------------------
private static boolean storeHistory(Vector history,String persistentHistoryKey)
//-------------------------------------------------------------------
{
	if (persistentHistoryKey != null){
		String s = "";
		for (int i = 0; i<history.size(); i++){
			if (i != 0) s += "\n";
			s += File.fixupPath(history.get(i).toString());
		}
		try{
			saveConfigInfo(s,"FileHistory\\"+persistentHistoryKey,0);//Io.SAVE_IN_REGISTRY);
		}catch(Exception e){
			return false;
		}
	}
	return true;
}

//-------------------------------------------------------------------
private static void putInHistory(Vector history,File ret,boolean onlyDir)
//-------------------------------------------------------------------
{
	if (onlyDir)
		if (!ret.isDirectory()) 
				ret = ret.getParentFile();
	if (history != null){
		String where = File.fixupPath(ret.getFullPath());
		boolean found = false;
		for (int i = 0; i<history.size(); i++){
			if (where.equalsIgnoreCase(toFullPath(history.get(i)))){
				found = true;
				break;
			}
		}
		if (!found) history.insertElementAt(ret,0);
		while(history.size() > historySize && history.size() > 0)
			history.removeElementAt(history.size()-1);
	}
}

/**
 * Add a file or directory to the saved history.
 * @param historyKey the name of the history key.
 * @param fileOrDirectory a file or directory.
 * @return true if it saved successfully, false if not.
 */
//===================================================================
public static boolean addToSavedHistory(String historyKey,File fileOrDirectory)
//===================================================================
{
	try{
		if (fileOrDirectory == null) return true;
		Vector v = new mVector(mString.split(getConfigInfo("FileHistory\\"+historyKey),'\n'));
		putInHistory(v,fileOrDirectory,true);
		return storeHistory(v,historyKey);
	}catch(Exception e){
		return false;
	}
}
//===================================================================
public boolean canExit(int value)
//===================================================================
{
	amExiting = true;
	if (value == IDCANCEL || ((type & 3) != SAVE) || ((type & NO_CONFIRM_OVERWRITE) != 0)) return true;
	final File f = existsAs(afile.getNew(file));
	if (f == null) return true;
	new Task(){
		protected void doRun(){
			int ret = new MessageBox("Confirm Overwrite","Overwrite existing file?\n"+f.getFileExt(),MBYESNOCANCEL).execute();
			if (ret == IDCANCEL) close(IDCANCEL);
			else if (ret == IDYES) close(IDOK);
			else {
				file = f.getFileExt();
				focusToFile();
				amExiting = false;
			}
		}
	}.start();
	return false;
}

//-------------------------------------------------------------------
void fileAction()
//-------------------------------------------------------------------
{
	File nf = afile.getNew(parent,file);
	boolean updated = false;
	if (file.indexOf(':') != -1 || file.startsWith("\\") || file.startsWith("/")){
		if (file.endsWith(":")) file += "/";
		nf = afile.getNew(null,file);
		file = nf.getFullPath();
		updated = true;
	}
	if (nf.isDirectory() || file.indexOf('*') != -1) {
		setFile(nf);
		return;
	}
	if (file.indexOf('*') != -1) return;
	if (existsAs(nf) == null)
		if ((type & FILE_MUST_EXIST) != 0){
			showError("Bad File","That file does not exist."
					//+"\n"+nf
					,true);
			return;
		}
	if (!updated) file = afile.getNew(parent,file).getFullPath();
	if (((type & 3) == DIRECTORY_SELECT) && !nf.isDirectory())
		return;
	if ((type & 3) != BROWSE)
		exit(IDOK);
	if ((type & LAUNCHER_TYPE) == LAUNCHER_TYPE)
		runFile(nf);
}
//===================================================================
public void dirUp()
//===================================================================
{
	if (parent != null){
		File up = parent.getParentFile();
		if (up != null) setFile(up);
	}
/*
	int l = levels.size();
	if (l <= 1) return;
	setFile((File)levels.get(l-2));
*/
}
//===================================================================
public void action(FieldTransfer ft,Editor f)
//===================================================================
{
	if (amExiting||inList) return;
	if (ft.fieldName.equals("dirUp")){
		dirUp();
	}else if (ft.fieldName.equals("acceptSelection")){// || ft.fieldName.equals("file")){
		//System.out.println("Accept selection!");
		if (!newFileInfo(true)){
			if ((type & ACCEPT_ANY) != 0){
				file = afile.getNew(parent,file).getFullPath();
				exit(IDOK);
			}
		}
	}else if (ft.fieldName.equals("file")){
		f.fromControls(ft.fieldName);
		fileAction();
	}else if (ft.fieldName.equals("newFolder")){
		doNewFolder(f);
	}else if (ft.fieldName.equals("fileTable")){
		fileAction();
	}else if (ft.fieldName.equals("goBack")){
		if (lastInHistory > 0)
			setFile(afile.getNew(backForward.get(lastInHistory-1).toString()),true,-1);
	}else if (ft.fieldName.equals("goForward")){
		if (lastInHistory < backForward.size()-1)
			setFile(afile.getNew(backForward.get(lastInHistory+1).toString()),true,+1);
	}
}
//===================================================================
private File getNewFileFolder(Editor f,String type)
//===================================================================
{
	String newDir = new InputBox("New "+type).input(getFrame(),"New"+type,20);
	if (newDir == null) return null;
	File dir = afile.getNew(parent,newDir);
	if (dir.exists()){
		showError(type+" already exists.",true);
		return null;
	}
	return dir;
}
//===================================================================
void doNewFolder(Editor f)
//===================================================================
{
	File dir = getNewFileFolder(f,"Folder");
	if (dir == null) return;
	if (!dir.createDir()){
		showError("The folder could not be created!",true);
		return;
	}
	setFile(parent);
	if (/*((type & 3) == DIRECTORY_SELECT) && */((type & DIRECTORY_TREE) != 0)){
		tree.getTreeTableModel().reExpandSelectedNode();
	}
}
//===================================================================
void doNewFile(Editor f)
//===================================================================
{
	File file = getNewFileFolder(f,"File");
	if (file == null) return;
	try{
		OutputStream s = file.toWritableStream(true);
		s.close();
	}catch(IOException e){
		showError("The file could not be created!",true);
		return;
	}
	setFile(parent);
}

//===================================================================
public void switchMultiSelect(boolean on)
//===================================================================
{
	if (toolMenu != null){
		MenuItem mi = toolMenu.getMenu().findItem("Multiple Select",true);
		if (mi == null) return;
		if (on) mi.modifiers |= mi.Checked;
		else mi.modifiers &= ~mi.Checked;
		table.multiSelect = on;
	}
	/*
	if (!on && table.getSelectedIndexes().length > 1) {
		table.clearSelectedCells(null);
		table.repaintNow();
	}
	*/
}
private File selected;

//===================================================================
public void fieldEvent(FieldTransfer ft,Editor f,Object ev)
//===================================================================
{
	if (inList) return;
	if (ft.fieldName.equals("fileTable")){
		if (ev instanceof TableEvent){
			TableEvent te = (TableEvent)ev;
			if (te.type == te.CELL_CLICKED){
				int idx = files.getFileIndex(te.row,te.col);
				if (idx != -1){
					File nf = selected;
					if (nf == null) nf = selected = afile.getNew(parent,foundFiles[idx]);
					else nf.set(parent,foundFiles[idx]);
					boolean selectedByKey = ((te.flags & TableEvent.FLAG_SELECTED_BY_ARROWKEY) != 0);
					if (nf.isDirectory() && !table.multiSelect && !selectedByKey) setFile(nf);
					else{
						file = foundFiles[idx];
						f.toControls("file");
						if ((type & QUICK_SELECT) != 0 && ((type & 3) != BROWSE) && !nf.isDirectory() && !selectedByKey) {
							file = nf.getFullPath();
							exit(IDOK);
						}
						focusToFile();
						newFileSelected(nf);
						//table.addToSelection(te.row,te.col);
						//table.repaintCell(te.row,te.col);
					}
				}
			}else if (te.type == te.CELL_DOUBLE_CLICKED){
				fileAction();				
			}
		}/*
	}else if (ft.fieldName.equals("drives")){
		if (ev instanceof MenuEvent){
			MenuEvent me = (MenuEvent)ev;
			if (me.type == me.SELECTED){
				
			}
		}*/
	}else if (ft.fieldName.equals("dirTree")){
		if (ev instanceof TreeEvent){
			TreeEvent te = (TreeEvent)ev;
			File f2 = ((FileTree)tree.getTableModel()).getFileAt(te.selectedLine);
			if (f2 != null) {
				//Vm.debug("Selected: "+f2);
				setFile(f2);
			}
		}
		/*
	}else if (ft.fieldName.equals("toolField")){
		if (ev instanceof MenuEvent){
			MenuEvent me = (MenuEvent)ev;
			if (me.type == me.SELECTED){
				String which = ((MenuItem)me.selectedItem).label;
				if (toolMenu.getMenu().checkOnlyOne(viewChoices,which,true)){
					files.setShowDetails(which.equals(viewChoices[0]));
					lastOptions.viewOptions = (which.equals(viewChoices[0])) ? 0 : FileChooserOptions.VIEW_NAME_ONLY;
					table.update(true);
				}else if (toolMenu.getMenu().checkOnlyOne(sortChoices,which,true)){
					if (which.equals(sortChoices[0])) sortBy = 0;
					else if (which.equals(sortChoices[1])) sortBy = File.LIST_BY_DATE;
					else if (which.equals(sortChoices[2])) sortBy = File.LIST_BY_TYPE;
					else if (which.equals(sortChoices[3])) sortBy = File.LIST_BY_SIZE;
					lastOptions.sortOptions = sortBy | descending;
					setFile(parent);
				}else if (which.equals("Descending")){
					MenuItem mi = toolMenu.getMenu().findItem(which,true);
					mi.modifiers ^= mi.Checked;
					if ((mi.modifiers & mi.Checked) != 0) descending = File.LIST_DESCENDING;
					else descending = 0;
					lastOptions.sortOptions = sortBy | descending;
					setFile(parent);
				}else if (which.equals("Folder")){
					doNewFolder(f);
				}
			}			
		}*/
	}
}
public static void eveMain(String args[])
{
	//System.out.println("eveMain()");
}

//===================================================================
public static void main(String args[])
//===================================================================
{
	//System.out.println("main()");
	Application.startApplication(args);
	//System.out.println("Now running main!");
	FileChooser fc = new FileChooser(BROWSE|EXPLORER_TYPE|DIRECTORY_TREE,baseDir);
	fc.exitSystemOnClose = true;
	//fc.windowFlagsToSet |= WindowConstants.FLAG_ALWAYS_ENABLED;
	fc.show(); 
}
//##################################################################
}
//##################################################################



