package eve.ui.filechooser;
import java.io.IOException;
import java.io.InputStream;
import java.util.Vector;

import eve.fx.Color;
import eve.fx.Dimension;
import eve.fx.DrawnIcon;
import eve.fx.IImage;
import eve.fx.ImageDecoder;
import eve.io.File;
import eve.io.FileChooserParameters;
import eve.sys.Device;
import eve.sys.Handle;
import eve.sys.Task;
import eve.sys.Vm;
import eve.ui.Application;
import eve.ui.GuiStyle;
import eve.ui.ImageControl;
import eve.ui.Label;
import eve.util.FormattedDataSource;

//##################################################################
public class ImageFileChooser extends FileChooser{
//##################################################################

ImageControl preview;
IImage lastImage;
Dimension lastSize;
Label info;

private void setup()
{
	try{
		Dimension d = getPreferredSize(null);
		d.width += 100;
		setPreferredSize(d.width,d.height);
		String msk = "*.png,*.bmp,*.gif,*.jpg,*.jpeg";
		msk+= " - Image Files";
		addMask(msk);	
		addMask(allFilesMask);
		extraPanel.setBorder(GuiStyle.inputEdge,3);
		extraPanel.addLast(new Label("Image Preview:"),HSTRETCH,DONTFILL|CENTER);
		extraPanel.addLast(info = new Label(" "),HSTRETCH,FILL);
		info.anchor = CENTER;
		extraPanel.addLast((preview = new ImageControl(null)).setPreferredSize(50,50));
		//preview.backGround = Color.Sand;
	}catch(RuntimeException e){
		e.printStackTrace();
		throw e;
	}
}
//===================================================================
public ImageFileChooser(FileChooserParameters fcp)
//===================================================================
{
	super(fcp,true);
	setup();
}
//===================================================================
public ImageFileChooser()
//===================================================================
{
	this(OPEN,baseDir);
}
//===================================================================
public ImageFileChooser(int type,String initial)
//===================================================================
{
	this(type,initial,Vm.newFileObject());
}
//===================================================================
public ImageFileChooser(int type,String initial,File aFileModel)
//===================================================================
{
	super(type|EXTRA_CONTROL,initial,aFileModel);
	setup();
}
//===================================================================
public void resizeTo(int width,int height)
//===================================================================
{
	super.resizeTo(width,height);
	setImage(lastImage,lastSize);
}
//-------------------------------------------------------------------
void setImage(IImage img,Dimension trueSize)
//-------------------------------------------------------------------
{
	IImage old = lastImage;
	lastImage = img;
	Dimension d = preview.getSize(null);
	if (img == null) {
		if (previewHandle == null){
			img = new DrawnIcon(DrawnIcon.CROSS,d.width,d.height,Color.Black);
			String txt = "No Preview";
			if (trueSize != null)
				txt += " ("+trueSize.width+" x "+trueSize.height+")";
			info.setText(txt);
		}else{
			img = null;
			info.setText("Creating...");
		}
	}else
		if (trueSize == null)
			info.setText(img.getWidth()+" x "+img.getHeight());
		else
			info.setText(trueSize.width+" x "+trueSize.height);
	lastSize = trueSize;
	if (old != null && old != img) old.free();
	preview.setImage(img);
}
//File previewed;

Handle previewHandle;
InputStream previewInput;
String previewName;

int num;
protected synchronized void stopPreview()
{
	if (previewHandle != null){
		//Vm.debug("Closing: "+previewName+" "+previewInput.hashCode());
		previewHandle.stop(0);
		try{
			previewInput.close();
		}catch(Exception e){}
		previewHandle = null;
		previewInput = null;
		previewName = null;
	}
}
//-------------------------------------------------------------------
protected synchronized void newFileSelected(final File f)
//-------------------------------------------------------------------
{
	if (f != null)
		if (f.isDirectory())
			return;
	//previewed = f;
	stopPreview();
	try{
		final InputStream in = f.toReadableStream();
		previewInput = in;
		previewName = f.getFileExt();
		//System.out.println("Get: "+previewName+" - "+previewInput.hashCode());
		final Dimension s = preview.getSize(null);
		final Dimension trueSize = new Dimension();
		previewHandle = new Task(){
			protected void doRun(){
				String name = f.getFileExt();
				IImage got = null;
				try{
					got = getImage(in,name,s.width,s.height,
							true,
							trueSize);
					synchronized(ImageFileChooser.this){
						if (previewHandle == this) {
							//if (got == null) new Exception("Got null: "+name).printStackTrace();
							stopPreview();
							setImage(got,trueSize);
						}else{
							//if (got == null) Vm.debug("Got null normal: "+name);
							if (got != null) got.free();
							throw new Exception();
						}
					}
				}catch(Exception e){
					//Vm.debug("Abandoned: "+name+" is: "+got);
					//e.printStackTrace();
					synchronized(ImageFileChooser.this){
						//System.out.println("Abandoned: "+name);
						if (previewHandle == this) stopPreview();
					}
				}
			}
		}.start();
		setImage(null,null);
	}catch(IOException e){
		setImage(null,null);
		return;
	}
}

/**
* This is the maximum size of an image to preview or get in bytes.
**/
//public static int imageSizeLimit = Vm.isMobile() ? 500000 : 5000000;

//===================================================================
public static IImage getImage(File f,boolean limitedSize)
//===================================================================
{
	return getImage(f,0,0,limitedSize,null);
}
public static IImage getImage(InputStream in, String imageName, int width, int height, boolean dontScaleOnDesktop,Dimension trueSize)
{
	try{
		FormattedDataSource fds = new FormattedDataSource();
		fds.set(in);
		fds.name = imageName;
		return ImageDecoder.decodeScaledPicture(fds, width, height, !((!Device.isMobile() && dontScaleOnDesktop) || width == 0 || height == 0), true, true, trueSize);
	}catch(Exception e){
		return null;
	}
}
//===================================================================
public static IImage getImage(File f,int width,int height, boolean dontScaleOnDesktop,Dimension trueSize)
//===================================================================
{
	try{
		return getImage(f.toReadableStream(),f.getFileExt(),width,height,dontScaleOnDesktop,trueSize);
	}catch(IOException e){
		return null;
	}
}
/**
 * This will execute an ImageFileChooser and return the chosen image.
 * @param title The title for the displayed ImageFileChooser
 * @param initial The intial directory - may be null.
 * @param dontScaleOnDesktop true to indicate that the image should be loaded at
 * full size on the desktop system.
 * @return The Image selected or null if cancelled.
 */
//===================================================================
public static IImage getImage(String title,String initial,boolean dontScaleOnDesktop)
//===================================================================
{
	ImageFileChooser ifc = new ImageFileChooser(OPEN,initial);
	ifc.title = title;
	if (ifc.execute() == IDCANCEL) return null;
	return getImage(ifc.getChosenFile(),dontScaleOnDesktop);
}
/**
* This can be called after the user has selected a file.
**/
//===================================================================
public IImage getChosenImage(boolean limitedSize)
//===================================================================
{
	return getImage(getChosenFile(),limitedSize);
}
/**
* This can be called after the user has selected a file or files.
**/
//===================================================================
public IImage[] getChosenImages(boolean limitedSize)
//===================================================================
{
	Vector v = new Vector();
	File [] all = getAllChosenFiles();
	for (int i = 0; i<all.length; i++){
		IImage im = getImage(all[i],limitedSize);
		if (im != null) v.add(im);
	}
	IImage[] ret = new IImage[v.size()];
	v.copyInto(ret);
	return ret;
}
//===================================================================
public static void main(String args[])
//===================================================================
{
	Application.startApplication(args);
	FileChooser fc = new ImageFileChooser(BROWSE|EXPLORER_TYPE|DIRECTORY_TREE,
			baseDir
			);
	fc.exitSystemOnClose = true;
	fc.show();
}
//##################################################################
}

//##################################################################

