package eve.ui.launcher;
import java.io.IOException;

import eve.fx.Color;
import eve.fx.IconAndText;
import eve.fx.Insets;
import eve.fx.gui.Cursor;
import eve.fx.gui.WindowConstants;
import eve.io.File;
import eve.io.FileChooserParameters;
import eve.sys.Device;
import eve.sys.IConsole;
import eve.sys.Registry;
import eve.sys.Type;
import eve.sys.Vm;
import eve.sys.VmServer;
import eve.ui.Application;
import eve.ui.ButtonPullDownMenu;
import eve.ui.CellPanel;
import eve.ui.Control;
import eve.ui.Form;
import eve.ui.Gui;
import eve.ui.IWebBrowser;
import eve.ui.IconMenuForm;
import eve.ui.Menu;
import eve.ui.MenuBar;
import eve.ui.MenuItem;
import eve.ui.MessageBox;
import eve.ui.PullDownMenu;
import eve.ui.ReportException;
import eve.ui.TabbedPanel;
import eve.ui.VMOptions;
import eve.ui.data.Editor;
import eve.ui.filechooser.FileChooser;
//import eve.ui.table.data.registry.RegistryViewForm;
import eve.util.mString;
//##################################################################
public class VMLauncher extends Editor{
//##################################################################

private static Type web = new Type("eve.ui.formatted.data.WebBrowser");
private static Type notepad = new Type("eve.ui.data.Notepad");

static void execServer(Control f)
{
	if (serverConsole != null)
		serverConsole.execute();
}
static void execBrowse(Control f)
{
	Form.showWait(f.getWindow());
	FileChooserParameters fcp = new FileChooserParameters();
	fcp.set(fcp.TYPE,fcp.TYPE_BROWSE);
	fcp.set(fcp.TITLE,"File System");
	Form fc = Gui.getBestFileChooser(fcp,File.getNewFile("\\"));
	fc.execute();
}
static boolean execOptions(Control f)
{
	try{
		if (f != null) Form.showWait(f.getWindow());
		VMOptions vo = VMOptions.getVMOptions();
		VMOptions tv = (VMOptions)vo.getCopy();
		Form ed = tv.getEditor(0);
		int ret = ed.execute(); 
		if (ret == IDCANCEL) return true;
		vo.copyFrom(tv);
		vo.save(f);
		vo.apply();
		return true;
	}catch(Throwable t){
		new ReportException(t).execute();
		return false;
	}
}
static void execAbout(Control f)
{
	Form.showWait(f.getWindow());
	int v = Vm.getVersion();
	String version = ""+(v/100)+".";
	v = v%100;
	if (v < 10) version += "0";
	version += v;
	version += Vm.getVersionPostFix();
	if (!new Type("eve.ui.filechooser.FileChooser").exists()) version = "Micro-"+version;
	
	//about += "<center><a href=http://www.ewesoft.com/mobile.htm>www.ewesoft.com</a></center><p>";
	if (Gui.hasPen && web.exists()){
		String about = "<head><title>About Eve VM</title></head><body><b><center>Eve Virtual Machine Version&nbsp;"+version+"<br>by&nbsp;Michael&nbsp;L&nbsp;Brereton, June&nbsp;2008</center></b><p>";
		IWebBrowser wb = (IWebBrowser)web.newInstance();
		((Form)wb).exec(f.isModal());
		wb.setHtml(about,"about:Ewe",null);
	}else{
		String about = "Eve Virtual Machine,\nVersion "+version+"\n\nby Michael L Brereton,\nJune 2008\n";
		about += "\nwww.ewesoft.com\n";
		MessageBox mb = new MessageBox("About Eve VM",about,MBOK);
		mb.execute();
		Form.cancelWait();
	}
}

//===================================================================
public VMLauncher()
//===================================================================
{
	this(null);
}

public AppLauncher launcher;

private void backgroundImage(Control c)
{
	c.backGround = Color.LighterBlue;
	c.setScaledBackgroundImage(loadImage("eve/eve64.png"),0.10);
	//c.setTiledBackgroundImage(loadImage("eve/evesmall.png"),.25);
}
static Form serverConsole;
//===================================================================
public VMLauncher(Form welcomeScreen)
//===================================================================
{
	IConsole ic = VmServer.getIConsole();
	if (ic instanceof Control){
		Form f = new Form();
		f.title = ic instanceof Form ? ((Form)ic).title : "Server Console";
		f.setPreferredSize(600,400);
		f.maximizeOnPDA();
		f.addLast((Control)ic);
		serverConsole = f;
		f.dismantleOnClose = false;
		//VmServer.setConsole(new PrintWriter(c.getWriter()));
	}
	resizable = !Device.isMobile();
	dontAddSIPFiller = true;
	boolean onTop = Gui.screenIs(Gui.WIDE_SCREEN);
	VMOptions vmoptions = VMOptions.getVMOptions();
	vmoptions.applySystemOptions();
	launcher = new AppLauncher(Gui.hasSoftKeys); 
	launcher.setPreferredSize(640,480);
	if (Gui.hasSoftKeys){
		inheritSoftKeys = true;
		IconMenuForm f = new IconMenuForm(launcher,0);
		f.cancelButtonName = "Exit";
		f.windowFlagsToSet |= WindowConstants.FLAG_MAXIMIZE_ON_PDA;
		f.title = "App Launcher";
		f.exitSystemOnClose = true;
		addLast(f).setTag(TAG_INSETS,new Insets(2,2,2,2));
		backgroundImage(this);
		//setSoftKeyBarFor(null,f.getSoftKeyBarFor(null));
		f.modifyAll(Transparent,0);
		return;
	}
	TabbedPanel mt = (TabbedPanel)addTabbedPanel(false)[0];
	mt.modifyCardOptions(mt.OPTION_AUTO_SCROLL,0);
	CellPanel cp = new CellPanel();
	backgroundImage(cp);
	cp.setPreferredSize(100,100);
	cp.addLast(launcher.getBestScroller()).setTag(cp.TAG_INSETS,new Insets(2,2,2,2));
	cp.modifyAll(Transparent,0,false);
	mt.addCard(cp,"Applications",null).iconize("eve/programsmall.png");
	//
	String initial = "\\";
	int type = FileChooser.BROWSE|FileChooser.LAUNCHER_TYPE;
	initial = Registry.getUserHomeDirectory();//vmoptions.launcherStart;
	if (vmoptions.launcherTree) type |= FileChooser.DIRECTORY_TREE;
	else type &= ~FileChooser.DIRECTORY_TREE;
	Type ty = new Type("eve.ui.filechooser.FileChooser");
	if (ty.exists())
		mt.addCard((Control)ty.newInstance("(ILjava/lang/String;)",new Object[]{new Integer(type),initial}),"File System",null).iconize("eve/ChooseFile.png");
	//
	//mt.modifyCardOptions(mt.OPTION_AUTO_SCROLL,0);
	mt.addCard(vmoptions.getEditor(0),"VM Options",null).iconize("eve/optionssmall.png");
	cp = mt.getExtraControls(true);
	String utils = "";
	//if (LaunchPanel.launcher != null) 
	utils = "Add Launch Item$a";
	if (serverConsole != null) utils += "|Server Console$c";
	if (notepad.exists()) utils += "|Text Editor$t";
	utils += "|Remote App Server";
	utils += "|Remote App Loader";
	utils += "|Remove Installation";
	if (web.exists()) utils += "|HTML Viewer$h";
	
	if (launcher.registryView.exists() && Registry.getLocalRegistry() != null) utils+= "|Registry View$r";
	if (!Device.isMobile()){
		utils+="|-|Run EveSync$s|EveSync Emulator$m";
	}
	//utils += "|Registry View$r";
	/*
	if (Reflection.forName("apps.eweconfig.EweConfig") != null) utils += "|EweSync Configuration$c";
	if (!Device.isMobile()) utils += "|-|Zaurus EweSync Server|EweSync Emulator";
	*/
	String [] extra = VMOptions.getExtraActions();
	if (extra.length != 0){
		utils += "|-";
		for (int i = 0; i<extra.length; i++)
			utils += "|"+extra[i];
	}
	utils += 	"|-|About Eve VM$e|Exit$x";
	Menu um = new Menu(mString.split(utils),"Tools");
	PullDownMenu pdm = new ButtonPullDownMenu("",um);
	pdm.arrowDirection = mt.tabLocation == mt.NORTH ? Down : Up;
	pdm.borderWidth = 0; pdm.borderStyle = BDR_NOBORDER;
	pdm.image = loadImage("eve/evesmall.png");
	pdm.setToolTip(new IconAndText(pdm.image,"Tools",null));
	MenuBar mb = new MenuBar();
	mb.addMenu(pdm);
	cp.addLast(mb);
	addField(mb,"tools");
}

public MenuItem tools;

//===================================================================
public void action(String field,Editor ed)
//===================================================================
{
	if (field.equals("Add Launch Item"))
		launcher.doAdd();
	else if (field.equals("Server Console")){
		execServer(this);
	}else if (field.equals("Remote App Server")){
		launcher.apps.runAppServer();
	}else if (field.equals("Remote App Loader")){
		launcher.loadAppAction();
	}else if (field.equals("Remove Installation")){
		new RemoveInstallation().execute();
	}
	else if (field.equals("About Eve VM")){
		execAbout(this);
	}else if (field.equals("Exit")){
		exit(0);
		//Vm.exit(0);
	}else if (field.startsWith("EweSync C")){
		try{
			Type t = new Type("apps.eweconfig.EweConfig");
			if (t.exists())
				((Form)t.newInstance()).execute();
		}catch(Throwable t){}
	}else if (field.equals("Text Editor") && notepad.exists()){
		Form rf = (Form)notepad.newInstance();
		rf.setPreferredSize(640,480);
		rf.execute();
	}else if (field.equals("HTML Viewer") && web.exists()){
		Form rf = (Form)web.newInstance();
		//rf.htmlProperties.set("maxImageSize",new ewe.fx.Dimension(16,16));
		//rf.htmlProperties.setBoolean("allowImages",false);
		//rf.htmlProperties.setBoolean("allowAnimatedImages",false);
		rf.setPreferredSize(640,480);
		rf.execute();
	}else if (field.equals("Registry View")){
		launcher.runRegistry();
		//new RegistryViewForm(Registry.getLocalRegistry()).execute();
	/*}else if (field.equals("Zaurus EweSync Server")){
		try{
			ewe.sys.Vm.execEwe(new String[]{"ewe.io.RemoteConnection","zaurus"},null);
		}catch(ewe.io.IOException e){
		}*/
	}else if (field.equals("EveSync Emulator")||field.equals("Run EveSync")){
		try{
			Device.getPathToVm();
			String cmd = field.equals("EveSync Emulator") ? "-emulator" : "-activesyncdesktop";
			String s = Registry.getVmCommandLine("eve.net.RemoteServer "+cmd,null);
			if (s != null) Vm.execCommandLine(s);
			//Vm.execCommandLine(new String[]{"ewe.io.RemoteConnection","emulator"},null);
		}catch(IOException e){
			new ReportException(e).execute();
		}
	
	}else
		VMOptions.doExtraAction(field,getFrame());
}
static boolean dontForceSoft = false;//true; 
public static void eveMain(String[] args)
{
	boolean serverCommand = args.length != 0 && args[0].equals("-servercommand");
	if (serverCommand){
		Vm.eveMainSetNoVisibleWindow();
	}
	if (!dontForceSoft)
		Gui.simulateSoftkeysOnPDA();
}
static{
	Application.mainTitle = "Eve VM Launcher";
}

static boolean unbuffered = false;

public static void main(String [] args)
{
	Application.startApplication(args);
	boolean serverCommand = args.length != 0 && args[0].equals("-servercommand");
	if (serverCommand){
		Application.mainApp.setCursor(Cursor.DEFAULT_CURSOR); 
		Application.mainApp.forceVisibleOnToFront = true;
	}
	/*
	new Thread(){
		public void run(){
			try{
				sleep(10000);
				Vm.exit(0);
			}catch(Exception e){}
		}
	}.start();
	*/
	try{
		//new Notepad().execute();
		//execOptions(null);
		//System.exit(0);
		//System.out.println("PD:"+File.getProgramDirectory());
		Form f = new VMLauncher();
		if (unbuffered) f.modify(DontBuffer,0);
		f.title = Application.mainTitle;
		f.exitSystemOnClose = true;
		f.execute();
		f.waitUntilClosed();
	}catch(Exception e){
		e.printStackTrace();
	}
}
//##################################################################
}
//##################################################################

