/*
 * Created on May 8, 2006
 *
 * Michael L Brereton - www.ewesoft.com
 * 
 * 
 */
package eve.ui.print;

import eve.fx.Brush;
import eve.fx.Color;
import eve.fx.Graphics;
import eve.fx.Point;
import eve.fx.Rect;
import eve.fx.points.PageFormat;
import eve.fx.points.PointDrawable;
import eve.fx.points.PointGraphics;
import eve.fx.points.PointRect;
import eve.sys.Countdown;
import eve.ui.Canvas;
import eve.ui.DragContext;
import eve.ui.event.PenEvent;

/**
 * @author Michael L Brereton
 *
 */
//####################################################
public class PointGraphicsViewer extends Canvas {
	
public double zoom = 1;
public PageFormat pageFormat = new PageFormat();
public PointDrawable drawable;
public double zoomStep = 0.5;

public void setDrawable(PointDrawable d, PageFormat pf, double useZoom)
{
	pageFormat = pf;
	drawable = d;
	if (useZoom == -1){
		if (width > 0 && height > 0) {
			fit();
			return;
		}
		else useZoom = 1;
	}
	zoom = useZoom;
	zoomChanged();
	repaintNow();
}
protected PointDrawable getDrawable()
{
	return drawable;
}
{
	modify(WantDrag,0);
	if (origin == null) origin = new Point();
}
private static final PointRect or = new PointRect();
public synchronized void doPaint(Graphics g, Rect area)
{
	g.setColor(getBackground());
	g.fillRect(area.x,area.y,area.width,area.height);
	Rect r = getCachedRect(area.x,area.y,area.width,area.height);
	double dpi = zoom*72;
	double sc = 1/zoom;
	if (pageFormat != null){ 
		int leftWidth = (int)((pageFormat.fullPageWidth-origin.x*sc)/sc);
		int leftHeight = (int)((pageFormat.fullPageHeight-origin.y*sc)/sc);
		r.set(0,0,leftWidth,leftHeight);
		r.getIntersection(area,r);
	}else{
		r.set(area);
	}
	try{
		if (r.width <= 0 || r.height <= 0) return;
		PointGraphics pg = new PointGraphics(g,width,height,dpi,dpi);
		pg.dontFreeGraphics = true;
		pg.setBrush(Color.White,Brush.SOLID);
		PointRect pr = PointRect.getCached();
		pg.startScreenUpdate(origin.x,origin.y);
		pg.getScreenUpdateRect(r.x,r.y,r.width,r.height,pr,origin.x,origin.y);
		pg.paintRect(pr.x,pr.y,pr.width,pr.height,true,false);
		try{
			PointDrawable pd = getDrawable();
			pd.draw(pg,pr);
		}finally{
			pr.cache();
		}
		pg.free();
	}catch(Exception e){
		e.printStackTrace();
	}finally{
		r.cache();
	}
}

private void zoomChanged()
{
	int i = modify(Invisible,0);
	virtualSize = new Rect(0,0,(int)(pageFormat.fullPageWidth*zoom),(int)(pageFormat.fullPageHeight*zoom));
	//virtualSize = new Rect(0,0,(int)(info.width*zoom),(int)(info.height*zoom));
	checkScrolls();
	restore(i,Invisible);
}
/*
public void setImage(ImageData im)
{
	setImage(im,0,true);
}
public void setImage(ImageData im, double zoom, boolean repaint)
{
	info = ImageDataInfo.toImageDataInfo(im);
	if (zoom <= 0 && width > 0){
		double xs = (double)width/info.width;
		double ys = (double)height/info.height;
		zoom = xs < ys ? xs : ys;
	}
	if (zoom <= 0) zoom = 1;
	this.zoom = zoom;
	source = im;
	int i = modify(Invisible,0);
	setOrigin(0,0,null);
	zoomChanged();
	restore(i,Invisible);
	if (repaint) repaintNow();
}
*/
public void setZoom(double newZoom)
{
	if (newZoom <= 0) newZoom = 1;
	zoom = newZoom;
	int i = modify(Invisible,0);
	zoomChanged();
	restore(i,Invisible);
	repaintNow();
}
public void fit()
{
	double xs = (double)width/pageFormat.fullPageWidth;
	double ys = (double)height/pageFormat.fullPageHeight;
	zoom = xs < ys ? xs : ys;
	zoomChanged();
	repaintNow();
}
public void zoomIn(double step)
{
	if (step < 0) step = 0.5;
	setZoom(zoom*(1+step));
}
public void zoomOut(double step)
{
	if (step < 0) step = 0.5;
	setZoom(zoom/(1+step));
}
public void onPenEvent(PenEvent ev)
{
	if (ev.type == PenEvent.SCROLL_UP){
		zoomIn(zoomStep);
	}else if (ev.type == PenEvent.SCROLL_DOWN){
		zoomOut(zoomStep);
	}else
		super.onPenEvent(ev);
}

public void startDragging(DragContext dc)
{
	oldOrigin.set(origin);
}
Point moved = new Point();
Point oldOrigin = new Point();
Rect screenRect = new Rect();

Countdown repaintTick = new Countdown();

public boolean scroll(int dx,int dy,Point moved)
{
	boolean ret = super.scroll(dx,dy,moved);
	if (ret) repaintTick.requestTick(this,200);
	return ret;
}
public void ticked(Object from, long elapsed)
{
	repaint();
}
public void dragged(DragContext dc)
{
	int dx = dc.curPoint.x-dc.prevPoint.x;
	int dy = dc.curPoint.y-dc.prevPoint.y;
	scroll(-dx,-dy,moved);
}

public void stopDragging(DragContext dc)
{
	
}
/*
protected void pointerEvent(SurfacePointerEvent ev)
{
	if (ev.type == ev.SCROLL_DOWN){
		zoom /= 2;
		repaintNow();
	}else if (ev.type == ev.SCROLL_UP){
		zoom *= 2;
		repaintNow();
	}else if (ev.type == ev.PEN_DOWN){
		dragging = null;
		penPressed(ev);
		if (dragging != null)
			startDragging(dragging);
	}else if (ev.type == ev.PEN_UP){
		if (dragging != null) {
			dragging.nowAt(ev.x,ev.y);	
			stopDragging(dragging);
			dragging = null;
		}
		else penReleased(ev);
	}else if (ev.type == ev.PEN_MOVE){
		if (dragging != null){
			if (!dragging.nowAt(ev.x,ev.y)) return;
			dragged(dragging);
		}else penMoved(ev);
	}
}
*/
}
//####################################################
