/*
 * Created on May 8, 2006
 *
 * Michael L Brereton - www.ewesoft.com
 * 
 * 
 */
package eve.ui.print;

import java.io.IOException;
import java.util.Vector;

import eve.fx.Color;
import eve.fx.Font;
import eve.fx.IImage;
import eve.fx.Insets;
import eve.fx.points.DataCell;
import eve.fx.points.FontManager;
import eve.fx.points.PageFormat;
import eve.fx.points.PointGraphics;
import eve.fx.points.PrintCellPanel;
import eve.fx.points.FontManager.ActiveFonts;
import eve.fx.print.PdfPrinter;
import eve.fx.print.PrintOptions;
import eve.fx.print.PrintPages;
import eve.fx.print.PrintPreview;
import eve.fx.print.Printable;
import eve.fx.print.PrintablePage;
import eve.fx.print.Printer;
import eve.fx.print.PrinterControl;
import eve.fx.print.PrinterJob;
import eve.fx.print.PrinterProperties;
import eve.fx.print.RandomPrintable;
import eve.io.File;
import eve.sys.ClassCounter;
import eve.sys.Handle;
import eve.sys.Registry;
import eve.sys.TimeOut;
import eve.ui.Application;
import eve.ui.Button;
import eve.ui.CellPanel;
import eve.ui.Frame;
import eve.ui.Gui;
import eve.ui.InputStack;
import eve.ui.ProgressAndControl;
import eve.ui.ProgressBarForm;
import eve.ui.ReportException;
import eve.ui.Window;
import eve.ui.data.Editor;
import eve.util.mVector;

/**
 * @author Michael L Brereton
 *
 */
//####################################################
public class PrintForm extends Editor {

	public Handle printingHandle;
	public Printable toPrint;
	public boolean dontExitOnPrint;
	public boolean dontExitAfterPrint;
	public File outputFileForPreview;
	public File tempDirForPreview;
	public boolean useNonNativeDialog = false;
	public PrinterProperties printerProperties = new PrinterProperties();
	protected CellPanel mainForm;
	protected ProgressAndControl progress;
	/**
	 * 
	 */

	public int previewDPI = 72;
	public int drawDPI = 72;//300;
	public String pageRange = "ALL";
	private CellPanel bb;
	protected Button addButton(String name, String action)
	{
		return addButton(name,action,(IImage)null);
	}
	protected Button addButton(String name, String action, IImage image)
	{
		Button b = new Button(name);
		if (image != null) Gui.iconize(b,image,true,null);
		if (bb == null){
			bb = new CellPanel();
			progress.controls.addNext(bb).setCell(HSTRETCH).setControl(DONTFILL|CENTER);	
		}
		bb.addNext(b).setTag(TAG_INSETS,new Insets(0,3,0,0));
		//if (buttons == null) buttons = new Vector();
		//buttons.add(b);
		addField(b,action);
		return b;
	}
	public PrintForm(String title)
	{
		this(title,0);
	}
	public static final int OPTION_NO_CANCEL = 0x1;
	public static final int OPTION_NO_PREVIEW = 0x2;
	public static final int OPTION_NO_PRINT = 0x4;
	public static final int OPTION_NO_PRINT_NOW = 0x8;
	public static final int OPTION_SHOW_PAGE_RANGE = 0x10;
	public static final int OPTION_EXTRA_PDF_PREVIEW = 0x20;
	
	public PrintForm(String title,int options)
	{
		this.title = title;
		addLast(mainForm = new CellPanel());
		InputStack is = new InputStack();
		is.columns = 2;
		if ((options & OPTION_SHOW_PAGE_RANGE) != 0){
			is.setText("Options");
			addField(is.addComboBox("Page Range:",new String[]{"ALL"},0),"pageRange");
			addLast(is).setCell(HSTRETCH);
		}
		/*
		is.setText("Preview");
		addField(is.addComboBox("Preview DPI",new String[]{"72","150","300","600"},0),"previewDPI");
		addField(is.addComboBox("Drawing DPI",new String[]{"72","150","300","600"},0),"drawDPI");
		addLast(is).setCell(HSTRETCH);
*/
		ProgressAndControl pc = progress = new ProgressAndControl();
		addLast(pc).setCell(HSTRETCH);
		if ((options & OPTION_NO_PRINT_NOW) == 0) addButton("Print Now","printNow");
		if ((options & OPTION_NO_PRINT) == 0) addButton("Print...","print");
		if ((options & OPTION_NO_PREVIEW) == 0) {
			Button b = addButton("Preview","preview");
		}
		if ((options & OPTION_EXTRA_PDF_PREVIEW) != 0) {
			Button b = addButton("PDF Preview","pdfPreview");
		}
		//Menu m = new Menu(new String[]{"Preview Without Saving","Save Preview Images"},"Preview Options");
		//b.setMenu(m);
		if ((options & OPTION_NO_CANCEL) == 0) cancel = addButton("Cancel","doCancel");
		//doButtons(CANCELB);
	}

	protected PageFormat getPageFormat()
	{
		return null;
	}
	protected PrintOptions getPrintOptions()
	{
		pageRange = pageRange.trim();
		if (pageRange.equalsIgnoreCase("ALL") || pageRange.length() == 0)
			return null;
		PrintOptions po = new PrintOptions();
		po.rangesFromString(pageRange);
		return po;
	}
	/**
	 * If you do not override getPrintable() then this method 
	 * is called. Override this method and call the addPage()
	 * method multiple times to add pages to the PrintForm.
	 */
	protected void addPages()
	{
	}
	/**
	 * You can call this multiple times within the addPages() method. It will create
	 * a PrintPages Object (if this is the first call to this method)
	 * and set it to be the Printable object for this PrintForm - i.e. the "toPrint"
	 * field. It will then
	 * add the specified PrintablePage to the PrintPages object.<p>
	 * Do not call this if you have set the "toPrint" field to be an Object that is
	 * not of type PrintPages, otherwise an IllegalStateException will be thrown. 
	 * @param p the PrintablePage to print.
	 * @throws IllegalStateException if the toPrint field is not null and is not
	 * an instance of PrintPages.
	 */
	protected void addPage(PrintablePage p) throws IllegalStateException
	{
		if (toPrint != null && !(toPrint instanceof PrintPages))
			throw new IllegalStateException();
		if (toPrint == null) toPrint = new PrintPages();
		((PrintPages)toPrint).addFullPage(p);
	}
	protected void doPrint(Printer job, boolean isPreview)
	{
		Printable p = getPrintable();
		if (p == null){
			return;
			//Gui.flashMessage("Nothing to print!",this);
			//printingHandle = new Handle(new IllegalArgumentException("Nothing to print"));
		}else
			printingHandle = job.print(p,getPageFormat(),getPrintOptions());
		nowPrinting(printingHandle,isPreview);
	}
	/**
	 * By default this will exit with value IDOK for a non-preview, and for 
	 * a preview it will display a progress bar and then display the preview form.
	 * @param printingHandle the handle to the printing task.
	 * @param isPreview true if it is a PrintPreview.
	 */
	protected void nowPrinting(Handle printingHandle, boolean isPreview)
	{
		if (!isPreview && !dontExitOnPrint) {
			exit(IDOK);
		}
		ProgressBarForm pbf = new ProgressBarForm();
		pbf.showMainTask = false;
		pbf.title = isPreview ? "Generating Preview..." : "Printing";
		pbf.setPreferredSize(200, -1);
		pbf.showStop = true;
		pbf.execute(printingHandle);
		if (printingHandle.error != null)
			printingHandle.error.printStackTrace();
		if (!isPreview){
			if (!dontExitAfterPrint) exit(IDOK);
			return; 
		}
		PreviewViewer f = new PreviewViewer();
		f.doButtons(OKB);
		f.title = "Print Preview";
		Object rv = printingHandle.returnValue;
		if (rv instanceof Vector){
			f.setImages((Vector)rv);
		}else if (rv instanceof File){
			f.setImages((File)rv);
		}else{
			//System.out.println(rv);
			f.setImages(new Vector());
		}
		f.execute();
	}
	protected boolean amPrinting = false;
	
	/**
	 * This is used to get the default PrinterJob.
	 * @return the PrinterJob for printing.
	 * @throws IOException on error.
	 */
	protected Printer getPrinterJob() throws IOException
	{
		return PrinterJob.getPrinterJob(null,printerProperties);
	}
	public void print_action(Editor ed)
	{
		boolean owned = true;
		synchronized(this){
			if (amPrinting) return;
			amPrinting = true;
		}
		try{
			if (useNonNativeDialog){
				final PrinterProperties pp = PrinterJob.getPrinterProperties(null);
				final Handle h = PrinterJob.printerDialog(pp);
				owned = false;
				new Thread(){
					public void run(){
						try{
							if (!h.waitOnSuccess(TimeOut.Forever)){
								//System.out.println("Print cancelled!");
								return;
							}
							try{
								PrinterControl pc = PrinterControl.getOrCreatePrinterControl(pp);
								printerProperties.setPrinterControl(pc);
								PrinterJob pj = PrinterJob.getPrinterJob(pp.getChosen(),printerProperties);
								doPrint(pj,false);
							}catch(IOException e){
								
							}
						}finally{
							synchronized(PrintForm.this){
								amPrinting = false;
							}
						}
					}
				}.start();
				return;
			}
			final Printer pj = getPrinterJob();
			final Handle h = pj.printDialog(ed.getWindow().getWindowSurface());
			owned = false;
			new Thread(){
				public void run(){
					try{
						if (!h.waitOnSuccess(TimeOut.Forever)){
							return;
						}
						doPrint(pj,false);
					}finally{
						synchronized(PrintForm.this){
							amPrinting = false;
						}
					}
				}
			}.start();
		}catch(Exception e){
			new ReportException(e).execute();
		}finally{
			if (owned) synchronized(this){
				amPrinting = false;
			}
		}
	}
	protected PrintPreview getPrintPreview()
	{
		PrintPreview pp = new PrintPreview();
		pp.previewToFile = outputFileForPreview != null || tempDirForPreview != null;
		pp.outputFile = outputFileForPreview;
		pp.tempDir = tempDirForPreview;
		if (previewDPI < 10) previewDPI = 72;
		pp.setImageDPI(previewDPI);
		pp.setDrawDPI(drawDPI);
		return pp;
	}
	
	ClassCounter cc;
	int checkPlease = 0;
	public void preview_action(Editor ed)
	{
		/*
		System.gc(); //FIXME
		if (cc == null) cc = new ClassCounter();
		else {
			String got = cc.getDifferencesToNow();
			cc.countNow();
			Vm.debug(got);
			if (checkPlease != 0){
				System.out.println("Track!-------------------------------");
				System.out.println(Vm.trackObject(checkPlease,0));
				//checkPlease = 0;
			}
		}
		*/
		synchronized(this){
			if (amPrinting) return;
			amPrinting = true;
		}
		new Thread(){
			public void run(){
				try{
					Printable p = getPrintable();
					if (p == null) return;
					p.adjustPrinterProperties(printerProperties);
					if (p instanceof RandomPrintable && outputFileForPreview == null && tempDirForPreview == null){
						PageFormat pf = getPageFormat();
						if (((RandomPrintable)p).countPages(pf) < 1) {
							Gui.flashMessage("No pages to preview.",PrintForm.this);
							return;
						}
						PrintPageViewer ppv = new PrintPageViewer()
						/*{
							protected void finalize(){
								Vm.debug("PPV Finalize");
							}
						}*/
						;
						ppv.title = "Print Preview";
						ppv.setPrintable((RandomPrintable)p,pf);
						if (true){
							ppv.exec();
							if (checkPlease == 0)
								checkPlease = 0;//Vm.getTrackKey(ppv.getWindow().getSurface());
							ppv.waitUntilClosed();
						}else{
							ppv.execute();
						}
						return;
					}
					PrintPreview pp = getPrintPreview();
					doPrint(pp,true);
					}catch(Exception e){
						e.printStackTrace();
					}finally{
						synchronized(PrintForm.this){
							amPrinting = false;
						}
					}
			}
		}.start();
	}
	public void printNow_action(Editor ed)
	{
		boolean owning = true;
		synchronized(this){
			if (amPrinting) return;
			amPrinting = true;
		}
		try{
			final Printer pj = getPrinterJob();
			owning = false;
			new Thread(){
				public void run(){
					try{
						doPrint(pj,false);
					}finally{
						synchronized(PrintForm.this){
							amPrinting = false;
						}
					}
				}
			}.start();
		}catch(Exception e){
			new ReportException(e).execute();
		}finally{
			if (owning) amPrinting = false;
		}
	}
	/**
	 * This method must return the Printable object to print.
	 * By default if the toPrint field is not null, then it will return
	 * that value. If toPrint is null then toPrint is set to a new
	 * PrintPages() object and the addPages() method is then called.
	 * Therefore if you do not override getPrintable() and you do not
	 * set the toPrint value, then you can override addPages() and
	 * add pages to add pages to the PrintPages.
	 * @return the Printable object that will be printed.
	 */
	protected Printable getPrintable()
	{
		if (toPrint == null){
			toPrint = new PrintPages();
			addPages();
		}
		return toPrint;
	}
	/**
	 * Show on the progress bar and wait for a task to complete on the PrintForm. While the task is operating
	 * the mainForm is disabled.
	 * @param h the Handle to the task.
	 * @param title a name for the task to be displayed in the progress bar.
	 * @return true if the task ended with the Success bit set, false if not.
	 */
	public boolean doTask(Handle h, String title)
	{
		progress.doTask(h,title,mainForm).waitUntilCompletion();
		if ((h.check() & h.Success) == 0) return false;
		return true;
	}
	/*
	public static void main(String[] args)
	{
		Application.startApplication(args);
		new PrintForm().execute();
		System.exit(0);
	}
	*/
	/**
	 * This executes a PrintPageViewer in a new window.
	 * @param title the title to display.
	 * @param printable the RandomPrintable to preview.
	 * @param pageFormat the PageFormat to use - this may be null, in which
	 * case a default PageFormat is used.
	 * @return the return value from the execute() method.
	 */
	public static int preview(String title, RandomPrintable p,PageFormat pf)
	{
		return PrintPageViewer.preview(title,p,pf);
	}
	/**
	 * This executes a PrintPageViewer in a new window.
	 * @param title the title to display.
	 * @param printable a single PrintablePage to view.
	 * @param widthInInches the page width in inches.
	 * @param heightInInches the page height in inches.
	 * @return the return value from the execute() method.
	 */
	public static int preview(String title, PrintablePage printable, double widthInInches, double heightInInches)
	{
		PrintPages pp = new PrintPages();
		pp.addFullPage(printable);
		return preview(title,pp,PageFormat.getForSize(widthInInches*72,heightInInches*72));
	}
	public void pdfPreview_action(Editor ed)
	{
		synchronized(this){
			if (amPrinting) return;
			amPrinting = true;
		}
		new Thread(){
			public void run(){
				try{
					Printable p = getPrintable();
					if (p == null) return;
					PdfPrinter pj = new PdfPrinter();
					PageFormat pf = getPageFormat();
					File output = File.getNewFile().createTempFile("Preview",".pdf", null);
					pj.outputFile = output;
					Handle h = pj.print(p,pf,null);
					ProgressBarForm pbf = new ProgressBarForm();
					pbf.showMainTask = false;
					pbf.title = "Generating Preview...";
					pbf.setPreferredSize(200, -1);
					pbf.showStop = true;
					pbf.execute(h);
					//h.waitUntilCompletion();
					if (!h.succeeded()) h.throwRuntimeError("Printing failed");
					else{
						Registry.executeShellCommand(output.getAbsolutePath());
					}
						return;
					}catch(Exception e){
						e.printStackTrace();
					}finally{
						synchronized(PrintForm.this){
							amPrinting = false;
						}
					}
			}
		}.start();
	}
	public static void pdfPreview(File output, String title, PrintablePage printable, double widthInInches, double heightInInches)
	throws IOException
	{
		PrintPages pp = new PrintPages();
		pp.addFullPage(printable);
		PageFormat pf = PageFormat.getForSize(widthInInches*72,heightInInches*72);
		PdfPrinter pj = new PdfPrinter();
		if (output == null) output = File.getNewFile().createTempFile("Preview",".pdf", null);
		pj.outputFile = output;
		Handle h = pj.print(pp,pf,null);
		h.waitUntilCompletion();
		if (!h.succeeded()) h.throwRuntimeError("Printing failed");
		else{
			Registry.executeShellCommand(output.getAbsolutePath());
		}
	}
	public static Handle print(Frame parent,Printable toPrint, PageFormat pf, boolean showDialog, boolean showProgress)
	{
		try{
			Printer pj = PrinterJob.getPrinterJob(null);
			Window w = parent == null ? Application.mainApp : parent.getWindow();
			if (showDialog){
				Handle h = pj.printDialog(w.getWindowSurface());
				if (!h.waitOnSuccess(TimeOut.Forever)) return new Handle(Handle.Aborted,null);
			}
			Handle h = pj.print(toPrint,pf,null);
			if (showProgress)
				ProgressBarForm.execute("Printing...",h);
			return h;
		}catch(IOException e){
			return new Handle(e);
		}
	}
	public static Handle print(PrintablePage printable, double widthInInches, double heightInInches)
	{
		PrintPages pp = new PrintPages();
		pp.addFullPage(printable);
		return print(null,pp,PageFormat.getForSize(widthInInches*72,heightInInches*72),true,true);
	}
	
	public static void main(String[] args) throws IOException
	{
		Application.startApplication(args);
		Vector names = new Vector(), des = new Vector();
		mVector.addAll(names, PrinterJob.getPrinterNames(des));
		System.out.println(names+" - "+des);
		StringBuffer def = new StringBuffer();
		String defName = PrinterJob.getDefaultPrinter(def);
		System.out.println(defName+" - "+def);
		PrintForm pf = new PrintForm("Printer Test",0);
		pf.useNonNativeDialog = true;
		PrintPages pt = new PrintPages();
		pf.toPrint = pt;
		final FontManager fm = new FontManager(); 
		PrinterJob.globalFontManager = fm;
		fm.addTimes("serif");
		pt.addFullPage(new PrintablePage(){

			protected void composePage(PointGraphics pg, PageFormat format,
					int pageNumber) {
				ActiveFonts af = fm.assignTo(pg);
				PrintCellPanel pcp = addMarginedPanel(72,72,72,72);
				pcp.font = af.get("serif", Font.PLAIN,40);
				PrintCellPanel p = new PrintCellPanel();
				DataCell dc = new DataCell("Hello there!");
				dc.setSpacing(72/16);
				dc.backGround = Color.LightBlue;
				dc.borderRect = dc.RECT_SIDE_FULL;
				dc.anchor = CENTER;
				p.addLast(dc).setCell(DONTSTRETCH);
				pcp.addLast(p);
			}
			
		});
		pf.dontExitAfterPrint = pf.dontExitOnPrint = true;
		pf.execute();
		Application.exit(0);
	}	
}

//####################################################
