package eve.ui.table;
import java.util.Vector;

import eve.data.TreeNode;
import eve.fx.IImage;

//##################################################################
public abstract class TreeModelAdapter extends TreeTableModel{
//##################################################################

//===================================================================
public TreeModelAdapter()
//===================================================================
{
	init((long)-1 << 32,(byte)(IsNode|CanExpand));
}
//-------------------------------------------------------------------
protected Vector objects = new Vector();
//-------------------------------------------------------------------
//-------------------------------------------------------------------
protected int newNodeObject(Object node)
//-------------------------------------------------------------------
{
	objects.add(node);
	return objects.size()-1;
}
//-------------------------------------------------------------------
protected Object getParentObject(int whichLine)
//-------------------------------------------------------------------
{
	long ref = objectAt(whichLine);
	int oi = (int)(ref >> 32);
	if (oi < 0) return rootObject;
	else return objects.get(oi);
}
//-------------------------------------------------------------------
protected int getIndexFor(int whichLine)
//-------------------------------------------------------------------
{
	return (int)objectAt(whichLine) & 0xffffffff;
}

//===================================================================
public Object getDataObjectAt(int whichLine)
//===================================================================
{
	return createObjectFor(getParentObject(whichLine),getIndexFor(whichLine));
}
//===================================================================
public int doExpand(int whichLine)
//===================================================================
{
	if (!canDoExpand(whichLine)) return 0;
	Object newNode = createObjectFor(getParentObject(whichLine),getIndexFor(whichLine));
	if (newNode == null) return 0;
	long parentIndex = (long)newNodeObject(newNode) << 32;
	long [] refs = getChildIndexes(newNode);
	if (refs == null) refs = new long[0];
	byte [] flags = new byte[refs.length];
	for (int i = 0; i<refs.length; i++) {
		refs[i] |= parentIndex;
		flags[i] = CanExpand|IsNode;
	}
	adjustFlags(newNode,refs,flags);
	return expand(whichLine,refs,flags);
}
/**
 * This calls and returns createObject(null,0).
 * @return the root object for the model.
 */
public Object getRoot()
{
	return createObjectFor(null,0);
}
public Object getObjectAtAddress(int [] address)
{
	Object obj = getRoot();
	for (int i = 0; i<address.length; i++){
		long[] got = getChildIndexes(obj);
		int idx = address[i];
		if (idx < 0 || idx >= got.length) return null;
		obj = createObjectFor(obj,(int)(got[idx] & 0xffffffff));
		if (obj == null) return null;
	}
	return obj;
}
/**
* Use this to create a new node object for a node which is about to be
* expanded.
**/
//-------------------------------------------------------------------
protected abstract Object createObjectFor(Object parent,int childIndex);
//-------------------------------------------------------------------
/**
* This is used to get an array of indexes for each of the children of the parent
* Object. By default this will call getChildCount() and then create an incremental
* list of indexes from 0 to the count-1.
* <p>
* These indexes are to be used by you to reference that child relative to the parent.
**/
//-------------------------------------------------------------------
protected long [] getChildIndexes(Object parent)
//-------------------------------------------------------------------
{
	int size = getChildCount(parent);
	long [] refs = new long[size];
	for (int i = 0; i<size; i++) refs[i] = i;
	return refs;
}
/**
* If you do not override getChildIndexes() you must implement this.
**/
//-------------------------------------------------------------------
protected int getChildCount(Object parent)
//-------------------------------------------------------------------
{
	return 0;
}
/**
* Use this to change the flags for the child indexes of the specified parent.
* The flags are created with each child being set as CanExpand and IsNode. By default
* this will leave the entries as they are.
**/
//-------------------------------------------------------------------
protected void adjustFlags(Object parent,long [] indexes,byte [] flags){}
//-------------------------------------------------------------------

//===================================================================
public String getDisplayString(int forLine,TreeNode node)
//===================================================================
{
	return getDisplayString(getParentObject(forLine),getIndexFor(forLine));
}
//===================================================================
public IImage getIcon(int forLine,TreeNode node)
//===================================================================
{
	return getIcon(getParentObject(forLine),getIndexFor(forLine));
}
//===================================================================
public byte getFlags(int forLine,TreeNode node)
//===================================================================
{
	return getFlags(getParentObject(forLine),getIndexFor(forLine),(byte)flagsAt(forLine));
}
/**
* Override this to get the display string for the child of the parent.
**/
//-------------------------------------------------------------------
protected String getDisplayString(Object parentNode,int childIndex)
//-------------------------------------------------------------------
{
	if (parentNode == null) return "Tree";
	return parentNode.toString()+"("+childIndex+")";	
}
/**
* Override this to get the icon for the child of the parent.
**/
//-------------------------------------------------------------------
protected IImage getIcon(Object parentNode,int childIndex)
//-------------------------------------------------------------------
{
	return null;
}
/**
* This is only called if "dynamicCanExpand" is set true. This will then be called each time
* the node is to be displayed.<p>
* Override this to return the flags (IsNode and/or CanExpand) for the child of the parent.
**/
//-------------------------------------------------------------------
protected byte getFlags(Object parentNode,int childIndex,byte savedFlags)
//-------------------------------------------------------------------
{
	return savedFlags;
}


//##################################################################
}
//##################################################################

