package eve.util;
/**
* This is a utility for storing an expandable array of integers. Similar
* to a Vector but for int values instead.
**/
//##################################################################
public class IntArray implements Copyable, Encodable{
//##################################################################
/**
* This is the array of integers itself. This can be changed while the
* array is being used.
**/
public int [] data; //Do not move this! It should be the first variable.
/**
* This specifies the number of valid values in the array. 
**/
public int length = 0;// Do not move this! It should be the second variable.
/**
* This specifies the number of values to increase by when expansion is
* needed. If it is zero or negative then everytime an increase is needed
* the size of the data will be doubled (this is the default).
**/
public int growSize;// Do not move this! It should be the third variable.

//===================================================================
public Object getCopy()
//===================================================================
{
	IntArray ia = new IntArray();
	ia.length = length;
	ia.data = new int[length];
	if (length != 0) System.arraycopy(data,0,ia.data,0,length);
	ia.growSize = growSize;
	return ia;
}

/**
* Creates an IntArray with an initial size of 100 and a grow size of 100.
**/
//===================================================================
public IntArray(){this(100,-1);}
//===================================================================
/**
* Creates an IntArray with the specified initial size and grow size.
**/
//===================================================================
public IntArray(int initialSize,int growSize)
//===================================================================
{
	this.growSize = growSize;
	data = new int[initialSize];
}
/**
* Clears the IntArray completely.
**/
//===================================================================
public void clear() {length = 0;}
//===================================================================
/**
* Append a value to the end of the array.
**/
//===================================================================
public void add(int value)
//===================================================================
{
	if (length == data.length){
		int growSize = this.growSize;
		if (growSize <= 0) growSize = length+1;
		if (growSize > 100000) growSize = 100000;
		int [] dt = new int[length+growSize];
		System.arraycopy(data,0,dt,0,length);
		data = dt;
	}
	data[length++] = value;
}
//===================================================================
public void append(int value)
//===================================================================
{
	add(value);
}
/**
* This adds space to the IntArray at the specified index, increasing the 
* length value by num. After this is
* done you can safely write into the data starting at index "where" for
* a length of num.
**/
//===================================================================
public boolean makeSpace(int where,int num)
//===================================================================
{
	if (num <= 0) return true;
	if (length+num > data.length) {
		int toAdd = num-(data.length-length);
		int growSize = this.growSize;
		if (growSize <= 0) growSize = data.length+1;
		if (growSize > 100000) growSize = 100000;
		if (toAdd < growSize) toAdd = growSize;
		int [] dt = new int[data.length+toAdd];
		System.arraycopy(data,0,dt,0,length);
		data = dt;
	}
	if (where < length)
		System.arraycopy(data,where,data,where+num,length-where);
	length += num;
	return true;
}

//===================================================================
public void insert(int[] values,int offset,int length,int where)
//===================================================================
{
	int myLength = this.length;
	try{
		makeSpace(where,length);
		System.arraycopy(values,offset,data,where,length);
	}catch(RuntimeException e){
		eve.sys.Vm.debug(where+", "+values.length+", "+offset+", "+length);
		eve.sys.Vm.debug(myLength+", "+data.length+", "+where+", "+this.length);
		throw e;
	}
}
//===================================================================
public void append(int[] values,int offset,int length)
//===================================================================
{
	insert(values,offset,length,this.length);
}

/**
* Insert a value at the specified index.
**/
//===================================================================
public void insert(int value,int index)
//===================================================================
{
	int toAdd = 0;
	if (index <= length) toAdd = 1;
	else toAdd = (index-length)+1;
	if (length+toAdd > data.length) {
		int growSize = this.growSize;
		if (growSize <= 0) growSize = data.length+1;
		if (growSize > 100000) growSize = 100000;
		if (toAdd < growSize) toAdd = growSize;
		int [] dt = new int[length+toAdd];
		System.arraycopy(data,0,dt,0,length);
		data = dt;
	}
	if (index < length) System.arraycopy(data,index,data,index+1,length-index);
	data[index] = value;
	length++;
}
static boolean hasNative = true;
native int nativeIndexOf(int value);
/**
* Find the index of the specified value. Returns -1 if it could not
* be found.
**/
//===================================================================
public int indexOf(int value)
//===================================================================
{
	if (hasNative) try{
		return nativeIndexOf(value);
	}catch(SecurityException e){
		hasNative = false;
	}catch(UnsatisfiedLinkError e){
		hasNative = false;
	}
	for (int i = 0; i<length; i++)
		if (data[i] == value) return i;
	return -1;
}
/**
* Remove the value if it is in the array.
**/
//===================================================================
public void remove(int value)
//===================================================================
{
	while(true){
		int i = indexOf(value);
		if (i == -1) return;
		removeAtIndex(i);
	}
}

/**
 * Remove the value at the specified index.
 */
//===================================================================
public void removeAtIndex(int index)
//===================================================================
{
	if (index == -1) return;
	System.arraycopy(data,index+1,data,index,length-index-1);
	length--;
}
/**
* Return a copy of the array of integers whose length exactly holds
* all added integers.
**/
//===================================================================
public int [] toIntArray()
//===================================================================
{
	int [] ret = new int[length];
	System.arraycopy(data,0,ret,0,length);
	return ret;
}
/**
* Copy all values into the destination array at the specified offset.
**/
//===================================================================
public void copyInto(int [] dest,int offset)
//===================================================================
{
	System.arraycopy(data,0,dest,offset,length);
}

/**
* Append all values in this IntArray to the specified destination array. 
* If the destination array is null it is treated as an array of zero
* length. A new int array is returned which holds the new array of integers.
**/
//===================================================================
public int [] appendTo(int [] dest) {return appendTo(dest,false);}
//===================================================================
/**
* Append all values in this IntArray to the specified destination array with
* the option of reversing the order of integers in this IntArray before appending.
* If the destination array is null it is treated as an array of zero
* length. A new int array is returned which holds the new array of integers.
**/
//===================================================================
public int [] appendTo(int [] dest,boolean reverse)
//===================================================================
{
	if (dest == null) dest = new int[0];
	int dl = dest.length;
	int [] ret = new int[dl+length];
	System.arraycopy(dest,0,ret,0,dl);
	if (reverse)
		for (int i = length-1; i>=0; i--) ret[dl++] = data[i];
	else System.arraycopy(data,0,ret,dl,length);
	return ret;
}
/**
 * Make enough space to receive incoming data up to a maximum of toReceive bytes.
 * This does not affect the length value, it just ensures that after it is called you
 * can safely write starting at data[length] for a length of toReceive bytes.
 *
 * After calling this and receiving the data, call received(int received) to indicate
 * how many bytes were actually received and placed into the buffer.
 *
 * @param toReceive the maximum number of bytes to receive.
 * @param append if this is false the length is set to zero.
 * @return The current length.
 */
//===================================================================
public int prepareReceive(int toReceive,boolean append)
//===================================================================
{
	if (!append) setLength(0);
	ensureCapacity(length+toReceive);
	return length;
}
/**
 * Call this after calling prepareReceive() and after having received some or zero data.
 * It updates the number of bytes in the array.
 * @param received the number of bytes actually received.
 */
//===================================================================
public void received(int received)
//===================================================================
{
	if (received > 0) length += received;
}
/**
* This sets the length of the valid data within the array and will either truncate or extend
* the array as necessary.
**/
//===================================================================
public void setLength(int len) 
//===================================================================
{
	if (len < 0) throw new IllegalArgumentException();
	if (len <= length) length = len;
	else if (len > length){
		ensureCapacity(len);
		Utils.zeroArrayRegion(data,length,len-length);
		length = len;
	}
}
/**
* Moves the data into a new data array of the specified length - which
* must be greater than or equal to the number of elements in the array.
**/
//-------------------------------------------------------------------
protected void toNewLength(int newLength)
//-------------------------------------------------------------------
{	
	int[] nd = new int[newLength];
	if (length != 0) System.arraycopy(data,0,nd,0,length);
	data = nd;
}
/**
* This makes sure the data array has at least totalBytes in it. It does
* not affect the data within it or the length value.
**/
//===================================================================
public void ensureCapacity(int newCapacity)
//===================================================================
{
	if (data.length >= newCapacity) return;
	if (newCapacity < (data.length*2+2)) newCapacity = data.length*2+2;
	toNewLength(newCapacity);
}
/**
 * Put an array of integers as a printable string of values separated
 * by commas. No other characters are placed in the StringBuffer.
 * @param data
 * @param start
 * @param len
 * @param dest
 */
public static void toString(int[] data, int start, int len, StringBuffer dest)
{
	for (int i = 0; i<len; i++){
		if (i != 0) dest.append(',');
		dest.append(data[start+i]);
	}
}
/**
 * Put an array of integers as a printable string of values separated
 * by commas and enclosed in []. 
 * @param data
 * @param start
 * @param len
 */
public static String toString(int[] data, int start, int len)
{
	StringBuffer sb = new StringBuffer();
	sb.append('[');
	if (data != null)
		toString(data,0,len,sb);
	sb.append(']');
	return sb.toString();
}
public String toString()
{
	return toString(data,0,length);
}
//##################################################################
}
//##################################################################

