package eve.util;
/**
 * A StringCollator is used to compare, order and search Strings.
 * 
 * @author Mike
 *
 */
public interface StringCollator {

	/**
	 * An option for any compare() operation.
	 */
	public static final int OPTION_IGNORE_CASE = 0x1;
	/**
	 * An option for the compare() method that compares two
	 * character arrays. It indicates that the comparison should
	 * return zero if the two parameter starts with the one parameter.
	 */
	public static final int OPTION_STARTS_WITH = 0x2;
	/**
	 * Compare two arrays of characters - or check if the second string
	 * starts with the first one.
	 * @param one the first array of characters.
	 * @param oneOffset the start of the characters in the first array.
	 * @param oneLength the number of characters in the first array.
	 * @param two the second array of characters.
	 * @param twoOffset the start of the characters in the second array.
	 * @param twoLength the number of characters in the second array.
	 * @param options any of the OPTION_XXX values OR'ed together.
	 * @return a positive value if one compares to be greater than two,
	 * a negative value if one compares to be less than two and 0 if one
	 * is considered equal to two OR if two starts with one and the OPTION_STARTS_WITH
	 * option is set.
	 */
	public int compare(char[] one, int oneOffset, int oneLength,char[] two, int twoOffset, int twoLength, int options);
	/**
	 * Compare two characters.
	 * @param one the first character.
	 * @param two the second character.
	 * @param options any of the OPTION_XXX values OR'ed together.
	 * @return 0 if the characters are considered equal, a positive value
	 * if one is greater than two, a negative value if two is greater than one.
	 */
	public int compare(char one, char two, int options);
	
}
