/*
 * Created on Apr 1, 2006
 *
 * Michael L Brereton - www.ewesoft.com
 * 
 */
package eve.util;

import java.util.NoSuchElementException;
import java.util.Vector;

/**
 * @author Michael L Brereton
 * This is a <b>simple re-usable</b> Iterator/Enumerator for iterating through the
 * elements of a Vector in a way that is safe (will not generate null or index exceptions)
 * but is not guaranteed to return consistent values if the Vector changes while
 * being iterated.
 */
//####################################################
public class VectorIterator extends IteratorEnumerator {
	private int nextIndex;
	private Object nextObject;
	private Vector vector;
	/**
	 * 
	 */
	public VectorIterator() {
	}
	public VectorIterator(Vector v){
		set(v);
	}
	private void zero()
	{
		nextIndex = -1;
		nextObject = null;
		vector = null;
	}
	
	public VectorIterator set(Vector v)
	{
		zero();
		vector = v;
		if (v != null) synchronized(v){
			int sz = v.size();
			if (sz == 0) v = null;
			else {
				nextObject = v.get(0);
				nextIndex = 1;
			}
		}
		return this;
	}

	public boolean hasNext()
	{
		return nextIndex != -1;
	}
	public Object next()
	{
		if (nextIndex == -1) throw new NoSuchElementException();
		Object toGo = nextObject;
		synchronized(vector){
			int sz = vector.size();
			if (sz <= nextIndex) zero();
			else {
				nextObject = vector.get(nextIndex);
				nextIndex++;
			}	
		}
		return toGo;
	}
}

//####################################################
