package eve.util;
import eve.sys.Locale;
import eve.sys.LocaleFormat;
import eve.sys.Vm;

import java.text.ParseException;
import java.util.Vector;
/**
* This provides many additional utility String manipulation routines.
**/
//##################################################################
/**
 * @author Michael L Brereton
 *
 */
public class mString {
//##################################################################
/**
* Splits up a String using the specified separator and add the substrings
* to the destination Vector after clearing the Vector. 
* It returns the destination vector and will
* create a new one if the provided one is null.
**/
//===================================================================
public static java.util.Vector split(String what,char separator,java.util.Vector dest)
//===================================================================
{
	return SubString.split(what,separator,dest);
/*	
	if (dest == null) dest = new Vector();
	dest.clear();
	if (what == null) return dest;
	if (what.length() == 0) return dest;
	StringBuffer sb = new StringBuffer();
	for (int i = 0;i<what.length(); i++){
		char c = what.charAt(i);
		if (c == separator) {
			dest.add(sb.toString());
			sb = new StringBuffer();
		}else{
			sb.append(c);
		}
	}
	String s = sb.toString();
	dest.add(s);
	return dest;
*/
}
/**
* Splits up a String using the specified separator and returns an
* array of sub Strings.
**/
//===================================================================
public static String [] split(String what,char separator)
//===================================================================
{
	StringList sl = StringList.getCached(SubString.class);
	sl.zero();
	sl.split(what,separator);
	String[] ret = sl.toStringArray();
	sl.cache();
	return ret;
}

/**
 * Possibly split a String into a set of sub-strings if one of a set of separators
 * is used.
 * @param what the String to possibly separate.
 * @param possibleSeparators the list of possible separator characters.
 * @return a String[] if a separator was present, null if no separator was present.
 */
public static String[] checkSplit(String what, String possibleSeparators)
{
	int len = possibleSeparators.length();
	for (int i = 0; i<len; i++){
		char ch = possibleSeparators.charAt(i);
		if (what.indexOf(ch) != -1)
			return split(what,ch);
	}
	return null;
}
/**
* This splits a String using '|' as the separator.
**/
//===================================================================
public static String [] split(String what)
//===================================================================
{
	return split(what,'|');
}
//===================================================================
public static void toBytes(String what,byte [] dest,int offset) {toBytes(what,dest,offset,false);}
//===================================================================
public static void toBytes(String what,byte [] dest,int offset,boolean ascii)
//===================================================================
{
	if (what == null) return;
	int l = what.length();
	char [] chars = what.toCharArray();
	for (int i = 0; i<l; i++) {
		char ch = chars[i];
		if (ascii) dest[i+offset] = (byte)(ch & 0xff);
		else{
			dest[i*2+offset+1] = (byte)(ch & 0xff);
			dest[i*2+offset] = (byte)((ch >> 8)&0xff);
		}
	}
}
//===================================================================
public static byte [] toAscii(String what)
//===================================================================
{
	if (what == null) return new byte[0];
	byte [] dest = new byte[what.length()];
	toBytes(what,dest,0,true);
	return dest;
}
//===================================================================
public static String fromAscii(byte [] bytes,int start,int length)
//===================================================================
{
	if (length <= 0) return "";
	char [] c = new char[length];
	for (int i = 0; i<length; i++)
		c[i] = (char)((int)bytes[start+i] & 0xff);
	return new String(c);
}
//===================================================================
public static byte [] toBytes(String what)
//===================================================================
{
	if (what == null) return new byte[0];
	byte [] dest = new byte[what.length()*2];
	toBytes(what,dest,0);
	return dest;
}
//===================================================================
public static String fromBytes(byte [] bytes,int offset,int stringLength)
//===================================================================
{
	char [] ch = new char[stringLength];
	for (int i = 0; i<stringLength; i++){
		char c = (char)bytes[i*2+offset+1];
		c |= (char)(bytes[i*2+offset]<<8)&0xff;
		ch[i] = c;
	}
	return new String(ch);
}
//===================================================================
public static String fromBytes(byte [] bytes) {return fromBytes(bytes,0,bytes.length/2);}
//===================================================================
public static CharArray toString(Object what, CharArray dest)
{
	if (dest == null) dest = new CharArray();
	if (what instanceof Stringable)
		return ((Stringable)what).toString(dest);
	return dest.append(toString(what));
}
public static String toString(Object what)
//===================================================================
{
	if (what == null) return "";
	else if (what instanceof String) return (String)what;
	else if (what instanceof Object[]) {
		//mVector v = new mVector();  < Probably should remove these two lines.
		//v.addElement((Object[])what);
		return new mVector((Object[])what).toString();
	}
	else {
		Class c = what.getClass();
		if (c.isArray()){
			return c.getComponentType()+"["+
			java.lang.reflect.Array.getLength(what)
			+"]";
		}else return what.toString();
	}
}
public static final int Uppercase = 1, Lowercase = 2;
//==================================================================
public static void changeCase(char [] chars,int toWhat)
//==================================================================
{
	int l = chars.length;
	for (int i = 0; i<l; i++) {
		char ch = chars[i];
		if (toWhat == Uppercase) 
			if (ch >= 'a' && ch <= 'z') chars[i] = (char)(ch+'A'-'a');
		if (toWhat == Lowercase)
			if (ch >= 'A' && ch <= 'Z') chars[i] = (char)(ch+'a'-'A');
	}
}
//==================================================================
public static String changeCase(String what,int toWhat)
//==================================================================
{
	if (what == null) return null;
	char [] chars = what.toCharArray();
	changeCase(chars,toWhat);
	return new String(chars);
}
//==================================================================
public static int compare(String one,String two) {return compare(one,two,false);}
public static int compare(String one,String two,boolean ignoreCase)
//==================================================================
{
	if (one == null)
		if (two == null) return 0;
		else return -1;
	else
		if (two == null) return 1;
	
	char [] ones = one.toCharArray();
	char [] twos = two.toCharArray();
	
	if (ignoreCase) {
		changeCase(ones,Uppercase);
		changeCase(twos,Uppercase);
	}
	int len = ones.length, len2 = twos.length;
	for (int i = 0; i<len; i++) {
		if (i >= len2) return 1;
		int diff = ones[i]-twos[i];
		if (diff != 0) return diff;
	}
	if (len < len2) return -1;
	else if (len > len2) return 1;
	else return 0;
}
//===================================================================
public static int indexOf(String s,char what)
//===================================================================
{
	if (s == null) return -1;
	int w = (int)what;
	for (int i = 0; i<s.length(); i++) 
		if (s.charAt(i) == w) return i;
	return -1;
}
/**
* This returns the string to the left of the specified character (not including the character). 
* If the character is not found then it will return the entire string.
**/
//===================================================================
public static String leftOf(String s,char c)
//===================================================================
{
	int idx = s.lastIndexOf(c);
	if (idx == -1) return s;
	else return s.substring(0,idx);
}
/**
* This returns the string to the right of the specified character (not including the character). 
* If the character is not found then it will return an empty string.
**/
//===================================================================
public static String rightOf(String s,char c)
//===================================================================
{
	int idx = s.lastIndexOf(c);
	if (idx == -1) return "";
	else return s.substring(idx+1);
}
/**
* This removes a trailing "/" or "\" character.
**/
//===================================================================
public static String removeTrailingSlash(String s)
//===================================================================
{
	if (s == null) return s;
	char [] all = Vm.getStringChars(s);
	int len = all.length;
	if (len <= 0) return s;
	if (all[len-1] == '/' || all[len-1] == '\\')
		return new String(all,0,len-1);
	else
		return s;
}
/**
 * Split command line arguments using spaces, but keeping together
 * words with spaces in them that are enclosed in quotes.
 * @param args the String to split.
 * @param preserveQuotes if this is true then the quotes are preserved,
 * otherwise they are removed.
 * @return the individual commands.
 */
public static String [] splitCommand(String args,boolean preserveQuotes)
//===================================================================
{
	Vector v = new Vector();
	if (args != null){
		char [] chars = Vm.getStringChars(args);
		int st = 0;
		char quote = 0;
		for (int i = 0; i<chars.length; i++){
			char c = chars[i];
			if (c == '"'){
				if (quote == 0) quote = c;
				else if (quote == c) quote = 0;
				continue;
			}
			if ((c == ' ' || c == '\t') && (quote == 0)) {
				if (st == i) {
					st = i+1;
					continue;
				}else{
					v.add(args.substring(st,i));
					st = i+1;
				}
			}
		}
		if (st != chars.length) v.add(args.substring(st,chars.length));
	}
	String [] ret = new String[v.size()];
	//Vm.debug(v.toString());
	v.copyInto(ret);
	if (!preserveQuotes){
		for (int i = 0; i<ret.length; i++){
			if (ret[i].startsWith("\"")) ret[i] = ret[i].substring(1);
			if (ret[i].endsWith("\"")) ret[i] = ret[i].substring(0,ret[i].length()-1);
		}
	}
	return ret;
}
/**
* Copy the characters from the String into the destination in the most
* memory effecient way.
* @param data The string data.
* @param dataStart The starting character in the String.
* @param destination The destination character array.
* @param destinationStart The starting point in the destination array.
* @param length The number of characters to copy.
*/
//===================================================================
public static void copyInto(String data,int dataStart,char[] destination,int destinationStart,int length)
//===================================================================
{
	char [] dc = Vm.getStringChars(data);
	System.arraycopy(dc,dataStart,destination,destinationStart,length);
}
/**
* Copy the characters from the String into the destination in the most
* memory effecient way.
* @param data The string data.
* @param destination The destination character array.
* @param destinationStart The starting point in the destination array.
 */
//===================================================================
public static void copyInto(String data,char[] destination,int destinationStart)
//===================================================================
{
	copyInto(data,0,destination,destinationStart,data.length());
}
/**
 * Convert an Object to a String, formatted for the current locale.
 * If the Object implements LocaleFormatted, the format() method
 * will be called, otherwise a simple toString() is called.
 * @param data the data Object to convert to a String.
 * @param loc the locale to use for formatting. If it is null, the default
 * Locale is used.
 * @return the Object converted to a Locale specific String.
 */
public static String toString(Object data,Locale loc)
{
	if (data instanceof LocaleFormatted)
		return LocaleFormat.format((LocaleFormatted)data,null,loc);
	else
		return toString(data);
}
/**
 * Convert an Object to a String, formatted for the current locale.
 * If the Object implements LocaleFormatted, the format() method
 * will be called, otherwise a simple toString() is called.
 * @param data the data Object to convert to a String.
 * @param dest the destination CharArray. If it is null a new one is 
 * created and returned.
 * @param loc the locale to use for formatting. If it is null, the default
 * Locale is used.
 * @return the dest CharArray or a new one if it was null.
 */
public static CharArray toString(Object data,CharArray dest,Locale loc)
{
	return toString(data,dest,null,loc);
}
/**
 * Convert an Object to a String, formatted for the current locale.
 * If the Object implements LocaleFormatted, the format() method
 * will be called, otherwise a simple toString() is called.
 * @param data the data Object to convert to a String.
 * @param dest the destination CharArray. If it is null a new one is 
 * created and returned.
 * @param loc the locale to use for formatting. If it is null, the default
 * Locale is used.
 * @param format a format specifier - usually a String.
 * @return the dest CharArray or a new one if it was null.
 */
public static CharArray toString(Object data,CharArray dest,Object format,Locale loc)
{
	if (dest == null) dest = new CharArray();
	if (data instanceof LocaleFormatted)
		return LocaleFormat.format(dest, (LocaleFormatted)data, format, loc);
	else
		toString(data,dest);
	return dest;
}
/**
 * Format a LocaleFormatted to a String.
 * @param lf the LocaleFormatted object to format.
 * @param loc the Locale to use for formatting.
 * @return the LocaleFormatted object formatted as a String.
 */
public static String localeFormat(LocaleFormatted lf, Locale loc)
{
	return LocaleFormat.format(lf, null, loc);
}
/**
 * Format a LocaleFormatted to a String.
 * @param lf the LocaleFormatted object to format.
 * @param format a format specifier - usually a String.
 * @param loc the Locale to use for formatting.
 * @return the LocaleFormatted object formatted as a String.
 */
public static String localeFormat(LocaleFormatted lf, Object format, Locale loc)
{
	return LocaleFormat.format(lf, format, loc);
}
/**
 * Parse a String to a LocaleFormatted Object.
 * @param lf the LocaleFormatted object that will be parsing.
 * @param loc the Locale that was used for formatting.
 * @throws java.text.ParseException on a parse error.
 */
public static void localeParse(LocaleFormatted lf, Locale loc, String formatted)
throws ParseException
{
	LocaleFormat.parse(formatted, lf, null, loc);
}
//##################################################################
}
//##################################################################

