/*
 * ProcessManager.java
 *
 * Created on 23 febbraio 2009, 17.37
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package romas;

import java.util.Hashtable;
import java.util.Vector;
import java.util.logging.*;

import romas.com.data.TableManager;
import romas.com.data.tables.ACTIONS;
import romas.com.data.tables.ACTIONS_PROCESS;
import romas.com.data.tables.TABLE;
import romas.com.data.tables.foreignkeys.ACTION_SCOPE;
import romas.com.data.Structures.*;
import romas.services.ActionService;
import romas.util.LogWriter;
import eve.database.*;


/**
 *
 * @author APaliani
 */
public class ProcessManager extends Thread{
    
    public Hashtable services; // servizi attivi
    public Hashtable runningScopeAction; // memorizza l'azioni in corso per ogni scope
    private int CLOCK_ACTION = 10; // tempo di attesa default prima di ogni esecuzione dei processi
    private volatile boolean isTerminated = false;
   
          
    /** Creates a new instance of ProcessManager
     * @param clockAction
     */
    public ProcessManager(int clockAction) {
        services = new Hashtable();
        CLOCK_ACTION = clockAction;
    }
    
    
    /**
     *Arresta il servizio
     */
    public void terminate()
    {
        isTerminated = true;
    }
    
    
    /**
     *Termina l'esecuzione di un processo
     * @param actionProcess
     * @return
     */
    public boolean killActionThread(ACTIONS_PROCESS actionProcess)
    {
        if(actionProcess.ACTION != null){
            return actionProcess.ACTION.terminate();
        }

        return false;
    }
   
    
    /**
     *Aggiunge un servizio alla lista dei servizi
     * @param service
     * @throws Exception
     */
    public void addService(ActionService service)throws Exception
    {
        Object put = services.put(service.getScope(), service);
    }

    
    /**
     *Archivia ed elimina le azioni concluse
     */
    private void deleteAndArchive(Vector vActions)throws Exception{
        // verifica che ci sia un risultato
        if((vActions != null) && (vActions.size() > 0))
        {
            LogWriter.writeLog(Level.FINE, "deleting and archive actions ...");
            
            // scorre le azioni da processare
            for(int i = 0; i < vActions.size(); i++)
            {
                ACTIONS_PROCESS actionProcess = (ACTIONS_PROCESS)vActions.get(i);

                LogWriter.writeLog(Level.FINE, "deleting ACTION : [" + TABLE.toString(actionProcess) + "]...");
                
                if(actionProcess.ACTION != null) actionProcess.ACTION.terminate();

                TableManager.deleteAndArchiveACTION_PROCESS(actionProcess);    
            }                        
        }
    }


    /**
     *Ricrea le azioni STOPPED
     */
    private void recreateStoppedActions(Vector vActions)throws Exception{
        // verifica che ci sia un risultato
        if((vActions != null) && (vActions.size() > 0)){
            LogWriter.writeLog(Level.FINER, "Recreate ["+ vActions.size() +"] stopped actions ...\n");

            // scorre le azioni da ricreare
            for(int i = 0; i < vActions.size(); i++){
                ACTIONS_PROCESS actionProcess = (ACTIONS_PROCESS)vActions.get(i);

                // imposto lo stato CREATE all'azione per essere riprocessata al prossimo ciclo
                LogWriter.writeLog(Level.FINER, "RE-Create action:  ["+ TABLE.toString(actionProcess) +"] ...");
                actionProcess.STATE_ACTION_PROCESS = ACTION_STATE.CREATED_ACTION_STATE;
                TableManager.updateACTION_PROCESS(actionProcess);
            }
        }
    }


    /**
     * Prepara le azioni da processare per ogni scope
     *
     * @param vActions
     * @throws java.lang.Exception
     */
    private Vector preProcessActions(EntriesView evCreated){
        if(evCreated == null) return null;
        if(evCreated.size() == 0) return null;

        Vector vListPreprocess = new Vector();
        ACTIONS_PROCESS actionProcess = null;
        try{
            // scorre gli scope
            EntriesView evScopes = TableManager.searchData(TableManager.dataACTION_SCOPE_TABLE, "By PriorityScope", false);
            if((evScopes != null) && (evScopes.size() >0)){
                for(int i=0; i<evScopes.size(); i++){
                    ACTION_SCOPE scope = (ACTION_SCOPE)evScopes.getData(i);

                    // inizializza la lista delle azioni in esecuzione per ogni scope
                    runningScopeAction = new Hashtable();

                    // per ogni scope rileva le ACTIONS da processare
                    //EntriesView evActionsScopes = TableManager.searchSubData(TableManager.dataACTIONS_PROCESS_TABLE, evCreated, "By Scope", scope.NAME_SCOPE, false);

                    for(int j=0; j<evCreated.size(); j++){

                        actionProcess = (ACTIONS_PROCESS)evCreated.getData(j);

                        // filtra lo SCOPE
                        if(actionProcess.SCOPE_ACTIONS_PROCESS.equals(scope.NAME_SCOPE)){
                            // processa un'azione con lo stesso nome di ogni SCOPE per ogni ciclo
                            if(!runningScopeAction.containsKey(actionProcess.NAME_SCOPE_ACTION)){
                                // memorizza l'azione da processare ed il nome
                                boolean add = vListPreprocess.add(actionProcess);

                                LogWriter.writeLog(Level.FINE, "preprocessed action: ["+ TABLE.toString(actionProcess) +"] ...");
                                Object put = runningScopeAction.put(actionProcess.NAME_SCOPE_ACTION, actionProcess);
                            }
                        }
                    }
                }
            }
        }
        catch(Exception ex){
            if (actionProcess != null){
                LogWriter.writeLog(Level.WARNING, ex.getMessage() + " - in action: ["+ TABLE.toString(actionProcess) +"] ...");
            }else{
                LogWriter.writeLog(Level.WARNING, ex.getMessage());
            }
        }
        return vListPreprocess;
    }

    
    /**
     *Processa la prima azione in stato CREATE in ordine di priorit per ogni scope
     */
    private void processActions(Vector vActions){
        LogWriter.writeLog(Level.FINER, "processing ACTION_PROCESS_TABLE ...");

        ACTIONS_PROCESS actionProcess = null;
        try{
            // verifica che ci siano azioni da processare
            if((vActions != null) && (vActions.size() > 0)){

                LogWriter.writeLog(Level.FINE, "to processing ["+ vActions.size() +"] actions ...");
                // scorre le azioni da processare
                for(int i = 0; i < vActions.size(); i++){

                    actionProcess = (ACTIONS_PROCESS)vActions.get(i);
                    LogWriter.writeLog(Level.FINER, TABLE.toString(actionProcess));

                    // verifica lo stato dell'azione
                    if(actionProcess.STATE_ACTION_PROCESS == ACTION_STATE.CREATED_ACTION_STATE){

                        // update database state
                        LogWriter.writeLog(Level.FINER, "Init action:  ["+ TABLE.toString(actionProcess) +"] ...");
                        actionProcess.STATE_ACTION_PROCESS = ACTION_STATE.INITIATED_ACTION_STATE;
                        TableManager.updateACTION_PROCESS(actionProcess);

                        // CHECK CLEAR ALL ACTION
                        if(actionProcess.COMMAND_ACTIONS_PROCESS.equals("CLEAR_ALL")){
                            if(actionProcess.SCOPE_ACTIONS_PROCESS != null && !actionProcess.SCOPE_ACTIONS_PROCESS.equals("")){
                                if(actionProcess.NAME_ACTION_ACTIONS_PROCESS != null  &&  !actionProcess.NAME_ACTION_ACTIONS_PROCESS.equals("")){
                                    TableManager.clearAllActions(actionProcess.SCOPE_ACTIONS_PROCESS, actionProcess.NAME_ACTION_ACTIONS_PROCESS);
                                    LogWriter.writeLog(Level.INFO, "CLEAR ALL ACTIONS:  ["+ actionProcess.SCOPE_ACTIONS_PROCESS.toString() + ", " + actionProcess.NAME_ACTION_ACTIONS_PROCESS.toString() +"] !");
                                }
                                else{
                                    TableManager.clearAllActions(actionProcess.SCOPE_ACTIONS_PROCESS);
                                    LogWriter.writeLog(Level.INFO, "CLEAR ALL ACTIONS:  ["+ actionProcess.SCOPE_ACTIONS_PROCESS.toString() + "] !");
                                }
                            }
                            else{
                                TableManager.clearAllActions();
                                LogWriter.writeLog(Level.INFO, "CLEAR ALL ACTIONS !");
                            }

                            // exit from this iteration
                            return;
                        }

                        // verifica azioni in corso
                        if(!findMostImportantActions(actionProcess)){
                            executeProcessAction(actionProcess);
                        }
                        else{
                            // imposto lo stato STOPPED all'azione per essere ricreata al prossimo ciclo
                            LogWriter.writeLog(Level.FINER, "STOPPED action:  ["+ TABLE.toString(actionProcess) +"] ...");
                            actionProcess.STATE_ACTION_PROCESS = ACTION_STATE.STOPPED_ACTION_STATE;
                            TableManager.updateACTION_PROCESS(actionProcess);
                        }
                    }
                    else{
                        LogWriter.writeLog(Level.FINE, "Skipped action, action scope is busy:  ["+ TABLE.toString(actionProcess) +"] ...");
                    }
                }

                LogWriter.writeLog(Level.FINE, "end processing ! \n");
            }
        }
        catch(Exception ex){
            if (actionProcess != null){
                LogWriter.writeLog(Level.WARNING, ex.getMessage() + " - in action: ["+ TABLE.toString(actionProcess) +"] ...");
            }else{
                LogWriter.writeLog(Level.WARNING, ex.getMessage());
            }
        }
    }
    
    
    /**
     * Verifica se esistono azioni simili in corso con priority maggiore
     */
    private boolean findMostImportantActions(ACTIONS_PROCESS actions_process)throws Exception{
        LogWriter.writeLog(Level.FINE, "finding ...");
        EntriesView evActions = findSimilarActions(actions_process);
        if(evActions != null){
            for(int i=0; i<evActions.size(); i++){

                ACTIONS_PROCESS temp_similarAction = (ACTIONS_PROCESS)evActions.get(i).getData();

                LogWriter.writeLog(Level.FINER, "check similar ["+ TABLE.toString(temp_similarAction) +"]...");

                // ricerca la ACTIONS_PROCESS_TABLE per verificare lo stato attuale
                EntriesView evList = TableManager.searchData(TableManager.dataACTIONS_PROCESS_TABLE, "By DateString", temp_similarAction.DATE_STRING_ACTION_PROCESS, false);

                ACTIONS_PROCESS similarAction = null;
                if((evList != null) && (evList.size() >0)){                    
                    similarAction = (ACTIONS_PROCESS)(evList.getData(0));
                }

                //se esiste una azione simile ma non la medesima
                if ((similarAction != null) && !similarAction.DATE_STRING_ACTION_PROCESS.equals(actions_process.DATE_STRING_ACTION_PROCESS)){
                    
                    // se l'azione nel ciclo di vita                                        
                    if((similarAction.STATE_ACTION_PROCESS == ACTION_STATE.CREATED_ACTION_STATE) ||
                       (similarAction.STATE_ACTION_PROCESS == ACTION_STATE.INITIATED_ACTION_STATE) ||
                       (similarAction.STATE_ACTION_PROCESS == ACTION_STATE.RUNNING_ACTION_STATE)){

                        // se l'azione simile  in esecuzione e di priorit inferiore la elimino e la ricreo
                        if((similarAction.STATE_ACTION_PROCESS == ACTION_STATE.RUNNING_ACTION_STATE)
                                && (similarAction.PRIORITY_ACTIONS_PROCESS < actions_process.PRIORITY_ACTIONS_PROCESS)){
                            killAndRecreate(similarAction);
                        }

                        // se l'azione simile  processata e di priority uguale o maggiore annullo l'azione in corso
                        if((similarAction.PRIORITY_ACTIONS_PROCESS >= actions_process.PRIORITY_ACTIONS_PROCESS) &&
                            (similarAction.STATE_ACTION_PROCESS == ACTION_STATE.RUNNING_ACTION_STATE)){
                            LogWriter.writeLog(Level.WARNING, "Similar action ["+ similarAction.DATE_STRING_ACTION_PROCESS +"] Most Importan ["+ actions_process.DATE_STRING_ACTION_PROCESS +"]");
                            //Object put = runningScopeAction.put(similarAction.NAME_SCOPE_ACTION, similarAction);
                            return true;
                        }
                    }
                }
            }
        }
        return false;
    }
    
           
    /**
     *Arresta e ricrea l'azione
     */
    private void killAndRecreate(ACTIONS_PROCESS actionProcess)throws Exception{        
        LogWriter.writeLog(Level.FINE, "killing ...");
        if(killActionThread(actionProcess)){
            LogWriter.writeLog(Level.WARNING, "Killed process thread for ACTION ["+ TABLE.toString(actionProcess) +"]");
            recreateOrdinaryAction(actionProcess);
        }
        else{
            LogWriter.writeLog(Level.WARNING, "Cannot kill process thread for ACTION ["+ TABLE.toString(actionProcess) +"]");
        }
    }
    

    /**
     *Ritorna la lista delle azioni relative al tipo
     */
    private EntriesView getTypeActions(String fullNameAction, int type)throws Exception{
        // rileva tutte le azioni con un certo nome
        EntriesView evAll = TableManager.searchData(TableManager.dataACTIONS_TABLE, "By RelativeFullName", fullNameAction, false);
        
        if (evAll.size() > 0){            
            EntriesView evActions = TableManager.searchSubData(TableManager.dataACTIONS_TABLE, evAll, "By Type", type, false);
            return evActions;
        }
        return null;
    }
    
    
    /**
     *Crea le azioni alternative
     */
    private void createAlternativeActions(ACTIONS_PROCESS actionProcess, int type)throws Exception{
        LogWriter.writeLog(Level.FINE, "creating alternative action ...");
        EntriesView evAlternative = getTypeActions(actionProcess.FULLNAME_ACTION_ACTIONS_PROCESS, ACTION_TYPE.ALTERNATIVE_ACTION_TYPE);

        if(evAlternative != null){
            for(int i=0; i< evAlternative.size(); i++){
                ACTIONS action = (ACTIONS)evAlternative.get(i).getData();

                ACTIONS_PROCESS alternativeAction = new ACTIONS_PROCESS(action);

                alternativeAction.TYPE_ACTIONS_PROCESS = type;
                alternativeAction.PRIORITY_SCOPE_ACTIONS_PROCESS = actionProcess.PRIORITY_SCOPE_ACTIONS_PROCESS;
                alternativeAction.ID_OWNER_ACTION_PROCESS = actionProcess.ID_PROCESS_ACTIONS_PROCESS;
                alternativeAction.RELATIVE_FULLNAME_ACTION_PROCESS = actionProcess.FULLNAME_ACTION_ACTIONS_PROCESS;
                alternativeAction.NUMBER_ITERATION_ACTION_PROCESS =  actionProcess.NUMBER_ITERATION_ACTION_PROCESS;

                TableManager.addNew(alternativeAction, TableManager.dataACTIONS_PROCESS_TABLE);
            }
        }
        else{
            LogWriter.writeLog(Level.FINE, "Alternative actions NOT FOUND for ACTION_PROCESS ["+ TABLE.toString(actionProcess) +"]");
            recreateOrdinaryAction(actionProcess);
        }
    }
    
    
    /**
     *Crea l'azione opposta
     */
    private void createOppositeAction(ACTIONS_PROCESS actionProcess, int type)throws Exception{
        LogWriter.writeLog(Level.FINE, "creating oppoiste action...");
        EntriesView evOpposite = getTypeActions(actionProcess.RELATIVE_FULLNAME_ACTION_PROCESS, ACTION_TYPE.OPPOSITE_ACTION_TYPE);
        if(evOpposite != null){
            ACTIONS action = (ACTIONS)evOpposite.get(0).getData();

            ACTIONS_PROCESS oppositeAction = new ACTIONS_PROCESS(action);

            oppositeAction.TYPE_ACTIONS_PROCESS = type;
            oppositeAction.PRIORITY_SCOPE_ACTIONS_PROCESS = actionProcess.PRIORITY_SCOPE_ACTIONS_PROCESS;
            oppositeAction.ID_OWNER_ACTION_PROCESS = actionProcess.ID_PROCESS_ACTIONS_PROCESS;
            oppositeAction.RELATIVE_FULLNAME_ACTION_PROCESS = actionProcess.FULLNAME_ACTION_ACTIONS_PROCESS;
            oppositeAction.NUMBER_ITERATION_ACTION_PROCESS =  actionProcess.NUMBER_ITERATION_ACTION_PROCESS;

            TableManager.addNew(oppositeAction, TableManager.dataACTIONS_PROCESS_TABLE);
        }
        else{
            LogWriter.writeLog(Level.WARNING, "Opposite action NOT FOUND for ACTION_PROCESS ["+ TABLE.toString(actionProcess) +"]");
            recreateOrdinaryAction(actionProcess);
        }
    }
    
    
    /**
     *Ricrea l'azione ordinaria
     */
    private void recreateOrdinaryAction(ACTIONS_PROCESS actionProcess)throws Exception{
        LogWriter.writeLog(Level.FINE, "recreating ordinary ...");
        ACTIONS_PROCESS ordinaryAction = new ACTIONS_PROCESS(actionProcess);
        TableManager.addNew(ordinaryAction, TableManager.dataACTIONS_PROCESS_TABLE);
    }
    
    
    /**
     *Ricrea un'azione a seconda del tipo
     */
    private void recreateActionByType(ACTIONS_PROCESS actionProcess)throws Exception{
        LogWriter.writeLog(Level.FINE, "recreating by type ...");
        
        switch(actionProcess.TYPE_ACTIONS_PROCESS)
        {
            case ACTION_TYPE.ORDINARY_ACTION_TYPE:
                // creazione azioni alternative alla ordinaria TYPE = ALTERNATIVE
                createAlternativeActions(actionProcess, ACTION_TYPE.ALTERNATIVE_ACTION_TYPE);
                break;
            
            case ACTION_TYPE.ALTERNATIVE_ACTION_TYPE:
                // creazione azione opposta alla ordinaria TYPE = OPPOISTE
                createOppositeAction(actionProcess, ACTION_TYPE.OPPOSITE_ACTION_TYPE);
                break;
                
            case ACTION_TYPE.OPPOSITE_ACTION_TYPE:
                // creazione azioni alternative alla ordinaria TYPE = ORDINARY
                // crea azione alternativa TYPE = ORDINARY
                createAlternativeActions(actionProcess, ACTION_TYPE.ORDINARY_ACTION_TYPE);
                break;
        }
    }
    
    
    /**
     *Cerca azioni simili
     *     
     */
    private EntriesView findSimilarActions(ACTIONS_PROCESS actions_process)throws Exception{
        LogWriter.writeLog(Level.FINE, "finding similar : By NameScopeAction ...");
        //By Scope
        EntriesView evNameScopeAction = TableManager.searchData(TableManager.dataACTIONS_PROCESS_TABLE, "By NameScopeAction", actions_process.NAME_SCOPE_ACTION, false);
        return evNameScopeAction;
    }
    


    /**
     * Restituisce il vettore degli elementi
     * @param evList
     * @return
     * @throws java.lang.Exception
     */
    private Vector getVectorFromList(EntriesView evList)throws Exception{
        Vector vList = new Vector();

        if((evList != null) && (evList.size() >0)){
            for(int i=0; i<evList.size(); i++){
                boolean add = vList.add(evList.getData(i));
            }
        }
        return vList;
    }


    /**
     *Ordina la lista dei processi
     * @param action_state
     * @return
     * @throws Exception
     */
    public EntriesView getSortedListACTIONS_PROCESSByState(int action_state)throws Exception{
        // ricerca delle ACTIONS_PROCESS_TABLE
        return TableManager.searchData(TableManager.dataACTIONS_PROCESS_TABLE, "By State_Priority", action_state, false);
    }
    

    /**
     *Esegue l'azione e memorizza il thread che la esegue.
     *
     *@param actionProcess  Azione da processare
     * @throws Exception
     */
    public void executeProcessAction(ACTIONS_PROCESS actionProcess)throws Exception{       
        if(actionProcess.STATE_ACTION_PROCESS == ACTION_STATE.INITIATED_ACTION_STATE){           
            if(services.containsKey(actionProcess.SCOPE_ACTIONS_PROCESS)){
                ActionService service = (ActionService)services.get(actionProcess.SCOPE_ACTIONS_PROCESS);
                service.executeAction(actionProcess);            
            }
            else{
                LogWriter.writeLog(Level.SEVERE, "Service scope ["+ actionProcess.SCOPE_ACTIONS_PROCESS +"] not found for ACTIONS_PROCESS ["+ actionProcess.FULLNAME_ACTION_ACTIONS_PROCESS +"]!");
                //throw new Exception("Service scope ["+ actionProcess.SCOPE_ACTIONS_PROCESS +"] not found !");
                TableManager.deleteACTION_PROCESS(actionProcess);
            }
        }        
    }


    /**
     *Imposta la cancellazione della ACTIONS_PROCESS
     * @param actionProcess
     */
    protected void doDeleting(ACTIONS_PROCESS actionProcess){
        LogWriter.writeLog(Level.FINE, TABLE.toString(actionProcess));
        actionProcess.STATE_ACTION_PROCESS = ACTION_STATE.DELETING_ACTION_STATE;
        TableManager.updateACTION_PROCESS(actionProcess);               
    }
    

    @Override
    public void run(){
        try{
            LogWriter.writeLog(Level.INFO, "START ProcessManager ...\n");
            while(!isTerminated){
                // elimina le azioni concluse e le archivia se in sospeso
                deleteAndArchive(getVectorFromList(getSortedListACTIONS_PROCESSByState(ACTION_STATE.DELETING_ACTION_STATE)));

                //imposta a CREATE le azioni SOPPED
                recreateStoppedActions(getVectorFromList(getSortedListACTIONS_PROCESSByState(ACTION_STATE.STOPPED_ACTION_STATE)));

                // processa le azioni create
                processActions(preProcessActions(getSortedListACTIONS_PROCESSByState(ACTION_STATE.CREATED_ACTION_STATE)));

                // wait CLOCK TIME
                Thread.sleep(CLOCK_ACTION);
            }
        }
        catch(Exception ex){
            LogWriter.writeLog(Level.SEVERE, ex.getMessage());
            //ProgramMain.start();
        }
    }
}
