/*
 * ProgramMain.java
 *
 * Created on 23 febbraio 2009, 18.09
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package romas;

import java.util.Vector;
import java.util.Hashtable;
import java.util.logging.Level;
import java.util.List;
import java.io.File;
import java.net.*;

import romas.com.data.tables.NOTIFICATIONS_MAP;
import romas.services.*;
import romas.com.data.*;
import romas.util.LogWriter;
import romas.services.resources.*;

/**
 *
 * @author APaliani
 */
public class ProgramMain{

    static String series = "ALPHA SERIES";
    static String release = "2011.06.14.2027 test ";
    static String currentVersion = " [ " + series + "  " + release + " ] ";

    static ProcessManager processManager;

    static Hashtable resources;
    static String currentDir;
    
    /** Creates a new instance of ProgramMain */
    public ProgramMain() {
    }

    
    /**
     *Inizializzazione del sistema
     */
    private static void initSystem()throws Exception
    {
        String USER_DIR_KEY = "user.dir";
        currentDir = System.getProperty(USER_DIR_KEY);

        // elimina tutti files del database e del log per evitare corruzioni
        File dir = new File(currentDir);
        try{
            File[] files = dir.listFiles();
            for(int i=0; i< files.length; i++)
            {
                File file = files[i];

                if(file.isFile())
                {
                    String fileName = file.getName();
                    if((!fileName.endsWith(".bat")) && (!fileName.endsWith(".log")) && (!fileName.endsWith(".jar")))
                    {
                        if(file.delete()){
                            LogWriter.writeLog(Level.CONFIG, "Deleted file: " + file.getName());
                        }
                        else{
                            LogWriter.writeLog(Level.WARNING, "Cannot delete file: " + file.getName());
                        }
                    }                
                }
            }
        }
        catch(Exception ex)
        {
            System.err.println(ex.getMessage());
        }
        
        try
        {
            Configuration config = new Configuration();
            config.loadPropertiesFile("config" + File.separator + "romas.properties");
            String log_Level_name = config.getPropertiesValue("Log_Level_name");
            LogWriter.init("romas.log", log_Level_name);

            int clockAction = Integer.parseInt(config.getPropertiesValue("Clock_Action"));
            processManager = new ProcessManager(clockAction);
        }
        catch(Exception ex)
        {
            LogWriter.init("romas.log", "ALL");
            LogWriter.writeLog(Level.WARNING, "Cannot create configured log !");
        }
        
        LogWriter.writeLog(Level.INFO, "\n\n******************************************************************************************************************\n\n" + 
                                       "START ROMAS " + currentVersion + " ON DIR: " + currentDir + " ........................\n\n" +
                                       "******************************************************************************************************************\n\n"); 
        

        // create all tables
        TableManager.createAllTables();

    }


    
    /**
     *Avvio dei servizi 
     *
     * @param idThreadMain    id del thread principale
     * @throws Exception      Intercetta tutte le eccezioni
     */
    private static void startServices(long idThreadMain)throws Exception
    {
        LogWriter.writeLog(Level.INFO, "RUN SERVICES ........................\n");
        System.out.println("\007\007\007");

        // Servizio di risorse
        ResourcesService resourceService = new ResourcesService(resources);

        // esplorazione directory dei servizi
        File serviceDir = new File(currentDir + File.separator + "services");

        String[] children = serviceDir.list();
        if (children == null)
        {
            // Either dir does not exist or is not a directory
            LogWriter.writeLog(Level.WARNING, "MISSING DIRECTORY: "+ serviceDir.getAbsolutePath() +"\n");
        }
        else
        {
            if (children.length ==0)
            {
                // Either dir does not exist or is not a directory
                LogWriter.writeLog(Level.WARNING, "THERE ARE NO SERVICES INTO DIRECTORY: "+ serviceDir.getAbsolutePath() +"\n");
            }
            else
            {
                for (int i=0; i<children.length; i++)
                {
                    // Nome del file
                    String filename = children[i];
                    if(filename.endsWith(".jar"))
                    {
                        System.out.println("\007");
                        LogWriter.writeLog(Level.INFO, "Find service plugin : "+ filename);
                        String fileFullName = serviceDir + File.separator + filename;
                        List files = PackageUtils.getClasseNamesInPackage(fileFullName, "romas.services");
                        
                        LogWriter.writeLog(Level.FINE, "# classes : "+ files.size());

                        if(files.size() > 0)
                        {
                            boolean loaded = false;
                            URL url = new URL("jar:file:/"+ fileFullName +"!/");
                            URLClassLoader ucl = new URLClassLoader(new URL[] { url });

                            for(int fi=0; fi< files.size(); fi++)
                            {
                                String className = files.get(fi).toString();

                                LogWriter.writeLog(Level.FINE, "service class : "+ className);

                                try
                                {
                                    Class myClassService = ucl.loadClass(className);
                                    
                                    Object objectService = myClassService.newInstance();

                                    LogWriter.writeLog(Level.FINE, "service class new instance ! ");
                                    if(objectService instanceof romas.services.ActionService)
                                    {
                                        ActionService service = ((ActionService)objectService);
                                        service.resources = resources;
                                        processManager.addService(service);
                                        loaded = true;
                                        LogWriter.writeLog(Level.INFO, "Loaded service class: "+ className +"\n");
                                    }

                                }
                                catch(Exception e)
                                {
                                    LogWriter.writeLog(Level.FINE, " error on instance class "+ className +" : " + e.getMessage());
                                    //e.printStackTrace();
                                }
                            }

                            if(!loaded)
                            {
                                LogWriter.writeLog(Level.WARNING, "THERE ARE NO CLASS LOADED INTO JAR: "+ fileFullName +"\n");
                            }
                        }
                        else
                        {
                            LogWriter.writeLog(Level.WARNING, "THERE ARE NO SERVICES CLASS INTO JAR: "+ fileFullName +"\n");
                        }
                    }
                }
            }
        }
    }
    
    
    /**
     *Notifica la richiesta di una ACTION
     * @param scope
     * @param action
     * @param command
     * @param params
     * @param active
     * @throws Exception
     */
    public static void notifyAction(String scope, String action, String command, Vector params, boolean active)throws Exception
    {
        String actionFULLNAME = scope + "_" +action + "_" + command;
        NOTIFICATIONS_MAP notify = new NOTIFICATIONS_MAP();        
        notify.PARAMS_COMMAND_EVENT = params;
        notify.NOTIFICATOR = scope;
        notify.FULLNAME_EVENT = actionFULLNAME;
        notify.ACTION_EVENT = action;
        notify.COMMAND_EVENT = command;
        notify.NUMBER_ITERATION_NOTIFY = 1;
        notify.ACTIVE = active;
        
        if(processManager != null)
        {
            if(processManager.services != null)
            {
                if(processManager.services.containsKey(scope))
                {
                    ((NotificationService)processManager.services.get(scope)).notifyAction(notify);
                }
            }
        }
    }
    
    
    private static void start(){
        // Set max priority on thread main
        Thread threadMain = Thread.currentThread();
        threadMain.setPriority(Thread.MAX_PRIORITY); 
        
        //inizializzazione del sistema
        try{
            initSystem();
        }
        catch(Exception ex)
        {
            LogWriter.writeLog(Level.SEVERE, "Init System ERROR: " + ex.getMessage());
            System.exit(0);
        }        
        
        
        // avvio dei servizi
        try{
            startServices(threadMain.getId());
        }
        catch(Exception ex)
        {
            LogWriter.writeLog(Level.SEVERE, "Start services ERROR: " + ex.getMessage());
            System.exit(0);
        }


        // Avvio del sistema
        try{            
            processManager.setPriority(Thread.MAX_PRIORITY);

            // Avvio delle direttive principali
            // DirectiveService
            System.out.println("\007\007\007");

            processManager.addService(new DirectiveService());
            // Avvio del Process Manager
            processManager.start();

            //Thread.sleep(1000); // necessario per non andare in concorrenza con le direttive principale

            // caricamento delle direttive principali
            notifyAction("DIRECTIVE", "LOAD", "DEFAULT", null, true);

            //Thread.sleep(3000); // necessario per caricare le direttive principale

            System.out.println("\007");
            Vector vparams = new Vector();
            vparams = new Vector();
            vparams.add("ROMAS  "+ currentVersion + " , AVVIATO");
            notifyAction("SPEAKER", "SPEAK", "TEXT", vparams, true);
        }
        catch(Exception ex)
        {
            LogWriter.writeLog(Level.SEVERE, "Start services ERROR: " + ex.getMessage());
            System.exit(0);
        }
    }


    /**
     *Metodo principale che avvia il sitema
     * @param args
     */
    public static void main(String[] args)
    {
        start();
    }
    
}
