/*
 * Configuration.java
 *
 * Created on 24 febbraio 2009, 10.18
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package romas.com.data;

import java.util.Properties;
import java.util.Vector;
import java.io.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import java.util.logging.*;

import romas.com.data.tables.*;
import romas.util.LogWriter;
import romas.util.Utility;
import romas.com.data.Structures.*;

import eve.database.*;

/**
 *
 * @author APaliani
 */
public class Configuration {
    
    Properties properties;
    Document docNotifications;
    Document docActions;
    
    public Configuration()
    {
        properties = new Properties();
    }
    
    
    /** Creates a new instance of Configuration */
    public Configuration(Logger logger) 
    {
        properties = new Properties();        
    }
    
    
    /*************************************************************************************************/
    /***************************************  file xml ***********************************************/
    /*************************************************************************************************/
    
    /**
     *Inizializza il file delle notifiche
     */
    public void loadNotificationsXMLFile(String fileName)throws Exception
    {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setValidating(false);
        //factory.setIgnoringElementContentWhitespace(true);
        docNotifications = factory.newDocumentBuilder().parse(fileName);
    }

    
    /**
     *Inizializza il file delle actions
     */
    public void loadActionsXMLFile(String fileName)throws Exception
    {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setValidating(false);        
        //factory.setIgnoringElementContentWhitespace(true);
        docActions = factory.newDocumentBuilder().parse(fileName);
    }

    
    /**
     *Legge le actions e le memorizza nel database
     */
    public void setActionField(Node node, ACTIONS actionRECORD)throws Exception
    {        
        if(node.getNodeName().equals("SCOPE"))
        {
            actionRECORD.SCOPE_ACTION = node.getTextContent();
        }

        if(node.getNodeName().equals("NAME"))
        {
            actionRECORD.NAME_ACTION = node.getTextContent();
        }

        if(node.getNodeName().equals("COMMAND"))
        {
            actionRECORD.COMMAND_ACTION = node.getTextContent();
        }        
        
        if(node.getNodeName().equals("INSTRUCTION"))
        {
            actionRECORD.INSTRUCTION_ACTION = node.getTextContent();
        }

        if(node.getNodeName().equals("ITERATION"))
        {
            actionRECORD.ITERATION_ACTION = Integer.valueOf(node.getTextContent()).intValue();
        }
        
        if(node.getNodeName().equals("PRIORITY"))
        {
            actionRECORD.PRIORITY_ACTION = Integer.valueOf(node.getTextContent()).intValue();
        }
        
        if(node.getNodeName().equals("PARAMS"))
        {
            if(node.hasChildNodes())
            {
                NodeList paramsList = node.getChildNodes();
                for(int pIn=0; pIn < paramsList.getLength(); pIn++)
                {
                    Node paramNode = paramsList.item(pIn);
                    actionRECORD.PARAMS_ACTION.add(actionRECORD.PARAMS_ACTION.size(), (Object)paramNode.getTextContent());
                }
            }
        }
    }
    
    
    /**
     *Legge le azioni disponibili per il servizio
     *dal file ServiceName_actions.xml
     */
     public void readActions(String serviceName)throws Exception
     {
        LogWriter.writeLog(Level.CONFIG, "Read ["+ serviceName +"] actions ...");
        
        NodeList azioni = docActions.getElementsByTagName("Action");        
        for (int i=0; i<azioni.getLength(); i++)
        {            
            Element azione = (Element)azioni.item(i);

            // ORDINARY
            Node nodeOrdinary = azione.getElementsByTagName("ORDINARY").item(0);
            ACTIONS ordinaryAction = new ACTIONS(ACTION_TYPE.ORDINARY_ACTION_TYPE);
            
            for(int iOrd = 0; iOrd < nodeOrdinary.getChildNodes().getLength(); iOrd++)
            {
                Node node = nodeOrdinary.getChildNodes().item(iOrd);
                setActionField(node, ordinaryAction);
            }          
            String fullNameOrdinaryAction = ordinaryAction.SCOPE_ACTION + "_" + ordinaryAction.NAME_ACTION + "_" + ordinaryAction.COMMAND_ACTION;
            ordinaryAction.FULLNAME_ACTION = fullNameOrdinaryAction;
            TableManager.addNew(ordinaryAction, TableManager.dataACTIONS_TABLE);

            //ALTERNATIVE
            NodeList alternative = azione.getElementsByTagName("ALTERNATIVE");            
            for(int aIn=0; aIn < alternative.getLength(); aIn ++)
            {
                Node alternativeNode = alternative.item(aIn);
                ACTIONS actionAlternative = new ACTIONS(ACTION_TYPE.ALTERNATIVE_ACTION_TYPE);
                actionAlternative.FULLNAME_RELATIVE_TYPE_ACTION = fullNameOrdinaryAction;
                for(int indexAl = 0; indexAl < alternativeNode.getChildNodes().getLength(); indexAl++)
                {
                    Node node = alternativeNode.getChildNodes().item(indexAl);
                    setActionField(node, actionAlternative);
                }
                String fullNameAlternativeAction = actionAlternative.SCOPE_ACTION + "_" + actionAlternative.NAME_ACTION + "_" + actionAlternative.COMMAND_ACTION;
                actionAlternative.FULLNAME_ACTION = fullNameAlternativeAction;               
                TableManager.addNew(actionAlternative, TableManager.dataACTIONS_TABLE);                
            }
            
            //OPPOSITE
            Node nodeOpposite = azione.getElementsByTagName("OPPOSITE").item(0);
            if(nodeOpposite != null && nodeOpposite.hasChildNodes())
            {
                ACTIONS actionOpposite = new ACTIONS(ACTION_TYPE.OPPOSITE_ACTION_TYPE);
                actionOpposite.FULLNAME_RELATIVE_TYPE_ACTION = fullNameOrdinaryAction;
                for(int iOpp = 0; iOpp < nodeOpposite.getChildNodes().getLength(); iOpp++)
                {
                    Node node = nodeOpposite.getChildNodes().item(iOpp);
                    setActionField(node, actionOpposite);
                }                        
                String fullNameOppositeAction = actionOpposite.SCOPE_ACTION + "_" + actionOpposite.NAME_ACTION + "_" + actionOpposite.COMMAND_ACTION;
                actionOpposite.FULLNAME_ACTION = fullNameOppositeAction;               
                TableManager.addNew(actionOpposite, TableManager.dataACTIONS_TABLE);
            }
        }
     }
    
    
    /**
     *Legge le notifiche e le memorizza nella tabella
     *
     */
    public void readNotifications(String scope)throws Exception
    {
        LogWriter.writeLog(Level.CONFIG, "Read ["+ scope +"] notifications ... ");
              
        // legge le sottoscrizioni
        NodeList sottoscrizioni = docNotifications.getElementsByTagName("Subscription");

        if(sottoscrizioni.getLength() == 0)
        {
            LogWriter.writeLog(Level.CONFIG, "There are no notifications for the scope ["+ scope +"] ! ");
            return;
        }
        
        for (int i=0; i<sottoscrizioni.getLength(); i++) 
        {
            NOTIFICATIONS_MAP notification = new NOTIFICATIONS_MAP();
            notification.ID_SIGNATURE = Utility.getDateTimeID() + String.valueOf(i);
            notification.NOTIFICATOR = scope;
            
            Element sottoscrizione = (Element)sottoscrizioni.item(i);

            //<Active>true</Active> 
            try{
                Node nodeActive = (Node)sottoscrizione.getElementsByTagName("Active").item(0);            
                notification.ACTIVE = Boolean.valueOf(nodeActive.getTextContent().toLowerCase()).booleanValue();
            }
            catch(Exception ex)
            {
                // le notifiche sono attive se non diversamente specificato
                notification.ACTIVE = true;
            }
            
            //<Subscriber>SPEAK</Subscriber> 
            Node nodeSubscriber = (Node)sottoscrizione.getElementsByTagName("Subscriber").item(0);            
            notification.SUBSCRIBER = nodeSubscriber.getTextContent();            
            
            //<Event>
            Node nodeEvent = (Node)sottoscrizione.getElementsByTagName("Event").item(0);
            
            Node actionEventNode = nodeEvent.getChildNodes().item(0);            
            //    <Action>MOTION_DETECTION</Action> 
            notification.ACTION_EVENT = actionEventNode.getTextContent();
            //    <Command>DETECTED</Command> 
            Node commandEventNode = nodeEvent.getChildNodes().item(1);
            notification.COMMAND_EVENT = commandEventNode.getTextContent();
            notification.FULLNAME_EVENT = notification.NOTIFICATOR + "_" + notification.ACTION_EVENT + "_" + notification.COMMAND_EVENT;
            //</Event>

            //<Notification>
            Node nodeNotification = (Node)sottoscrizione.getElementsByTagName("Notification").item(0);
            Node actionNotificationNode = nodeNotification.getChildNodes().item(0);
            //    <Action>SPEAKER</Action> 
            notification.ACTION_NOTIFY = actionNotificationNode.getTextContent();
            //    <Command>TEXT</Command> 
            Node commandNotificationNode = nodeNotification.getChildNodes().item(1);
            notification.COMMAND_NOTIFY = commandNotificationNode.getTextContent();
            
            notification.FULLNAME_NOTIFY = notification.SUBSCRIBER + "_" + notification.ACTION_NOTIFY + "_" + notification.COMMAND_NOTIFY;

            // ricarica la ACTION per leggere il numero di iterazioni
            EntriesView evAction = TableManager.searchData(TableManager.dataACTIONS_TABLE, "By FullName", notification.FULLNAME_NOTIFY, false);
            notification.NUMBER_ITERATION_NOTIFY = ((ACTIONS)evAction.get(0).getData()).ITERATION_ACTION;
            
            //    <Params>
            
            NodeList paramsList = nodeNotification.getChildNodes().item(2).getChildNodes();
            int numParams = paramsList.getLength();
            if(numParams > 0)
            {
                notification.PARAMS_COMMAND_NOTIFY = new Vector(numParams);
                for(int j=0; j<numParams; j++)
                {
                    //        <Param>RILEVATO MOVIMENTO</Param> 
                    notification.PARAMS_COMMAND_NOTIFY.insertElementAt(paramsList.item(j).getTextContent(), notification.PARAMS_COMMAND_NOTIFY.size());
                }
            }
            //    </Params>
            //</Notification>
            
            TableManager.addNew(notification, TableManager.dataNOTIFICATIONS_MAP_TABLE);
        }
    }
    
    
    /**
     *Legge le direttive dal file di configurazione
     * @throws Exception
     */
    @SuppressWarnings("unchecked")
    public void readDefaultDirectives()throws Exception
    {
        LogWriter.writeLog(Level.CONFIG, "Read default directives ...");        
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setValidating(false);
        File filexml = new File("config" + File.separator + "Directives.xml");
        docNotifications = factory.newDocumentBuilder().parse(filexml);        
        
        // legge le direttive
        NodeList direttive = docNotifications.getElementsByTagName("Directive");
        
        for (int i=0; i<direttive.getLength(); i++) 
        {
            Element direttiva = (Element)direttive.item(i);
            boolean enabled = Boolean.valueOf(direttiva.getAttribute("enabled")).booleanValue();
            boolean run_on_start = Boolean.valueOf(direttiva.getAttribute("run_on_start")).booleanValue();            
            
            Node nodeName = direttiva.getElementsByTagName("Name").item(0);            
            String nameDirective = nodeName.getTextContent();
            LogWriter.writeLog(Level.CONFIG, "Read directive : ["+ nameDirective +"] ...");
            Node nodeActions = direttiva.getElementsByTagName("Actions").item(0);            
            NodeList actions = nodeActions.getChildNodes();
            
            int numActions = actions.getLength();
            if(numActions > 0)
            {
                for(int j=0; j<numActions; j++)
                {
                    DIRECTIVES directive = new DIRECTIVES();
                    directive.NAME_DIRECTIVE = nameDirective;
                    directive.ENABLED_DIRECTIVE = enabled;
                    directive.START_ON_RUN_DIRECTIVE = run_on_start;
                    
                    Node action = actions.item(j);
                   
                    directive.SEQUENCE_ACTION_DIRECTIVE = Integer.valueOf((action.getChildNodes().item(0)).getTextContent()).intValue();
                    directive.SCOPE_NAME_ACTIONS_PROCESS = (action.getChildNodes().item(1)).getTextContent();
                    directive.SCOPE_PRIORITY_ACTION_PROCESS = Integer.valueOf((action.getChildNodes().item(2)).getTextContent()).intValue();
                    directive.NAME_ACTION_ACTIONS_PROCESS = (action.getChildNodes().item(3)).getTextContent();
                    directive.COMMAND_ACTIONS_PROCESS = (action.getChildNodes().item(4)).getTextContent();
                    directive.PRIORITY_ACTIONS_PROCESS = Integer.valueOf((action.getChildNodes().item(5)).getTextContent()).intValue();
                    directive.NUMBER_ITERATION_ACTION_DIRECTIVE = Integer.valueOf((action.getChildNodes().item(6)).getTextContent()).intValue();
                    
                    int numParamsAction = (action.getChildNodes().item(7)).getChildNodes().getLength();
                    if(numParamsAction > 0)
                    {
                        directive.PARAMS_ACTIONS_PROCESS = new Vector(numParamsAction);
                        for(int k = 0; k < numParamsAction; k++)
                        {                            
                            directive.PARAMS_ACTIONS_PROCESS.add((action.getChildNodes().item(7)).getChildNodes().item(k).getTextContent());
                        }
                        
                    }
                    
                    TableManager.addNew(directive, TableManager.dataDIRECTIVES_TABLE);
                }
            }
            
        
        }
        
        //tableManager.closeDatabase(dataDIRECTIVES_TABLE);
    }    
    
    /*************************************************************************************************/
    /*****************************  file properies ***********************************************/
    /************************************************************************************************
     * @param fileName
     */
    
    public void loadPropertiesFile(String fileName){
        try
        {            
            properties.load(new FileInputStream(fileName));

        }catch(Exception ex){}
    }
    
    public void storePropertiesFile(String fileName){
        try
        {
            properties.store(new FileOutputStream(fileName), null);
        }catch(Exception ex){}       
    }
    
    public String getPropertiesValue(String keyValue){
        try{
            String value = properties.getProperty(keyValue);
            return value;
        }catch(Exception ex){}
        return null;
    }

    public void putPropertiesValue(String key, String value){
        try{
            properties.setProperty(key, value);            
        }catch(Exception ex){}        
    }    
}
