/*
 * TableManager.java
 *
 * Created on 25 febbraio 2009, 15.14
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package romas.com.data;

import eve.database.*;
import java.util.Enumeration;

import romas.com.data.tables.*;
import romas.com.data.tables.foreignkeys.*;
import java.util.logging.*;
import java.util.Vector;
import romas.util.LogWriter;
import romas.com.data.Structures.*;

/**
 *
 * @author APaliani
 */
public class TableManager {
    
    // variabile di log
    
    public static final String NOTIFICATIONS_MAP_TABLE = "NOTIFICATIONS_MAP_TABLE";
    public static final String DIRECTIVES_TABLE = "DIRECTIVES_TABLE";
    public static final String ACTION_SCOPE_TABLE = "ACTION_SCOPE_TABLE";
    public static final String ACTIONS_PROCESS_TABLE = "ACTIONS_PROCESS_TABLE";
    public static final String ACTIONS_PROCESS_HISTORY_TABLE = "ACTIONS_PROCESS_HISTORY_TABLE";
    public static final String ACTIONS_TABLE = "ACTIONS_TABLE";
    
   
    public static Database dataACTION_SCOPE_TABLE;
    public static Database dataACTIONS_TABLE;
    public static Database dataACTIONS_PROCESS_TABLE;
    public static Database dataACTIONS_PROCESS_HISTORY_TABLE;
    public static Database dataDIRECTIVES_TABLE;
    public static Database dataNOTIFICATIONS_MAP_TABLE;
    
    public static volatile boolean lock = false;
    
    /** Creates a new instance of TableManager */
    public TableManager() {
    }         

    
    /**
     *Aggiunge nuovo record
     */
    public static int addNew(Object newObject, Database database)throws Exception{
        try{
            while(lock){
                Thread.sleep(100);
            }            
            lock = true;
            
            int index = addObject(newObject, database);
            
            LogWriter.writeLog(Level.FINER, "Insert - " + TABLE.toString(newObject));
            return index;
        }
        finally{
            lock = false;
        }
    }

    
    static int addObject(Object newObject, Database database)throws Exception{
        FoundEntries fe = database.getEmptyEntries(); 
        int index = fe.addData(newObject);

        saveDatabase(database);
        return index;
    }
    
    
    static void deleteObject(int indexObject, Database database)throws Exception{
        FoundEntries fe = database.getEmptyEntries(); 
        fe.delete(indexObject);        
    }
    
    
    /**
     *Elimina un record
     */
    public static void delete(int indexObject, Object object, Database database)throws Exception{        
        try{
            while(lock){
                Thread.sleep(100);
            }
            lock = true;
            deleteObject(indexObject, database);
            LogWriter.writeLog(Level.FINER, "Deleted - " + TABLE.toString(object));
        }
        finally{
            lock = false;
        }
    }
   
    
    /**
     *Ritorna un record
     */
    public Object getData(int indexObject, Database database)throws Exception{
        try{
            while(lock){
                Thread.sleep(100);
            }
            lock = true;
            FoundEntries fe = database.getEmptyEntries();
            return fe.getData(indexObject);
        }
        finally{
            lock = false;
        }
    }
    
    
    /**
     *Modifica un record esistente
     */
    public int modify(int idObject, Object object, Database database)throws Exception{
        try{
            while(lock){
                Thread.sleep(100);
            }
            lock = true;
            FoundEntries fe = database.getEmptyEntries(); 
            LogWriter.writeLog(Level.FINER, "Update - " + TABLE.toString(object));
            return fe.setData(idObject, object);
        }
        finally{
            lock = false;
        }
    }
        
    static EntriesView search(Database database, String sortName, Object param, boolean descending)throws Exception{
        int sortId = database.findSort(sortName); 
        FoundEntries data = database.getFoundEntries(null, sortId); 

        if(data == null) return null;

        if(data.size() == 0)return null;

        EntriesView entriesView = data.getEmptyView();
        entriesView = entriesView.search(null, param);        
        entriesView.sort(descending);
        return entriesView;        
    }
    
    
    static EntriesView getAll(Database database, String sortName, boolean descending)throws Exception{
        int sortId = database.findSort(sortName); 
        FoundEntries data = database.getFoundEntries(null, sortId); 

        if(data == null) return null;
        if(data.size() == 0)return null;

        EntriesView entriesView = data.getEmptyView(); 
        entriesView.getFullFiew();
        entriesView.sort(descending);
        return entriesView;
    }
    
    /**
     *Ricerca un sottoinsieme di dati eseguendo un filtro ed ordina la lista per il sort name specificato
     * @param database
     * @param sortName
     * @param param
     * @param descending
     * @return
     * @throws Exception
     */
    public static EntriesView searchData(Database database, String sortName, Object param, boolean descending)throws Exception{
        try{
            while(lock){
                Thread.sleep(100);
            }
            lock = true;            

            return search(database, sortName, param, descending);
        }
        finally{
            lock = false;
        }
    }
    
    
    public static void deleteAndArchiveACTION_PROCESS(ACTIONS_PROCESS actionProcess){
        try{
            while(lock){
                Thread.sleep(100);
            }
            lock = true;            
                    
            // create history
            ACTIONS_PROCESS_HISTORY actionProcessHistory = new ACTIONS_PROCESS_HISTORY(actionProcess);

            // retrive all action process histroy
            EntriesView evHistory = TableManager.getAll(TableManager.dataACTIONS_PROCESS_HISTORY_TABLE, "By DateString", false);
            if(evHistory != null){

                // stored last 5000 ACTIONS
                if(evHistory.size() > 5000){
                    // elimino primo record inserito
                    TableManager.deleteObject(evHistory.size(), TableManager.dataACTIONS_PROCESS_HISTORY_TABLE);
                }
            }
            // aggiungo utlimo record
            TableManager.addObject(actionProcessHistory, TableManager.dataACTIONS_PROCESS_HISTORY_TABLE);
            TableManager.m_deleteACTION_PROCESS(actionProcess);        
        }
        catch(Exception ex){
            LogWriter.writeLog(Level.SEVERE, ex.getMessage());
        }
        finally{
            lock = false;
        }
        
    }
    
    
    /**
     *Ricerca tutti i dati ed ordina la lista per il sort name specificato
     * @param database
     * @param sortName
     * @param descending
     * @return 
     * @throws Exception
     */
    public static EntriesView searchData(Database database, String sortName, boolean descending)throws Exception{
        try{
            while(lock){
                Thread.sleep(100);
            }
            lock = true;
            return getAll(database, sortName, descending);
        }
        catch(Exception ex){
            LogWriter.writeLog(Level.SEVERE, ex.getMessage());
        }
        finally{
            lock = false;
        }
        return null;
    }
    

    /**
     *Ritorna un sub set di dati
     * @param database
     * @param dataSearch
     * @param sortName
     * @param param
     * @param descending
     * @return
     * @throws Exception
     */
    public static EntriesView searchSubData(Database database, EntriesView dataSearch, String sortName, Object param, boolean descending)throws Exception{
        try{
            while(lock){
                Thread.sleep(100);
            }
            lock = true;
            FoundEntries foundData = dataSearch.getFoundEntries().getSubSet(dataSearch);

            int sortId1 = database.findSort(sortName); 
            foundData.sort(null, sortId1);
            EntriesView shortData = foundData.getEmptyView();
            foundData.getFullView(shortData);
            if(shortData.size() > 0){
                shortData = shortData.search(null, param, descending);
                return shortData;
            }
            return null;
        }
        finally{
            lock = false;
        }
    }

            
    /**
     *Crea nuovo database
     */
    public void createDatabase(String tableName, Object object)throws Exception{
        try{
            while(lock){
                Thread.sleep(100);
            }
            lock = true;
            deleteDatabase(tableName);
            DatabaseManager.initializeDatabase(null, tableName, object);
        }
        finally{
            lock = false;
        }
    }
    
    /**
     *Cancella il database
     */
    static void deleteDatabase(String tableName)throws Exception{
        try{
            while(lock){
                Thread.sleep(100);
            }
            lock = true;
            if(DatabaseManager.databaseExists(null, tableName)){
                LogWriter.writeLog(Level.FINER, "Deleting database ... [" + tableName + "]");
                Database database = DatabaseManager.openDatabase(null, tableName, "rw");
                if(database != null){
                    database.delete();
                    LogWriter.writeLog(Level.FINER, "Database deleted:  [" + tableName + "] !");
                }
            }
        }
        finally{
            lock = false;
        }
    }
    
    /**
     *Apre database
     */
    public Database openDatabase(String tableName) throws Exception        {        
        try{
            while(lock){
                Thread.sleep(100);
            }
            lock = true;
            LogWriter.writeLog(Level.FINER, "Opening database ... [" + tableName + "]");
            return DatabaseManager.openDatabase(null, tableName, "rw");
        }
        finally{
            lock = false;
        }
    }
    

    /**
     *Apre database
     */
    static Database openNewDatabase(String tableName, Object object) throws Exception        {        
        LogWriter.writeLog(Level.FINER, "Opening new database ... [" + tableName + "]");
        deleteDatabase(tableName);
        DatabaseManager.initializeDatabase(null, tableName, object);
        return DatabaseManager.openDatabase(null, tableName, "rw");
    }
    
    
    /**
     *Chiude il database
     */
    static void closeDatabase(Database database)throws Exception{
        try{
            while(lock){
                Thread.sleep(100);
            }
            lock = true;
            LogWriter.writeLog(Level.FINER, "Closing database ... ");
            if(database != null){
                database.close();
                LogWriter.writeLog(Level.FINER, "Database closed !");
            }
        }
        finally{
            lock = false;
        }        
    }
    
    /**
     *Chiude il database
     */
    static void saveDatabase(Database database)throws Exception
    {
        LogWriter.writeLog(Level.FINER, "Saving database ... ");
        if(database != null)
        {
            database.save();
            LogWriter.writeLog(Level.FINER, "Database saved !");
        }
    }
    
    
    public static void saveDB(Database database)
    {
        try
        {
            while(lock)
            {
                Thread.sleep(100);
            }
            lock = true;
            saveDatabase(database);
        }
        catch(Exception ex)
        {
            LogWriter.writeLog(Level.SEVERE, ex.getMessage());
        }
        finally
        {
            lock = false;
        }                
    }

    
    /**
     *Creazione di tutte le tabelle
     */
    public static void createAllTables()throws Exception
    {      
        LogWriter.writeLog(Level.FINER, "Creating Tables processing ...");
        /***************************************************************************************************/
        /*    TABELLE  DI ELABORAZIONE   */
        /***************************************************************************************************/
        //ACTIONS_TABLE
        dataACTIONS_TABLE = openNewDatabase(ACTIONS_TABLE, new ACTIONS());        
        //closeDatabase(dataACTIONS_TABLE);
        
        //ACTIONS_PROCESS_TABLE
        dataACTIONS_PROCESS_TABLE = openNewDatabase(ACTIONS_PROCESS_TABLE, new ACTIONS_PROCESS());
        //closeDatabase(dataACTIONS_PROCESS_TABLE);
        
        //ACTIONS_PROCESS_HISTORY_TABLE
        dataACTIONS_PROCESS_HISTORY_TABLE = openNewDatabase(ACTIONS_PROCESS_HISTORY_TABLE, new ACTIONS_PROCESS_HISTORY());
        //closeDatabase(dataACTIONS_PROCESS_HISTORY_TABLE);
        
        //DIRECTIVES_TABLE
        dataDIRECTIVES_TABLE = openNewDatabase(DIRECTIVES_TABLE, new DIRECTIVES());
        //closeDatabase(dataDIRECTIVES_TABLE);

        //NOTIFICATIONS_MAP_TABLE
        dataNOTIFICATIONS_MAP_TABLE = openNewDatabase(NOTIFICATIONS_MAP_TABLE, new NOTIFICATIONS_MAP());
        //closeDatabase(dataNOTIFICATIONS_MAP_TABLE);
        
        
        LogWriter.writeLog(Level.FINER, "Creating foreignkeys tables !");
        /***************************************************************************************************/
        /*    TABELLE  DI RIFERIMENTO  */
        /***************************************************************************************************/
        //SCOPE_TABLE
        dataACTION_SCOPE_TABLE = openNewDatabase(ACTION_SCOPE_TABLE, new ACTION_SCOPE());
        Enumeration enumSCOPE = ACTION_SCOPE.getList().elements();
        while (enumSCOPE.hasMoreElements()) {  
            Object element = enumSCOPE.nextElement();
            if(element != null)
            {
                ACTION_SCOPE scope = (ACTION_SCOPE)element;  
                addNew(scope, dataACTION_SCOPE_TABLE);
            }
        }        
        //closeDatabase(dataSCOPE_TABLE);        
    }
    
        
    /**
     *Aggiornamento della ACTION_PROCESS_TABLE
     * @param actionProcess 
     */
    public static void updateACTION_PROCESS(ACTIONS_PROCESS actionProcess){
        try{
            while(lock){
                Thread.sleep(100);
            }
            lock = true;
            LogWriter.writeLog(Level.FINER, "Updating ACTION_PROCESS_TABLE for ACTION ["+ actionProcess.DATE_STRING_ACTION_PROCESS +"] ...");
            
            EntriesView ev = search(dataACTIONS_PROCESS_TABLE, "By DateString", actionProcess.DATE_STRING_ACTION_PROCESS, false);
            if((ev != null) && (ev.size() > 0)){

                // Update data
                int count = 0;
                for(int j=0; j<ev.size(); j++){
                    ACTIONS_PROCESS temp_actionProcess = (ACTIONS_PROCESS)ev.getData(j);
                    if(temp_actionProcess.DATE_STRING_ACTION_PROCESS.equals(actionProcess.DATE_STRING_ACTION_PROCESS)){
                        ev.setData(j, actionProcess);
                        ev.delete(j);
                        int nI = addObject(actionProcess, TableManager.dataACTIONS_PROCESS_TABLE);
                        LogWriter.writeLog(Level.FINER, "Updated ACTION_PROCESS_TABLE - Old: [index: "+ j +"] ["+ TABLE.toString(temp_actionProcess) +"] !");
                        LogWriter.writeLog(Level.FINER, "Updated ACTION_PROCESS_TABLE - New: [index: "+ nI +"] ["+ TABLE.toString(actionProcess) +"] !");
                        count++;
                    }
                }
                if(count > 1){
                    LogWriter.writeLog(Level.WARNING, "More ACTION_PROCESS [" + count + "] for one identificator:  " + actionProcess.DATE_STRING_ACTION_PROCESS);
                }
                else{
                    LogWriter.writeLog(Level.FINE, "ACTION_PROCESS_TABLE not found - Latest available information :  " + TABLE.toString(actionProcess));
                }

            }
        }
        catch(Exception ex){
            LogWriter.writeLog(Level.WARNING, ex.getMessage() + " - DATA :  " + TABLE.toString(actionProcess));
        }
        finally{
            lock = false;
        }        
    }
    

    /**
     *Cancellazione di una ACTION dalla ACTION_PROCESS_TABLE
     * @param actionProcess
     */
    public static void deleteACTION_PROCESS(ACTIONS_PROCESS actionProcess){
        LogWriter.writeLog(Level.FINER, " Deleting action ["+ actionProcess.DATE_STRING_ACTION_PROCESS +"] ");
        try{
            while(lock){
                Thread.sleep(100);
            }
            lock = true;
            
            m_deleteACTION_PROCESS(actionProcess);
        }
        catch(Exception ex){
            LogWriter.writeLog(Level.SEVERE, ex.getMessage() + " - DATA:  " + TABLE.toString(actionProcess));
        }
        finally{
            lock = false;
        }            
    }


    /**
     * Check if exist
     * @param DATE_STRING
     * @return
     * @throws java.lang.Exception
     */
    public static ACTIONS_PROCESS getACTION_PROCESS_exist(String DATE_STRING)throws Exception{
        try{
            while(lock){
                Thread.sleep(100);
            }
            lock = true;

            EntriesView ev = search(dataACTIONS_PROCESS_TABLE, "By DateString", DATE_STRING, false);
            if(ev != null){
                if(ev.size() > 0){
                    ACTIONS_PROCESS actionProcess = (ACTIONS_PROCESS)ev.get(0).getData();
                    LogWriter.writeLog(Level.FINER, "ACTIONS_PROCESS_TABLE ["+ DATE_STRING +"] exist!");
                    return actionProcess;
                }
            }

        }
        catch(Exception ex){
            LogWriter.writeLog(Level.SEVERE, ex.getMessage() + " - DATE_STRING :  " + DATE_STRING);
        }
        finally{
            lock = false;
        }

        return null;
    }

    /**
     * Check if exist
     * @param DATE_STRING
     * @return
     * @throws java.lang.Exception
     */
    public static boolean checkACTION_PROCESS_exist(String DATE_STRING)throws Exception{
        try{
            while(lock){
                Thread.sleep(100);
            }
            lock = true;

            EntriesView ev = search(dataACTIONS_PROCESS_TABLE, "By DateString", DATE_STRING, false);
            if(ev != null){
                if(ev.size() > 0){
                    LogWriter.writeLog(Level.FINER, "ACTIONS_PROCESS_TABLE ["+ DATE_STRING +"] already exist!");
                    return true;
                }
            }

        }
        catch(Exception ex){
            LogWriter.writeLog(Level.SEVERE, ex.getMessage() + " - DATE_STRING :  " + DATE_STRING);
        }
        finally{
            lock = false;
        }

        return false;
    }

    /**
     * Private method for delete ACTION
     *
     * @param actionProcess
     * @throws java.lang.Exception
     */
    static void m_deleteACTION_PROCESS(ACTIONS_PROCESS actionProcess)throws Exception{
        LogWriter.writeLog(Level.FINER, "Deleting ACTIONS_PROCESS_TABLE for ACTION ["+ actionProcess.DATE_STRING_ACTION_PROCESS +"] ...");

        EntriesView ev = search(dataACTIONS_PROCESS_TABLE, "By DateString", actionProcess.DATE_STRING_ACTION_PROCESS, false);
        if(ev != null){
            for(int i = 0; i < ev.size(); i++){
                ACTIONS_PROCESS actionDeleting = (ACTIONS_PROCESS)ev.get(i).getData();
                if(actionDeleting.DATE_STRING_ACTION_PROCESS.equals(actionProcess.DATE_STRING_ACTION_PROCESS))
                {
                    ev.delete(i);
                    saveDatabase(dataACTIONS_PROCESS_TABLE);
                    LogWriter.writeLog(Level.FINER, "Deleted ACTIONS_PROCESS_TABLE for ACTION ["+ TABLE.toString(actionProcess) +"] !");
                }
                else
                {
                    LogWriter.writeLog(Level.WARNING, " CANNOT DELETE ACTIONS_PROCESS ["+ actionDeleting.DATE_STRING_ACTION_PROCESS +"] <> ["+ actionProcess.DATE_STRING_ACTION_PROCESS +"]:  ");
                }
            }
        }
        else
        {
            LogWriter.writeLog(Level.FINE, " CANNOT DELETE ACTIONS_PROCESS :  " + TABLE.toString(actionProcess));
        }        
    }


    /**
     * Delete ALL action by SCOPE
     *
     * @param scope
     * @return
     * @throws java.lang.Exception
     */
    public static boolean clearAllActions(String scope)throws Exception
    {
        LogWriter.writeLog(Level.FINER, "Clear Scope actions : ["+ scope +"] ...");
        //By Scope
        EntriesView evScope = searchData(TableManager.dataACTIONS_PROCESS_TABLE, "By Scope", scope, false);

        if(evScope == null) {
            return true;
        }

        if(evScope.size() > 0)
        {
            for(int i=0; i<evScope.size(); i++)
            {
                deleteACTION_PROCESS((ACTIONS_PROCESS)evScope.getData(i) );
            }
        }

        return true;
    }


    /**
     *
     *
     * @param scope
     * @param actionName
     * @return
     * @throws java.lang.Exception
     */
    public static boolean clearAllActions(String scope, String actionName)throws Exception
    {
        LogWriter.writeLog(Level.FINER, "Clear Scope actions : ["+ scope +"] ...");
        //By Scope
        EntriesView evScope = searchData(TableManager.dataACTIONS_PROCESS_TABLE, "By Scope", scope, false);

        if(evScope == null) {
            return true;
        }

        if(evScope.size() > 0)
        {
            EntriesView evAction = searchSubData(TableManager.dataACTIONS_PROCESS_TABLE, evScope, "ByName", actionName, false);

            if(evAction == null) {
                return true;
            }

            if(evAction.size() > 0)
            {
                for(int i=0; i<evAction.size(); i++)
                {
                    deleteACTION_PROCESS((ACTIONS_PROCESS)evAction.getData(i) );
                }
            }
        }

        return true;
    }


    /**
     * 
     * @return
     * @throws java.lang.Exception
     */
    public static boolean clearAllActions()throws Exception
    {
        LogWriter.writeLog(Level.FINER, "Clear ALL actions ...");
        //By Scope
        EntriesView evAll = searchData(TableManager.dataACTIONS_PROCESS_TABLE, "By Scope", false);

        if(evAll == null) {
            return true;
        }

        if(evAll.size() > 0)
        {
            for(int i=0; i<evAll.size(); i++)
            {
                deleteACTION_PROCESS((ACTIONS_PROCESS)evAll.getData(i) );
            }
        }

        return true;
    }


}

