
package romas.mobile.video.images;


import eve.ui.*;
import eve.sys.*;
import eve.fx.*;
import eve.ui.game.AniImage;
import eve.io.PNGEncoder;
import java.io.*;
import java.io.InputStream;
import java.net.*;
import romas.mobile.video.ui.VideoScreen;


public class BannerSlider
{

    Form form = null;
    MenuItem stringitem = null;
    public Choice devices;
            
    private Socket socket = null;
    private InputStream is;
    private OutputStream os;
    private boolean running = false;
    
    private final int ERROR = 2;
    private final int WARNING = 1;
    private final int INFO = 0;
    
    private int nDevice = 0;
    private int nPortServer = 0;
    private String destAddr;
    
    private VideoScreen video;
    
    public BannerSlider(VideoScreen displayVideo, String destAddr, int port){
        this.destAddr = destAddr;
        this.nPortServer = port;
        this.video = displayVideo;
    } 
    
    private void showMessage(String message, int type){
        boolean debug = false;
        String sLabel = "Debug: ";        
        switch (type){
            case ERROR:     sLabel = "ERROR: ";
                break;                
            /**
            case INFO:      sLabel = "INFO: ";   
                break;
            case WARNING:   sLabel = "WARNING: ";   
                break;
             */
        }
        
        if(debug)
        {
            new MessageBox(sLabel, message, MessageBox.MBOK).execute();            
        }
        else
        {
            //System.out.println(sLabel + message);
        }
    }
        
    private void sendMessageLength(int lenght)throws Exception{        
        byte[] byteArray = new byte[4];
        byteArray[0] = (byte)((lenght >> 24) & 0xFF);
        byteArray[1] = (byte)((lenght >> 16) & 0xFF);
        byteArray[2] = (byte)((lenght >> 8) & 0xFF);
        byteArray[3] = (byte)(lenght & 0xFF);

        os.write(byteArray);   
        os.flush();
    }
    
    private void sendMessage(String sMessage) throws Exception{        
        showMessage("Send Message ..." + sMessage,INFO);
        sendMessageLength(sMessage.length());
        os.write(sMessage.getBytes());
        os.flush();
    }
      
    private int getLenght()throws Exception{
        /**
        while(is.available() <4){
            Thread.sleep(300);
        }
         */
        byte[] lenBuf = new byte[4];
        is.read(lenBuf, 0, 4);        
        int MAX_LENGTH = arr4int (lenBuf, 0);        
        return MAX_LENGTH;    
    }
    
    public int arr4int (byte[] arr, int start) {  
        /*
        showMessage("arr[start] : " + arr[start], INFO);
        showMessage("arr[start+1] : " + arr[start+1], INFO);
        showMessage("arr[start+2] : " + arr[start+2], INFO);
        showMessage("arr[start+3] : " + arr[start+3], INFO);
        */
        int b1 = (int)arr[start]*256*256*256;
        int b2 = (int)arr[start+1]*256*256;
        int b3 = (int)arr[start+2]*256;
        int b4 = (int)arr[start+3];
        int nInt = b1 + b2 + b3 + b4;
        return nInt;       
    }      
    
    private String getMessage() throws Exception{
               
        String reply = "";             
        showMessage("Wait message ...", INFO);
        /**
        while (!(is.available() > 0)){
            Thread.sleep(300);
        }
        */
        int MAX_LENGTH = getLenght();

        byte[] buf = new byte[MAX_LENGTH];

        int total = 0;

        while (total < MAX_LENGTH) {
            int count = is.read(buf, total, MAX_LENGTH - total);
            if (count < 0) {
                break;
            }
            total += count;
        }

        reply = new String(buf, 0, MAX_LENGTH);
        
        showMessage("Message received !", INFO);
        
        return reply;
    }

    public void setRunning(boolean running){
        this.running = running;
    }

    private void emptyInputStream() throws Exception{
        // empty input stream
        if (is.available() > 0){
            showMessage("Empty input stream: " + is.available() + " bytes", INFO);
            while(is.available() > 0){
                int b = is.read();
            }
        }
    }
    
    public void playVideo(){
        try{
            running = true;
            while(running) {
                sendMessage("PLAY");

                try{
                    renderingImage();
                }
                catch(Exception ex){
                    showMessage("Image refused by: " + ex.getMessage(), WARNING);
                }
                
                //emptyInputStream();
            }
        }catch(Exception ex){
            showMessage("Exception play(): " + ex.getMessage(), ERROR);
        }
    }
    
    public void stopVideo(){
        try{
            running = false;
            sendMessage("STOP");
        }
        catch(Exception ex){
            showMessage("Exception stop(): " + ex.getMessage(), ERROR);
        }
    }

    public void recordStart(){
        try{
            sendMessage("RECORD_START");
            playVideo();
        }
        catch(Exception ex){
            showMessage("Exception stop(): " + ex.getMessage(), ERROR);
        }
    }

    public void recordStop(){
        try{
            sendMessage("RECORD_STOP");
            playVideo();
        }
        catch(Exception ex){
            showMessage("Exception stop(): " + ex.getMessage(), ERROR);
        }
    }
    
    
    public void exit(){
        try{
            running = false;
            sendMessage("EXIT");
        }
        catch(Exception ex){
            showMessage("Exception exit(): " + ex.getMessage(), ERROR);
        }
    }
      
    public void saveScreenShot(String cameraWeb){
        try {
                String fileName = cameraWeb.replace(' ', '_') + ".png";
                //FileWriter fw = new FileWriter(fileName, true);                
                PNGEncoder png = new PNGEncoder(); 
                png.writeImage(new FileOutputStream(fileName), video.videoImage.getImageData());
                png.writeEnd(); 
                png.closeOutput();         
                new MessageBox("ScreenShot", new File(fileName).getAbsolutePath(), MessageBox.MBOK).execute();
        }
        catch (IOException e)
        {
            new MessageBox("Error", e.toString(), MessageBox.MBOK).show();
        }
    }   
        
    private void connect(){
        try {
            showMessage("CONNECTION TO ..." + destAddr, INFO);
           
            socket = new Socket(destAddr, nPortServer);
            //RemoteConnection conn = RemoteConnection.getNewConnection();
            //socket = conn.connectToHost(destAddr, nPortServer, eve.sys.TimeOut.Forever);
            showMessage("CONNECTED to server >> " + socket.getInetAddress() + ":" + socket.getPort(), INFO);
                        
            os = socket.getOutputStream();
            is = socket.getInputStream(); 
            running = true;
            
        } catch (Exception ex) {
            showMessage(ex.getMessage(), ERROR);
        }        
    }
        
    public void connectToRemoteServer() {          
        try {
            if(video != null){            
                
                connect();
                
                Thread.sleep(300);
                
                String message = getMessage();
                
                String msSetup = message.substring(0,5);

                if (msSetup.equals("SETUP")){
                    String nDevices = message.substring(6,8);
                    nDevice = Integer.parseInt(nDevices);                  
                }
                else{
                    showMessage("Message from server: " + message, INFO);
                }
            }
        }
        catch (Exception e) {
            showMessage("Exception doReceive(): " + e.getMessage(), ERROR);
        }
    }
    
    public void setDevice(int index){
        try{
            String sNumDevices = String.valueOf(index);
            if (index < 10){
                sNumDevices = "0" + sNumDevices;
            }         
            sendMessage("SETDEVICE," + sNumDevices);
        }catch (Exception e) {
            showMessage("Exception setDevice(): " + e.getMessage(), ERROR);
        }
    }
    
    public void buildChoiseDevice(String listCameraEmpty){
        int newNDevices = (nDevice + 1);
        String[] items = new String[newNDevices];
        items[0] = listCameraEmpty;
        for(int i=1; i < newNDevices; i++){
            items[i] = "Camera " + (i); 
        }

        devices = new Choice(items, 0);
    }
    
    private ImageData getImage() throws Exception{
        ImageData imageData = null;
        try{      
            /**
            int lenData = 0;            
            try{
                lenData = getLenght();
                if (lenData > 20000 || lenData < 1){
                    throw new Exception("("+ lenData +") not in 1 - 20000 byte ");
                }
            }catch(Exception exi){
                throw new Exception("get lenght image error; " + exi.getMessage());
            }
            */
            byte[] bytedata = new byte[60000];

            int count = 0;
            while(is.read(bytedata) == -1)
            {
                Thread.sleep(10);
                count++;
                if (count == 200)return null;
            }
                       
            ImageBytes imageByte = new ImageBytes(bytedata);
            imageData = imageByte.pngDecode().getImageData();
                        
            showMessage("Recived image " + imageData.getImageWidth() + " * " + imageData.getImageHeight() + ", size: " + bytedata.length + " bytes", INFO);
            
        }catch(Exception ex){
            throw new Exception(ex.getMessage());
        }
        return imageData;
    }    
    
    private void renderingImage() throws Exception{
        //Image 
        showMessage("Wait image ...", INFO);
        ImageData imageData = getImage();
        if (imageData != null)
        {            
            video.setImage(new AniImage(imageData));           
            showMessage("Rendering frame OK !", INFO);
        }
        else
        {
            showMessage("Rendering >> KO << !", WARNING);
        }
    }
   
    public void shutDown(){
        running = false;
    }
    
}