/*
 * ActionService.java
 *
 * Created on 23 febbraio 2009, 12.15
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package romas.services;

import java.util.Hashtable;
import java.io.File;
import java.util.logging.*;

import romas.com.data.*;
import romas.com.data.tables.TABLE;
import romas.com.data.tables.ACTIONS_PROCESS;
import romas.com.data.tables.foreignkeys.ACTION_SCOPE;
import romas.util.LogWriter;
import romas.com.data.Structures.ACTION_STATE;


/**
 *
 * @author APaliani
 */
public abstract class ActionService {
        
    public Configuration config;
    public Hashtable resources;
    
    /** Creates a new instance of ActionService */
    public ActionService() {
    }
    
    
    /**
     *Inizializzazione del servizio
     */
    protected void init(String serviceName)throws Exception{
        String USER_DIR_KEY = "user.dir";
        String currentDir = System.getProperty(USER_DIR_KEY);        
        
        config = new Configuration();
        config.loadPropertiesFile("config" + File.separator + "services" + File.separator + serviceName + ".properties");
        
        config.loadNotificationsXMLFile(currentDir + File.separator + "config" + File.separator + "services" + File.separator + serviceName + "_notifications.xml");
        config.loadActionsXMLFile(currentDir + File.separator + "config" + File.separator + "services" + File.separator + serviceName + "_actions.xml");
        LogWriter.writeLog(Level.INFO, String.format("Init service [%s] version [%s] scope [%s]... ", serviceName, getCurrentVersion(), getScope()));
        
        setActionList(serviceName);
    }
    
            
    /**
     *Metodo da implementare per eseguire il lavoro delle azioni del servizio
     */
    protected abstract boolean actionWork(ACTIONS_PROCESS actionProcess);

    
    /**
     *Ritorna la priorit del Servizio
     *per impostare la priorit del thread
     */
    public abstract int getPriorityScope();


    /**
     *
     * @return current version service
     */
    public abstract String getCurrentVersion();

        
    /**
     *Ritorna l'ambiente dove opera il servizio
     */
    public final String getScope()throws Exception{
        String scope = config.getPropertiesValue("SCOPE");
        if(!ACTION_SCOPE.getList().containsKey(scope)){
            String msg = String.format("SCOPE [%s] not managed !", scope);
            //logger.log(Level.SEVERE, msg);
            throw new Exception(msg);
        }
        return scope;
    }
    
    
    /**
     *Imposta la lista delle azioni disponibili 
     *lette dal file di configurazione
     */
    public final void setActionList(String serviceName)throws Exception{
        config.readActions(serviceName);        
    }
        
    
    /**
     *Esegue il metodo actionWork in un thread
     */
    public final void executeAction(ACTIONS_PROCESS actionProcess)throws Exception{
        // esegue soltanto le azioni in stato INITIATED
        if(actionProcess.STATE_ACTION_PROCESS == ACTION_STATE.INITIATED_ACTION_STATE){
            LogWriter.writeLog(Level.FINE, "Exec action : " + TABLE.toString(actionProcess));

            Action actionThread = new Action(actionProcess){
                @Override
                public boolean executeWork(ACTIONS_PROCESS actionP){
                    return actionWork(actionP);
                }
            };     

            // essendo le priorit da 0 a 10 nel sistema assegno priorit MAX al servizio con priorit 0
            actionThread.setPriority(Thread.MAX_PRIORITY - this.getPriorityScope());
            actionProcess.ID_PROCESS_ACTIONS_PROCESS = actionThread.getThreadId();
            actionProcess.ACTION = actionThread;
            actionProcess.STATE_ACTION_PROCESS = ACTION_STATE.RUNNING_ACTION_STATE;

            // update database
            TableManager.updateACTION_PROCESS(actionProcess);

            // avvio del thread
            actionThread.start();        
        }
        else{
            LogWriter.writeLog(Level.WARNING, "Cannot execute ACTION on STATE: " + actionProcess.STATE_ACTION_PROCESS);
        }
    }   
}