/*
 * NotificationService.java
 *
 * Created on 23 febbraio 2009, 12.16
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package romas.services;

import java.util.logging.*;
import java.util.List;

import eve.database.*;

import romas.util.*;
import romas.com.data.*;
import romas.com.data.tables.*;
import romas.com.data.Structures.*;



/**
 *
 * @author APaliani
 */
public abstract class NotificationService extends ActionService implements Runnable{
    
    public volatile String m_action;    
    private String m_serviceName;
    public int sleepTime = 10;
    private boolean m_terminated = false;
    private EntriesView m_notifications;
        
    /** Creates a new instance of NotificationService */
    public NotificationService(String serviceName)throws Exception {
        m_serviceName = serviceName;
        initService();
        //LogWriter.writeLog(Level.INFO, "["+ serviceName +"] !");
    }
    

    public String getCurrentVersion()
    {
        return "2011.05.26.0115";
    }

    /**
     *Ritorna il nome del servizio
     */
    public String getServiceName(){
        return m_serviceName;
    }
    
    
    /**
     *Inizializza il servizio   
     */
    private void initService()throws Exception{
        // inizializza il servizio base
        init(m_serviceName);
        
        // carica il file delle notifiche
        storeNotifications();
    }
    
    
    /**
     *Inserisce nella NOTIFICATION_MAP_TABLE LE NOTIFICHE DEL SERVIZIO
     */
    private void storeNotifications()throws Exception{
        // memorizza le notifiche dal file al database
        this.config.readNotifications(this.getScope());

        // memorizza le notifiche in una variabile di cache
        m_notifications = TableManager.searchData(TableManager.dataNOTIFICATIONS_MAP_TABLE, "By Notificator", this.getScope(), false);
    }
    

    /**
     * Ritorna tutte le notifiche del servizio
     * @return
     * @throws java.lang.Exception
     */
    public EntriesView getNotifications()throws Exception{
        return m_notifications;
    }



    /**
     *Metodo da implementare per automatizzare la notifica
     */
    public abstract List<NOTIFICATIONS_MAP> prepareAction();
    
    
    /**
     * Esegue la notifica dell'azione scrivendo nella ACTION_PROCESS_TABLE
     *
     * @param notify
     * @throws Exception
     */
    public final void notifyAction(NOTIFICATIONS_MAP notify)throws Exception{
        LogWriter.writeLog(Level.FINE, TABLE.toString(notify));
        
        notify.NOTIFICATOR = super.getScope();
        notify.FULLNAME_EVENT = super.getScope() + "_" + notify.ACTION_EVENT + "_" + notify.COMMAND_EVENT;
        
        final NOTIFICATIONS_MAP _notify = notify;
                
        Runnable notifyWorkThread = new Runnable(){
            public void run(){
                checkNotify(_notify);
            }
        };

        Thread notifyThread = new Thread(notifyWorkThread);     
        notifyThread.setPriority(this.getPriorityScope());        
        notifyThread.start();
    }



    
    /**
     *Scrive la ACTION_PROCESS_TABLE se esiste una sottoscrizione attiva alla notifica
     * @param notify
     */
    public final void checkNotify(NOTIFICATIONS_MAP notify){
        try{
            // esecuzione dell'azione richiesta se attiva
            // infatti un'azione pu essere notificata soltanto per scatenarne altre
            if(notify.ACTIVE){

                int priority = getPriority(notify.FULLNAME_EVENT);
                ACTIONS_PROCESS actionRequest = new ACTIONS_PROCESS(this.getPriorityScope(),
                                                                    notify.NOTIFICATOR,
                                                                    notify.FULLNAME_EVENT,
                                                                    notify.ACTION_EVENT,
                                                                    notify.COMMAND_EVENT,
                                                                    notify.PARAMS_COMMAND_EVENT,
                                                                    priority,
                                                                    0,
                                                                    ACTION_TYPE.ORDINARY_ACTION_TYPE,
                                                                    0,
                                                                    ACTION_STATE.CREATED_ACTION_STATE,
                                                                    "",
                                                                    Utility.getDateTimeID(),
                                                                    notify.NUMBER_ITERATION_NOTIFY
                                                                    );
                TableManager.addNew(actionRequest, TableManager.dataACTIONS_PROCESS_TABLE);
            }

            // verifico se ci sono azioni sottoscritte alla notifica
            // qui bisogna fare attenzione alle notifiche se scritte male generano malfunzionamenti
            // impostare bene sottoscrittore e notificante
            EntriesView viewNotifications = TableManager.searchData(TableManager.dataNOTIFICATIONS_MAP_TABLE, "By Event", notify.FULLNAME_EVENT, false);
            if((viewNotifications != null) && (viewNotifications.size() > 0)){
                for(int i=0; i < viewNotifications.size(); i++){
                    NOTIFICATIONS_MAP viewNotify = (NOTIFICATIONS_MAP)viewNotifications.get(i).getData();

                    if(viewNotify.ACTIVE){
                        if((viewNotify.PARAMS_COMMAND_NOTIFY != null) && (notify.PARAMS_COMMAND_EVENT != null)){
                            // vengono aggiunti i parametri dell'evento ai parametri della notifica
                            viewNotify.PARAMS_COMMAND_NOTIFY.addAll(notify.PARAMS_COMMAND_EVENT);
                        }

                        int priority = getPriority(viewNotify.FULLNAME_NOTIFY);

                        // esecuzione delle azioni di notifica
                        ACTIONS_PROCESS actionNotify = new ACTIONS_PROCESS(this.getPriorityScope(),
                                                                            viewNotify.SUBSCRIBER,
                                                                            viewNotify.FULLNAME_NOTIFY,
                                                                            viewNotify.ACTION_NOTIFY,
                                                                            viewNotify.COMMAND_NOTIFY,
                                                                            viewNotify.PARAMS_COMMAND_NOTIFY,
                                                                            priority,
                                                                            0,
                                                                            ACTION_TYPE.ORDINARY_ACTION_TYPE,
                                                                            0,
                                                                            ACTION_STATE.CREATED_ACTION_STATE,
                                                                            "",
                                                                            Utility.getDateTimeID(),
                                                                            viewNotify.NUMBER_ITERATION_NOTIFY
                                                                            );

                        TableManager.addNew(actionNotify, TableManager.dataACTIONS_PROCESS_TABLE);
                    }
                }
            }
            
        }
        catch(Exception ex){
            LogWriter.writeLog(Level.SEVERE, ex.getMessage());
        }
    }
    

    /**
     *Ritorna la priorit della ACTION
     */
    public int getPriority(String actionFULLNAME)throws Exception{
        EntriesView view = TableManager.searchData(TableManager.dataACTIONS_TABLE, "By FullName", actionFULLNAME, false);
        if(view.size() > 0) {
            ACTIONS action = (ACTIONS)view.get(0).getData();
            
            return action.PRIORITY_ACTION;
        }
        else{
            LogWriter.writeLog(Level.FINE, "Action FULLNAME_ACTION ["+ actionFULLNAME +"] not found in ACTIONS_TABLE !");
            return 0;
        }
    }
    
    
    /**
     *Metodo che esegue il lavoro
     */
    public void doWork() throws Exception{
        try{
            List notifications = prepareAction();
            if(notifications != null){

                if (notifications.size() > 0){
                    for(int i=0; i<notifications.size(); i++){
                        NOTIFICATIONS_MAP notify = (NOTIFICATIONS_MAP)notifications.get(i);
                        if(notify != null){
                            notifyAction(notify);
                        }            
                    }
                }
            }
        }
        catch(Exception ex){
            throw ex;
        }
    }
    
    
    /**
     * Metodo che esegue il servizio
     */
    public void run(){
        try{
            while(!m_terminated){
                
                // lavora
                doWork();
                
                // attesa
                Thread.sleep(sleepTime);
            }
        }
        catch(Exception ex){
            LogWriter.writeLog(Level.SEVERE, ex.getMessage());
        }        
    }
 
    
    /**
     *Arresta il server
     */
    public void stopService(){
        m_terminated = true;        
    }
}
