/*
 * ActuatorService.java
 *
 * Created on 23 marzo 2009, 16.12
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package romas.services.actuator;

import romas.services.actuator.Definition.*;
import romas.com.data.tables.TABLE;
import romas.com.data.tables.ACTIONS_PROCESS;
import romas.com.data.tables.NOTIFICATIONS_MAP;
import romas.com.data.Structures.SERVICE_TYPE;
import romas.services.NotificationService;
import romas.util.LogWriter;
import romas.util.comunication.port.Serial;
import romas.util.comunication.controller.servo.*;

import java.util.logging.Level;
import java.util.List;



/**
 *
 * @author APaliani
 */
public class ActuatorService extends NotificationService{
    
    Serial  serialPort;
    Head    headRobot;
    int     speed = Robot.getSpeedServoValue(SPEED.NORMAL);
    int     time  = Robot.getTimeServoValue(TIME.NORMAL);
    int     steps = 1;
    int     actualPulse     = 0;
    int     requiredPulse   = 1;
    boolean serialPortResponse = false;

    volatile boolean lock = false;
    
    /** Creates a new instance of ActuatorService
     * @throws Exception
     */
    public ActuatorService() throws Exception{
        super("ActuatorService");
        try{
            //serial port
            String portName = super.config.getPropertiesValue("port_name");
            int baud = Integer.valueOf(super.config.getPropertiesValue("baud_rate")).intValue();

            // create instance of serial port
            serialPort = new Serial(portName, baud, false){
                @Override
                public void processMessage(byte[] buffer) {
                    checkServoMovement(buffer);
                }
            };

            if(serialPort == null) throw new Exception("Cannot Init ActuatorService !");

            // initit Robot Head
            activeHead();
            
        }catch(Exception ex){
            throw new Exception("ActuatorService: " + ex.getMessage());
        }
    }

    public String getCurrentVersion()
    {
        return "2011.05.26.0110";
    }


    /**
     * 
     * @param buffer
     */
    private void checkServoMovement(byte[] buffer){        
        if(buffer.length > 0){
            String resultString = new String(buffer);

            //check query pulse
            try{
                int nPos = buffer[0] & 0xff;
                actualPulse = nPos * 10;
                serialPortResponse = true;
                return;
            }catch(Exception ex){
                actualPulse = 0;
            }

        }
        else{
            LogWriter.writeLog(Level.FINER, "Buffer empty !");
            lock = false;
        }
    }


    /**
     * Active servo Head
     */
    void activeHead(){
        ServoConfiguration servoConfig_H = new ServoConfiguration(super.config.getPropertiesValue("servo_head_H"));
        ServoConfiguration servoConfig_V = new ServoConfiguration(super.config.getPropertiesValue("servo_head_V"));

        // servo configuration
        headRobot = new Head(serialPort, servoConfig_H, servoConfig_V);
    }


    /**
     *For action builder
     *
     * @return the NOTIFICATIONS_MAP record
     */
    public List<NOTIFICATIONS_MAP> prepareAction() {
        return null;
    }



    /**
     * Excecute service actions
     *
     * @param actionProcess : ACTIONS_PROCESS_TABLE record
     * @return true if execute action
     */
    protected boolean actionWork(ACTIONS_PROCESS actionProcess) {
        if(actionProcess == null) return false;

        try{

            while(lock){
                Thread.sleep(50);
            }
            lock = true;

            actualPulse = 0;
            requiredPulse = 1;

            LogWriter.writeLog(Level.FINE, TABLE.toString(actionProcess));

            if(actionProcess.PARAMS_ACTIONS_PROCESS != null && actionProcess.PARAMS_ACTIONS_PROCESS.size() > 0){
                steps = Integer.valueOf(actionProcess.PARAMS_ACTIONS_PROCESS.get(0).toString()).intValue();
                requiredPulse = 0;
            }

            // Set movement speed servo
            if(actionProcess.NAME_ACTION_ACTIONS_PROCESS.equals("SET_SPEED")){
                SPEED  _Speed = Definition.getSpeed(actionProcess.COMMAND_ACTIONS_PROCESS);
                speed  = Robot.getSpeedServoValue(_Speed);
                requiredPulse = 0;
            }

            // Set movement time servo
            if(actionProcess.NAME_ACTION_ACTIONS_PROCESS.equals("SET_TIME")){
                TIME _Time = Definition.getTime(actionProcess.COMMAND_ACTIONS_PROCESS);
                time  = Robot.getTimeServoValue(_Time);
                requiredPulse = 0;
            }

            // Manage Head actions
            if(actionProcess.NAME_ACTION_ACTIONS_PROCESS.equals("HEAD")){

                requiredPulse = headRobot.getRequiredPulse(actionProcess);

                int counter = 0;
                //check servo position store the info to actualPulse variable
                while(!(actualPulse == requiredPulse) && counter < 50){
                    // move limb
                    headRobot.actuateAction(actionProcess, speed, steps);

                    // send command
                    serialPort.sendCommand(true, time);

                    Thread.sleep(time);

                    LogWriter.writeLog(Level.FINER,"actualPulse: ["+actualPulse+"] == requiredPulse: ["+requiredPulse+"]");
                    LogWriter.writeLog(Level.FINER, "Executed action ["+ counter +"]: " + TABLE.toString(actionProcess));
                    serialPortResponse = false;

                    if(!serialPortResponse){
                        headRobot.sendQueryPosition(actionProcess);
                        // wait serial response
                        Thread.sleep(5);
                    }
                    counter++;
                }
            }

        }catch(Exception ex){
            LogWriter.writeLog(Level.SEVERE, ex.getMessage());
            return false;
        }
        finally{
            lock = false;
        }

        boolean result = (actualPulse == requiredPulse);

        if(!result){LogWriter.writeLog(Level.WARNING, "Action not executed: " + TABLE.toString(actionProcess));}

        return result;
    }
    
    public int getPriorityScope() {
        return SERVICE_TYPE.SECONDARY_SERVICE;
    }
}
