/*
 * ClientStreamSocket.java
 *
 * Created on 24 febbraio 2009, 16.11
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package romas.services.comunications;

import java.net.*;
import java.util.Vector;
import java.util.Date;
import java.util.logging.Level;
import java.util.StringTokenizer;
import romas.com.data.tables.NOTIFICATIONS_MAP;
import romas.util.LogWriter;
/**
 *
 * @author APaliani
 */
    // Thread che dialoga con il client
class ClientComunicationSocket extends Thread{
        
        ComunicationService server;
        Socket clientSocket;

        public int lastImageLength = 0;
        public int indexDeviceSelected = 0;

        final static int INIT = 0;
        final static int READY = 1;
        final static int PLAYING = 2;
        //rtsp message types
        final static int SETUP = 3;
        final static int PLAY = 4;
        final static int PAUSE = 5;
        final static int RECORD_START = 6;
        final static int RECORD_STOP = 7;
        final static int EXIT = 8;
        
        int state;       
        int cError = 0;
      
        String fileMovName;
        Vector inputFiles;
        String cameraName;
        Date dateFirstDetectedMotion;
        String dir = System.getProperty("user.dir");
        String rootFile = null;
        
        /**
         *Costruttore
         */
        public ClientComunicationSocket(Socket clientSocket, ComunicationService server){
            try{
                this.server = server;
                this.clientSocket = clientSocket;
                state = SETUP;

                sendMessage("SETUP");

//                if (server.getAvailableActions().size() == 0){
//                    // send message no device available
//                    sendMessage("NO ACTIONS AVAILABLE");
//                }
            }catch(Exception ex){
                LogWriter.writeLog(Level.SEVERE, "Cannot inizialize ClientComunicationSocket: " + ex.getMessage());
            }    
        }


        public void processMessage(String message){
            try{
                // Costruisce la Action da notific in relazione al messaggio ricevuto
                StringTokenizer tokens = new StringTokenizer(message, ",");

                NOTIFICATIONS_MAP notify = new NOTIFICATIONS_MAP();
                notify.ACTIVE = true;
                try{
                    // SCOPE
                    String scope = tokens.nextToken();
                    notify.SUBSCRIBER = scope;
                    //ACTION
                    notify.ACTION_EVENT = tokens.nextToken();
                    //COMMAND
                    notify.COMMAND_EVENT = tokens.nextToken();
                    //PARAM_1
                    java.util.Vector params = new java.util.Vector();
                    params.add(tokens.nextToken());
                    notify.PARAMS_COMMAND_EVENT = params;
                }catch(Exception ex){}
                this.server.notifyAction(notify);
            }
            catch(Exception ex){
                
            }
        }
        
        /**
         *Avvia il thread di comunicazione con il client
         */
        public void run() {            
            try
            {
                // wait response from client
                while(state == SETUP){
                    sendMessage("SETUP");
                    state = parse_request();               
                }

                while(state != EXIT){                  
                    state = parse_request();                    
                }

                if(clientSocket != null){ 
                    closeClientConnection();
                }
            }
            catch(Exception ex)
            {
                LogWriter.writeLog(Level.SEVERE, ex.getMessage());
            }                            
        }
        

        /**
         *Chiude la connessione verso il client
         */
        private void closeClientConnection()throws Exception{
            if(clientSocket != null)
            {
                if(!clientSocket.isClosed())
                {
                    clientSocket.close();
                }                    
            }

            LogWriter.writeLog(Level.WARNING, "Close connection by Client : " + clientSocket.getInetAddress());
            
            this.join();
        }
        
        
        /**
         *Rileva il messaggio dall'output stream del client
         */
        private String getMessage() throws Exception{

            String reply = "";      

            while (!(clientSocket.getInputStream().available()>0) ){
                sleep(5);
            }

            int MAX_LENGTH = getlength();

            byte[] buf = new byte[MAX_LENGTH];

            int total = 0;

            while (total < MAX_LENGTH) {
                int count = clientSocket.getInputStream().read(buf, total, MAX_LENGTH - total);
                if (count < 0) {
                    break;
                }
                total += count;
            }

            reply = new String(buf, 0, MAX_LENGTH);

            return reply;
        }                
        
        /**
         *Decodifica la richiesta del client
         */
        private int parse_request(){
            int request_type = -1;
            String RequestLine = "?";
            try{
              
              //parse request line and extract the request_type:
              RequestLine = getMessage();

              if (!RequestLine.equals("")){
                  
                  StringTokenizer tokens = new StringTokenizer(RequestLine, ";");
                  String request_string = tokens.nextToken();

                  //convert to request_type structure:
                  if ((new String(request_string)).compareTo("SETUP") == 0)
                    request_type = SETUP;
                  else if ((new String(request_string)).compareTo("EXIT") == 0)
                    request_type = EXIT;
                  // processa la richiesta
                  else processMessage(request_string);

                  LogWriter.writeLog(Level.FINE, "Request from "+ clientSocket.getInetAddress() +" : " + RequestLine);
              }
            }
            catch(Exception ex)
            {
                LogWriter.writeLog(Level.SEVERE, "ClientComunicationSocket "+ clientSocket.getInetAddress() +" exception in parse request: " + ex.getMessage());
                request_type = EXIT;
            }
            return(request_type);
        }                

        /**
         * Invia le azioni disponibili
         */
        private void sendAvailableActions(){
            try{

            }catch(Exception ex){
                LogWriter.writeLog(Level.SEVERE, "Error in send message: " + ex.getMessage());
            }
        }
        
        private int getlength()throws Exception{
            byte[] lenBuf = new byte[4];
            clientSocket.getInputStream().read(lenBuf, 0, 4);
            int MAX_LENGTH = arr4int (lenBuf, 0);
            return MAX_LENGTH;    
        }

        public int arr4int (byte[] arr, int start) {        
            int b1 = (int)arr[start]*256*256*256;
            int b2 = (int)arr[start+1]*256*256;
            int b3 = (int)arr[start+2]*256;
            int b4 = (int)arr[start+3];
            int nInt = b1 + b2 + b3 + b4;
            return nInt;       
        }    
                        
        private void sendMessage(String sMessage){
            try{
                int length = sMessage.length();
                sendMessageLength(length);
                clientSocket.getOutputStream().write(sMessage.getBytes());                
                clientSocket.getOutputStream().flush();
                LogWriter.writeLog(Level.FINE, "Send message '"+ sMessage +"' to : " + clientSocket.getInetAddress() + ", port: " + clientSocket.getPort() + ", message length: " + length);
            
            }catch(Exception ex){
                LogWriter.writeLog(Level.SEVERE, "Error in send message: " + ex.getMessage());
            }
        }
                
        private void sendMessageLength (int length)throws Exception{        
            byte[] byteArray = new byte[4];
            byteArray[0] = (byte)((length >> 24) & 0xFF);
            byteArray[1] = (byte)((length >> 16) & 0xFF);
            byteArray[2] = (byte)((length >> 8) & 0xFF);
            byteArray[3] = (byte)(length & 0xFF);
            
            clientSocket.getOutputStream().write(byteArray);  
            //clientSocket.getOutputStream().flush();
            LogWriter.writeLog(Level.FINE, "Send message length: " + arr4int(byteArray, 0));
        }
    }
