/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package romas.services.sensor;

/**
 *
 * @author apaliani
 */

import romas.com.data.Structures.SERVICE_TYPE;
import romas.com.data.tables.TABLE;
import romas.com.data.tables.ACTIONS_PROCESS;
import romas.com.data.tables.NOTIFICATIONS_MAP;
import romas.services.NotificationService;
import romas.util.LogWriter;
import java.util.ArrayList;

import java.util.logging.Level;
import java.util.List;
import java.util.Hashtable;
import romas.services.resources.boards.Arduino2009;

public class SensorService extends NotificationService{

    Arduino2009 m_arduino2009;

    public SensorService() throws Exception{
        super("SensorService");        
    }

    public String getCurrentVersion()
    {
        return "2011.06.13.0049";
    }

    /**
     *For action builder
     * Viene utilizzato questo metodo per rilevare gli allarmi dei sensori
     *
     * @return the NOTIFICATIONS_MAP record
     */
    public List<NOTIFICATIONS_MAP> prepareAction() {
        List<NOTIFICATIONS_MAP> listNotifications;

        try
        {
            m_arduino2009 = (Arduino2009)super.resources.get("arduino2009");

            if(m_arduino2009 != null)
            {
                if(!m_arduino2009.busy && !m_arduino2009.sValue.equals("") && !m_arduino2009.processedSensor.equals(""))
                {
                    listNotifications = new ArrayList();

                    // notifico il valore campionato
                    NOTIFICATIONS_MAP notify = new NOTIFICATIONS_MAP();
                    notify.NOTIFICATOR = this.getScope();
                    notify.ACTIVE = false; // notifica non attiva perchè non deve essere eseguita un'azione ma soltanto notificato un evento
                    notify.ACTION_EVENT = m_arduino2009.processedSensor;
                    notify.COMMAND_EVENT = "VALUE";
                    java.util.Vector params = new java.util.Vector();
                    params.add(m_arduino2009.sValue);
                    notify.PARAMS_COMMAND_EVENT = params;

                    listNotifications.add(notify);

                    return listNotifications;
                }

            }
        }catch(Exception ex){
            LogWriter.writeLog(Level.SEVERE, ex.getMessage());
        }finally{
            if(m_arduino2009 != null)
            {
                // per annullare l'allarme azzero i valori campionati
                m_arduino2009.sValue = "";
                m_arduino2009.processedSensor = "";
                listNotifications = null;
            }
        }
        return null;
    }




    /**
     * Viene utilizzato questo metodo per effettuare delle interrogazioni
     * puntuali sui sensori
     *
     * Execute action
     * @param actionProcess
     * @return
     */
    protected boolean actionWork(ACTIONS_PROCESS actionProcess) {

        m_arduino2009 = (Arduino2009)resources.get("arduino2009");
        
        if(actionProcess == null) return false;

        LogWriter.writeLog(Level.FINE, TABLE.toString(actionProcess));

        String sensorName = config.getPropertiesValue(actionProcess.NAME_ACTION_ACTIONS_PROCESS);
        String commandName = config.getPropertiesValue(sensorName);

        try{
            // verifico se la scheda  impegnata da un altro servizio
            while(m_arduino2009.busy){
                Thread.sleep(50);
            }
            m_arduino2009.busy = true;
            m_arduino2009.sendCommand(commandName);

            // verifico se la scheda ha elaborato il dato
            while(m_arduino2009.lock){
                Thread.sleep(50);
            }

            // verifico se sto esaminando il sensore richiesto
            if(commandName.equals(m_arduino2009.processedSensor)){
                // la notifica avviene in automatico sfruttando il listener del servizio
                // ritorna il risultato
                return true;
            }

        }catch(Exception ex){
            LogWriter.writeLog(Level.SEVERE, ex.getMessage());
        }
        finally{
            // comunque libero la scheda
            m_arduino2009.busy = false;
        }
        // ritorna false se non  andata a buon fine la campionatura
        return false;
    }


    /**
     * Set priority scope
     */
    public int getPriorityScope() {
        return SERVICE_TYPE.PRIMARY_SERVICE;
    }
}
