/*
 * ClientStreamSocket.java
 *
 * Created on 24 febbraio 2009, 16.11
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package romas.services.videostream;

import java.net.*;
import java.util.Vector;
import java.util.Date;
import java.util.logging.Level;
import javax.media.*;
import java.util.StringTokenizer;
import java.text.SimpleDateFormat;
import java.io.File;
import java.awt.image.BufferedImage;
import java.io.ByteArrayOutputStream;
import javax.imageio.ImageIO;

import romas.com.data.tables.NOTIFICATIONS_MAP;
import romas.util.LogWriter;
/**
 *
 * @author APaliani
 */
    // Thread che dialoga con il client
class ClientStreamSocket extends Thread{
        
        VideoStreamService server;
        Socket clientSocket;

        public int lastImageLength = 0;
        public int indexDeviceSelected = 0;

        final static int INIT = 0;
        final static int READY = 1;
        final static int PLAYING = 2;
        //rtsp message types
        final static int SETUP = 3;
        final static int PLAY = 4;
        final static int PAUSE = 5;
        final static int RECORD_START = 6;
        final static int RECORD_STOP = 7;
        final static int EXIT = 8;
        
        int state;       
        int cError = 0;
        Device device;        
        String fileMovName;
        Vector inputFiles;
        String cameraName;
        Date dateFirstDetectedMotion;
        String dir = System.getProperty("user.dir");
        String rootFile = null;
        
        /**
         *Costruttore
         */
        public ClientStreamSocket(Socket clientSocket, VideoStreamService server){
            
            try{
                
                this.server = server;
                this.clientSocket = clientSocket;
                state = SETUP;
                if (server.deviceList.size() == 0){
                    // send message no device available
                    sendMessage("NO DEVICES AVAILABLE");
                }              
            }catch(Exception ex){
                LogWriter.writeLog(Level.SEVERE, "Cannot inizialize ClientStreamSocket: " + ex.getMessage());
            }    
        }
        
        
        /**
         *Avvia il thread di comunicazione con il client
         */
        public void run() {            
            try
            {
                // wait response from client
                while(state == SETUP){
                    sendMessage("SETUP," + server.getDeviceNumber());
                    state = parse_request();               
                }

                int c = 0;
                while(state != EXIT){                  
                    switch (state){
                        case PLAY:
                            sendImage();
                            break;

                        case RECORD_START:
                            startSavingVideo();
                            //state = PLAY;
                            break;

                        case RECORD_STOP:
                            stopSavingVideo();
                            //state = PLAY;
                            break;
                            
                        default:
                            Thread.sleep(300);
                            break;
                    }     

                    state = parse_request();                    
                }

                if(clientSocket != null){ 
                    closeClientConnection();
                }
            }
            catch(Exception ex)
            {
                LogWriter.writeLog(Level.SEVERE, ex.getMessage());
            }                            
        }
        

        /**
         *Chiude la connessione verso il client
         */
        private void closeClientConnection()throws Exception{
            if(clientSocket != null)
            {
                if(!clientSocket.isClosed())
                {
                    clientSocket.close();
                }                    
            }

            LogWriter.writeLog(Level.WARNING, "Close connection by Client : " + clientSocket.getInetAddress());
            
            this.join();
        }
        
        /**
         *Imposta il dispositivo video dal quale perelevare il flusso
         */
        private void setDevice(){
            try{
                device = (Device)this.server.deviceList.get(indexDeviceSelected);      
                cameraName = String.valueOf(indexDeviceSelected + 1);
            }catch(Exception ex){
                LogWriter.writeLog(Level.SEVERE, "Error in set device: " + ex.getMessage());
            }
        }
        
        /**
         *Rileva il messaggio dall'output stream del client
         */
        private String getMessage() throws Exception{

            String reply = "";      

            int waitCount = 0;
            while (!(clientSocket.getInputStream().available()>0) ){
                sleep(50);
                waitCount++;
                if(waitCount == 20){
                    LogWriter.writeLog(Level.WARNING, "force PLAY message !");
                    return "PLAY";
                }
            }

            int MAX_LENGTH = getlength();

            byte[] buf = new byte[MAX_LENGTH];

            int total = 0;

            while (total < MAX_LENGTH) {
                int count = clientSocket.getInputStream().read(buf, total, MAX_LENGTH - total);
                if (count < 0) {
                    break;
                }
                total += count;
            }

            reply = new String(buf, 0, MAX_LENGTH);

            return reply;
        }                
        
        /**
         *Decodifica la richiesta del client
         */
        private int parse_request(){
            int request_type = -1;
            String RequestLine = "?";
            try{
              
              //parse request line and extract the request_type:
              RequestLine = getMessage();

              if (!RequestLine.equals("")){
                  
                  StringTokenizer tokens = new StringTokenizer(RequestLine, ",");
                  String request_type_string = tokens.nextToken();

                  //convert to request_type structure:
                  if ((new String(request_type_string)).compareTo("SETUP") == 0)
                    request_type = SETUP;
                  else if ((new String(request_type_string)).compareTo("PLAY") == 0){
                    request_type = PLAY;
                  }
                  else if ((new String(request_type_string)).compareTo("SETDEVICE") == 0){
                    request_type = PAUSE;
                    indexDeviceSelected = Integer.parseInt(tokens.nextToken()) - 1;
                    setDevice();
                  }
                  
                  else if ((new String(request_type_string)).compareTo("PAUSE") == 0)
                    request_type = PAUSE;
                  else if ((new String(request_type_string)).compareTo("EXIT") == 0)
                    request_type = EXIT;
                  else if ((new String(request_type_string)).compareTo("RECORD_START") == 0)
                    request_type = RECORD_START;
                  else if ((new String(request_type_string)).compareTo("RECORD_STOP") == 0)
                    request_type = RECORD_STOP;

                  LogWriter.writeLog(Level.FINE, "Request from "+ clientSocket.getInetAddress() +" : " + RequestLine);
              }
            }
            catch(Exception ex)
            {
                LogWriter.writeLog(Level.SEVERE, "ClientStreamSocket "+ clientSocket.getInetAddress() +" exception in parse request: " + ex.getMessage());
                request_type = EXIT;
            }
            return(request_type);
        }                
        
        /**
         *Salva il frame in una immagine
         */
        public String saveScreenShot(Buffer buffer)
        {
            SimpleDateFormat timeFormat =  new SimpleDateFormat("HHmmssS");
            SimpleDateFormat dateFormat =  new SimpleDateFormat("yyyyMMdd");

            Date date = new Date();

            File dirJpg = new File(rootFile);
            if(!dirJpg.exists() && !dirJpg.isDirectory())
            {
                dirJpg.mkdir();
            }

            String fileImageName = rootFile + File.separator + timeFormat.format(date) + ".jpg";
            ManageBufferImage.convertAndSaveJPEGFile(buffer, fileImageName); 
            return fileImageName; 
        }        
        
        /**
         *Rileva i byte del frame
         */
        private byte[] getDataImageFrame() throws Exception{        
            byte[] dataImage = new byte[0];
                    
            java.awt.Image image = null;
            Buffer buffer = null;
            
            while (buffer == null){
                buffer = device.frameAccess.getBuffer();
                sleep(1);
            }
            
            if (buffer != null){
                image = ManageBufferImage.bufferToImage(buffer);
                
                // if recording add frame to file
                if(inputFiles != null)
                {
                    inputFiles.add(saveScreenShot(buffer));
                }
                
                BufferedImage img = (BufferedImage)image;
                ByteArrayOutputStream bas = new ByteArrayOutputStream();
                ImageIO.write(img, "jpg", bas);
                dataImage = bas.toByteArray();
                
                checkMotionDetection(dataImage.length, buffer);
            }
            
            return dataImage;
        }
                
        /**
         *Invia un frame al client
         */
        private synchronized void sendImage(){            
            try{                                
                byte[] dataImage = getDataImageFrame();

                int len = dataImage.length;

                if (len > 0)
                {
                    if (!clientSocket.isClosed())
                    {                        
                        // send immage
                        clientSocket.getOutputStream().write(dataImage);
                        clientSocket.getOutputStream().flush();
                        
                        LogWriter.writeLog(Level.FINE, "Send frame to : " + clientSocket.getInetAddress() + ", port: " + clientSocket.getPort() + ",  length (byte): " + String.valueOf(len));
                        cError = 0;
                    }
                    else 
                    {
                        closeClientConnection();
                    }
                }
                else{

                    LogWriter.writeLog(Level.FINE, "Frame zero to : " + clientSocket.getInetAddress() + ", port: " + clientSocket.getPort());
                }
            }catch(Exception ex){
                LogWriter.writeLog(Level.SEVERE, "Error in send frame: " + ex.getMessage() + " to : " + clientSocket.getInetAddress() + ", port: " + clientSocket.getPort());
                cError++;
                if (cError == 50)
                {
                    try
                    {
                        closeClientConnection();
                    }
                    catch(Exception exc)
                    {
                        LogWriter.writeLog(Level.SEVERE, exc.getMessage());
                    }
                }
                else
                {
                    LogWriter.writeLog(Level.FINE, "Re send image after error ... ");
                    sendImage();
                }
            }                
        }
                
        /**
         *Rileva la presenza di movimento
         */
        private void checkMotionDetection(int len, Buffer buffer)throws Exception{
            if(lastImageLength > 0)
            {
                int marginLength = Integer.parseInt(String.valueOf(Math.round(lastImageLength * 0.05)));
                int downLength = lastImageLength - marginLength;
                int upLength = lastImageLength + marginLength;
                                
                if((len < downLength) || (len > upLength))
                {      
                    if(dateFirstDetectedMotion == null) dateFirstDetectedMotion = new Date();                    
                    
                    int comparedDate = 1;//(new Date()). - dateFirstDetectedMotion;
                    
                    if(comparedDate > 0)
                    {   
                        // notify alert image changing
                        NOTIFICATIONS_MAP notify = new NOTIFICATIONS_MAP();
                        notify.ACTION_EVENT = "MOTION_DETECTION";
                        notify.COMMAND_EVENT = "DETECTED";
                        
                        notify.PARAMS_COMMAND_EVENT.add("Camera " + cameraName.toString());
                        notify.PARAMS_COMMAND_EVENT.add(buffer);
                        server.notifyAction(notify);                        
                    }
                }
            }
            lastImageLength = len;
        }
                
        public void startSavingVideo(){
            
            try
            {                            
                SimpleDateFormat timeFormat =  new SimpleDateFormat("HHmmssS");
                SimpleDateFormat dateFormat =  new SimpleDateFormat("yyyyMMdd");
                
                Date date = new Date();
                rootFile = dir + File.separator + "mov" + File.separator + cameraName + File.separator + dateFormat.format(date);
                File dirMov = new File(rootFile);
                if(!dirMov.exists() && !dirMov.isDirectory())
                {
                    dirMov.mkdir();
                }

                fileMovName = rootFile + File.separator + timeFormat.format(date) + ".mov";

                inputFiles = new Vector();
            }
            catch(Exception ex)
            {
                LogWriter.writeLog(Level.SEVERE, ex.getMessage());
            }   
        }

        public void stopSavingVideo() {
            try{                
                JpegImagesToMovie imageToMovie = new JpegImagesToMovie();
                
                imageToMovie.doIt(320, 240, 15, inputFiles, JpegImagesToMovie.createMediaLocator(fileMovName));
                
                inputFiles = null;
            }
            catch(Exception ex)
            {
                LogWriter.writeLog(Level.SEVERE, ex.getMessage());
            }               
        }        
        
        private int getlength()throws Exception{
            byte[] lenBuf = new byte[4];
            clientSocket.getInputStream().read(lenBuf, 0, 4);
            int MAX_LENGTH = arr4int (lenBuf, 0);
            return MAX_LENGTH;    
        }

        public int arr4int (byte[] arr, int start) {        
            int b1 = (int)arr[start]*256*256*256;
            int b2 = (int)arr[start+1]*256*256;
            int b3 = (int)arr[start+2]*256;
            int b4 = (int)arr[start+3];
            int nInt = b1 + b2 + b3 + b4;
            return nInt;       
        }    
                        
        private void sendMessage(String sMessage){
            try{
                
                int length = sMessage.length();
                sendMessageLength(length);
                clientSocket.getOutputStream().write(sMessage.getBytes());                
                clientSocket.getOutputStream().flush();
                LogWriter.writeLog(Level.FINE, "Send message '"+ sMessage +"' to : " + clientSocket.getInetAddress() + ", port: " + clientSocket.getPort() + ", message length: " + length);
            
            }catch(Exception ex){
                LogWriter.writeLog(Level.SEVERE, "Error in send message: " + ex.getMessage());
            }
        }
                
        private void sendMessageLength (int length)throws Exception{        
            byte[] byteArray = new byte[4];
            byteArray[0] = (byte)((length >> 24) & 0xFF);
            byteArray[1] = (byte)((length >> 16) & 0xFF);
            byteArray[2] = (byte)((length >> 8) & 0xFF);
            byteArray[3] = (byte)(length & 0xFF);
            
            clientSocket.getOutputStream().write(byteArray);  
            //clientSocket.getOutputStream().flush();
            LogWriter.writeLog(Level.FINE, "Send message length: " + arr4int(byteArray, 0));
        }
    }
