/*
 * Speaker.java
 *
 * Created on 24 febbraio 2009, 11.18
 *
 * Classe speicializzata all'utilizzo del sintetizzatore eSpeak !
 */

package romas.services.voice;


import romas.util.LogWriter;

import java.util.logging.Level;
import java.util.LinkedList;
import java.util.Queue;

/**
 *
 * @author Armanod Paliani
 */

/**
 *
 * @author paliani
 */
public class Speaker extends Thread{

    String m_eSpeakFileAbsolutePath = "";
    String m_eSpeakSpeed = "";
    String m_eSpeakTone = "";
    String m_eSpeakLang = "";
    
    public boolean isBusy = false;

    private int m_sleepTime = 300;

    public enum TYPE{
        TEXT,
        FILE
    }

    private volatile Queue queue;


    class TextDoSpeak
    {
        public String message;        
        public TYPE type;

        private TextDoSpeak(String message, TYPE type) {
            this.message = message;
            this.type = type;
        }
    }


    /** Creates a new instance of Speaker */
    public Speaker(String eSpeakFileAbsolutePath, 
                    String eSpeak_Tone,
                    String eSpeak_Speed,
                    String eSpeak_Lang,
                    int sleepTime) {
        this.m_eSpeakFileAbsolutePath   = eSpeakFileAbsolutePath;
        this.m_eSpeakTone               = eSpeak_Tone;
        this.m_eSpeakSpeed              = eSpeak_Speed;
        this.m_eSpeakLang               = eSpeak_Lang;
            

        m_sleepTime = sleepTime;
        queue = new LinkedList();
    }


    /**
     * Add speking message
     *
     * @param message
     * @param type
     */
    public boolean push(String message, TYPE type)
    {
        queue.contains(type);
        boolean added = queue.add(new TextDoSpeak(message, type));
        return added;
    }


    /**
     * Metodo che esegue il servizio
     */
    @Override
    public void run()
    {
        try{
            while(true){
                Object textObj = queue.poll();
                if(textObj != null){
                    switch(((TextDoSpeak)textObj).type){
                        case TEXT:
                            speakText(((TextDoSpeak)textObj).message);
                            break;

                        case FILE:
                            speakTextFile(((TextDoSpeak)textObj).message);
                            break;
                    }
                }

                // attesa
                Thread.sleep(m_sleepTime);
            }
        }
        catch(Exception ex){
            LogWriter.writeLog(Level.SEVERE, ex.getMessage());
        }
    }
    

    /**
     *Speak text
     */
    private boolean speakText(String text)throws Exception{
        int returnValue = -1;
        try {
            LogWriter.writeLog(Level.INFO, text);
            
            while(isBusy)
            {
                Thread.sleep(300);
            }
            isBusy = true;

            String commandline = m_eSpeakFileAbsolutePath + " -vit+f4 -s160 \" "+ text +" \" ";
            Runtime rt = Runtime.getRuntime();
            Process process = rt.exec(commandline);
            returnValue = process.waitFor();

        } catch (Exception e) {
            throw e;
        }        
        finally
        {
            isBusy = false;
            return (returnValue == 0);
        }
    }
    
    
    /**
     *Speak text file
     */
    private boolean speakTextFile(String fileName)throws Exception{
        int returnValue = -1;
        try 
        {
            String commandline = m_eSpeakFileAbsolutePath + " -vit+m4 -s 120 -f"+ fileName ;
            Runtime rt = Runtime.getRuntime();
            Process process = rt.exec(commandline);
            returnValue = process.waitFor();
        } catch (Exception e) {
            throw e;
        }        
        finally
        {
            isBusy = false;
            return (returnValue == 0);
        }
    }
}
